#ifndef UART_APP_H
#define UART_APP_H

#include "uart_register.h"

typedef enum {
    FIVE_BITS,
    SIX_BITS,
    SEVEN_BITS,
    EIGHT_BITS
} UartBitsNum4Char;

typedef enum {
    NO_STOP_BIT,
    ONE_STOP_BIT,
    ONE_HALF_STOP_BIT,
    TWO_STOP_BIT
} UartStopBitsNum;

typedef enum {
    PARITY_EVEN         = 2,
    PARITY_ODD          = 3
} UartParity;

typedef enum {
    BIT_RATE_9600       = 9600,
    BIT_RATE_19200      = 19200,
    BIT_RATE_38400      = 38400,
    BIT_RATE_57600      = 57600,
    BIT_RATE_74880      = 74880,
    BIT_RATE_115200     = 115200,
    BIT_RATE_230400     = 230400,
    BIT_RATE_460800     = 460800,
    BIT_RATE_921600     = 921600
} UartBautRate;

typedef enum {
    NONE_CTRL,
    HARDWARE_CTRL,
    XON_XOFF_CTRL
} UartFlowCtrl;

typedef enum {
    EMPTY,
    UNDER_WRITE,
    WRITE_OVER
} RcvMsgBuffState;

typedef struct {
    uint32              RcvBuffSize;
    uint8               *pRcvMsgBuff;
    uint8               *pWritePos;
    uint8               *pReadPos;
    uint8               TrigLvl;
    RcvMsgBuffState     BuffState;
} RcvMsgBuff;

typedef struct {
    uint32      TrxBuffSize;
    uint8       *pTrxBuff;
} TrxMsgBuff;

typedef enum {
    BAUD_RATE_DET,
    WAIT_SYNC_FRM,
    SRCH_MSG_HEAD,
    RCV_MSG_BODY,
    RCV_ESC_CHAR,
} RcvMsgState;

typedef enum {
    NONE_BITS = 0,
    ODD_BITS  = 0,
    EVEN_BITS = BIT4
} UartParityMode;

typedef enum {
    STICK_PARITY_DIS   = 0,
    STICK_PARITY_EN    = BIT3 | BIT5
} UartExistParity;

typedef struct {
    UartBautRate        baut_rate;
    UartBitsNum4Char    data_bits;
    UartExistParity     exist_parity;
    UartParityMode      parity;
    UartStopBitsNum     stop_bits;
    UartFlowCtrl        flow_ctrl;
    RcvMsgBuff          rcv_buff;
    TrxMsgBuff          trx_buff;
    RcvMsgState         rcv_state;
    int                         received;
    int                         buff_uart_no;
} UartDevice;

#endif

