;**********************************************************************
;   This file is a basic code template for assembly code generation   *
;   on the PIC12F683. This file contains the basic code               *
;   building blocks to build upon.                                    *
;                                                                     *
;   Refer to the MPASM User's Guide for additional information on     *
;   features of the assembler (Document DS33014).                     *
;                                                                     *
;   Refer to the respective PIC data sheet for additional             *
;   information on the instruction set.                               *
;                                                                     *
;**********************************************************************
;                                                                     *
;    Filename:	   Panel.asm                                          *
;    Date:                                                            *
;    File Version:                                                    *
;                                                                     *
;    Author:                                                          *
;    Company:                                                         *
;                                                                     *
;                                                                     *
;**********************************************************************
;                                                                     *
;    Files Required: P12F683.INC                                      *
;                                                                     *
;**********************************************************************
;                                                                     *
;    Notes:                                                           *
;                                                                     *
;**********************************************************************

	list      p=12F683        ; list directive to define processor
	#include <p12F683.inc>    ; processor specific variable definitions

	errorlevel  -302, -305    ; suppress messages from list file

	__CONFIG   _FCMEN_OFF & _IESO_OFF & _CP_OFF & _CPD_OFF & _BOD_OFF & _MCLRE_ON & _WDT_OFF & _PWRTE_ON & _INTRC_OSC_NOCLKOUT 

; '__CONFIG' directive is used to embed configuration word within .asm file.
; The lables following the directive are located in the respective .inc file.
; See data sheet for additional information on configuration word settings.

;**********************************************************************
; Constants

OT_NMI	EQU	TRISIO0
OT_RST	EQU	TRISIO1
IN_RST	EQU	GP2
IN_NMI	EQU	GP4
OT_ENB	EQU	TRISIO5

FL_ENB	EQU	0			; ENB output active

FOSC_4	EQU	d'1000000'		; Instruction clock (Hz)
OUT_SG	EQU	d'25'			; Signal width (ms) (max 65)

;**********************************************************************
; Variables

	CBLOCK	0x70
FLAG_S
SAVE_W
SAVE_S
PORT_S
COUN_T
	ENDC

;**********************************************************************
; Reset

	ORG     0x000             
	CLRF	GPIO			; Clear GPIO
	MOVLW 	1<<CM2|1<<CM1|1<<CM0
    	MOVWF	CMCON0			; Turn off comparators
	GOTO    INIT_D
	
;**********************************************************************
; Interrupt

	MOVWF	SAVE_W			; Save context
	SWAPF	STATUS,W
	MOVWF	SAVE_S

	BTFSS	INTCON,GPIF		; GPIO interrupt?
	GOTO	TIM_IT			; No!
	MOVF	GPIO,W			; Clear mismatch
	MOVWF	PORT_S			; Store value
	BCF	INTCON,GPIF		; Clear GPIO interrupt flag
	BTFSS	T1CON,T1CKPS1		; Debouncing?
	GOTO	TIM_IT			; No!
	BCF	T1CON,TMR1ON		; Stop timer
	BTFSS	PORT_S,IN_RST		; RST button pressed?
	GOTO	STA_TM			; Yes!
	BTFSC	PORT_S,IN_NMI		; NMI button pressed?
	GOTO	EXIT_I			; No!
STA_TM	MOVLW	high (~(FOSC_4/d'1000'*OUT_SG))
	MOVWF	TMR1H
	MOVLW	low (~(FOSC_4/d'1000'*OUT_SG))
	MOVWF	TMR1L
	BSF	T1CON,TMR1ON		; Start timer
TIM_IT	BTFSS	PIR1,TMR1IF		; Timer interrupt?
	GOTO	EXIT_I			; No!
	BCF	T1CON,TMR1ON		; Stop timer
	BCF	PIR1,TMR1IF		; Clear timer interrupt flag
	BTFSS	T1CON,T1CKPS1		; Debouncing?
	GOTO	SG_END			; Not!

	banksel	TRISIO
	BTFSC	PORT_S,IN_RST		; RST button pressed?
	GOTO	CK_NMI			; No!
	BTFSC	PORT_S,IN_NMI		; NMI button also pressed?
	GOTO	BU_RST			; No!
	BCF	TRISIO,OT_ENB		; Yes! Assert ENB output
	BSF	FLAG_S,FL_ENB		; Signal: ENB active!
	GOTO	S_TIME
BU_RST	BCF	TRISIO,OT_RST		; Assert RST output
	GOTO	S_TIME
CK_NMI	BTFSC	PORT_S,IN_NMI		; NMI button pressed?
	GOTO	EXIT_I			; No!
	BCF	TRISIO,OT_NMI		; Yes! Assert NMI output
S_TIME	
	banksel	T1CON			; Set signal width = 
	BCF	T1CON,T1CKPS1		; 1/4 of debounce time
INI_CN	MOVLW	0x08
	BTFSC	FLAG_S,FL_ENB		; Release time for ENB
	MOVLW	0x032			; is times longer
	MOVWF	COUN_T			; Initialize counter
	GOTO	STA_TM			; Start timer
SG_END	
	banksel	TRISIO
	MOVLW	0xff
	MOVWF	TRISIO			; Release outputs
	banksel	T1CON
	BTFSS	PORT_S,IN_RST		; RST button still pressed?
	GOTO	INI_CN			; Yes!
	BTFSS	PORT_S,IN_NMI		; NMI button still pressed?
	GOTO	INI_CN			; Yes!
	DECFSZ	COUN_T			; No!
	GOTO	STA_TM			; Repeat the test
	BSF	T1CON,T1CKPS1		; Set debounce timing
	BCF	FLAG_S,FL_ENB		; Signal: ENB inactive!

EXIT_I	SWAPF	SAVE_S,W		; Restore context
	MOVWF	STATUS
	SWAPF	SAVE_W
	SWAPF	SAVE_W,W
	RETFIE

;**********************************************************************
; Initialization

INIT_D	
	banksel	ANSEL
	CLRF	ANSEL			; Inputs are digital
	MOVLW	1<<IOC4|1<<IOC2	        ; Lines 2 and 4 are inputs
	MOVWF	IOC			; with interrupt on change
	BCF	OPTION_REG,NOT_GPPU	; and weak pull-up
	BSF	PIE1,TMR1IE		; Enable timer 1 interrupt
	banksel	GPIO

	CLRF	FLAG_S			; Clear all flags

; Normal timing:  Signal = OUT_SG,   Press = x4,   Release = x8
; Double timing:  x2

	MOVLW	1<<T1CKPS1;|1<<T1CKPS0 	; Normal timing
	MOVWF	T1CON			; Set timer 1 prescaler
	MOVF	GPIO,W			; Clear mismatch status
	MOVWF	PORT_S			; Save port state
	MOVLW	1<<GIE|1<<PEIE|1<<GPIE	; Enable interrupts
	MOVWF	INTCON

;**********************************************************************
; Main loop

M_LOOP	BTFSC	T1CON,TMR1ON		; Timer running?
	GOTO	M_LOOP			; Yes! Wait until it stops
	SLEEP				; Wait for button press
	NOP
	GOTO	M_LOOP

;**********************************************************************

	END                       

