;**********************************************************************
;   This file is a basic code template for assembly code generation   *
;   on the PIC12F683. This file contains the basic code               *
;   building blocks to build upon.                                    *
;                                                                     *
;   Refer to the MPASM User's Guide for additional information on     *
;   features of the assembler (Document DS33014).                     *
;                                                                     *
;   Refer to the respective PIC data sheet for additional             *
;   information on the instruction set.                               *
;                                                                     *
;**********************************************************************
;                                                                     *
;    Filename:	   Panel.asm                                          *
;    Date:                                                            *
;    File Version:                                                    *
;                                                                     *
;    Author:                                                          *
;    Company:                                                         *
;                                                                     *
;                                                                     *
;**********************************************************************
;                                                                     *
;    Files Required: P12F683.INC                                      *
;                                                                     *
;**********************************************************************
;                                                                     *
;    Notes:                                                           *
;                                                                     *
;**********************************************************************

	list      p=12F683        	; list directive to define processor
	#include <p12F683.inc>    	; processor specific variable definitions

	errorlevel  -302, -305, -311	; suppress messages from list file

	__CONFIG   _FCMEN_OFF & _IESO_OFF & _CP_OFF & _CPD_OFF & _BOD_OFF & _MCLRE_OFF & _WDT_OFF & _PWRTE_ON & _INTRC_OSC_NOCLKOUT 

; '__CONFIG' directive is used to embed configuration word within .asm file.
; The lables following the directive are located in the respective .inc file.
; See data sheet for additional information on configuration word settings.

;**********************************************************************
; Constants

NMI_O	EQU	GP0
RST_O	EQU	GP1
RST_I	EQU	GP2
ENB_I	EQU	GP3
NMI_I	EQU	GP4
ENB_O	EQU	GP5
INP_M	EQU	1<<NMI_I|1<<ENB_I|1<<RST_I

FL_DEB	EQU	0			; Deboucing
FL_SIG	EQU	1			; Signal
FL_REL	EQU	2			; Release

FOSC_4	EQU	.1000000		; Instruction clock (Hz)
TM_PER	EQU	.5			; Timing period = 5ms

;**********************************************************************
; Variables

	CBLOCK	0x70
FLAG_S
PORT_S
COUN_T
	ENDC

;**********************************************************************
; Reset

	ORG     0x000             
	MOVLW 	1<<CM2|1<<CM1|1<<CM0
    	MOVWF	CMCON0			; Turn off comparators
	banksel	ANSEL
	GOTO    INIT_D
	
;**********************************************************************
; Interrupt

	MOVF	GPIO,W			; Clear mismatch
	MOVWF	PORT_S			; Store port state
	ANDLW	INP_M			; Set Z if no button is
	XORLW	INP_M			; being pressed
	BTFSS	INTCON,GPIE		; IOC interrupt enabled?
	GOTO	TIM_IT			; No!
	BTFSS	INTCON,GPIF		; IOC interrupt?
	GOTO	TIM_IT			; No!
	BCF	INTCON,GPIF		; Clear IOC interrupt flag
	BTFSC	STATUS,Z		; Button pressed?
	RETFIE				; No!
	BCF	INTCON,GPIE		; Disable IOC interrupt
	BSF	INTCON,PEIE		; Enable timer interrupt
	BSF	FLAG_S,FL_DEB		; Set: Debouncing!
DEB_TM	MOVLW	.64
	MOVWF	COUN_T
STA_TM	BCF	T1CON,TMR1ON		; Stop timer
	MOVLW	high (~(FOSC_4/.1000*TM_PER))
	MOVWF	TMR1H
	MOVLW	low (~(FOSC_4/.1000*TM_PER))
	MOVWF	TMR1L
	BCF	PIR1,TMR1IF		; Clear timer interrupt flag
	BSF	T1CON,TMR1ON		; Start timer
	RETFIE
TIM_IT	BTFSS	INTCON,PEIE		; Timer interrupt enabled?
	RETFIE				; No!
	BTFSS	PIR1,TMR1IF		; Timer interrupt?
	RETFIE				; No!
	BTFSS	FLAG_S,FL_DEB		; Debouncing?
	GOTO	SG_TIM			; No!
	DECFSZ	COUN_T			; Debouncing done?
	GOTO	STA_TM			; Not yet!
	BCF	FLAG_S,FL_DEB		; Set: Debouncing done!
	BTFSC	STATUS,Z		; Button still pressed?
	GOTO	BU_END			; No!
	BSF	FLAG_S,FL_SIG		; Set: Generating signal!
	banksel	TRISIO
	XORLW	1<<NMI_I|1<<RST_I
	BTFSS	STATUS,Z		; NMI and RST pressed?
	GOTO	CK_ENB			; No!
	MOVLW	INP_M|1<<NMI_O|1<<ENB_O|1<<RST_O
	MOVWF	TRISIO			; Deactivate nENB
	XORLW	1<<RST_O		; Assert nRST
	MOVWF	TRISIO
	XORLW	1<<NMI_O		; Assert nNMI
	MOVWF	TRISIO
	BSF	COUN_T,7
SYS_LP	DECFSZ	COUN_T
	GOTO	SYS_LP			; Wait 384us
	XORLW	1<<ENB_O		; Assert nENB
	GOTO	RST_ON
CK_ENB	MOVF	PORT_S,W
	IORLW	INP_M|1<<NMI_O|1<<RST_O
	BTFSC	PORT_S,ENB_I		; ENB button pressed?
	GOTO	CK_RST			; No!
	XORLW	1<<ENB_O		; Toggle nENB
	GOTO	NO_RST
CK_RST	BTFSC	PORT_S,RST_I		; RST button pressed?
	GOTO	CK_NMI			; No!
	XORLW	1<<RST_O		; Assert nRST
RST_ON	BSF	COUN_T,6
	GOTO	SG_OUT
CK_NMI	BTFSS	PORT_S,NMI_I		; NMI button pressed?
	XORLW	1<<NMI_O		; Assert nNMI
NO_RST	BSF	COUN_T,0
SG_OUT	MOVWF	TRISIO
	banksel	GPIO
	GOTO	STA_TM
SG_TIM	BTFSS	FLAG_S,FL_SIG		; Generating signal?
	GOTO	BU_REL			; No!
	MOVF	PORT_S,W
	IORLW	INP_M|1<<NMI_O		; Deactivate nNMI
	banksel	TRISIO
	DECFSZ	COUN_T
	GOTO	SG_OUT
	IORLW	1<<RST_O		; Deactivate nRST
	BCF	FLAG_S,FL_SIG		; Set: Signal done!
	BSF	FLAG_S,FL_REL		; Set: Button release!
	GOTO	RST_ON
BU_REL	BTFSS	FLAG_S,FL_REL		; Waiting for button release?
	GOTO	BU_END			; No!
	BTFSS	STATUS,Z		; Button still pressed?
	GOTO	DEB_TM			; Yes!
	DECFSZ	COUN_T			; Release done?
	GOTO	STA_TM			; Not yet!
	BCF	FLAG_S,FL_REL		; Set: Button done!	
BU_END	BCF	INTCON,PEIE		; Disable timer interrupt
	BCF	INTCON,GPIF		; Clear IOC interrupt flag
	BSF	INTCON,GPIE		; Enable IOC interrupt
	RETFIE	

;**********************************************************************
; Initialization

INIT_D	CLRF	ANSEL			; Inputs are digital
	MOVLW	INP_M
	MOVWF	IOC
	MOVWF	WPU
	BCF	OPTION_REG,NOT_GPPU	; Enable weak pull-up
	BSF	PIE1,TMR1IE		; Enable timer 1 interrupt
	banksel	GPIO
	CLRF	GPIO			; Clear latches
	CLRF	FLAG_S			; Clear flags
	MOVLW	1<<GIE|1<<GPIE		; Enable IOC interrupt
	MOVWF	INTCON

;**********************************************************************
; Main loop

M_LOOP	BTFSC	INTCON,GPIE		; Waiting for button?
	SLEEP				; Yes. Sleep!
	GOTO	M_LOOP

;**********************************************************************
;
	END                       

