;**********************************************************************
;   This file is a basic code template for assembly code generation   *
;   on the PIC10F200. This file contains the basic code               *
;   building blocks to build upon.                                    *
;                                                                     *
;   Refer to the MPASM User's Guide for additional information on     *
;   features of the assembler (Document DS33014).                     *
;                                                                     *
;   Refer to the respective PIC data sheet for additional             *
;   information on the instruction set.                               *
;                                                                     *
;**********************************************************************
;                                                                     *
;    Filename:	    Button.asm                                        *
;    Date:                                                            *
;    File Version:                                                    *
;                                                                     *
;    Author:                                                          *
;    Company:                                                         *
;                                                                     * 
;                                                                     *
;**********************************************************************
;                                                                     *
;    Files Required: P10F322.INC                                      *
;                                                                     *
;**********************************************************************
;                                                                     *
;    Notes:                                                           *
;                                                                     *
;**********************************************************************

	list      p=10F322      ; List directive to define processor
	#include <p10F322.inc>  ; Processor specific variable definitions
	errorlevel    -311,-202 ; Suppress the message: Operand of HIGH
				;      operator was larger than H'FFFF'

	__CONFIG _FOSC_INTOSC & _LVP_OFF & _MCLRE_OFF & _WDTE_OFF
; '__CONFIG' directive is used to embed configuration word within .asm file.
; The lables following the directive are located in the respective .inc file. 
; See respective data sheet for additional information on configuration word.

;**********************************************************************
; Constants

PA_RST	EQU	RA0
PA_ENB	EQU	RA1
PA_NMI	EQU	RA2
PA_BUT	EQU	RA3

FL_ACT	EQU	0
FL_SIG	EQU	1
FL_REL	EQU	2

FOSC_4	EQU	.2000000		; Instruction Clock = 8MHz / 4
TM_PER	EQU	.5			; Timing period = 5ms
PRE_SC	EQU	.64			; Timer prescaler = 1:64

;***** VARIABLE DEFINITIONS

	cblock	0x040
FLAG_S	
COUN_T
	endc

;**********************************************************************
	ORG     0x000 

	CLRF	ANSELA			; No analog inputs
	MOVLW	1<<WPUA3		; Pull-up enabled only on
	MOVWF	WPUA			; button
	GOTO    INIT_D
	

;**********************************************************************
; Interrupt
	
TM_INT	BTFSS	INTCON,TMR0IE		; Timer interrupt enabled?
	GOTO	IO_CHG			; No!
	BTFSS	INTCON,TMR0IF		; Timer interrupt?
	GOTO	IO_CHG			; No!
; Timer interrupt occurs every 5ms
	BTFSS	FLAG_S,FL_REL		; Timing button release?
	GOTO	TM_SIG			; No!
	BTFSS	PORTA,PA_BUT		; If the button is still
	GOTO	CNT_DN			; pressed, restart counting
	DECFSZ	COUN_T,F
	GOTO	INI_TM
; The button was released for 200ms
	BCF	INTCON,TMR0IE		; Disable timer interrupt
	BCF	FLAG_S,FL_REL		; Signal: Button released!
	BCF	FLAG_S,FL_ACT		; Signal: Go to sleep
	RETFIE
TM_SIG	BTFSS	FLAG_S,FL_SIG		; Timing output pulse?
	GOTO	TM_CNT			; No!
; The output signal was active for 5ms
	BSF	TRISA,PA_RST		; Deactivate nRST and nNMI
	BSF	TRISA,PA_NMI
	BCF	FLAG_S,FL_SIG
CNT_DN	BSF	FLAG_S,FL_REL		; Wait for button release!
	MOVLW	.40			; Set counter for 200ms
	MOVWF	COUN_T
	GOTO	INI_TM
TM_CNT	INCF	COUN_T,F		; Increment counter
	BTFSC	STATUS,Z		; 256 counts?
	GOTO	TIM_MX			; Yes
	MOVF	COUN_T,W		; Wait 200ms for debouncing
	ADDLW	-.40
	BTFSC	STATUS,C
	BTFSS	PORTA,PA_BUT		; Button still pressed?
	GOTO	INI_TM			; Yes
; First release bounce
	BSF	FLAG_S,FL_SIG
	MOVF	COUN_T,W
	ADDLW	-.100			; Button press longer than
	BTFSS	STATUS,C		; 500ms?
	GOTO	TS_NMI
	BCF	TRISA,PA_RST		; Yes! Activate nRST
	GOTO	INI_TM
TS_NMI	BCF	TRISA,PA_NMI		; No! Activate nNMI
	GOTO	INI_TM
TIM_MX	MOVLW	1<<PA_ENB		; Toggle ENB output
	XORWF	TRISA,F
	GOTO	CNT_DN
; Check if the button was pressed
IO_CHG	BTFSS	INTCON,IOCIE		; Interrupt on change enabled?
	RETFIE				; No!
	BTFSS	INTCON,IOCIF		; Interrupt on change?
	RETFIE				; No!
; Start counting periods of 5ms
	BCF	INTCON,IOCIE		; Disable further IOC
	CLRF	COUN_T			; Initialize the counter
	BSF	INTCON,TMR0IE		; Enable timer interrupt
; Initialize timer for 5ms
INI_TM	MOVLW	low(~(FOSC_4/d'1000'*TM_PER/PRE_SC))
	MOVWF	TMR0
	BCF	INTCON,TMR0IF		; Clear interrupt flag
	RETFIE

;**********************************************************************
; Initialization

INIT_D	MOVLW	1<<PS2|1<<PS0	        ; Pull-up enabled, timer 0 
	MOVWF	OPTION_REG		; prescaler =1:64
	BSF	IOCAN,IOCAN3		; Button interrupt on falling edge
	CLRF	FLAG_S
	CLRF	PORTA			; Outputs active low
	MOVLW	1<<GIE			; Enable interrupts
	MOVWF	INTCON

;**********************************************************************
; Main loop


M_LOOP	BTFSC	FLAG_S,FL_ACT		; Go to sleep?
	GOTO	M_LOOP			; Not yet!
	MOVF	PORTA,W			; Clear mismatch
	NOP
	CLRF	IOCAF			; Clear interrupt on change flag 
	BSF	INTCON,IOCIE		; Enable button interrupt
	SLEEP				; Wait for button press
	NOP
	BSF	FLAG_S,FL_ACT		; Disable sleep
	GOTO	M_LOOP


;**********************************************************************
	END				; Directive 'end of program'

