;******************************************************************************
;   This file is a basic template for assembly code for a PIC18F4550. Copy    *
;   this file into your project directory and modify or add to it as needed.  *
;                                                                             *
;   Refer to the MPASM User's Guide for additional information on the         *
;   features of the assembler.                                                *
;                                                                             *
;   Refer to the PIC18Fx455/x550 Data Sheet for additional                    *
;   information on the architecture and instruction set.                      *
;                                                                             *
;******************************************************************************
;                                                                             *
;    Filename:      Serial adaptor for the ZX Spectrum Interface 1-bis V2.1   *
;    Date:          16 Apr 2011                                               *
;    File Version:  2.0                                                       *
;                                                                             *
;    Author:                                                                  *
;    Company:                                                                 *
;                                                                             * 
;******************************************************************************
;                                                                             *
;    Files Required: P18F4550.INC                                             *
;                                                                             *
;******************************************************************************
	LIST P=18F4550		; Directive to define processor
	#include <P18F4550.INC>	; Processor specific variable definitions
	errorlevel	-311	; Suppress the message: Operand of HIGH
				;      operator was larger than H'FFFF'
;------------------------------------------------------------------------------
;
; CONFIGURATION WORD SETUP
;
; The 'CONFIG' directive is used to embed the configuration word within the 
; .asm file. The lables following the directive are located in the respective 
; .inc file.  See the data sheet for additional information on configuration 
; word settings.
;
;------------------------------------------------------------------------------

	CONFIG	PLLDIV = 5		; Divide by 5 (20 MHz oscillator input)
	CONFIG	CPUDIV = OSC1_PLL2	; [OSC1/OSC2 Src: /1][96 MHz PLL Src: /2]
	CONFIG	USBDIV = 2         	; USB from the 96 MHz PLL divided by 2
	CONFIG	FOSC = HSPLL_HS		; HS oscillator, PLL enabled, HS used by USB
	CONFIG	FCMEN = OFF		; Fail-Safe Clock Monitor disabled
	CONFIG	IESO = OFF		; Oscillator Switchover mode disabled
	CONFIG	PWRT = ON		; PWRT disabled
	CONFIG	BOR = ON		; Brown-out Reset enabled in hardware only
	CONFIG	BORV = 2		; Brown-out Voltage medium setting
	CONFIG	VREGEN = ON		; USB voltage regulator enabled
	CONFIG	WDT = OFF		; HW Disabled - SW Controlled
	CONFIG	WDTPS = 512		; Watch-dog timer set to 512*4ms = 2s
	CONFIG	MCLRE = OFF             ; RE3 input pin enabled; MCLR disabled
	CONFIG	LPT1OSC = OFF		; Timer1 configured for higher power operation
	CONFIG	PBADEN = OFF		; PORTB<4:0> pins are digital I/O on Reset
	CONFIG	CCP2MX = OFF		; CCP2 input/output is multiplexed with RB3
	CONFIG	STVREN = OFF		; Stack full/underflow will not cause Reset
	CONFIG	LVP = OFF		; Single-Supply ICSP disabled
	CONFIG	ICPRT = OFF		; ICPORT disabled
	CONFIG	XINST = OFF		; Instruction set extension disabled
	CONFIG	DEBUG = OFF 		; RB6 and RB7 are general purpose I/O pins
	CONFIG	CP0 = OFF		; Block 0 (000800-001FFFh) not code-protected
	CONFIG	CP1 = OFF		; Block 1 (002000-003FFFh) not code-protected
	CONFIG	CP2 = OFF		; Block 2 (004000-005FFFh) not code-protected
	CONFIG	CP3 = OFF		; Block 3 (006000-007FFFh) not code-protected
	CONFIG	CPB = OFF		; Boot block (000000-0007FFh) not code-protected
	CONFIG	CPD = OFF		; Data EEPROM not code-protected
	CONFIG	WRT0 = OFF		; Block 0 (000800-001FFFh) not write-protected
	CONFIG	WRT1 = OFF		; Block 1 (002000-003FFFh) not write-protected
	CONFIG	WRT2 = OFF		; Block 2 (004000-005FFFh) not write-protected
	CONFIG	WRT3 = OFF		; Block 3 (006000-007FFFh) not write-protected
	CONFIG	WRTB = OFF		; Boot block (000000-0007FFh) not write-protected
	CONFIG	WRTC = OFF		; Configuration (300000-3000FFh) not write-protected
	CONFIG	WRTD = OFF		; Data EEPROM not write-protected
	CONFIG	EBTR0 = OFF		; Block 0 not protected from table reads
	CONFIG	EBTR1 = OFF		; Block 1 not protected from table reads
	CONFIG	EBTR2 = OFF		; Block 2 not protected from table reads
	CONFIG	EBTR3 = OFF		; Block 3 not protected from table reads
	CONFIG	EBTRB = OFF		; Boot block not protected from table reads

;------------------------------------------------------------------------------
; CONSTANT DEFINITIONS

; I/O Port bits
B_PSDI	equ	RB0			; SPI Data In
B_PSCK	equ	RB1			; SPI Clock
B_NSTB	equ	RB2			; Parallel port Strobe
B_NFCS	equ	RB3			; Flash memory select
B_NSCS	equ	RB4			; External SPI select
B_NCRD	equ	RB5			; Flash card detect
B_USEN	equ	RB6			; USB power sense
B_EINT	equ	RB7			; External SPI interrupt
C_DRDY	equ	RC2			; SPP Data Ready
E_NCTS	equ	RE0			; Async CTS to server
E_NSDC	equ	RE1			; SD Card not present
E_NSRV	equ	RE2			; Server not connected
E_NRTS	equ	RE3			; Async RTS from server
; Clocks
FOSC_4	equ	d'12000000'		; Instruction cycle (Hz)
BAUD_R	equ	d'115200'		; RS232 Baud rate (Hz)
; Communication State (COM_ST) bits
CO_ACT	equ	0			; Communication in progress
CO_HED	equ	1			; Header / not Data
CO_TRA	equ	2			; Transmit / not Receive
CO_SER	equ	3			; Serial / not Parallel
CO_CHG	equ	4			; Device status has changed
; Timer 0 State (TIM_ST) bits
TM_SBD	equ	0			; USB power debounce
TM_CRD	equ	1			; SDC detect debounce
TM_PSP	equ	2			; PSP timer running
TM_USB	equ	3			; USB timer running
TM_SDC	equ	4			; SDC timer running
TM_RTS	equ	5			; RTS timer running
; ZX Interface 1-Bis constants
SA_SIG	equ	0x40			; Serial Adaptor signature
NS_ATC	equ	0xf1			; Client Command Code
E_COMM	equ	0x09			; Error 'Communication error'
E_NSUP	equ	0x07			; Error 'Feature not supported'
E_NPRE	equ	0x10			; Error 'Microdrive not present'
E_WRPR	equ	0x0e			; Error 'Drive write protected'

; USB related constants

; Processor specific
USB_MB	equ	2			; LOG2 of first USB memory bank
; BDnSTAT bits (DS39632E Register 17-5, 17-6) missing from .INC file 
BSTALL	equ	2			; Stall
DTSEN	equ	3			; Data Sync Enable
DTS	equ	6			; Data Sync Parity
UOWN	equ	7			; USB Owned
PID0	equ	2			; PID bit 0
PID1	equ	3			; PID bit 1
PID2	equ	4			; PID bit 2
PID3	equ	5			; PID bit 3

; Application specific
; General
CF_MAX	equ	1			; Maximum configuration number
IF_MAX	equ	1			; Maximum interface number
EP_MAX	equ	2			; Maximum endpoint number
; CC Endpoint Numbers
EP_CTL	equ	1			; CC Control Endpoint number
EP_DAT	equ	EP_MAX+1-EP_CTL		; CC Data Endpoint number
; Maximum packet size (binary logarithm)
PS_CTL	equ	3			; LOG2 of control packet size
PS_DAT	equ	6			; LOG2 of data packet size
; USB State (USB_ST) bits
SB_SPW	equ	0			; USB Self Powered
SB_RWU	equ	1			; USB Remote Wake-up enabled
SB_DEF	equ	2			; USB in Default state
SB_CFG	equ	3			; USB in Configured state
; Other
TB_SEL	equ	0			; Table selector bit
 
; Standard USB Constants
; Token PID Types (USB 2.0 Table 8-1)
PID_SU	equ	b'1101'			; SETUP Token
; Request Type Bits (USB 2.0 Table 9-2)
RT_ITF	equ	0			; Interface
RT_ENP	equ	1			; Endpoint
RT_CLA	equ	5			; Class
RT_VEN	equ	6			; Vendor
RT_DIR	equ	7			; Data Direction IN
; Standard Request Codes (USB 2.0 Table 9-4)
R_GSTA	equ	0x00			; Get Status
R_CFTR	equ	0x01			; Clear Feature
R_SFTR	equ	0x03			; Set Feature
R_SADR	equ	0x05			; Set Address
R_GDES	equ	0x06			; Get Descriptor
R_GCFG	equ	0x08			; Get Configuration
R_SCFG	equ	0x09			; Set Configuration
R_GITF	equ	0x0a			; Get Interface
R_SITF	equ	0x0b			; Set Interface
; Standard Descriptor Types (USB 2.0 Table 9-5)
DT_DEV	equ	0x01			; Device
DT_CFG	equ	0x02			; Configuration
DT_STR	equ	0x03			; String
DT_ITF	equ	0x04			; Interface
DT_ENP	equ	0x05			; Endpoint
; Feature Selectors (USB 2.0 Table 9-6)
DF_RWU	equ	0x01			; Device Remote Wake-up
EF_HLT	equ	0x00			; Endpoint HALT
; Endpoint Status bits (USB 2.0 Figure 9-6)
GS_HLT	equ	0			; HALT
; Standard Configuration Attributes (USB 2.0 Table 9-10)
CA_RWU	equ	1<<7|1<<5		; Remote Wake-up
CA_SPW	equ	1<<7|1<<6		; Self powered
; Standard Endpoint Address bits (USB 2.0 Table 9-13)
EA_DIR	equ	7			; Direction IN
; Standard Endpoint Attributes (USB 2.0 Table 9-13)
EA_BLK	equ	0x02			; Bulk
EA_INT	equ	0x03			; Interrupt
; CC Device and Interface (Sub)Classes (CDC 1.2 Table 2,3,4,6)
DC_NUS	equ	0x00			; Not Used
DC_COM	equ	0x02			; Communication Device Class
IC_COM	equ	0x02			; Communication Interface Class
IC_DAT	equ	0x0a			; Data Interface Class
IS_ACM	equ	0x02			; Abstract Control Interface Subclass
; CC Protocols (CDC 1.2 Table 5,7)
PC_NOP	equ	0x00			; No protocol
PC_ATC	equ	0x01			; AT Commands V.250
; CC Descriptor Types (CDC 1.2 Table 12)
DT_CSI	equ	0x24			; CS Interface Descriptor
; CC Functional Descriptor Subtypes (CDC 1.2 Table 13)
FD_HED	equ	0x00			; Header Descriptor
FD_CMA	equ	0x01			; Call Management Descriptor
FD_ACM	equ	0x02			; Abstract Control Descriptor
FD_UNI	equ	0x06			; Union Descriptor
; CC Request Codes (CDC 1.2 Table 19)
C_SECO	equ	0x00			; Send Encapsulated Command
C_GERP	equ	0x01			; Get Encapsulated Response
C_SLCO	equ	0x20			; Set Line Coding
C_GLCO	equ	0x21			; Get Line Coding
C_SCLS	equ	0x22			; Set Control Line State
; CC Notifications (CDC 1.2 Table 20)
N_SSTA	equ	0x20			; Serial State
; CC Abstract Control Management Capability Bits (PSTN 1.2 Table 4)
CA_AC1	equ	1			; C_SLCO, C_GLCO, C_SCLS, N_SSTA
; CC Control Line State Bits (CLN_ST) (PSTN 1.2 Table 18)
CS_DTR	equ	0			; DTR
CS_RTS	equ	1			; RTS
; CC Serial Line State Bits (PSTN 1.2 Table 31)
SS_DCD	equ	0			; DCD
SS_DSR	equ	1			; DSR

; SD card related constants
CB_TRN	equ	6			; Command Transfer bit
CB_HCS	equ	6			; 
R1_IDL	equ	0			; R1 'Idle State' bit
R1_ILC	equ	2			; R1 'Illegal Command' bit
C8_VHS	equ	b'0001'			; CMD8 VHS field
C8_PAT	equ	0xaa			; CMD8 check pattern
OC_CCS	equ	6
OC_PWU	equ	7
SD_DST	equ	0xfe			; DATA_START_TOKEN
SD_DAC	equ	0x05			; DATA_ACCEPTED
CSD_HC	equ	6
; SD card Commands
SD_SWA	equ	0xf8			; 'Set Write Address'
SD_IDF	equ	0xf9			; 'Identify Card'
SD_BRD	equ	0xfa			; 'Read Block'
SD_BWR	equ	0xfb			; 'Write Block'
SD_EFA	equ	0xfc			; 'Set First Erase Address'
SD_ELA	equ	0xfd			; 'Set Last Erase Address'
SD_ERA	equ	0xfe			; 'Erase Blocks'
; SD card State (SDC_ST) bits
SD_VER	equ	0			; SD card Version 2.0
SD_HCS	equ	1			; SD card High Capacity
SD_DET	equ	2			; Detect SD card
SD_PRE	equ	3			; SD card Present
SD_EXE	equ	4			; Execute SDC command

;------------------------------------------------------------------------------
; MACRO INSTRUCTION DEFINITIONS

timing	macro	ms
	MOVLW	high (-(FOSC_4/d'1000'*ms/h'100'))
	MOVWF	TMR0H
	MOVLW	low (-(FOSC_4/d'1000'*ms/h'100'))
	endm

copybl	macro	src,dst,num
	local	i = 0
	while	i < num
	MOVFF	src+i,dst+i
i ++
	endw
	endm
;------------------------------------------------------------------------------
; VARIABLE DEFINITIONS

	cblock	0x000			; Variables in access RAM
COM_ST					; Communication State 
TIM_ST					; Timer 0 State
CHG_ST					; Last PORTB value
USB_ST					; USB State
CLN_ST					; Control Line State
SDC_ST					; SD card State
CNT_LO					; Byte counter low
CNT_HI					; Byte counter high
EP_OFS					; Offset of last active EP
TB_CNT					; Table read byte counter
; USB Request Header
RQ_TYP,RQ_NUM,RQ_VAL,RQ_DTY,RQ_IDX:2,RQ_LNL,RQ_LNH
; SD card variables
SD_SZ0,SD_SZ1,SD_SZ2,SD_SZ3		; SD card size 
CM_CNT,RP_CNT,DT_CNL,DT_CNH		; SD Command counters
; SD card Command
SD_CMD,SD_AD3,SD_AD2,SD_AD1,SD_AD0,SD_CRC
	endc

	cblock	(1<<USB_MB)*0x100	; Variables in USB RAM
; Buffer Descriptor Table
BDT0_O,EP0O_L:3,BDT0_I,EP0I_L:0x13
; Endpoint 0 buffers
BUF0_O:1<<PS_CTL			; EP0 OUT buffer
BUF0_I:1<<PS_CTL			; EP0 IN buffer
; CC Notification Header 
NT_TYP,NT_NUM,NT_VAL:2,NT_IDX:2,NT_LEN:2
; CC Notification Data
NT_DAT:4
; CC Line Coding data
LN_COD:7
; Main communication buffer
CMD_CD,BLN_LO,BLN_HI			; Communication Header Block
DATA_B:0x202				; Communication Data Block
	endc

;------------------------------------------------------------------------------
; RESET

	ORG	0x0000
	CLRF	COM_ST			; Clear the Device State
	CLRF	TIM_ST			; Clear the Timer State
	GOTO	MAIN_C

;------------------------------------------------------------------------------
; HIGH PRIORITY INTERRUPT

	BTFSC	INTCON3,INT2IF		; SPP Strobe interrupt?		
	BRA	SPP_IT			; Yes!
	BTFSC	INTCON,TMR0IF		; Timer 0 Interrupt?
	BRA	TMR_IT			; Yes!
	BTFSS	INTCON,RBIF		; PORTB Interrupt on change?
	BRA	SDC_IT			; No!
; PORTB Interrupt on Change can be triggered by:
; - Connecting or disconnecting the USB cable
; - Inserting or ejecting the SD card
	MOVF	PORTB,W			; Get the new PORTB value
	XORWF	CHG_ST,W		; Set the changed bits
	XORWF	CHG_ST			; Change old to new value
	BCF	INTCON,RBIF		; Clear interrupt flag
	BTFSS	WREG,B_USEN		; USB power?
	BRA	CRD_IT			; No!
; USB cable (dis)connected
	BSF	TIM_ST,TM_SBD		; Signal: USB power!
WAI_50	timing	.50			; and wait for 50ms
WAIT_W	RCALL	STA_TM
	RETFIE	FAST
CRD_IT	BTFSS	WREG,B_NCRD		; SD card change?
	BRA	EXT_IT			; No!
; SD card inserted / ejected 
	BSF	TIM_ST,TM_CRD		; Signal: SD card change!
	timing	.1000
	BRA	WAIT_W			; Wait for 1s
EXT_IT	BTFSS	WREG,B_EINT
	RETFIE	FAST			; No!
; External SPI Interrupt
	BTFSC	CHG_ST,B_EINT		; Falling edge?
	RETFIE	FAST			; No!
	RETFIE	FAST
; Timer 0 Overflow Interrupt :
; - USB cable (dis)connect de-bounce
; - SD card insert/eject de-bounce
; - USB RESUME signal timing
; - USB Communication Stage time-out
; - SD card time-out
; - PSP Communication Stage time-out
; - Wait for the assertion of the RTS line
TMR_IT	BCF	T0CON,TMR0ON		; Stop timer
	BCF	INTCON,TMR0IF		; Clear interrupt flag
	BTFSS	UCON,RESUME		; USB resume?
	BRA	SBD_TO			; No!
; The USB RESUME signal must be active for 5ms
	BCF	UCON,RESUME		; Stop RESUME signal
	BRA	UEP_PT
SBD_TO	BTFSS	TIM_ST,TM_SBD		; USB power?
	BRA	CRD_TO			; No!
; The USB cable connection state is assessed after 50ms
	BCF	TIM_ST,TM_SBD		; Signal: Not USB power!
	RCALL	USB_SE			; Attach / detach USB
	RETFIE	FAST
CRD_TO	BTFSS	TIM_ST,TM_CRD		; SD card change?
	BRA	SDC_TO			; No!
; The SD card state is assessed after 1s
	BCF	TIM_ST,TM_CRD		; Signal: No SD card change!
	BSF	SDC_ST,SD_DET		; Signal: Detect SD card!
	RETFIE	FAST
SDC_TO	BTFSS	TIM_ST,TM_SDC		; SD card timer running?
	BRA	USB_TO			; No!
	BCF	TIM_ST,TM_SDC		; Signal: Time-out
	RETFIE	FAST
USB_TO	BTFSC	TIM_ST,TM_USB		; USB time-out error?
; USB Communication Stage time-out occurs after 1s
	BRA	CO_ERR			; Yes! 
	BTFSC	TIM_ST,TM_PSP		; PSP time-out error?
; PSP Communication Stage time-out occurs after 1s
	BRA	PSP_ER
	BTFSC	TIM_ST,TM_RTS		; Waiting for RTS?
; The RTS line is re-checked every 50ms
	BRA	RES_TX			; Yes!
	RETFIE	FAST			; No other timer events!
; SPI Interrupt
SDC_IT	BTFSS	PIE1,SSPIE		; SPI interrupts enabled?
	BRA	ARX_IT			; No!
	BCF	PIE1,SSPIE		; Disable the SPI interrupt
	BRA	STRT_T			; Return response
; Check if the server has set the DTR line
CK_DTR	BCF	LATE,E_NSRV		; Switch on the Server LED
	BTFSS	CLN_ST,CS_DTR		; DTR set?
NO_DTR	BSF	LATE,E_NSRV		; Switch off the Server LED
	BSF	COM_ST,CO_CHG
	RETFIE	FAST	
; USART Interrupt
ARX_IT	BTFSS	PIE1,RCIE		; RX interrupt enabled?
	BRA	ATX_IT			; No!
	BTFSC	PIR1,RCIF		; USART Receive Interrupt?
; USART Receive Interrupt
	BRA	NEW_CO			; Yes!
ATX_IT	BTFSS	PIE1,TXIE		; TX interrupt enabled?
	BRA	USB_IT			; No!
	BTFSC	PIR1,TXIF		; USART Transmit Interrupt?
; USART Transmit interrupt
	BRA	TST_CN			; Yes!
USB_IT	BTFSS	PIE2,USBIE		; USB interrupt enabled?
	RETFIE	FAST			; No!
	BTFSS	PIR2,USBIF		; USB interrupt?
	RETFIE	FAST			; No!
; USB Interrupt
; Call the USB Interrupt Service Routine to handle all USB events
; except for the CC Data Transfers Complete interrupts
	RCALL	USB_IS			; Service USB interrupt
	BTFSS	USB_ST,SB_CFG		; USB in 'Configured State'?
	BRA	NO_DTR			; No!
; FSR0 was initialized at the start of the USB Communication Stage with
; the address of the corresponding CC Data (IN / OUT) endpoint descriptor
; Only transfers of that same endpoint are handled until the stage ends
	MOVF	FSR0L,W			; Expected EP offset
	CPFSEQ	EP_OFS			; Does it match?
	BRA	CK_DTR			; No! 
	INCF	FSR0L			; Point to the packet size
; The first 3 bytes of the first packet received are the Communication
; Header, containing the size of an eventual Data Block
	BTFSS	COM_ST,CO_HED		; Header Block needed?
	BRA	USB_CO			; No!
	MOVF	INDF0,W			; Size of the received packet
	BZ	UEP_SZ			; Allow for zero-length packets
	SUBLW	2			; The first received packet must
	BC	TMR_IT			; be at least 3 bytes long
	RCALL	TOT_LN			; Get the total length 
; Move the buffer pointer after the last byte transferred
USB_CO	MOVF	POSTINC0,W		; Update the buffer address
	ADDWF	POSTINC0		; by adding the length of the
	MOVLW	0			; current transfer.
	ADDWFC	POSTDEC0
	DECF	FSR0L			; Point again to the packet size
; Check if there is more data to be transferred
	MOVF	INDF0,W			; Subtract the packet size of
	SUBWF	CNT_LO			; the current transfer from the
	MOVLW	0			; remaining length.
	SUBWFB	CNT_HI
	BNC	TMR_IT			; This should never happen!
	BNZ	UEP_SZ
	CPFSEQ	CNT_LO			; If there is still data to be
	BRA	UEP_SZ			; transferred, arm the endpoint
; All data has been transferred. If the last transmitted packet was 
; full-sized, send a zero-length packet.
	BTFSS	EP_OFS,DIR		; IN transfer ?
	BRA	END_CO			; No!
	BTFSC	INDF0,PS_DAT		; If the last packet was full,
	BRA	UEP_ST			; send a zero-length packet
; End of the USB Communication Stage
	BRA	END_CO
;
; Parallel Port Strobe (INT2) Interrupt
SPP_IT	BCF	INTCON3,INT2IF		; Clear the interrupt flag		
; One byte has been transferred
NEW_CO	BTFSC	COM_ST,CO_ACT		; New communication stage?
	BRA	CON_CO			; No!
; Start of a new Reception Stage
	MOVLW	2			; Header Block length - 1
; Start of a new Communication Stage
STA_CO	BSF	COM_ST,CO_ACT		; Communication started!
	BSF	COM_ST,CO_HED		; Signal: Header Block needed!
	MOVWF	CNT_LO			; Copy the Header Block length
	CLRF	CNT_HI			; to the byte counter
	LFSR	FSR0,CMD_CD		; Reset the buffer address
;	BTFSC	COM_ST,CO_SER		; Serial communication?
;	BRA	CON_CO			; Yes
;	RCALL	TIM_1S			; Start timer
;	BSF	TIM_ST,TM_PSP		; Signal: PSP timer running
CON_CO	BTFSC	COM_ST,CO_TRA		; Transmission in progress?
	BRA	TST_CN			; Yes!
; One byte has been received
	BTFSC	COM_ST,CO_SER		; Serial reception?
	BRA	UART_R			; Yes!
; One byte has been received over the Parallel Port
	MOVFF	PORTD,POSTINC0		; Copy received byte to buffer
; Test if the byte counter has reached 0
TST_CN	TSTFSZ	CNT_LO			; Test if the byte counter has
	BRA	DEC_LO			; reached 0.
	TSTFSZ	CNT_HI
	BRA	DEC_HI
; All data has been transferred. Check if there is a Data Block, of up
; to 512 bytes, to be transferred after the 3-byte Header Block
	BTFSS	COM_ST,CO_HED		; Data Block finished?
	BRA	END_CO			; Yes
; Prepare to transfer the Data Block
	BCF	COM_ST,CO_HED		; Signal: Header Block done!
	MOVFF	BLN_LO,CNT_LO		; Copy the Data Block size
	MOVFF	BLN_HI,CNT_HI		; to the byte counter and
	BRA	TST_CN			; test again
; One byte has been received over the USART
UART_R	MOVF	RCSTA,W			; USART Receive Status
	MOVFF	RCREG,POSTINC0		; Copy received byte to buffer
	ANDLW	1<<FERR|1<<OERR		; Error?
	BZ	TST_CN			; No!
; Communication Error
; Send an error message to the client
CO_ERR	MOVLW	E_COMM			; Error Code
ERR_RT	MOVWF	CMD_CD
	CLRF	BLN_LO			; Set the Data Block size
	CLRF	BLN_HI			; to 0
	BSF	COM_ST,CO_SER           ; Parallel transmission
	BCF	COM_ST,CO_TRA		; follows!
; End of Communication Stage
END_CO	BCF	COM_ST,CO_ACT		; End of Communication Stage!
	BCF	T0CON,TMR0ON		; Stop timer
	CLRF	TIM_ST			; Clear all timers events
; Prepare the next Communication Stage
	BTG	COM_ST,CO_TRA		; Toggle Transmit/Receive flag
	BTFSS	COM_ST,CO_TRA		; Receive to follow?
	BRA	PREP_R			; Yes!
; A Transmit Stage is triggered immediately after each Receive Stage,
; within the same interrupt
	BTFSC	COM_ST,CO_SER		; Last reception over parallel port ?
	BRA	PREP_T			; No!
	MOVLW	NS_ATC			; Is the received code a
	CPFSLT	CMD_CD			; Client Command?
	BRA	TS_CLC			; Yes!
; Prepare for Transmission
PREP_T	BTG	COM_ST,CO_SER		; Toggle Serial/Parallel flag
	BTFSS	COM_ST,CO_SER		; Serial transmission?
	BRA	DIS_RX			; No!
	MOVLW	E_NPRE			; Signal error: 'Not present'
	BTFSC	LATE,E_NSRV		; if the server is not connected
	BRA	ERR_RT
	BTFSC	USB_ST,SB_CFG		; Transmit over USB?
	BRA	USB_TX			; Yes!
; Start USART Transmission
	BTFSS	RCSTA,SPEN		; USART enabled?
	RCALL	ASY_EN			; If not, enable now
	BSF	PIE1,TXIE		; Enable the TX interrupt
DIS_RX	BCF	PIE1,RCIE		; Disable the RX interrupt
STRT_T	MOVLW	3			; Size of the Header Block
	BRA	STA_CO			; Start transmission
; Execute a client command and send the response
TS_CLC	MOVLW	SD_SWA
	CPFSGT	CMD_CD			; SD card command?
	BRA	EX_CLC			; No!
	BSF	SDC_ST,SD_EXE		; Signal: Execute SDC command!
	RETFIE	FAST
EX_CLC	RCALL	CL_CMD			; Execute the command
	CLRF	CMD_CD			; Clear the error code
	BRA	STRT_T			; Start transmission
; Start USB Transmission
; Point FSR0 to the last byte of the CC Data IN endpoint BDT entry 
USB_TX	LFSR	FSR0,BDT0_I+(1<<ENDP0)*EP_DAT+(1<<DIR)-1
	RCALL	TOT_LN			; Get the total length
	BRA	SUS_CK
; Start USB Reception
; Point FSR0 to the last byte of the CC Data OUT endpoint BDT entry 
USB_RX	LFSR	FSR0,BDT0_O+(1<<ENDP0)*EP_DAT+(1<<DIR)-1
	BSF	COM_ST,CO_HED		; Signal: Need Header Block!
SUS_CK	BTFSS	UCON,SUSPND		; USB in Suspended Mode?
	BRA	UEP_PT			; No!
; If the USB is in Suspend Mode, try Remote Wake-up
	BTFSS	USB_ST,SB_RWU		; Remote Wake-up enabled?
	BRA	UEP_PT			; No!
; Activate the RESUME signal for 5ms
	BSF	UCON,RESUME		; Yes!  Activate the 
	timing	.5			; RESUME signal for 5ms
	BRA	WAIT_W
; Initialize the USB buffer address
UEP_PT	BSF	COM_ST,CO_ACT		; Communication started!
	MOVLW	high CMD_CD		; Initialize the buffer
	MOVWF	POSTDEC0		; address.
	MOVLW	low CMD_CD
	MOVWF	POSTDEC0
; Set the time-out for the USB Communication Stage to 1s
	RCALL	TIM_1S			; Start timer
	BSF	TIM_ST,TM_USB		; Signal: USB timer running!
; Establish the packet size
UEP_SZ	MOVLW	-(1<<PS_DAT)
	ANDWF	CNT_LO,W		; Clear Z if there is enough data
	IORWF	CNT_HI,W		; for a full-length packet
	MOVLW	1<<PS_DAT		; Full-length
; Allow full-length for the first packet to be received
	BTFSC	COM_ST,CO_HED		; Have Header Block?
	BRA	UEP_ST			; Not yet!
	BNZ	UEP_ST
	MOVF	CNT_LO,W		; Short packet!
UEP_ST	MOVWF	POSTDEC0		; Set the packet size
; Pass control of the endpoint to SIE
	MOVLW	1<<UOWN|1<<DTSEN
	BTFSC	INDF0,DTS		; Copy the DTS bit			
	BSF	WREG,DTS
	MOVWF	INDF0			; Pass control to SIE
	RETFIE	FAST
; Decrement the byte counter
DEC_HI	DECF	CNT_HI			; Decrement the byte counter
DEC_LO	DECF	CNT_LO
	BTFSS	COM_ST,CO_TRA		; Receive?
; One byte has already been received and copied to the buffer
	RETFIE	FAST			; Done!
; Transmit one byte
	BTFSC	COM_ST,CO_SER		; Serial transmission?
	BRA	RTS_CK			; Yes!
; Transmit one byte over the parallel port
	MOVF	POSTINC0,W		; Byte to transmit
	MOVWF	PORTA			; Bits <5:0> to RA<5:0>
	RLNCF	WREG			; Rotate bits <7:6> to
	RLNCF	WREG			; position <1:0>
	BSF	WREG,C_DRDY		; Signal: Data ready!
	MOVWF	PORTC			; Bits <7:6> to RC<1:0>
	RETFIE	FAST
; Return to check RTS again after a delay of 50ms
RES_TX	BCF	TIM_ST,TM_RTS		; Signal: RTS timer not running!
	BSF	PIE1,TXIE		; Enable the TX interrupt
; Check if the server is ready to receive over the USART
RTS_CK	BTFSS	PORTE,E_NRTS		; Is RTS asserted?
	BRA	UART_T			; Yes!
	BSF	TIM_ST,TM_RTS		; Signal: RTS timer running!
	BCF	PIE1,TXIE		; Disable the TX interrupt
	BRA	WAI_50			; Wait for 50ms
; Transmit one byte over the USART
UART_T	MOVFF	POSTINC0,TXREG		; Load the byte into the USART
	RETFIE	FAST			; Transmit Register
; PSP error
PSP_ER	BCF	TIM_ST,TM_PSP		; Signal: Time-out
	CLRF	COM_ST
; Prepare for reception
PREP_R	BTFSS	COM_ST,CO_SER		; Serial receive?
	BRA	CLR_PP			; No!
	BTFSC	USB_ST,SB_CFG		; Receive over USB?
	BRA	USB_RX			; Yes!
; Prepare for USART reception
	BCF	RCSTA,CREN		; Clear the USART error flags
	MOVF	RCREG,W			; Clear the USART FIFO
	MOVF	RCREG,W
	MOVF	RCREG,W
	BSF	RCSTA,CREN		; Re-enable USART receive
	BCF	PIE1,TXIE		; Disable the TX interrupt
	BSF	PIE1,RCIE		; Enable the RX interrupt
	RETFIE	FAST
; Prepare for Parallel Port reception
CLR_PP	RCALL	PP_RST			; Reset the SPP
	RETFIE	FAST
; Calculate the combined length of header and data
TOT_LN	MOVLW	3			; Length of header = 3
	ADDWF	BLN_LO,W		; Add to the length of the
	MOVWF	CNT_LO			; data block to get the
	MOVLW	0			; total length and copy it
	ADDWFC	BLN_HI,W		; to the byte counter
	MOVWF	CNT_HI
	BCF	COM_ST,CO_HED		; Signal: Header Block done!
	RETURN
; Client Command 
CL_CMD	CPFSEQ	CMD_CD			; 'Set Write Address'?
	BRA	CLC_EX			; No!
; Set Write Block Address
	CLRF	BLN_LO			; Set reponse length to 0
	LFSR	FSR1,CMD_24		; Command: WRITE_BLOCK
; Set up a SD card command string in RAM
CM_SET	RCALL	TAB_NW			; Copy the command string to RAM
	DECF	SD_CRC,W		; A CRC-7 value of 0, indicates
	BNZ	CDT_EN			; that a data address is required
; Copy the data address to the command string
; The data address must be given in bytes, for standard capacity SD cards
; and in blocks of 512 bytes for high capacity SD cards
	LFSR	FSR1,SD_AD3		; Point to the highest byte and
	BTFSC	SDC_ST,SD_HCS		; copy it only for a high capacity
	MOVFF	DATA_B+3,POSTINC1	; SD card
	MOVFF	DATA_B+2,POSTINC1	; In this way, the block address
	MOVFF	DATA_B+1,POSTINC1	; is implicitly multiplied by 256
	MOVFF	DATA_B+0,INDF1		; for standard capacity SD cards
	BTFSC	SDC_ST,SD_HCS		; For standard capacity cards,
	RETURN				; multiply the address again by 2
	RLCF	SD_AD0,W		; This clears the Carry Flag
	RLCF	SD_AD1			; Multiply by 2
	RLCF	SD_AD2
	RLCF	SD_AD3
CDT_EN	RETURN
CLC_EX	RETURN
; Start the SD card time-out counter
SDC_TM	BSF	TIM_ST,TM_SDC		; Signal: SD card timer is running
; Time-out 1s
TIM_1S	timing	.1000			; Start timer 0 for 1s
; Start Timer 0
STA_TM	BCF	T0CON,TMR0ON		; Stop timer 0
	BCF	INTCON,TMR0IF		; Clear interrupt flag
	MOVWF	TMR0L			; Load timing constant
	BSF	T0CON,TMR0ON		; Start timer 0
	RETURN
; Enable the USART
ASY_EN	CLRF	SSPCON1			; Disable SPI
	MOVLW	1<<TX|1<<RX		; Define RX and TX as inputs
	MOVWF	TRISC
	BCF	LATE,E_NCTS		; Signal: Clear to send!
	MOVLW	1<<TXEN|1<<BRGH
	MOVWF	TXSTA			; Enable USART transmit
	MOVLW	1<<SPEN|1<<CREN
	MOVWF	RCSTA			; Enable USART receive
	RETURN
;******************************************************************************
; USB

; USB initialization data
BD_EP0	dw	BDT0_O,BD_CTL-BD_EP0-4+(0x100<<TB_SEL)
; Buffer Descriptor Initial Data for EP0
	db	1<<UOWN|1<<BSTALL,1<<PS_CTL
	dw	BUF0_O
	db	1<<UOWN|1<<BSTALL,1<<PS_CTL
	dw	BUF0_I
BD_CTL	dw	BDT0_O+(1<<ENDP0)*EP_CTL,BD_DAT-BD_CTL-4+(0x100<<TB_SEL)
; Buffer Descriptor Initial Data for the CC Control EP
	dw	0,0
	db	1<<UOWN|1<<DTSEN,0x0a
	dw	NT_TYP
BD_DAT	dw	BDT0_O+(1<<ENDP0)*EP_DAT,NT_INI-BD_DAT-4+(0x100<<TB_SEL)
; Buffer Descriptor Initial Data for the CC Data EP
	db	0,0
	dw	CMD_CD
	db	0,0
	dw	CMD_CD
NT_INI	dw	NT_TYP,LC_INI-NT_INI-4+(0x100<<TB_SEL)
; Notification buffer Initial Data
	db	1<<RT_DIR|1<<RT_CLA|1<<RT_ITF,N_SSTA
	dw	0,IF_MAX-1
	dw	2,1<<SS_DSR|1<<SS_DCD
LC_INI	dw	LN_COD,CMD_CD-LN_COD+(0x100<<TB_SEL)
; CC Line Coding Initial Data
	db	low BAUD_R, high BAUD_R 
	db	upper BAUD_R, 0
	db	0,0,8	

; USB Descriptors Index Table
DE_PNT	dw	DE_DEV,DE_CFG,DE_ST0,DE_ST1,DE_ST2,DE_ST3

; USB Descriptors
DE_DEV	dw	BUF0_I,DE_CFG-DE_DEV-4	; Table header
; Device Descriptor
	db	0x12,DT_DEV		; bLength		18
					; bDescriptorType	Device
    	dw	0x0200			; bcdUSB		USB 2.0
	db	DC_COM,DC_NUS		; bDeviceClass		Communication
					; bDeviceSubClass	None
	db	DC_NUS,1<<PS_CTL	; bDeviceProtocol	None
					; bMaxPacketSize0	Packet Size
	dw    	0x04D8			; idVendor		Microchip
	dw	0x000A			; idProduct		PIC18F4550
	dw	0x0111			; bcdDevice		ZX Int1bis
	db    	1,2			; iManufacturer		String 1
					; iProduct		String 2
	db	3,CF_MAX		; iSerialNum		String 3
					; bNumConfigurations	1

DE_CFG	dw	BUF0_I,DE_ST0-DE_CFG-5	; Table header	
; Configuration (1) Descriptor
	db	0x09,DT_CFG		; bLength	     	9
					; bDescriptorType	Configuration
	dw	0x0043			; wTotalLength       	67
	db	IF_MAX+1,CF_MAX		; bNumInterfaces	2
					; bConfigurationValue	1
	db	0,CA_SPW|CA_RWU		; iConfiguration	None
					; bmAttributes       	Self-Pow,Rem Wake-up
	db	1,9			; bMaxPower          	2mA
; CDC Control Interface (0) Descriptor
					; bLength		9
	db	DT_ITF,IF_MAX-1		; bDescriptorType	Interface
					; bInterfaceNumber	0
	db	0,1			; bAlternateSetting	None
					; bNumEndpoints		1
	db	IC_COM,IS_ACM		; bInterfaceClass	Communication
					; bInterfaceSubClass	Abstract
	db	PC_ATC,0		; bInterfaceProtocol	AT Command
					; iInterface		None
; CDC Header Functional Descriptor
	db	5,DT_CSI		; bFunctionLength	5
					; bDescriptorType	CS Interface
	db	FD_HED,0x10		; bDescriptorSubtype	Header
					; bcdCDC		CDC 1.10
	db	0x01,4			;
; CDC Abstract Control Model Functional Descriptor
					; bFunctionLength	4
	db	DT_CSI,FD_ACM		; bDescriptorType	CS Interface
					; bDescriptorSubtype	Abstract Control
	db	1<<CA_AC1,5		; bmCapabilities	D1
; CDC Union Functional Descriptor
					; bFunctionLength	5
	db	DT_CSI,FD_UNI		; bDescriptorType	CS Interface
					; bDescriptorSubtype	Union
	db	IF_MAX-1,IF_MAX		; bmMasterInterface	0
					; bSlaveInterface0	1
; CDC Call Management Functional Descriptor
	db	5,DT_CSI		; bFunctionLength	5
					; bDescriptorType	CS Interface
	db	FD_CMA,0		; bDescriptorSubtype	Call Management
					; bmCapabilities	None
	db	IF_MAX,7		; bDataInterface	1
; CDC Notification Endpoint Descriptor
					; bLength		7
	db	DT_ENP,EP_CTL|1<<EA_DIR	; bDescriptorType	Endpoint
					; bEndpointAddress	Address
	db	EA_INT,LN_COD-NT_TYP	; bmAttributes		Interrupt
					; wMaxPacketSize	Packet Size
	db	0,2			; bInterval		2ms 
; CDC Data Interface (1) Descriptor
	db	9,DT_ITF		; bLength		9
					; bDescriptorType	Interface
	db	IF_MAX,0		; bInterfaceNumber	1
					; bAlternateSetting	None
	db	2,IC_DAT		; bNumEndpoints		2
					; bInterfaceClass	Data
	db	DC_NUS,PC_NOP		; bInterfaceSubClass	None
					; bInterfaceProtocol	None
	db	0,7			; iInterface		None
; CDC Data OUT Endpoint Descriptor
					; bLength		7
	db	DT_ENP,EP_DAT		; bDescriptorType	Endpoint
					; bEndpointAddress	Address
	db	EA_BLK,1<<PS_DAT	; bmAttributes		Bulk
					; wMaxPacketSize	Packet Size
	db	0,0			; bInterval		None
; CDC Data IN Endpoint Descriptor
	db	7,DT_ENP		; bLength		7
					; bDescriptorType	Endpoint
	db	EP_DAT|1<<EA_DIR,EA_BLK	; bEndpointAddress	Address
					; bmAttributes		Bulk
	db	1<<PS_DAT,0		; wMaxPacketSize	Packet Size
	db	0			; bInterval		None

DE_ST0	dw	BUF0_I,DE_ST1-DE_ST0-4	; Table header
; Language ID String Descriptor
	db	DE_ST1-DE_ST0-4,DT_STR	; bLength		Length
					; bDescriptorType	String
	dw	0x0409			; wLangID		English (US)

DE_ST1	dw	BUF0_I,DE_ST2-DE_ST1-4	; Table header
; Manufacturer String Descriptor 
	db	DE_ST2-DE_ST1-4,DT_STR	; bLength		Length
					; bDescriptorType	String
	dw	'M','i','c','r','o','c','h','i','p',' '
	dw	'T','e','c','h','n','o','l','o','g','y'
	dw	' ','I','n','c','.'

DE_ST2	dw	BUF0_I,DE_ST3-DE_ST2-4	; Table header
; Product String Descriptor 
	db	DE_ST3-DE_ST2-4,DT_STR	; bLength		Length
					; bDescriptorType	String
	dw	'P','I','C','1','8','F','4','5','5','0'

DE_ST3	dw	BUF0_I,DE_END-DE_ST3-4	; Table header
; Serial Number String Descriptor 
	db	DE_END-DE_ST3-4,DT_STR	; bLength		Length
					; bDescriptorType	String
	dw	'Z','X',' ','I','n','t','1','b','i','s'
DE_END

; Copy  data from a table in the program memory.
; The table header has the following structure:
; (Word)Destination Address 
; (Byte)Data Length
; (Byte)Selector: 0 = Descriptor, >0 = Other

; Copy from the table at address: (TBLPTR)
TAB_CO	RCALL	TB_DST			; Get the destination address
	BRA	TB_LEN
; Continue copying the same table from address (TBLPTR) if called
; from a Control IN Transfer interrupt, or otherwise copy from a new 
; descriptor table at address: (FSR1)
TAB_RD	BTFSC	EP_OFS,DIR		; Control IN Transfer?
	BRA	TB_TST			; Yes!
; Copy from the table at address: (FSR1)
TAB_NW	RCALL	TB_ADR			; New Table entry
TB_LEN	TBLRD*+				; Load the size of the
	MOVFF	TABLAT,TB_CNT		; table in the counter
; Descriptors are copied only (WREG) bytes at a time
	TBLRD*+				; Read the Table Selector
	TSTFSZ	WREG			
	BTFSC	TABLAT,TB_SEL		; Copy the entire table?
	MOVF	TB_CNT,W		; Yes!
TB_TST	TSTFSZ	TB_CNT			; End of table reached?
	BRA	TB_NXT			; Not yet!
	RETURN				; Yes!
; (WREG) returns the number of unavailable bytes
TB_NXT	TBLRD*+				; Copy one byte and decrement
	MOVFF	TABLAT,POSTINC1		; the 'available bytes' counter
	DECF	TB_CNT			; and the 'copied bytes'
	DECFSZ	WREG			; counter
	BRA	TB_TST			; All done?
	RETURN				; Yes!

; Index (WREG) into the table at FSR1
TB_IND	RLNCF	WREG
	ADDWF	FSR1L
	MOVLW	0
	ADDWFC	FSR1H
; New Table
TB_ADR	MOVFF	FSR1L,TBLPTRL		; Initialize the pointer for
	MOVFF	FSR1H,TBLPTRH		; a new table
TB_DST	TBLRD*+	
	MOVFF	TABLAT,FSR1L		; Load the destination
	TBLRD*+				; address from the table
	MOVFF	TABLAT,FSR1H
	RETURN
; Attach / Detach USB
USB_SE	RCALL	EP_DIS			; Disable all Endpoints
	MOVLW	1<<SB_SPW		; Signal: Self powered!
	MOVWF	USB_ST
	BSF	LATE,E_NSRV		; Signal: Server not connected!
	BSF	COM_ST,CO_CHG
	BTFSC	CHG_ST,B_USEN		; USB cable connected to host?
	BRA	USB_AT			; Yes!
; Detach USB
	BCF	UCON,SUSPND		; End Suspend Mode (17.2.1)
	BCF	UCON,USBEN		; Disable USB
	BCF	PIE2,USBIE		; Disable the USB interrupt
	BTFSS	PORTE,E_NRTS		; If the RTS line is asserted
	BCF	LATE,E_NSRV		; signal: Server connected!
	BRA	USB_CE
; Attach USB
USB_AT	BSF	UCON,USBEN		; Enable USB
	MOVLW	1<<IDLEIE|1<<TRNIE|1<<URSTIE
	MOVWF	UIE			; Unmask needed USB interrupts
	BSF	PIE2,USBIE		; Enable USB interrupts
WT_SE0	BTFSC	UCON,SE0		; Single Ended Zero State?
	BRA	WT_SE0			; Yes! Wait until it ends.
USB_CE	CLRF	UIR			; Clear all USB event flags
	BTFSC	UIR,ACTVIF		; Wait until the Bus Activity
	BRA	USB_CE			; event flag is stable
USB_CF	BCF	PIR2,USBIF		; Clear the USB interrupt flag
	RETURN
; Reset the Parallel Port
PP_RST	COMF	LATE,W
	ANDLW	1<<E_NSRV|1<<E_NSDC
	IORLW	SA_SIG&(1<<6)-1		; Load SA_SIG in
	MOVWF	PORTA			; RC<1:0>:RA<5:0>
      	MOVLW	SA_SIG>>6		; and signal: Data not ready!
	MOVWF	PORTC
	RETURN
; USB Interrupt Service Routine
USB_IS	SETF	EP_OFS			; Clear the last EP offset
	BTFSS	UCON,SUSPND		; USB in Suspend Mode?
	BRA	TS_IDL			; No!
; USB in Suspend Mode
; Only Bus Activity can generate an interrupt (DS39632e 17.5.1)
; Bus Activity interrupt
	BCF	UCON,SUSPND		; End USB Suspend Mode
	BCF	UIE,ACTVIE		; Disable Bus Activity Interrupt
	BRA	WT_SE0			; Return
TS_IDL	BTFSS	UIR,IDLEIF		; Idle Detect Interrupt?
	BRA	TS_RST			; No!
; Idle Detect interrupt
	BSF	UCON,SUSPND		; Enter the USB Suspend Mode
	BSF	UIE,ACTVIE		; Enable Bus Activity Interrupt
	BRA	USB_CE			; Return
TS_RST	BTFSS	UIR,URSTIF		; Reset interrupt?
	BRA	TS_TRN			; No!
; Reset interrupt
; Empty the USTAT FIFO. If the stack is not empty it takes 5 instruction
; cycles after clearing UIR for the TRNIF flag to be re-asserted
CL_UIR	CLRF	UIR			; Clear all USB event flags and
	NOP				; wait for 5 instruction cycles
	NOP
	NOP
	NOP
	NOP
	BTFSC	UIR,TRNIF		; Is the USTAT FIFO empty?
	BRA	CL_UIR			; Not yet!
	RCALL	EP_DIS			; Disable all endpoints
	LFSR	FSR1,BD_EP0		; Initialize the EP0 buffer
	RCALL	TAB_NW			; descriptor
	MOVLW	1<<EPHSHK|1<<EPOUTEN|1<<EPINEN
	MOVWF	UEP0			; Enable EP0 for control
	BCF	UCON,PKTDIS		; Enable packet processing
	MOVLW	1<<SB_DEF|1<<SB_SPW	; Signal: Default State!
	MOVWF	USB_ST
TS_TRN	BTFSS	UIR,TRNIF		; Transfer Complete interrupt?
	BRA	USB_CF			; No!
; Transfer Complete interrupt
; Identify the active endpoint
	MOVF	USTAT,W
	ANDLW	1<<ENDP3|1<<ENDP2|1<<ENDP1|1<<ENDP0|1<<DIR
	MOVWF	EP_OFS			; Offset of endpoint in BDT
	BCF	UIR,TRNIF		; Clear transfer complete flag
	LFSR	FSR1,BDT0_O		; Point FSR1 to BDT
	BTG	PLUSW1,DTS		; Toggle the data sync bit.
	ANDLW	1<<ENDP3|1<<ENDP2|1<<ENDP1|1<<ENDP0
	BNZ	USB_CF			; Not a Control Transfer!
; Control Transfer Complete interrupt
	BTFSS	EP_OFS,DIR		; Control IN Transfer?
	BRA	CK_STU			; No!
; Control IN Transfer Complete interrupt
	BTFSC	RQ_TYP,RT_DIR		; Control Read? 
	BRA	CDT_IN			; Yes!
; Status Stage of a Control Write or Non-Data Control Transfer
; Only 'Set Address' requests need to be handled
	MOVLW	R_SADR
	CPFSEQ	RQ_NUM			; Set Address Request?
	BRA	TS_TRN			; No!
; Status Stage of a 'Set Address' Setup Request
	MOVFF	RQ_VAL,UADDR		; Copy the new address
	BRA	TS_TRN
; Control Read Transfer Data Stage
; All implemented requests except for 'Get Descriptor' need to respond
; with only one (short) IN packet. After completion of the corresponding 
; transfer, returning to this point results in the EP0 IN endpoint being
; stalled, which prevents the host from requesting any further data
; The 'Get Descriptor' request uses this entry point repeatedly, after
; each completed transfer of a packet, until all required data has been
; sent, following which it also stalls the endpoint.
CDT_IN	LFSR	FSR1,BUF0_I		; Initialize the buffer address
; Check if all requested data has been transmitted
	MOVF	EP0I_L,W		; Size of the transmitted packet
	SUBWF	RQ_LNL			; Subtract the size of the sent 
	MOVLW	0			; packet from the requested
	SUBWFB	RQ_LNH			; remaining length
	BNC	RQ_ERR			; This should never happen!
	BNZ	CDI_SZ
	CPFSEQ	RQ_LNL			; Any more data requested?
	BRA	CDI_SZ			; Yes!
; All data has been sent
; If the last packet was full-length, send a zero-length packet
; or else stall the endpoint
	BTFSS	EP0I_L,PS_CTL		; Last packet was full-length?
	BRA	RQ_ERR			; No! Stall EP0 IN
; Establish the size of the packet to be sent
CDI_SZ	MOVLW	-(1<<PS_CTL)		; Clear Z if there is enough data
	ANDWF	RQ_LNL,W		; for a full-length packet
	IORWF	RQ_LNH,W
	MOVLW	1<<PS_CTL		; Full-length
	BNZ	CDI_FL
	MOVF	RQ_LNL,W		; Short packet
CDI_FL	MOVWF	EP0I_L			; Set the packet size
	BZ	CDI_ST			; Zero-length packet
; Copy the next (WREG) descriptor bytes from the table to the EP0 IN buffer
	RCALL	TAB_RD			; Copy (WREG) bytes to the buffer
	SUBWF	EP0I_L			; Reduce the packet size by the
	BNZ	CDI_ST			; number of bytes not available
; If no more data is available set the remaining response length to zero
CDI_ZR	CLRF	RQ_LNL			; More data requested than
	CLRF	RQ_LNH			; available!
; Pass control of the endpoint to SIE
CDI_ST	MOVLW	1<<UOWN|1<<DTSEN
	BTFSC	BDT0_I,DTS		; Copy the DTS bit
	BSF	WREG,DTS
CIN_EX	MOVWF	BDT0_I			; Pass control of EP0 IN to SIE
; Enable Packet Processing because this exit point is also used by
; every SETUP transfer
	BCF	UCON,PKTDIS		; Enable Packet Processing
	BRA	TS_TRN			; Done!
; Differentiate between an OUT and a SETUP packet
CK_STU	MOVF	BDT0_O,W		; Get the Token PID
	ANDLW	1<<PID3|1<<PID2|1<<PID1|1<<PID0
	XORLW	PID_SU<<PID0		; SETUP Token?
	BZ	SET_UP			; Yes!
; Control OUT Transfer Complete interrupt
; Only 'Set Line Coding' Data OUT packets need to be processed
	MOVLW	C_SLCO
	CPFSEQ	RQ_NUM			; Set Line Coding?
	BRA	CDO_ST			; No!
; Copy the received Line Coding Data
	copybl	BUF0_O,LN_COD,CMD_CD-LN_COD
; Prepare EP0 OUT for the next transfer
CDO_ST	RCALL	EP0O_A			; Re-arm EP0 OUT
	BRA	TS_TRN
; SETUP Transfer Complete interrupt
; Copy the Request Header to the access RAM
SET_UP	copybl	BUF0_O,RQ_TYP,SD_SZ0-RQ_TYP
	MOVLW	1<<DTS			; Reclaim control and
	MOVWF	BDT0_O			; set DATA1 for both EP0
	MOVWF	BDT0_I			; OUT and IN endpoints
; Setting EP0 OUT to DATA1 is necessary for both Control Read and
; Control Write transfers. Non-Data Control Transfers are not 
; affected while the next Setup Token will be accepted irrespective
; of the DTS bit state ( DS39632e Register 17-5)
	RCALL	EP0O_A			; Re-arm EP0 OUT and set the
	CLRF	EP0I_L			; EP0 IN packet size to 0
	CLRF	BUF0_I			; Clear the first response byte
; Differentiate between Setup Request types
	BTFSC	RQ_TYP,RT_CLA		; Class request?
	BRA	CLA_RQ			; Yes!
	BTFSS	RQ_TYP,RT_VEN		; Standard request?
	BRA	STD_RQ			; Yes!
; Vendor Setup Requests are not supported
EP_HLT	MOVLW	1<<UOWN|1<<BSTALL	; Stall EP0
	MOVWF	BDT0_O			; Stall EP0 OUT
	BRA	CIN_EX			; Stall EP0 IN

; Check the Request Recipient: Device, Interface or Endpoint
CK_RQR	BTFSC	RQ_TYP,RT_ITF		; Interface?
	BRA	CKR_OR			; Yes!
	BTFSS	RQ_TYP,RT_ENP		; Device?
; Device Recipient: nothing to check	  Yes!
	RETURN
; Endpoint Recipient
	RLNCF	RQ_IDX,W		; Point to the status byte
	RLNCF	WREG			; of the endpoint descriptor
	RLNCF	WREG			; 
	MOVWF	FSR1L
	MOVLW	(1<<ENDP0)*(EP_MAX+1)	; Lowest invalid offset
	BTFSS	USB_ST,SB_CFG		; If not in Configured State
	MOVLW	1<<ENDP0		; only EP0 is valid
	CPFSLT	FSR1L			; Valid endpoint index?
	BRA	CKR_ER			; No! Signal: Request Error!
	MOVLW	(1<<ENDP0)*EP_CTL	; The Control EP OUT is not
	CPFSEQ	FSR1L			; a valid endpoint
	RETURN	
	BRA	CKR_ER			; Signal: Request Error!
; The recipient must be an interface, otherwise signal: Request Error!
CKR_IF	BTFSS	RQ_TYP,RT_ITF		; Interface?
	BRA	CKR_ER			; No! Signal: Request Error!
; Setup Request Recipient of type 'Other' is not supported
CKR_OR	BTFSC	RQ_TYP,RT_ENP		; Other?
	BRA	CKR_ER			; Signal: Request Error!
; Interface Recipient
	BTFSS	USB_ST,SB_CFG		; Configured state?
	BRA	CKR_ER			; No! Signal: Request Error!
	MOVLW	IF_MAX			; Maximum interface number
	CPFSGT	RQ_IDX			; Valid interface number?
	RETURN				; Yes!
CKR_ER	POP				; Discard the return address 
; Request Error
RQ_ERR	MOVLW	1<<UOWN|1<<BSTALL	; Request Error action:
	BRA	CIN_EX			; Stall EP0 IN

; Communication Class Requests
CLA_RQ	RCALL	CKR_IF			; Recipient must be an interface
	TSTFSZ	RQ_IDX			; Recipient must be the
	BRA	RQ_ERR			; CC Control Interface (0)
	MOVF	RQ_NUM,W		; Request Code to WREG
	XORLW	C_SECO			; Send Encapsulated Command Request?
; Send Encapsulated Command
; Accept the Request but discard the received data
	BZ	CDI_ST			; Yes! Prepare Status
	XORLW	C_SECO^C_GERP		; Get Encapsulated Response Request?
; Get Encapsulated Response
; Clear the requested response length and send a zero-length packet
	BZ	CDI_ZR			; Yes!
	XORLW	C_GERP^C_SLCO		; Set Line Coding Request?
; Set Line Coding
	BZ	CDI_ST			; Yes! Prepare Status
	XORLW	C_SLCO^C_GLCO		; Get Line Coding Request?
	BNZ	TS_SLS			; No!
; Get Line Coding
; Copy the Line Coding data to the EP0 IN buffer
	copybl	LN_COD,BUF0_I,CMD_CD-LN_COD
	BRA	IN_SHO			; Send the Line Coding data
TS_SLS	XORLW	C_GLCO^C_SCLS		; Set Control Line State Request?
; No other requests are supported
NO_OTH	BNZ	EP_HLT			; Stall EP0
; Set Control Line State
	MOVFF	RQ_VAL,CLN_ST		; Copy the Line State byte
	BRA	IN_SHO			; Send Status
; Standard SETUP Requests
STD_RQ	MOVF	RQ_NUM,W		; Request Code to WREG
	XORLW	R_SADR			; Set Address Request?
; Set Address
	BZ	IN_SHO			; Yes! Send Status
	XORLW	R_SADR^R_GDES		; Get Descriptor Request?
	BNZ	TS_SCF			; No!
; Get Descriptor
	MOVLW	DT_ITF			; Only supported types: Device,
	CPFSLT	RQ_DTY			; Configuration and String
	BRA	RQ_ERR			; Signal: Request Error!
	MOVLW	DT_CFG			; Configuration Descriptor	
	CPFSEQ	RQ_DTY			; requested?
	BRA	GT_DES			; No!
	MOVLW	CF_MAX			; If yes, check the index
	CPFSLT	RQ_VAL			; Valid Configuration index?
	BRA	RQ_ERR			; No! Signal: Request Error!
GT_DES	LFSR	FSR1,DE_PNT-2		; Index into the table DE_PNT
	MOVF	RQ_DTY,W		; containing the descriptor
	ADDWF	RQ_VAL,W		; addresses
	RCALL	TB_IND			; Get the descriptor's address
	BRA	CDI_SZ			; and send the first packet
TS_SCF	XORLW	R_GDES^R_SCFG		; Set Configuration Request?
	BNZ	TS_GCF			; No!
; Set Configuration
	MOVLW	CF_MAX+1		; Test the Configuration number
	CPFSLT	RQ_VAL			; If it is invalid,
	BRA	RQ_ERR			; signal: Request Error!
	CLRF	WREG
	MOVFF	WREG,UEP0+EP_CTL	; Disable the CC Control EP
	MOVFF	WREG,UEP0+EP_DAT	; Disable the CC Data EP
	TSTFSZ	RQ_VAL			; Configuration number >0 ?
	BRA	SCF_CO			; Yes!
; Go back to Addressed Sate
	BCF	USB_ST,SB_CFG		; Signal: Not Configured State!
	BRA	IN_SHO			; Send Status
; Advance to Configured State
SCF_CO	LFSR	FSR1,BD_CTL		; Initialize the CC Control
	RCALL	TAB_NW			; EP Buffer Descriptor
	MOVLW	1<<EPHSHK|1<<EPINEN|1<<EPCONDIS
	MOVFF	WREG,UEP0+EP_CTL	; Enable the CC Control EP
	RCALL	TAB_CO			; Initialize the CC Data EP
	MOVLW	1<<EPHSHK|1<<EPOUTEN|1<<EPINEN|1<<EPCONDIS
	MOVFF	WREG,UEP0+EP_DAT	; Enable the CC Data EP
	RCALL	TAB_CO			; Initialize the Notification
	RCALL	TAB_CO			; Header and Line Coding Data
	BSF	USB_ST,SB_CFG		; Signal: Configured State!
	BRA	IN_SHO			; Send Status
TS_GCF	XORLW	R_SCFG^R_GCFG		; Get Configuration Request?
	BNZ	TS_GST			; No!
; Get Configuration
	BTFSC	USB_ST,SB_CFG		; Configured State?
	MOVLW	CF_MAX			; Yes! Return the configuration
IN_VAL	MOVWF	BUF0_I			; number or else return 0
; Send one short packet
IN_SHO	MOVFF	RQ_LNL,EP0I_L		; Requested packet size
	BRA	CDI_ST			; Send packet
TS_GST	XORLW	R_GCFG^R_GSTA		; Get Status Request?
	BNZ	TS_CFT			; No!
; Get Status	
	RCALL	CK_RQR			; Check Recipient
	CLRF	BUF0_I+1		; Clear the second response byte
	BTFSC	RQ_TYP,RT_ITF		; Interface?
	BRA	IN_SHO			; Yes! Nothing to report
	BTFSC	RQ_TYP,RT_ENP		; Endpoint?
	BRA	GST_EP			; Yes!
; Get Device Status
	MOVF	USB_ST,W		; Copy the 'Self Powered'
	ANDLW	1<<SB_RWU|1<<SB_SPW	; and 'Remote Wake-up' bits
	BRA	IN_VAL			; to the EP0 IN buffer
; Get Endpoint status
GST_EP	BTFSC	INDF1,BSTALL		; If the endpoint is stalled,
	BSF	BUF0_I,GS_HLT		; set the HLT bit
	BRA	IN_SHO			; Send the response
TS_CFT	XORLW	R_GSTA^R_CFTR		; Clear Feature Request?
	BNZ	TS_SFT			; No!
; Clear / Set Feature
CHG_FT	RCALL	CK_RQR			; Check Recipient
	MOVF	RQ_NUM,W		; Set Z for 'Clear' and
	XORLW	R_CFTR			; clear Z for 'Set'
	BTFSC	RQ_TYP,RT_ITF		; Interface?
	BRA	RQ_ERR			; Yes! Signal: Request Error!
	BTFSC	RQ_TYP,RT_ENP		; Endpoint?
	BRA	CSF_EP			; Yes!
; Clear / Set Remote Wake-up
	MOVLW	DF_RWU			; Device Remote Wake-up Feature?
	CPFSEQ	RQ_VAL
	BRA	RQ_ERR			; No! Signal: Request Error!
	BCF	USB_ST,SB_RWU		; Clear Remote Wake-up bit
	BZ	IN_SHO			; Set?
	BSF	USB_ST,SB_RWU		; Set Remote Wake-up bit
	BRA	IN_SHO			; Send status
; Clear / Set Endpoint HALT
CSF_EP	MOVLW	EF_HLT			; Endpoint Halt Feature?
	CPFSEQ	RQ_VAL
	BRA	RQ_ERR			; No! Signal: Request Error!
CSF_HT	RCALL	CHG_HT			; Change HALT and clear DTS bit
	BRA	IN_SHO			; Send status
TS_SFT	XORLW	R_CFTR^R_SFTR		; Set Feature Request?
; Set Feature
	BZ	CHG_FT			; Yes!
	XORLW	R_SFTR^R_GITF		; Get Interface Request?
	BNZ	TS_SIF			; No!
; Get Interface
	RCALL	CKR_IF			; Recipient must be an interface
	BRA	IN_SHO			; Return 0
TS_SIF	XORLW	R_GITF^R_SITF		; Set Interface Request?
	BNZ	NO_OTH			; No other requests are supported
; Set Interface
	RCALL	CKR_IF			; Recipient must be an interface
	TSTFSZ	RQ_VAL			; Accept only the alternate
	BRA	RQ_ERR			; interface 0
	MOVLW	(1<<ENDP0)*EP_CTL	; CC Control Interface (0)
	TSTFSZ	RQ_IDX			; CC Data Interface (1)?
	MOVLW	(1<<ENDP0)*EP_DAT	; Yes!
	MOVWF	FSR1L			; Point to corresponding OUT EP
	CLRF	WREG			; Clear! ( Set Z )
	RCALL	CHG_HT			; Clear the OUT EP HALT and DTS
	BSF	FSR1L,DIR		; Switch to the IN EP
	BRA	CSF_HT			; Clear the IN EP HALT and DTS
; Disable all used USB Endpoints and reset the Device Address
EP_DIS	CLRF	UEP0			; Disable EP0
	CLRF	UEP0+EP_CTL		; Disable CC Control EP
	CLRF	UEP0+EP_DAT		; Disable CC Data EP
	CLRF	UADDR			; Set default address
	RETURN
; Arm EP0 OUT
EP0O_A	MOVLW	1<<PS_CTL		; Set the packet length to the
	MOVWF	EP0O_L			; maximum declared size
	MOVLW	1<<UOWN|1<<DTSEN
	BTFSC	BDT0_O,DTS		; Copy the DTS bit	
	BSF	WREG,DTS
	MOVWF	BDT0_O			; Pass control of EP0 OUT to SIE
	RETURN
; Clear ( Z==1 ) / Set ( Z==0 ) the endpoint HALT and clear the DTS bit
CHG_HT	MOVLW	1<<DTSEN		; Clear the HALT and DTS bits
	BZ	CHT_UO			; Set?	
	BSF	WREG,BSTALL		; Yes! Set the HALT bit
CHT_UO	BTFSC	INDF1,UOWN		; Copy the UOWN bit
	BSF	WREG,UOWN
	MOVWF	INDF1
	RETURN

;------------------------------------------------------------------------------
; SD Card

; SD card command strings
; GO_IDLE_STATE
CMD_00	dw	CM_CNT,0x0a+(0x100<<TB_SEL)
	db	6,1					; Response R1
	dw	0
	db	d'00'|1<<CB_TRN,0,0,0,0,0x4a<<1|1				
; SEND_IF_COND
CMD_08	dw	CM_CNT,0x0a+(0x100<<TB_SEL)
	db	6,5					; Response R3
	dw	0
	db	d'08'|1<<CB_TRN,0,0,C8_VHS,C8_PAT,0x43<<1|1
; SEND_CSD
CMD_09	dw	CM_CNT,0x0a+(0x100<<TB_SEL)
	db	6,1					; Response R1
	dw	0x0012					; Data size: 16+2
	db	d'09'|1<<CB_TRN,0,0,0,0,0x57<<1|1
; SEND_CID
CMD_10	dw	CM_CNT,0x0a+(0x100<<TB_SEL)
	db	6,1					; Response R1
	dw	0x0012					; Data size: 16+2
	db	d'10'|1<<CB_TRN,0,0,0,0,0x0d<<1|1
; SET_BLOCKLEN
CMD_16	dw	CM_CNT,0x0a+(0x100<<TB_SEL)
	db	6,1					; Response R1
	dw	0
	db	d'16'|1<<CB_TRN,0,0,2,0,0x0a<<1|1
; READ_SINGLE_BLOCK
CMD_17	dw	CM_CNT,0x0a+(0x100<<TB_SEL)
	db	6,1					; Response R1
	dw	0x0202					; Data size: 512+2
	db	d'17'|1<<CB_TRN,0,0,0,0,0x00<<1|1	; Data address
; WRITE_BLOCK
CMD_24	dw	CM_CNT,0x0a+(0x100<<TB_SEL)
	db	6,1					; Response R1
	dw	0x0202					; Data size: 512+2
	db	d'24'|1<<CB_TRN,0,0,0,0,0x00<<1|1	; Data address
; ERASE_WR_BLK_START_ADDR
CMD_32	dw	CM_CNT,0x0a+(0x100<<TB_SEL)
	db	6,1					; Response R1
	dw	0
	db	d'32'|1<<CB_TRN,0,0,0,0,0x00<<1|1	; Data address
; ERASE_WR_BLK_END_ADDR
CMD_33	dw	CM_CNT,0x0a+(0x100<<TB_SEL)
	db	6,1					; Response R1
	dw	0
	db	d'33'|1<<CB_TRN,0,0,0,0,0x00<<1|1	; Data address
; ERASE
CMD_38	dw	CM_CNT,0x0a+(0x100<<TB_SEL)
	db	6,1
	dw	0xffff					; Response R1b
	db	d'38'|1<<CB_TRN,0,0,0,0,0x52<<1|1
; SD_SEND_OP_COND
ACM_41	dw	CM_CNT,0x0a+(0x100<<TB_SEL)
	db	6,1					; Response R1
	dw	0
	db	d'41'|1<<CB_TRN,1<<CB_HCS,0,0,0,0x3b<<1|1
; APP_CMD
CMD_55	dw	CM_CNT,0x0a+(0x100<<TB_SEL)
	db	6,1					; Response R1
	dw	0
	db	d'55'|1<<CB_TRN,0,0,0,0,0x32<<1|1
; READ_OCR
CMD_58	dw	CM_CNT,0x0a+(0x100<<TB_SEL)
	db	6,5					; Response R7
	dw	0
	db	d'58'|1<<CB_TRN,0,0,0,0,0x7e<<1|1
; CRC_ON_OFF
CMD_59	dw	CM_CNT,0x0a+(0x100<<TB_SEL)
	db	6,1					; Response R1
	dw	0
	db	d'59'|1<<CB_TRN,0,0,0,0,0x7B<<1|1


; Enable SPI
SPI_EN	BSF	LATE,E_NCTS		; Signal: Not clear to send!
	CLRF	RCSTA			; Disable USART
	CLRF	TRISC			; Set SDO and TX as outputs
	MOVLW	1<<SSPEN		; Enable SPI at 12MHz
	MOVWF	SSPCON1
; Read SPI buffer to (WREG) 
SPI_RD	SETF	WREG
; Write (WREG) to SPI buffer
SPI_WR	MOVWF	SSPBUF			; Write the byte to the SPI buffer
SPI_LP	BTFSS	SSPSTAT,BF		; Wait until the transmission ends
	BRA	SPI_LP
	MOVF	SSPBUF,W		; Read the received byte
	RETURN
; Send a SD card Command from the table at FSR1
SDC_CM	RCALL	CM_SET			; Set up the command string in RAM
CM_SND	BCF	LATB,B_NFCS		; Select the SD card	
	LFSR	FSR1,SD_CMD		; Command string address
; Send the command string
SCM_LP	MOVF	POSTINC1,W		; Command byte to WREG
	RCALL	SPI_WR			; Send command bytes
	DECFSZ	CM_CNT			
	BRA	SCM_LP	
	SETF	CM_CNT			; Number of trials: 255
; Wait for the response
SRW_LP	RCALL	SPI_RD
	INCFSZ	WREG			; First response byte?
	BRA	SCM_RP			; Yes!
	DECFSZ	CM_CNT			; No!
	BRA	SRW_LP			; Retry	
SCM_RP	DECF	WREG	
	LFSR	FSR1,SD_CMD		; Response store address
; Get the response
SRP_LP	MOVWF	POSTINC1		; Store response bytes	
	RCALL	SPI_RD			; Get next response byte
	DECFSZ	RP_CNT			; Any more response bytes?
	BRA	SRP_LP			; Yes!
	TSTFSZ	DT_CNL			; Done!
	RETURN
	TSTFSZ	DT_CNH			; If the command is a
	RETURN				; 'no-data' operation,
	BRA	SD_DSL			; de-select the SD card
; Detect the SD card
SDC_ID	CLRF	SDC_ST			; Clear the SD card status byte
	BSF	LATE,E_NSDC	
	BSF	COM_ST,CO_CHG
	BTFSC	CHG_ST,B_NCRD		; Is a card present?
	RETURN				; No!
; Start the SD card in SPI mode
	BTFSS	SSPCON1,SSPEN		; SPI enabled?        
	RCALL	SPI_EN			; If not, enable now
	MOVLW	.10			; Generate 80 clock pulses
	MOVWF	CM_CNT
CLK_LP	RCALL	SPI_RD
	DECFSZ	CM_CNT			
	BRA	CLK_LP
	BCF	LATB,B_NFCS		; Select the SD card	
	LFSR	FSR1,CMD_00		; Command: GO_IDLE_STATE
	RCALL	SDC_CM			; Send command
	MOVLW	1<<R1_IDL		; Idle state?
	CPFSEQ	SD_CMD
	RETURN				; No! Abort card detection
; Check the SD card's version and operating voltage
	LFSR	FSR1,CMD_08		; Command: SEND_IF_COND
	RCALL	SDC_CM			; Send command
	BTFSC	SD_CMD,R1_ILC		; Illegal command?
	BRA	SD_V1X			; Yes! Version 1.x card
	MOVLW	C8_PAT			; Check pattern			
	CPFSEQ	SD_AD0			; Does it match?
	RETURN				; No! Abort card detection
	MOVLW	C8_VHS			; Check Voltage
	CPFSEQ	SD_AD1			; Does it match?
	RETURN				; No! Abort card detection
	BSF	SDC_ST,SD_VER		; Signal: Version 2 SD card
; Initialize the SD card
SD_V1X	RCALL	SDC_TM
C41_LP	LFSR	FSR1,CMD_55		; Command: APP_CMD
	RCALL	SDC_CM			; Send command
	LFSR	FSR1,ACM_41		; Command: SEND_OP_COND
	RCALL	SDC_CM			; Send command
	BTFSC	SD_CMD,R1_ILC		; Illegal command?
	RETURN				; Yes! Not a SD card
	BTFSS	TIM_ST,TM_SDC		; Time-out?
	RETURN				; Yes! Abort card detection
	TSTFSZ	SD_CMD			; Card still in 'Idle State'?
	BRA	C41_LP			; Yes! Repeat command
; Differentiate between standard and high capacity SD cards
	RCALL	SDC_TM
C58_LP	LFSR	FSR1,CMD_58		; Command: READ_OCR
	RCALL	SDC_CM			; Send command
	BTFSS	TIM_ST,TM_SDC		; Time-out?
	RETURN				; Yes! Abort card detection
	BTFSS	SD_AD3,OC_PWU		; Power-up finished?
	BRA	C58_LP			; Not yet! Try again
	BTFSC	SD_AD3,OC_CCS		; High Capacity card?
	BSF	SDC_ST,SD_HCS		; Yes!
; Disable the CRC check
	LFSR	FSR1,CMD_59		; Command: CRC_ON_OFF
	RCALL	SDC_CM			; Send command
; Set block length to 512 bytes for standard capacity SD cards
	LFSR	FSR1,CMD_16		; Command: SET_BLOCKLEN
	BTFSS	SDC_ST,SD_HCS		; Execute only for standard
	RCALL	SDC_CM			; capacity cards
; Get the SD card's capacity
	LFSR	FSR1,CMD_09		; Command: SEND_CSD
	RCALL	SDC_RD			; Send command
	BTFSS	TIM_ST,TM_SDC		; Time-out?
	RETURN				; Yes! Abort card detection
	BTFSC	DATA_B,CSD_HC		; Version 2 CSD structure?
	BRA	CSD_V2			; Yes!
; Version 1 CSD structure
; Get C_SIZE*2^6 and multiply by 2^(C_SIZE_MULT+2+READ_BL_LEN-9-6)
	MOVF	DATA_B+0x08,W
	ANDLW	b'11000000'		; Bits <63:62>
	MOVWF	SD_SZ0
	MOVF	DATA_B+0x07,W		; Bits <71:64>  
	MOVWF	SD_SZ1
	MOVF	DATA_B+0x06,W
	ANDLW	b'00000011'		; Bits <73:72>
	MOVWF	SD_SZ2
	CLRF	SD_SZ3
	MOVF	DATA_B+0x05,W
	ANDLW	b'00001111'		; Bits <83:80>
	MOVWF	CM_CNT			; READ_BL_LEN
	RLCF	DATA_B+0x0a,W		; Bit  <47> to Carry Flag
	RLCF	DATA_B+0x09,W		; Bits <49:48>
	ANDLW	b'00000111'		; C_SIZE_MULT
	ADDLW	-.13			; Add 2-6-9 = -13
	ADDWF	CM_CNT			; Add READ_BL_LEN
	BZ	C_SMAL
	BC	CMP_SZ			; SD Cards capacities of less
C_SMAL	RETURN				; than 256 MB are not accepted
; Version 2 CSD structure
; Get C_SIZE*2^8 and multiply by 2^2
CSD_V2	CLRF	SD_SZ0
	MOVFF	DATA_B+0x09,SD_SZ1	; Bits <55:48>
	MOVFF	DATA_B+0x08,SD_SZ2	; Bits <63:56>
	MOVFF	DATA_B+0x07,SD_SZ3	; Bits <71:64> (<71:70> are 00)
	MOVLW	2
	MOVWF	CM_CNT
; Calculate the SD card's capacity, in blocks of 512 bytes
CMP_SZ	MOVLW	0
	INCF	SD_SZ0			; Add 1
	ADDWFC	SD_SZ1
	ADDWFC	SD_SZ2
	ADDWFC	SD_SZ3			; This clears the Carry Flag
SHL_LP	RLCF	SD_SZ0			; Multiply by 2^(CM_CNT)
	RLCF	SD_SZ1
	RLCF	SD_SZ2
	RLCF	SD_SZ3			; This clears the Carry Flag
	DECFSZ	CM_CNT			; ( Borrow = set )
	BRA	SHL_LP
	SUBWFB	SD_SZ0			; Subtract 1
	SUBWFB	SD_SZ1
	SUBWFB	SD_SZ2
	SUBWFB	SD_SZ3
; Get the SD card's ID
	LFSR	FSR1,CMD_10		; Command: SEND_CID
	RCALL	SDC_RD			; Send command
	BTFSS	TIM_ST,TM_SDC		; Time-out?
	RETURN				; Yes! Abort card detection
	MOVF	SDC_ST,W		; Set first response byte to:
	ANDLW	b'00000011'		; '1' - Ver 1.x standard SD card
	ADDLW	'1'			; '2' - Ver 2.0 standard SD card
	MOVWF	DATA_B			; '4' - High capacity SD card
	copybl  SD_SZ0,DATA_B+8,4	; Copy the SD card's capacity
	BSF	SDC_ST,SD_PRE		; Signal: The SD card is present
	BCF	PORTE,E_NSDC	
	RETURN
; SD card Read ( Commands: CMD9, CMD10 and CMD17 )
SDC_RD	RCALL	SDC_TM			; Start time-out counter
	RCALL	SDC_CM			; Send command
	LFSR	FSR1,DATA_B
; Wait for the DATA_START_TOKEN
TS_FBU	INCFSZ	WREG			; Is the SDI line still high?
	BRA	TS_DST			; Not any more!
	BTFSS	TIM_ST,TM_SDC		; Time-out?
	BRA	SD_DSL			; Yes! Abort read operation
	RCALL	SPI_RD			; Wait for non-floating line
	BRA	TS_FBU
TS_DST	DECF	WREG			; Is the received byte a
DST_LP	XORLW	SD_DST			; DATA_START_TOKEN?
	BZ	RD_DAT			; Yes
	BTFSS	TIM_ST,TM_SDC		; Time-out?
	BRA	SD_DSL			; Yes! Abort read operation
	RCALL	SPI_RD			; Wait for DATA_START_TOKEN
	BRA	DST_LP
; Receive data
DRC_HI	DECF	DT_CNH			; Decrement the byte counter
DRC_LO	DECF	DT_CNL
	MOVWF	POSTINC1		; Save the received byte
RD_DAT	RCALL	SPI_RD			; Read next byte
	TSTFSZ	DT_CNL			; Check the byte counter
	BRA	DRC_LO
	TSTFSZ	DT_CNH			; Done?
	BRA	DRC_HI			; Not yet!
SD_DSL	BSF	LATB,B_NFCS		; Deselect the SD card
	RETURN
; Execute SD card client commands
SDC_OP	BCF	SDC_ST,SD_EXE		; Signal: No SDC commands!
	BTFSS	SSPCON1,SSPEN		; SPI enabled?        
	RCALL	SPI_EN			; If not, enable now
	MOVF	CMD_CD,W
	CLRF	CMD_CD			; Set Error Code to 0
	CLRF	BLN_LO			; Set the response length to 0
	CLRF	BLN_HI
	XORLW	SD_IDF			; 'Identify Card' command ?
	BNZ	TST_PR			; No!
; Identify card
	RCALL	SDC_ID			; Detect the SD card
	BTFSS	SDC_ST,SD_PRE		; Is the SD card present?
	BRA	NOT_PR			; No! Signal error
	MOVLW	0x0c			; Send 8 bytes of SD card
	MOVWF	BLN_LO			; identifier followed by 
	BRA	SD_END			; 4 bytes of card capacity
; Check if the SD card is present
TST_PR	BTFSC	SDC_ST,SD_PRE		; Is the SD card present?
	BRA	TST_RD			; Yes!
; Error condition: 'Not present'
NOT_PR	MOVLW	E_NPRE			; Error: 'Not present'
SD_ERR	MOVWF	CMD_CD			; Set the error code
SD_END	BSF	LATB,B_NFCS		; Deselect the SD card
	BSF	PIE1,SSPIE		; Signal: SDC command done!
	RETURN
TST_RD	XORLW	SD_IDF^SD_BRD		; 'Read Block' Command?
	BNZ	TST_WR			; No!
; Read Block
	LFSR	FSR1,CMD_17		; Command: READ_SINGLE_BLOCK
	RCALL	SDC_RD			; Send command
	BTFSS	TIM_ST,TM_SDC		; Time-out?
	BRA	NOT_PR			; Yes! Signal error
	MOVLW	0x02			; Return 512 bytes of data
	MOVWF	BLN_HI
	BRA	SD_END
TST_WR	XORLW	SD_BRD^SD_BWR		; 'Write Block' Command?
	BNZ	TST_EF			; No!
; Write Block
; The command string and data address have already been set up by a
; previous 'Set Write Address' command
	RCALL	CM_SND			; Send command
	TSTFSZ	SD_CMD			; Are any error bits set?
	BRA	WR_PRO			; Yes! Signal error
	LFSR	FSR1,DATA_B		; Address of data to be written
	MOVLW	SD_DST			; First send a DATA_START_TOKEN
	BRA	WR_DAT
; Send data
DWC_HI	DECF	DT_CNH			; Decrement the byte counter
DWC_LO	DECF	DT_CNL
	MOVF	POSTINC1,W		; Get the next byte to transmit
WR_DAT	RCALL	SPI_WR			; Send one byte
	TSTFSZ	DT_CNL			; Test the byte counter
	BRA	DWC_LO
	TSTFSZ	DT_CNH			; Done?
	BRA	DWC_HI			; Not yet!
; Check if the data was accepted
	RCALL	SPI_RD
	ANDLW	0x0f			; Did the card respond with
	XORLW	SD_DAC			; a DATA_ACCEPTED token?
	MOVLW	2			; Yes! Wait for 1 second while
	BZ	W_BUSY			; the card is busy
; Error condition: 'Write protected'
WR_PRO	MOVLW	E_WRPR			; Error: 'Write protected'
	BRA	SD_ERR
TST_EF	XORLW	SD_BWR^SD_EFA		; 'Set First Erase Address'?
	BNZ	TST_EL			; No!
; Set First Erase Address
	LFSR	FSR1,CMD_32		; Command: ERASE_WR_BLK_START_ADDR
	BRA	SND_EA			; Send command
TST_EL	XORLW	SD_EFA^SD_ELA		; 'Set Last Erase Address'?
	BNZ	TST_ER			; No!
; Set Last Erase Address
	LFSR	FSR1,CMD_33		; Command: ERASE_WR_BLK_END_ADDR
SND_EA	RCALL	SDC_CM			; Send command
	TSTFSZ	SD_CMD			; Are any error bits set?
	BRA	WR_PRO			; Yes! Signal error
	BRA	SD_END
TST_ER	XORLW	SD_ELA^SD_ERA		; 'Erase Blocks' Command?
	BNZ	NOT_SU			; No other commands are supported
; Erase
	LFSR	FSR1,CMD_38		; Command: ERASE
	RCALL	SDC_CM			; Send command
	TSTFSZ	SD_CMD			; Are any error bits set?
	BRA	WR_PRO			; Yes! Signal error
	MOVLW	6			; Set time-out to 5 seconds
; Wait while the SD card is busy
W_BUSY	MOVWF	CM_CNT			; Timing duration + 1 (in seconds)
CHK_TM	DCFSNZ	CM_CNT			; Count seconds
	BRA	WR_PRO			; Time-out! Signal error
	RCALL	SDC_TM			; Restart timer
BSY_LP	BTFSS	TIM_ST,TM_SDC		; Is one second over?
	BRA	CHK_TM			; Yes! 
	RCALL	SPI_RD			; Card still busy?
	BZ	BSY_LP			; Yes! Check again!
	RCALL	SPI_RD			; Done! Send 8 clock pulses
	BRA	SD_END			; Deselect the SD card
; Unimplemented command
NOT_SU	MOVLW	E_NSUP			; Error: 'Feature not supported'
	BRA	SD_ERR			; Set the error code
;------------------------------------------------------------------------------
; MAIN PROGRAM
;
MAIN_C	MOVLB	1<<USB_MB		; Select first USB memory bank
; Set all inputs to digital
	MOVLW	1<<PCFG3|1<<PCFG2|1<<PCFG1|1<<PCFG0
	MOVWF	ADCON1			; No analog inputs
; Port E initialization	 
	MOVLW	1<<E_NCTS
	MOVWF	LATE			; Not clear to send
	MOVLW	1<<E_NRTS		; RTS is input
	MOVWF	TRISE
; Port B initialization
	MOVLW	1<<B_NSCS|1<<B_NFCS	; SPI devices deselected
	MOVWF	PORTB
	MOVLW	1<<B_EINT|1<<B_USEN|1<<B_NCRD|1<<B_NSTB|1<<B_PSDI	
      	MOVWF	TRISB			; Define inputs
	MOVLW	1<<NOT_RBPU		; No Pull-ups. nSTROBE triggers
	MOVWF	INTCON2			; INT2 interrupt on falling edge
; Port A and C initialization
	RCALL	PP_RST			; Reset PPORT
	CLRF	TRISA			; RA<5:0> are outputs
	CLRF	TRISC
; USART initialization
	BSF	BAUDCON,BRG16		; 16-bit Baud rate generator
	MOVLW	high (FOSC_4/BAUD_R-1)
	MOVWF	SPBRGH			; Baud rate: BAUD_R bit/s
	MOVLW	low (FOSC_4/BAUD_R-1)
	MOVWF	SPBRG
; Timer 0 initialization
	MOVLW	1<<T0PS2|1<<T0PS1|1<<T0PS0
	MOVWF	T0CON			; Timer 0 prescaler is 1:256
; SPI initialization
	MOVLW	1<<CKE			; Shift on falling edge
	MOVWF	SSPSTAT
; USB initialization
	MOVLW	1<<UPUEN|1<<FSEN	; Full speed, on-chip transreceiver,
	MOVWF	UCFG			; pull-up enabled, no ping-pong
; Start-up delay
	BSF	OSCCON,IDLEN		; Set Idle mode on SLEEP
	BSF	WDTCON,SWDTEN		; Start Watch-dog timer
	SLEEP				; Wait for 2s
	NOP
	BCF	WDTCON,SWDTEN		; Stop Watch-dog timer
	BSF	LATE,E_NSRV		; Switch off the Server LED
	BSF	LATE,E_NSDC		; Switch off the SDC LED
; Clear interrupt-on-change mismatch
	MOVF	PORTB,W			; Clear RB port mismatch condition
  	MOVWF	CHG_ST			; Save the value of the RB port.
; Signal: Detect SD Card
	BSF	SDC_ST,SD_DET
; Attach USB
	RCALL	USB_SE			; Attach USB
; Enable Interrupts
	MOVLW	1<<INT2IE		; Enable the SPP Strobe (INT2)
	MOVWF	INTCON3			; interrupt
	MOVLW	1<<GIE|1<<PEIE|1<<TMR0IE|1<<RBIE
	MOVWF	INTCON			; Enable all needed interrupts
	RCALL	SDC_TM
; Main loop
M_LOOP	BTFSC	COM_ST,CO_ACT		; Communication in progress?
	BRA	M_LOOP			; Yes! Wait until it ends
	BTFSC	SDC_ST,SD_DET		; Detect SD card?
	RCALL	SDC_ID			; Yes!
	MOVF	COM_ST,W
	ANDLW	1<<CO_ACT|1<<CO_TRA|1<<CO_SER|1<<CO_CHG
	XORLW	1<<CO_ACT|1<<CO_TRA|1<<CO_CHG
	BZ	M_EXEC
	BCF	COM_ST,CO_CHG
	RCALL	PP_RST
M_EXEC	BTFSC	SDC_ST,SD_EXE		; SDC command pending?
	RCALL	SDC_OP			; Execute SDC command
	SLEEP				; Wait for an interrupt
	NOP
	BRA	M_LOOP	
;------------------------------------------------------------------------------
; End of program
	END