;******************************************************************************
;                                                                             *
;   This file is a basic code template for code generation on the             *
;   PIC18F47J53. This file contains the basic code building blocks to build   *
;   upon.                                                                     *
;                                                                             *
;   Refer to the MPASM User's Guide for additional information on features    *
;   of the assembler.                                                         *
;                                                                             *
;   Refer to the respective data sheet for additional information on the      *
;   instruction set.                                                          *
;                                                                             *
;******************************************************************************
;                                                                             *
;    Filename:         SerMod.asm                                             *
;    Date:                                                                    *
;    File Version:                                                            *
;    Author:                                                                  *
;    Company:                                                                 *
;                                                                             *
;******************************************************************************
;                                                                             *
;    Files Required: P18F47J53.INC                                            *
;                                                                             *
;******************************************************************************
;                                                                             *
;    Notes:                                                                   *
;                                                                             *
;******************************************************************************
;                                                                             *
;    Revision History:                                                        *
;                                                                             *
;******************************************************************************

;------------------------------------------------------------------------------
; PROCESSOR DECLARATION

	LIST      P=PIC18F47J53	; list directive to define processor
 	#INCLUDE <P18F47J53.INC>; processor specific variable definitions
	errorlevel    -311,-202 ; Suppress the message: Operand of HIGH
				;      operator was larger than H'FFFF'
;------------------------------------------------------------------------------
;
; CONFIGURATION WORD SETUP
;
; The 'CONFIG' directive is used to embed the configuration word within the 
; .asm file. The lables following the directive are located in the respective 
; .inc file.  See the data sheet for additional information on configuration 
; word settings.
;
;------------------------------------------------------------------------------

     
     CONFIG WDTEN = OFF		;Watchdog Timer: Disabled - Controlled by SWDTEN bit
     CONFIG PLLDIV = 4		;Divide by 4 (16 MHz oscillator input)
     CONFIG CFGPLLEN = ON	;PLL Enabled
     CONFIG STVREN = OFF	;Stack Overflow/Underflow Reset: Disabled
     CONFIG XINST = OFF		;Extended Instruction Set: Disabled
     CONFIG CPUDIV = OSC1	;No CPU system clock divide
     CONFIG CP0 = OFF		;Program memory is not code-protected 
     CONFIG OSC = HSPLL		;HSPLL oscillator with PLL software controlled
     CONFIG SOSCSEL = DIG	;Digital (SCLKI) mode selected
     CONFIG CLKOEC = OFF	;CLKO output disabled on the RA6 pin
     CONFIG FCMEN = OFF		;Fail-Safe Clock Monitor: Disabled
     CONFIG IESO = OFF		;Internal External Oscillator Switch Over Mode: Disabled
     CONFIG WDTPS = 512		;Watch-dog timer set to 512*4ms
     CONFIG DSWDTOSC = INTOSCREF;DSWDT Reference Clock uses INTRC
     CONFIG RTCOSC = INTOSCREF	;RTCC Reference Clock INTRC
     CONFIG DSBOREN = OFF	;Deep Sleep BOR: Disabled
     CONFIG DSWDTEN = OFF	;Deep Sleep Watchdog Timer: Disabled
     CONFIG DSWDTPS = 2		;Deep Sleep Watchdog Postscaler:
     CONFIG IOL1WAY = ON        ;The IOLOCK bit (PPSCON<0>) can be set once
     CONFIG ADCSEL = BIT10	;ADC 10 or 12 Bit Select: 10 - Bit ADC Enabled
     CONFIG MSSP7B_EN = MSK5	;MSSP address masking: 5 Bit address masking mode
     CONFIG WPFP = PAGE_0	;Write/Erase Protect Page Start/End Location: Page 0
     CONFIG WPCFG = OFF		;Configuration Words page not erase/write-protected
     CONFIG WPDIS = OFF		;WPFP[6:0], WPEND, and WPCFG bits ignored
     CONFIG WPEND = PAGE_0	;Page 0 to WPFP<6:0> erase/write-protected
     CONFIG LS48MHZ = SYS48X8	;System clock at 48 MHz USB CLKEN divide-by is set to 8

;------------------------------------------------------------------------------
; CONSTANTS

; Clocks
FOSC_4	equ	d'12000000'		; Instruction cycle (Hz)
BAUD_R	equ	d'115200'		; Serial Baud rate (Hz)
; I/O Ports
MOUS_Y	equ	PMDOUT1L		; Kempston mouse Y port
MOUS_X	equ	PMDOUT1H		; Kempston mouse X port
KEM_JS	equ	PMDOUT1H		; Kempston joystick port
DV_STA	equ	PMDOUT2L		; Device state
MOUS_B	equ	PMDOUT2L		; Kempston mouse buttons port
PSP_DO  equ	PMDOUT2H		; PSP out
PSP_DI	equ	PMDIN2H			; PSP in
; I/O Port bits
A_NFCS	equ	RA2			; SD card select
;A_NRCS	equ	RA3			; Serial RAM select
B_JOY0	equ	RB0
B_JOY1	equ	RB1
B_JOY2  equ	RB2
B_JOY3	equ	RB3
B_USEN	equ	RB6			; USB power sense
B_NCRD	equ	RB7			; Flash card detect
C_NMSE	equ	RC1			; LED: Mouse not enabled
C_NSDC	equ	RC1                     ; LED: SD card not present
C_NSRV	equ	RC2                     ; LED: Server not connected
J_JOY4	equ	RC0			; Joystick button bit
PORT_J	equ	PORTC			; Joystick button port
TRIS_J	equ	PORT_J+0x12
; Device State (DV_STA) bits
;DV_BSY	equ	7			; Device busy
DV_DAT	equ	6			; Data ready
DV_SRV	equ	5                       ; Server connected
DV_SDC	equ	4			; SDC identified
DV_MSE	equ	3			; PS/2 mouse connected
DV_REZ	equ	2			; Rezerved
MSE_BL	equ	1			; Mouse left button inactive
MSE_BR	equ	0			; Mouse right button inactive
; Joystick port bits
JOY_RT	equ	0
JOY_LT	equ	1
JOY_FR	equ	4
; Communication State (COM_ST) bits
CO_ACT	equ	0			; Communication in progress
CO_HED	equ	1			; Header / not Data
CO_TRA	equ	2			; Transmit / not Receive
CO_SER	equ	3			; Serial / not Parallel
; Mouse State (MSE_ST) bits
MS_LEG	equ	0			; Mouse in legacy mode
MS_HLD	equ	1			; Mouse on hold
MS_PAK	equ	2			; Mouse packet in progress
MS_ACK	equ	3			; Expect PS/2 'Acknowledge'
MS_ENB	equ	4			; Mouse enabled
MS_JOY	equ	5			; Mouse as joystick
; Timer 0 State (TIM_ST) bits
TM_SBD	equ	0			; USB power debounce
TM_CRD	equ	1			; SDC detect debounce
TM_PSP	equ	2			; PSP timer running
TM_USB	equ	3			; USB timer running
TM_SDC	equ	4			; SDC timer running
TM_JOY	equ	5			; Joystick timer running
; Interface constants
NS_ATC	equ	0xf1			; Host Command Code
E_COMM	equ	0x09			; Error 'Communication error'
E_NSUP	equ	0x07			; Error 'Feature not supported'
E_NPRE	equ	0x10			; Error 'Microdrive not present'
E_WRPR	equ	0x0e			; Error 'Drive write protected'

; USB related constants

; Processor specific
; BDnSTAT bits (DS39632E Register 17-5, 17-6) missing from .INC file 
BSTALL	equ	2			; Stall
DTSEN	equ	3			; Data Sync Enable
DTS	equ	6			; Data Sync Parity
UOWN	equ	7			; USB Owned
PID0	equ	2			; PID bit 0
PID1	equ	3			; PID bit 1
PID2	equ	4			; PID bit 2
PID3	equ	5			; PID bit 3

; Application specific
; General
CF_MAX	equ	1			; Maximum configuration number
IF_MAX	equ	1			; Maximum interface number
EP_MAX	equ	2			; Maximum endpoint number
; CC Endpoint Numbers
EP_CTL	equ	1			; CC Control Endpoint number
EP_DAT	equ	EP_MAX+1-EP_CTL		; CC Data Endpoint number
; Maximum packet size (binary logarithm)
PA_CTL	equ	3			; LOG2 of control packet size
PA_DAT	equ	6			; LOG2 of data packet size
; USB State (USB_ST) bits
SB_SPW	equ	0			; USB Self Powered
SB_RWU	equ	1			; USB Remote Wake-up enabled
SB_DEF	equ	2			; USB in Default state
SB_CFG	equ	3			; USB in Configured state
SB_DTR	equ	4			; DTR line state
; Other
TB_SEL	equ	0			; Table selector bit
 
; Standard USB Constants
; Token PID Types (USB 2.0 Table 8-1)
PID_SU	equ	b'1101'			; SETUP Token
; Request Type Bits (USB 2.0 Table 9-2)
RT_ITF	equ	0			; Interface
RT_ENP	equ	1			; Endpoint
RT_CLA	equ	5			; Class
RT_VEN	equ	6			; Vendor
RT_DIR	equ	7			; Data Direction IN
; Standard Request Codes (USB 2.0 Table 9-4)
R_GSTA	equ	0x00			; Get Status
R_CFTR	equ	0x01			; Clear Feature
R_SFTR	equ	0x03			; Set Feature
R_SADR	equ	0x05			; Set Address
R_GDES	equ	0x06			; Get Descriptor
R_GCFG	equ	0x08			; Get Configuration
R_SCFG	equ	0x09			; Set Configuration
R_GITF	equ	0x0a			; Get Interface
R_SITF	equ	0x0b			; Set Interface
; Standard Descriptor Types (USB 2.0 Table 9-5)
DT_DEV	equ	0x01			; Device
DT_CFG	equ	0x02			; Configuration
DT_STR	equ	0x03			; String
DT_ITF	equ	0x04			; Interface
DT_ENP	equ	0x05			; Endpoint
; Feature Selectors (USB 2.0 Table 9-6)
DF_RWU	equ	0x01			; Device Remote Wake-up
EF_HLT	equ	0x00			; Endpoint HALT
; Endpoint Status bits (USB 2.0 Figure 9-6)
GS_HLT	equ	0			; HALT
; Standard Configuration Attributes (USB 2.0 Table 9-10)
CA_RWU	equ	1<<7|1<<5		; Remote Wake-up
CA_SPW	equ	1<<7|1<<6		; Self powered
; Standard Endpoint Address bits (USB 2.0 Table 9-13)
EA_DIR	equ	7			; Direction IN
; Standard Endpoint Attributes (USB 2.0 Table 9-13)
EA_BLK	equ	0x02			; Bulk
EA_INT	equ	0x03			; Interrupt
; CC Device and Interface (Sub)Classes (CDC 1.2 Table 2,3,4,6)
DC_NUS	equ	0x00			; Not Used
DC_COM	equ	0x02			; Communication Device Class
IC_COM	equ	0x02			; Communication Interface Class
IC_DAT	equ	0x0a			; Data Interface Class
IS_ACM	equ	0x02			; Abstract Control Interface Subclass
; CC Protocols (CDC 1.2 Table 5,7)
PC_NOP	equ	0x00			; No protocol
PC_ATC	equ	0x01			; AT Commands V.250
; CC Descriptor Types (CDC 1.2 Table 12)
DT_CSI	equ	0x24			; CS Interface Descriptor
; CC Functional Descriptor Subtypes (CDC 1.2 Table 13)
FD_HED	equ	0x00			; Header Descriptor
FD_CMA	equ	0x01			; Call Management Descriptor
FD_ACM	equ	0x02			; Abstract Control Descriptor
FD_UNI	equ	0x06			; Union Descriptor
; CC Request Codes (CDC 1.2 Table 19)
C_SECO	equ	0x00			; Send Encapsulated Command
C_GERP	equ	0x01			; Get Encapsulated Response
C_SLCO	equ	0x20			; Set Line Coding
C_GLCO	equ	0x21			; Get Line Coding
C_SCLS	equ	0x22			; Set Control Line State
; CC Notifications (CDC 1.2 Table 20)
N_SSTA	equ	0x20			; Serial State
; CC Abstract Control Management Capability Bits (PSTN 1.2 Table 4)
CA_AC1	equ	1			; C_SLCO, C_GLCO, C_SCLS, N_SSTA
; CC Control Line State Bits (CLN_ST) (PSTN 1.2 Table 18)
CS_DTR  equ     0                       ; DTR
CS_RTS	equ	1			; RTS
; CC Serial Line State Bits (PSTN 1.2 Table 31)
SS_DCD	equ	0			; DCD
SS_DSR	equ	1			; DSR

; SD card related constants
CB_TRN	equ	6			; Command Transfer bit
CB_HCS	equ	6			; 
R1_IDL	equ	0			; R1 'Idle State' bit
R1_ILC	equ	2			; R1 'Illegal Command' bit
C8_VHS	equ	b'0001'			; CMD8 VHS field
C8_PAT	equ	0xaa			; CMD8 check pattern
OC_CCS	equ	6
OC_PWU	equ	7
SD_DST	equ	0xfe			; DATA_START_TOKEN
SD_DAC	equ	0x05			; DATA_ACCEPTED
CSD_HC	equ	6
; SD card Commands
SD_SWA	equ	0xf8			; 'Set Write Address'
SD_IDF	equ	0xf9			; 'Identify Card'
SD_BRD	equ	0xfa			; 'Read Block'
SD_BWR	equ	0xfb			; 'Write Block'
SD_EFA	equ	0xfc			; 'Set First Erase Address'
SD_ELA	equ	0xfd			; 'Set Last Erase Address'
SD_ERA	equ	0xfe			; 'Erase Blocks'
; SD card State (SDC_ST) bits
SD_VER	equ	0			; SD card Version 2.0
SD_HCS	equ	1			; SD card High Capacity
SD_DET	equ	2			; Detect SD card
SD_PRE	equ	3			; SD card Present
SD_EXE	equ	4			; Execute SDC command

; PS/2 mouse related constants
PS_ACK	equ	0xfa			; Acknowledge
;PS_TST	equ	0xaa			; Self-test passed
;PS_MID	equ	0x00			; Mouse ID
MV_MBL	equ	0			; PS/2 mouse left button
MV_MBR	equ	1			; PS/2 mouse right button
;MV_MBM	equ	2			; PS/2 mouse middle button
MV_SGN	equ	5			; Y-movement sign bit
; PS/2 mouse commands
;PS_RST	equ	0xff			; Reset
;PS_DEF	equ	0xf6			; Set defaults: 100,2,1:1,disabled
PS_REP	equ	0xf4			; Enable data reporting
;PS_SAM	equ	0xf3			; Set sample rate: 10,20,40,60,80,100,200
;PS_RES	equ	0xe8			; Set resolution: 0,1,2,3 (log2)
;PS_SC2	equ	0xe7			; Set scaling 2:1
;PS_SC1	equ	0xe6			; Set scaling 1:1

;------------------------------------------------------------------------------
; MACRO INSTRUCTIONS

l_time	macro	ms
	MOVLW	high (~(FOSC_4/d'1000'*ms/h'100'))
	MOVWF	TMR0H
	MOVLW	low (~(FOSC_4/d'1000'*ms/h'100'))
	endm

s_time	macro	us
	MOVLW	high (~(FOSC_4/d'1000000'*us/h'8'))
	MOVWF	TMR1H
	MOVLW	low (~(FOSC_4/d'1000000'*us/h'8'))
	endm

copy_b	macro	src,dst,num
	local	i = 0
	while	i < num
	MOVFF	src+i,dst+i
i ++
	endw
	endm

tab_ad	macro	ad
	MOVLW	low ad
	MOVWF	TBLPTRL
	MOVLW	high ad
	MOVWF	TBLPTRH
	MOVLW	upper ad
	MOVWF	TBLPTRU
	endm

;------------------------------------------------------------------------------
; VARIABLES

	cblock	0x000			; Variables in access RAM

; General variables
COM_ST					; Communication State 
TIM_ST					; Timer 0 State
CHG_ST					; Last PORTB value
USB_ST					; USB State
SDC_ST					; SD card State
CNT_LO					; Byte counter low
CNT_HI					; Byte counter high
EP_OFS					; Offset of last active EP
TB_CNT					; Table read byte counter
; PS/2 mouse variables
MSE_ST					; PS/2 mouse state
JS_SEN					; Mouse as joystick sensitivity
JS_RTC:2				; Joystick read timing constant
MWIN_X,MWIN_Y				; Mouse window size
MPOS_X,MPOS_Y				; Mouse coordinates
MOVE_B,MOVE_X,MOVE_Y			; PS/2 mouse movement packet
MS_POS					; Mouse position
; SD card variables
SD_SZ0,SD_SZ1,SD_SZ2,SD_SZ3		; SD card size 
CM_CNT,RP_CNT,DT_CNL,DT_CNH		; SD Command counters
; SD card Command
SD_CMD,SD_AD3,SD_AD2,SD_AD1,SD_AD0,SD_CRC
; USB Request Header
RQ_TYP,RQ_NUM,RQ_VAL,RQ_DTY,RQ_IDX:2,RQ_LNL,RQ_LNH
; Spare access RAM
SPARE
	endc

	cblock	0x050
; Main communication buffer
CMD_CD,AUX_CD,BLN_LO,BLN_HI		; Communication Header Block
DATA_B:0x400
DB_END					; Communication Data Block
	endc

	cblock	0xd00
; Buffer Descriptor Table
BDT0_O,EP0O_L:3,BDT0_I,EP0I_L:0x13
; Endpoint 0 buffers
BUF0_O:1<<PA_CTL			; EP0 OUT buffer
BUF0_I:1<<PA_CTL			; EP0 IN buffer
; CC Notification Header 
NT_TYP,NT_NUM,NT_VAL:2,NT_IDX:2,NT_LEN:2
; CC Notification Data
NT_DAT:4
; CC Line Coding data
LN_COD:7
LC_END
	endc

;------------------------------------------------------------------------------
; RESET

	org	0x0000
	CLRF	COM_ST			; Clear Communication State
	CLRF	TIM_ST			; Clear Timer State
	GOTO	INIT_D
	GOTO	HI_INT
	fill	0xffff,0x0018-$
	GOTO	HI_INT

;------------------------------------------------------------------------------
; TABLES

; USB initialization data
BD_EP0	dw	BDT0_O,BD_CTL-BD_EP0-4+(0x100<<TB_SEL)
; Buffer Descriptor Initial Data for EP0
	db	1<<UOWN|1<<BSTALL,1<<PA_CTL
	dw	BUF0_O
	db	1<<UOWN|1<<BSTALL,1<<PA_CTL
	dw	BUF0_I
BD_CTL	dw	BDT0_O+(1<<ENDP0)*EP_CTL,BD_DAT-BD_CTL-4+(0x100<<TB_SEL)
; Buffer Descriptor Initial Data for the CC Control EP
	dw	0,0
	db	1<<UOWN|1<<DTSEN,0x0a
	dw	NT_TYP
BD_DAT	dw	BDT0_O+(1<<ENDP0)*EP_DAT,NT_INI-BD_DAT-4+(0x100<<TB_SEL)
; Buffer Descriptor Initial Data for the CC Data EP
	db	0,0
	dw	CMD_CD
	db	0,0
	dw	CMD_CD
NT_INI	dw	NT_TYP,LC_INI-NT_INI-4+(0x100<<TB_SEL)
; Notification buffer Initial Data
	db	1<<RT_DIR|1<<RT_CLA|1<<RT_ITF,N_SSTA
	dw	0,IF_MAX-1
	dw	2,1<<SS_DSR|1<<SS_DCD
LC_INI	dw	LN_COD,LC_END-LN_COD+(0x100<<TB_SEL)
; CC Line Coding Initial Data
	db	low BAUD_R, high BAUD_R 
	db	upper BAUD_R, 0
	db	0,0,8	

; USB Descriptors Index Table
DE_PNT	db	low DE_DEV,high DE_DEV,upper DE_DEV,0
	db	low DE_CFG,high DE_CFG,upper DE_CFG,0
	db	low DE_ST0,high DE_ST0,upper DE_ST0,0
	db	low DE_ST1,high DE_ST1,upper DE_ST1,0
	db	low DE_ST2,high DE_ST2,upper DE_ST2,0
	db	low DE_ST3,high DE_ST3,upper DE_ST3,0

; USB Descriptors
DE_DEV	dw	BUF0_I,DE_CFG-DE_DEV-4	; Table header
; Device Descriptor
	db	0x12,DT_DEV		; bLength		18
					; bDescriptorType	Device
    	dw	0x0200			; bcdUSB		USB 2.0
	db	DC_COM,DC_NUS		; bDeviceClass		Communication
					; bDeviceSubClass	None
	db	DC_NUS,1<<PA_CTL	; bDeviceProtocol	None
					; bMaxPacketSize0	Packet Size
	dw    	0x04D8			; idVendor		Microchip
	dw	0xF6FC			; idProduct		IF1BIS
	dw	0x0111			; bcdDevice		PIC18F47J53
	db    	1,2			; iManufacturer		String 1
					; iProduct		String 2
	db	3,CF_MAX		; iSerialNum		String 3
					; bNumConfigurations	1

DE_CFG	dw	BUF0_I,DE_ST0-DE_CFG-5	; Table header	
; Configuration (1) Descriptor
	db	0x09,DT_CFG		; bLength	     	9
					; bDescriptorType	Configuration
	dw	0x0043			; wTotalLength       	67
	db	IF_MAX+1,CF_MAX		; bNumInterfaces	2
					; bConfigurationValue	1
	db	0,CA_SPW|CA_RWU		; iConfiguration	None
					; bmAttributes       	Self-Pow,Rem Wake-up
	db	1,9			; bMaxPower          	2mA
; CDC Control Interface (0) Descriptor
					; bLength		9
	db	DT_ITF,IF_MAX-1		; bDescriptorType	Interface
					; bInterfaceNumber	0
	db	0,1			; bAlternateSetting	None
					; bNumEndpoints		1
	db	IC_COM,IS_ACM		; bInterfaceClass	Communication
					; bInterfaceSubClass	Abstract
	db	PC_ATC,0		; bInterfaceProtocol	AT Command
					; iInterface		None
; CDC Header Functional Descriptor
	db	5,DT_CSI		; bFunctionLength	5
					; bDescriptorType	CS Interface
	db	FD_HED,0x10		; bDescriptorSubtype	Header
					; bcdCDC		CDC 1.10
	db	0x01,4			;
; CDC Abstract Control Model Functional Descriptor
					; bFunctionLength	4
	db	DT_CSI,FD_ACM		; bDescriptorType	CS Interface
					; bDescriptorSubtype	Abstract Control
	db	1<<CA_AC1,5		; bmCapabilities	D1
; CDC Union Functional Descriptor
					; bFunctionLength	5
	db	DT_CSI,FD_UNI		; bDescriptorType	CS Interface
					; bDescriptorSubtype	Union
	db	IF_MAX-1,IF_MAX		; bmMasterInterface	0
					; bSlaveInterface0	1
; CDC Call Management Functional Descriptor
	db	5,DT_CSI		; bFunctionLength	5
					; bDescriptorType	CS Interface
	db	FD_CMA,0		; bDescriptorSubtype	Call Management
					; bmCapabilities	None
	db	IF_MAX,7		; bDataInterface	1
; CDC Notification Endpoint Descriptor
					; bLength		7
	db	DT_ENP,EP_CTL|1<<EA_DIR	; bDescriptorType	Endpoint
					; bEndpointAddress	Address
	db	EA_INT,LN_COD-NT_TYP	; bmAttributes		Interrupt
					; wMaxPacketSize	Packet Size
	db	0,2			; bInterval		2ms 
; CDC Data Interface (1) Descriptor
	db	9,DT_ITF		; bLength		9
					; bDescriptorType	Interface
	db	IF_MAX,0		; bInterfaceNumber	1
					; bAlternateSetting	None
	db	2,IC_DAT		; bNumEndpoints		2
					; bInterfaceClass	Data
	db	DC_NUS,PC_NOP		; bInterfaceSubClass	None
					; bInterfaceProtocol	None
	db	0,7			; iInterface		None
; CDC Data OUT Endpoint Descriptor
					; bLength		7
	db	DT_ENP,EP_DAT		; bDescriptorType	Endpoint
					; bEndpointAddress	Address
	db	EA_BLK,1<<PA_DAT	; bmAttributes		Bulk
					; wMaxPacketSize	Packet Size
	db	0,0			; bInterval		None
; CDC Data IN Endpoint Descriptor
	db	7,DT_ENP		; bLength		7
					; bDescriptorType	Endpoint
	db	EP_DAT|1<<EA_DIR,EA_BLK	; bEndpointAddress	Address
					; bmAttributes		Bulk
	db	1<<PA_DAT,0		; wMaxPacketSize	Packet Size
	db	0			; bInterval		None

DE_ST0	dw	BUF0_I,DE_ST1-DE_ST0-4	; Table header
; Language ID String Descriptor
	db	DE_ST1-DE_ST0-4,DT_STR	; bLength		Length
					; bDescriptorType	String
	dw	0x0409			; wLangID		English (US)

DE_ST1	dw	BUF0_I,DE_ST2-DE_ST1-4	; Table header
; Manufacturer String Descriptor 
	db	DE_ST2-DE_ST1-4,DT_STR	; bLength		Length
					; bDescriptorType	String
	dw	'M','i','c','r','o','c','h','i','p',' '
	dw	'T','e','c','h','n','o','l','o','g','y'
	dw	' ','I','n','c','.'

DE_ST2	dw	BUF0_I,DE_ST3-DE_ST2-4	; Table header
; Product String Descriptor 
	db	DE_ST3-DE_ST2-4,DT_STR	; bLength		Length
					; bDescriptorType	String
	dw	'I','n','t','e','r','f','a','c','e','1','b','i','s'

DE_ST3	dw	BUF0_I,DE_END-DE_ST3-4	; Table header
; Serial Number String Descriptor 
	db	DE_END-DE_ST3-4,DT_STR	; bLength		Length
					; bDescriptorType	String
	dw	'P','I','C','1','8','F','4','7','J','5','3'
DE_END

; SD card command strings
; GO_IDLE_STATE
CMD_00	dw	CM_CNT,0x0a+(0x100<<TB_SEL)
	db	6,1					; Response R1
	dw	0
	db	d'00'|1<<CB_TRN,0,0,0,0,0x4a<<1|1				
; SEND_IF_COND
CMD_08	dw	CM_CNT,0x0a+(0x100<<TB_SEL)
	db	6,5					; Response R3
	dw	0
	db	d'08'|1<<CB_TRN,0,0,C8_VHS,C8_PAT,0x43<<1|1
; SEND_CSD
CMD_09	dw	CM_CNT,0x0a+(0x100<<TB_SEL)
	db	6,1					; Response R1
	dw	0x0012					; Data size: 16+2
	db	d'09'|1<<CB_TRN,0,0,0,0,0x57<<1|1
; SEND_CID
CMD_10	dw	CM_CNT,0x0a+(0x100<<TB_SEL)
	db	6,1					; Response R1
	dw	0x0012					; Data size: 16+2
	db	d'10'|1<<CB_TRN,0,0,0,0,0x0d<<1|1
; SET_BLOCKLEN
CMD_16	dw	CM_CNT,0x0a+(0x100<<TB_SEL)
	db	6,1					; Response R1
	dw	0
	db	d'16'|1<<CB_TRN,0,0,2,0,0x0a<<1|1
; READ_SINGLE_BLOCK
CMD_17	dw	CM_CNT,0x0a+(0x100<<TB_SEL)
	db	6,1					; Response R1
	dw	0x0202					; Data size: 512+2
	db	d'17'|1<<CB_TRN,0,0,0,0,0x00<<1|1	; Data address
; WRITE_BLOCK
CMD_24	dw	CM_CNT,0x0a+(0x100<<TB_SEL)
	db	6,1					; Response R1
	dw	0x0202					; Data size: 512+2
	db	d'24'|1<<CB_TRN,0,0,0,0,0x00<<1|1	; Data address
; ERASE_WR_BLK_START_ADDR
CMD_32	dw	CM_CNT,0x0a+(0x100<<TB_SEL)
	db	6,1					; Response R1
	dw	0
	db	d'32'|1<<CB_TRN,0,0,0,0,0x00<<1|1	; Data address
; ERASE_WR_BLK_END_ADDR
CMD_33	dw	CM_CNT,0x0a+(0x100<<TB_SEL)
	db	6,1					; Response R1
	dw	0
	db	d'33'|1<<CB_TRN,0,0,0,0,0x00<<1|1	; Data address
; ERASE
CMD_38	dw	CM_CNT,0x0a+(0x100<<TB_SEL)
	db	6,1
	dw	0xffff					; Response R1b
	db	d'38'|1<<CB_TRN,0,0,0,0,0x52<<1|1
; SD_SEND_OP_COND
ACM_41	dw	CM_CNT,0x0a+(0x100<<TB_SEL)
	db	6,1					; Response R1
	dw	0
	db	d'41'|1<<CB_TRN,1<<CB_HCS,0,0,0,0x3b<<1|1
; APP_CMD
CMD_55	dw	CM_CNT,0x0a+(0x100<<TB_SEL)
	db	6,1					; Response R1
	dw	0
	db	d'55'|1<<CB_TRN,0,0,0,0,0x32<<1|1
; READ_OCR
CMD_58	dw	CM_CNT,0x0a+(0x100<<TB_SEL)
	db	6,5					; Response R7
	dw	0
	db	d'58'|1<<CB_TRN,0,0,0,0,0x7e<<1|1
; CRC_ON_OFF
CMD_59	dw	CM_CNT,0x0a+(0x100<<TB_SEL)
	db	6,1					; Response R1
	dw	0
	db	d'59'|1<<CB_TRN,0,0,0,0,0x7b<<1|1

; Joystick read rate timing constants
JS_TMC	dw	~(FOSC_4/h'100'/.1)			;  1Hz
	dw	~(FOSC_4/h'100'/.2)			;  2Hz
	dw	~(FOSC_4/h'100'/.3)			;  3Hz
	dw	~(FOSC_4/h'100'/.4)			;  4Hz
	dw	~(FOSC_4/h'100'/.5)			;  5Hz
	dw	~(FOSC_4/h'100'/.6)			;  6Hz
	dw	~(FOSC_4/h'100'/.7)			;  7Hz
	dw	~(FOSC_4/h'100'/.8)			;  8Hz
	dw	~(FOSC_4/h'100'/.9)			;  9Hz
	dw	~(FOSC_4/h'100'/.10)			; 10Hz

; Joystick / mouse initial settings
JM_INI	dw	MSE_ST,MPOS_X-MSE_ST+(0x100<<TB_SEL)
	db	0,.10					; State, Sensitivity
	dw	~(FOSC_4/h'100'/.5)			; Timing constant
	db	.255,.175				; Window size

;------------------------------------------------------------------------------
; DEVICE INITALIZATION

INIT_D	
	banksel	ANCON0
; Set all inputs to digital
	SETF	ANCON0			; No analog inputs AN0-AN7
	MOVLW	1<<PCFG12|1<<PCFG11|1<<PCFG10|1<<PCFG9|1<<PCFG8
	MOVWF	ANCON1			; No analog inputs AN8-AN12
	banksel	DV_STA
	LFSR	FSR0,NT_DAT		; No CC data request active!
; Port B configuration
	BCF	INTCON2,NOT_RBPU	; Enable PORTB pull-up 
; Read PORTB
	MOVF	PORTB,W			; Clear RB port mismatch condition
  	MOVWF	CHG_ST			; Save the value of the RB port.
; Check if line JOY0 is low
	BTFSC	CHG_ST,B_JOY0		; Is JOY0 low?
	BRA	MAIN_C			; No! Continue initialization
;
; If the joystick is held to the right during start-up, the
; self-programming sequence is triggered

; Port C configuration
	SETF	PORTC			; LEDs are off
	MOVLW	1<<DT1|1<<CK1		; PS/2 clock & data 
	MOVWF	TRISC                   ; are inputs
; PSP configuration
	BSF	PMCONH,PMPEN		; Configure PSP in addressable mode
	MOVLW	1<<IRQM1|1<<IRQM0|1<<MODE0
	MOVWF	PMMODEH
; Timer 0 configuration
	MOVLW	1<<T0PS2|1<<T0PS1|1<<T0PS0
	MOVWF	T0CON			; Timer 0 prescaler is 1:256
; USB configuration
	MOVLW	1<<UPUEN|1<<FSEN	; Full speed, on-chip transreceiver,
	MOVWF	UCFG			; pull-up enabled, no ping-pong
; Attach USB
	RCALL	USB_SE			; Attach USB
	CLRF	DV_STA			; Signal: Server not connected!
	CLRF	INTCON			; Clear Interrupt flags
; Wait for connection to server
SRV_LP	RCALL	POLL_E
	BTFSS	DV_STA,DV_SRV		; Server connected?
	BRA	SRV_LP			; Not yet!
; The server is connected. Wait for 50ms
	RCALL	TIM_50
C_LOOP	BTFSS	INTCON,TMR0IF
	BRA	C_LOOP
; Send a data request to the server
B_LOOP	MOVLW	SD_IDF
	MOVWF	CMD_CD
	CLRF	BLN_LO
	CLRF	BLN_HI
	RCALL	USB_TX
T_LOOP	RCALL	POLL_E			; Transmission complete?
	BNC	T_LOOP			; Not yet!
	RCALL	USB_RX			; Wait for data from server
R_LOOP	RCALL	POLL_E			; Receipt complete?
	BNC	R_LOOP			; Not yet!
; Check if a block of 1024 bytes was received
	MOVLW	high .1024
	CPFSEQ	BLN_HI
	BRA	R_LOOP			; No more data. Loop indefinitely
; A data block was received
	CPFSLT	CMD_CD			; The command code field holds
	BRA	ST_LIN			; the destination page number,
	BRA	R_LOOP			; multiplied by 2. 
;
; Program a flash memory block of 1024 bytes
;
ST_LIN	CLRF	TBLPTRU			; Set the destination pointer	
	RLCF	CMD_CD,W		; to an address between:
	RLCF	TBLPTRU			; 0x00800 and 0x1fc00
	ANDLW	high -.1024
	MOVWF	TBLPTRH
	CLRF	TBLPTRL	
	BCF	PORTC,C_NSDC		; Switch on the SDC LED
; Erase a flash memory block of 1024 bytes
	BSF	EECON1,FREE		; Enable flash memory erase
	RCALL	WR_MEM			; Erase block
	TBLRD*-				; Adjust destination address
	LFSR	FSR0,DATA_B		; Set source address
; Write a flash memory block of 1024 bytes
	MOVLW	.16
	MOVWF	CM_CNT			; There are 16 lines
NX_LIN	MOVLW 	.64			; of 64 bytes each
	MOVWF	RP_CNT
WR_LIN	MOVFF   POSTINC0,TABLAT
	TBLWT+* 
	DECFSZ	RP_CNT			; Line done?
	BRA	WR_LIN			; Not yet!
	RCALL	WR_MEM			; Write line
	DECFSZ	CM_CNT			; All lines done?
	BRA	NX_LIN			; Not yet!
; The 1024 bytes block has been written
	BSF	PORTC,C_NSDC		; Switch off the SDC LED
	BRA	B_LOOP			; Loop back

; Write to flash memory
WR_MEM	BSF 	EECON1,WREN		; Enable write to flash memory
	MOVLW	0x55
	MOVWF	EECON2
	MOVLW	0xAA
	MOVWF	EECON2
	BSF	EECON1,WR		; Start erase operation
	NOP
	BCF 	EECON1,WREN		; Disable write to flash memory
	RETURN

;------------------------------------------------------------------------------
; USB SUBROUTINES

; USB Interrupt Service Routine
USB_IS	SETF	EP_OFS			; Clear the last EP offset
	BTFSS	UCON,SUSPND		; USB in Suspend Mode?
	BRA	TS_IDL			; No!
; USB in Suspend Mode
; Only Bus Activity can generate an interrupt (DS39632e 17.5.1)
; Bus Activity interrupt
	BCF	UCON,SUSPND		; End USB Suspend Mode
	BCF	UIE,ACTVIE		; Disable Bus Activity Interrupt
	BRA	WT_SE0			; Return
TS_IDL	BTFSS	UIR,IDLEIF		; Idle Detect Interrupt?
	BRA	TS_RST			; No!
; Idle Detect interrupt
	BSF	UCON,SUSPND		; Enter the USB Suspend Mode
	BSF	UIE,ACTVIE		; Enable Bus Activity Interrupt
	BRA	USB_CE			; Return
TS_RST	BTFSS	UIR,URSTIF		; Reset interrupt?
	BRA	TS_TRN			; No!
; Reset interrupt
; Empty the USTAT FIFO. If the stack is not empty it takes 5 instruction
; cycles after clearing UIR for the TRNIF flag to be re-asserted
CL_UIR	CLRF	UIR			; Clear all USB event flags and
	NOP				; wait for 5 instruction cycles
	NOP
	NOP
	NOP
	NOP
	BTFSC	UIR,TRNIF		; Is the USTAT FIFO empty?
	BRA	CL_UIR			; Not yet!
	RCALL	EP_DIS			; Disable all endpoints
	tab_ad	BD_EP0			; Initialize the EP0 buffer
	RCALL	TAB_NW			; descriptor
	MOVLW	1<<EPHSHK|1<<EPOUTEN|1<<EPINEN
	MOVWF	UEP0			; Enable EP0 for control
	BCF	UCON,PKTDIS		; Enable packet processing
	MOVLW	1<<SB_DEF|1<<SB_SPW	; Signal: Default State!
	MOVWF	USB_ST
TS_TRN	BTFSS	UIR,TRNIF		; Transfer Complete interrupt?
NO_TRN	BRA	USB_CF			; No!
; Transfer Complete interrupt
; Identify the active endpoint
	MOVF	USTAT,W
	ANDLW	1<<ENDP3|1<<ENDP2|1<<ENDP1|1<<ENDP0|1<<DIR
	MOVWF	EP_OFS			; Offset of endpoint in BDT
	BCF	UIR,TRNIF		; Clear transfer complete flag
	LFSR	FSR1,BDT0_O		; Point FSR1 to BDT
	BTG	PLUSW1,DTS		; Toggle the data sync bit.
	ANDLW	1<<ENDP3|1<<ENDP2|1<<ENDP1|1<<ENDP0
	BNZ	NO_TRN			; Not a Control Transfer!
; Control Transfer Complete interrupt
	banksel	BDT0_O
	BTFSS	EP_OFS,DIR		; Control IN Transfer?
	BRA	CK_STU			; No!
; Control IN Transfer Complete interrupt
	BTFSC	RQ_TYP,RT_DIR		; Control Read? 
	BRA	CDT_IN			; Yes!
; Status Stage of a Control Write or Non-Data Control Transfer
; Only 'Set Address' requests need to be handled
	MOVLW	R_SADR
	CPFSEQ	RQ_NUM			; Set Address Request?
	BRA	TS_TRN			; No!
; Status Stage of a 'Set Address' Setup Request
	MOVFF	RQ_VAL,UADDR		; Copy the new address
	BRA	TS_TRN
; Control Read Transfer Data Stage
; All implemented requests except for 'Get Descriptor' need to respond
; with only one (short) IN packet. After completion of the corresponding 
; transfer, returning to this point results in the EP0 IN endpoint being
; stalled, which prevents the host from requesting any further data
; The 'Get Descriptor' request uses this entry point repeatedly, after
; each completed transfer of a packet, until all required data has been
; sent, following which it also stalls the endpoint.
CDT_IN	LFSR	FSR1,BUF0_I		; Initialize the buffer address
; Check if all requested data has been transmitted
	MOVF	EP0I_L,W		; Size of the transmitted packet
	SUBWF	RQ_LNL			; Subtract the size of the sent 
	MOVLW	0			; packet from the requested
	SUBWFB	RQ_LNH			; remaining length
	BNC	RQ_ERR			; This should never happen!
	BNZ	CDI_SZ
	CPFSEQ	RQ_LNL			; Any more data requested?
	BRA	CDI_SZ			; Yes!
; All data has been sent
; If the last packet was full-length, send a zero-length packet
; or else stall the endpoint
	BTFSS	EP0I_L,PA_CTL		; Last packet was full-length?
	BRA	RQ_ERR			; No! Stall EP0 IN
; Establish the size of the packet to be sent
CDI_SZ	MOVLW	-(1<<PA_CTL)		; Clear Z if there is enough data
	ANDWF	RQ_LNL,W		; for a full-length packet
	IORWF	RQ_LNH,W
	MOVLW	1<<PA_CTL		; Full-length
	BNZ	CDI_FL
	MOVF	RQ_LNL,W		; Short packet
CDI_FL	MOVWF	EP0I_L			; Set the packet size
	BZ	CDI_ST			; Zero-length packet
; Copy the next (WREG) descriptor bytes from the table to the EP0 IN buffer
	RCALL	TAB_RD			; Copy (WREG) bytes to the buffer
	SUBWF	EP0I_L			; Reduce the packet size by the
	BNZ	CDI_ST			; number of bytes not available
; If no more data is available set the remaining response length to zero
CDI_ZR	CLRF	RQ_LNL			; More data requested than
	CLRF	RQ_LNH			; available!
; Pass control of the endpoint to SIE
CDI_ST	MOVLW	1<<UOWN|1<<DTSEN
	BTFSC	BDT0_I,DTS		; Copy the DTS bit
	BSF	WREG,DTS
CIN_EX	MOVWF	BDT0_I			; Pass control of EP0 IN to SIE
; Enable Packet Processing because this exit point is also used by
; every SETUP transfer
	BCF	UCON,PKTDIS		; Enable Packet Processing
	BRA	TS_TRN			; Done!
; Differentiate between an OUT and a SETUP packet
CK_STU	MOVF	BDT0_O,W		; Get the Token PID
	ANDLW	1<<PID3|1<<PID2|1<<PID1|1<<PID0
	XORLW	PID_SU<<PID0		; SETUP Token?
	BZ	SET_UP			; Yes!
; Control OUT Transfer Complete interrupt
; Only 'Set Line Coding' Data OUT packets need to be processed
	MOVLW	C_SLCO
	CPFSEQ	RQ_NUM			; Set Line Coding?
	BRA	CDO_ST			; No!
; Copy the received Line Coding Data
	copy_b	BUF0_O,LN_COD,LC_END-LN_COD
; Prepare EP0 OUT for the next transfer
CDO_ST	RCALL	EP0O_A			; Re-arm EP0 OUT
	BRA	TS_TRN
; SETUP Transfer Complete interrupt
; Copy the Request Header to the access RAM
SET_UP	copy_b	BUF0_O,RQ_TYP,8
	MOVLW	1<<DTS			; Reclaim control and
	MOVWF	BDT0_O			; set DATA1 for both EP0
	MOVWF	BDT0_I			; OUT and IN endpoints
; Setting EP0 OUT to DATA1 is necessary for both Control Read and
; Control Write transfers. Non-Data Control Transfers are not 
; affected while the next Setup Token will be accepted irrespective
; of the DTS bit state ( DS39632e Register 17-5)
	RCALL	EP0O_A			; Re-arm EP0 OUT and set the
	CLRF	EP0I_L			; EP0 IN packet size to 0
	CLRF	BUF0_I			; Clear the first response byte
; Differentiate between Setup Request types
	BTFSC	RQ_TYP,RT_CLA		; Class request?
	BRA	CLA_RQ			; Yes!
	BTFSS	RQ_TYP,RT_VEN		; Standard request?
	BRA	STD_RQ			; Yes!
; Vendor Setup Requests are not supported
EP_HLT	MOVLW	1<<UOWN|1<<BSTALL	; Stall EP0
	MOVWF	BDT0_O			; Stall EP0 OUT
	BRA	CIN_EX			; Stall EP0 IN
; Check the Request Recipient: Device, Interface or Endpoint
CK_RQR	BTFSC	RQ_TYP,RT_ITF		; Interface?
	BRA	CKR_OR			; Yes!
	BTFSS	RQ_TYP,RT_ENP		; Device?
; Device Recipient: nothing to check	  Yes!
	RETURN
; Endpoint Recipient
	RLNCF	RQ_IDX,W		; Point to the status byte
	RLNCF	WREG			; of the endpoint descriptor
	RLNCF	WREG			; 
	MOVWF	FSR1L
	MOVLW	(1<<ENDP0)*(EP_MAX+1)	; Lowest invalid offset
	BTFSS	USB_ST,SB_CFG		; If not in Configured State
	MOVLW	1<<ENDP0		; only EP0 is valid
	CPFSLT	FSR1L			; Valid endpoint index?
	BRA	CKR_ER			; No! Signal: Request Error!
	MOVLW	(1<<ENDP0)*EP_CTL	; The Control EP OUT is not
	CPFSEQ	FSR1L			; a valid endpoint
	RETURN	
	BRA	CKR_ER			; Signal: Request Error!
; The recipient must be an interface, otherwise signal: Request Error!
CKR_IF	BTFSS	RQ_TYP,RT_ITF		; Interface?
	BRA	CKR_ER			; No! Signal: Request Error!
; Setup Request Recipient of type 'Other' is not supported
CKR_OR	BTFSC	RQ_TYP,RT_ENP		; Other?
	BRA	CKR_ER			; Signal: Request Error!
; Interface Recipient
	BTFSS	USB_ST,SB_CFG		; Configured state?
	BRA	CKR_ER			; No! Signal: Request Error!
	MOVLW	IF_MAX			; Maximum interface number
	CPFSGT	RQ_IDX			; Valid interface number?
	RETURN				; Yes!
CKR_ER	POP				; Discard the return address 
; Request Error
RQ_ERR	MOVLW	1<<UOWN|1<<BSTALL	; Request Error action:
	BRA	CIN_EX			; Stall EP0 IN
; Communication Class Requests
CLA_RQ	RCALL	CKR_IF			; Recipient must be an interface
	TSTFSZ	RQ_IDX			; Recipient must be the
	BRA	RQ_ERR			; CC Control Interface (0)
	MOVF	RQ_NUM,W		; Request Code to WREG
	XORLW	C_SECO			; Send Encapsulated Command Request?
; Send Encapsulated Command
; Accept the Request but discard the received data
	BZ	CDI_ST			; Yes! Prepare Status
	XORLW	C_SECO^C_GERP		; Get Encapsulated Response Request?
; Get Encapsulated Response
; Clear the requested response length and send a zero-length packet
	BZ	CDI_ZR			; Yes!
	XORLW	C_GERP^C_SLCO		; Set Line Coding Request?
; Set Line Coding
	BZ	CDI_ST			; Yes! Prepare Status
	XORLW	C_SLCO^C_GLCO		; Get Line Coding Request?
	BNZ	TS_SLS			; No!
; Get Line Coding
; Copy the Line Coding data to the EP0 IN buffer
	copy_b	LN_COD,BUF0_I,LC_END-LN_COD
	BRA	IN_SHO			; Send the Line Coding data
TS_SLS	XORLW	C_GLCO^C_SCLS		; Set Control Line State Request?
; No other requests are supported
NO_OTH	BNZ	EP_HLT			; Stall EP0
; Set Control Line State
	BCF	USB_ST,SB_DTR		; Copy DTR bit to
	BTFSC	RQ_VAL,CS_DTR		; USB state
	BSF	USB_ST,SB_DTR
	BRA	IN_SHO			; Send Status
; Standard SETUP Requests
STD_RQ	MOVF	RQ_NUM,W		; Request Code to WREG
	XORLW	R_SADR			; Set Address Request?
; Set Address
	BZ	IN_SHO			; Yes! Send Status
	XORLW	R_SADR^R_GDES		; Get Descriptor Request?
	BNZ	TS_SCF			; No!
; Get Descriptor
	MOVLW	DT_ITF			; Only supported types: Device,
	CPFSLT	RQ_DTY			; Configuration and String
	BRA	RQ_ERR			; Signal: Request Error!
	MOVLW	DT_CFG			; Configuration Descriptor	
	CPFSEQ	RQ_DTY			; requested?
	BRA	GT_DES			; No!
	MOVLW	CF_MAX			; If yes, check the index
	CPFSLT	RQ_VAL			; Valid Configuration index?
	BRA	RQ_ERR			; No! Signal: Request Error!
GT_DES	tab_ad	(DE_PNT-4)		; Index into the table DE_PNT
	MOVF	RQ_DTY,W		; containing the descriptor
	ADDWF	RQ_VAL,W		; addresses
	RCALL	TB_IND			; Get the descriptor's address
	BRA	CDI_SZ			; and send the first packet
TS_SCF	XORLW	R_GDES^R_SCFG		; Set Configuration Request?
	BNZ	TS_GCF			; No!
; Set Configuration
	MOVLW	CF_MAX+1		; Test the Configuration number
	CPFSLT	RQ_VAL			; If it is invalid,
	BRA	RQ_ERR			; signal: Request Error!
	CLRF	WREG
	MOVFF	WREG,UEP0+EP_CTL	; Disable the CC Control EP
	MOVFF	WREG,UEP0+EP_DAT	; Disable the CC Data EP
	TSTFSZ	RQ_VAL			; Configuration number >0 ?
	BRA	SCF_CO			; Yes!
; Go back to Addressed Sate
	BCF	USB_ST,SB_CFG		; Signal: Not Configured State!
	BRA	IN_SHO			; Send Status
; Advance to Configured State
SCF_CO	tab_ad	BD_CTL			; Initialize the CC Control
	RCALL	TAB_NW			; EP Buffer Descriptor
	MOVLW	1<<EPHSHK|1<<EPINEN|1<<EPCONDIS
	MOVFF	WREG,UEP0+EP_CTL	; Enable the CC Control EP
	RCALL	TAB_NW			; Initialize the CC Data EP
	MOVLW	1<<EPHSHK|1<<EPOUTEN|1<<EPINEN|1<<EPCONDIS
	MOVFF	WREG,UEP0+EP_DAT	; Enable the CC Data EP
	RCALL	TAB_NW			; Initialize the Notification
	RCALL	TAB_NW			; Header and Line Coding Data
	BSF	USB_ST,SB_CFG		; Signal: Configured State!
	BRA	IN_SHO			; Send Status
TS_GCF	XORLW	R_SCFG^R_GCFG		; Get Configuration Request?
	BNZ	TS_GST			; No!
; Get Configuration
	BTFSC	USB_ST,SB_CFG		; Configured State?
	MOVLW	CF_MAX			; Yes! Return the configuration
IN_VAL	MOVWF	BUF0_I			; number or else return 0
; Send one short packet
IN_SHO	MOVFF	RQ_LNL,EP0I_L		; Requested packet size
	BRA	CDI_ST			; Send packet
TS_GST	XORLW	R_GCFG^R_GSTA		; Get Status Request?
	BNZ	TS_CFT			; No!
; Get Status	
	RCALL	CK_RQR			; Check Recipient
	CLRF	BUF0_I+1		; Clear the second response byte
	BTFSC	RQ_TYP,RT_ITF		; Interface?
	BRA	IN_SHO			; Yes! Nothing to report
	BTFSC	RQ_TYP,RT_ENP		; Endpoint?
	BRA	GST_EP			; Yes!
; Get Device Status
	MOVF	USB_ST,W		; Copy the 'Self Powered'
	ANDLW	1<<SB_RWU|1<<SB_SPW	; and 'Remote Wake-up' bits
	BRA	IN_VAL			; to the EP0 IN buffer
; Get Endpoint status
GST_EP	BTFSC	INDF1,BSTALL		; If the endpoint is stalled,
	BSF	BUF0_I,GS_HLT		; set the HLT bit
	BRA	IN_SHO			; Send the response
TS_CFT	XORLW	R_GSTA^R_CFTR		; Clear Feature Request?
	BNZ	TS_SFT			; No!
; Clear / Set Feature
CHG_FT	RCALL	CK_RQR			; Check Recipient
	MOVF	RQ_NUM,W		; Set Z for 'Clear' and
	XORLW	R_CFTR			; clear Z for 'Set'
	BTFSC	RQ_TYP,RT_ITF		; Interface?
	BRA	RQ_ERR			; Yes! Signal: Request Error!
	BTFSC	RQ_TYP,RT_ENP		; Endpoint?
	BRA	CSF_EP			; Yes!
; Clear / Set Remote Wake-up
	MOVLW	DF_RWU			; Device Remote Wake-up Feature?
	CPFSEQ	RQ_VAL
	BRA	RQ_ERR			; No! Signal: Request Error!
	BCF	USB_ST,SB_RWU		; Clear Remote Wake-up bit
	BZ	IN_SHO			; Set?
	BSF	USB_ST,SB_RWU		; Set Remote Wake-up bit
	BRA	IN_SHO			; Send status
; Clear / Set Endpoint HALT
CSF_EP	MOVLW	EF_HLT			; Endpoint Halt Feature?
	CPFSEQ	RQ_VAL
	BRA	RQ_ERR			; No! Signal: Request Error!
CSF_HT	RCALL	CHG_HT			; Change HALT and clear DTS bit
	BRA	IN_SHO			; Send status
TS_SFT	XORLW	R_CFTR^R_SFTR		; Set Feature Request?
; Set Feature
	BZ	CHG_FT			; Yes!
	XORLW	R_SFTR^R_GITF		; Get Interface Request?
	BNZ	TS_SIF			; No!
; Get Interface
	RCALL	CKR_IF			; Recipient must be an interface
	BRA	IN_SHO			; Return 0
TS_SIF	XORLW	R_GITF^R_SITF		; Set Interface Request?
	BNZ	NO_OTH			; No other requests are supported
; Set Interface
	RCALL	CKR_IF			; Recipient must be an interface
	TSTFSZ	RQ_VAL			; Accept only the alternate
	BRA	RQ_ERR			; interface 0
	MOVLW	(1<<ENDP0)*EP_CTL	; CC Control Interface (0)
	TSTFSZ	RQ_IDX			; CC Data Interface (1)?
	MOVLW	(1<<ENDP0)*EP_DAT	; Yes!
	MOVWF	FSR1L			; Point to corresponding OUT EP
	CLRF	WREG			; Clear! ( Set Z )
	RCALL	CHG_HT			; Clear the OUT EP HALT and DTS
	BSF	FSR1L,DIR		; Switch to the IN EP
	BRA	CSF_HT			; Clear the IN EP HALT and DTS
; Disable all used USB Endpoints and reset the Device Address
EP_DIS	CLRF	UEP0			; Disable EP0
	CLRF	UEP0+EP_CTL		; Disable CC Control EP
	CLRF	UEP0+EP_DAT		; Disable CC Data EP
	CLRF	UADDR			; Set default address
	RETURN
; Arm EP0 OUT
EP0O_A	MOVLW	1<<PA_CTL		; Set the packet length to the
	MOVWF	EP0O_L			; maximum declared size
	MOVLW	1<<UOWN|1<<DTSEN
	BTFSC	BDT0_O,DTS		; Copy the DTS bit	
	BSF	WREG,DTS
	MOVWF	BDT0_O			; Pass control of EP0 OUT to SIE
	RETURN
; Clear ( Z==1 ) / Set ( Z==0 ) the endpoint HALT and clear the DTS bit
CHG_HT	MOVLW	1<<DTSEN		; Clear the HALT and DTS bits
	BZ	CHT_UO			; Set?	
	BSF	WREG,BSTALL		; Yes! Set the HALT bit
CHT_UO	BTFSC	INDF1,UOWN		; Copy the UOWN bit
	BSF	WREG,UOWN
	MOVWF	INDF1
	RETURN

; Poll relevant events during self-programming
POLL_E	BTFSC	INTCON,RBIF		; PORTB change?
	RCALL	INT_CG			; Handle event
	BTFSC	INTCON,TMR0IF		; Timer 0 time-out?
	RCALL	INT_TM			; Handle event
	BCF	STATUS,C		; Signal: USB transfer not done!
	BTFSS	UCON,USBEN		; USB enabled?
	RETURN				; No!
	BTFSS	PIR2,USBIF		; USB event?
	RETURN				; No!
;
; USB Interrupt
;
; Call the USB Interrupt Service Routine to handle all USB events
; except for the CC Data Transfers Complete interrupts	
INT_SB	RCALL	USB_IS			; Service USB interrupt
	banksel DV_STA
; Handle CC Data Transfers Complete interrupts
; Return the Carry flag set on completion
	BTFSS	USB_ST,SB_CFG		; USB in 'Configured State'?
	BRA	NO_DTR			; No!
; FSR0 was initialized at the start of the USB Communication Stage with
; the address of the corresponding CC Data IN / OUT endpoint descriptor
; Only transfers of that same endpoint are handled until the stage ends
	MOVF	FSR0L,W			; Expected EP offset
	CPFSEQ	EP_OFS			; Does it match?
	BRA	CK_DTR			; No! 
	INCF	FSR0L			; Point to the packet size
; The first 3 bytes of the first packet received are the Communication
; Header, containing the size of an eventual Data Block
	BTFSS	COM_ST,CO_HED		; Header Block needed?
	BRA	USB_CO			; No!
	MOVF	INDF0,W			; Size of the received packet
	BZ	UEP_SZ			; Allow for zero-length packets
	SUBLW	DATA_B-CMD_CD-1		; The first received packet must
	BC	COM_ER			; be at least 4 bytes long
	RCALL	TOT_LN			; Get the total length 
; Move the buffer pointer after the last byte transferred
USB_CO	MOVF	POSTINC0,W		; Update the buffer address
	ADDWF	POSTINC0		; by adding the length of the
	MOVLW	0			; current transfer.
	ADDWFC	POSTDEC0
	DECF	FSR0L			; Point again to the packet size
; Check if there is more data to be transferred
	MOVF	INDF0,W			; Subtract the packet size of
	SUBWF	CNT_LO			; the current transfer from the
	MOVLW	0			; remaining length.
	SUBWFB	CNT_HI
	BNC	COM_ER			; This should never happen!
	BNZ	UEP_SZ
	CPFSEQ	CNT_LO			; If there is still data to be
	BRA	UEP_SZ			; transferred, arm the endpoint
; All data has been transferred. If the last transmitted packet was 
; full-sized, send a zero-length packet.
	BSF	STATUS,C		; Signal: Transfer done!	
	BTFSC	EP_OFS,DIR		; IN transfer ?
	BTFSS	INDF0,PA_DAT		; If the last packet was short,
	RETURN				; end communication stage, else
	BRA	UEP_ST			; send a zero-length packet

; Check if the server has set the DTR line
CK_DTR	BTFSS	USB_ST,SB_DTR		; DTR set?
	BRA	NO_DTR                  ; No!
	BSF	DV_STA,DV_SRV		; Signal: Server connected
	BCF	LATC,C_NSRV		; Switch on the Server LED
	BRA	USB_RT
NO_DTR	BCF	DV_STA,DV_SRV           ; Signal: Server not ready
	BSF	LATC,C_NSRV		; Switch off the Server LED
	BRA	USB_RT

; Start USB Transmission
; Point FSR0 to the last byte of the CC Data IN endpoint BDT entry 
USB_TX	LFSR	FSR0,BDT0_I+(1<<ENDP0)*EP_DAT+(1<<DIR)-1
	RCALL	TOT_LN			; Get the total length
	BRA	SRV_CK

; Start USB Reception
; Point FSR0 to the last byte of the CC Data OUT endpoint BDT entry 
USB_RX	LFSR	FSR0,BDT0_O+(1<<ENDP0)*EP_DAT+(1<<DIR)-1
	BSF	COM_ST,CO_HED		; Signal: Need Header Block!
SRV_CK	BTFSS	DV_STA,DV_SRV		; Check if the server is connected
	BRA	USB_NP			; Signal error if not
	BTFSS	UCON,SUSPND		; USB in Suspended Mode?
	BRA	UEP_AD			; No!
; If the USB is in Suspend Mode, try Remote Wake-up
	BTFSS	USB_ST,SB_RWU		; Remote Wake-up enabled?
	BRA	USB_NP			; No! Tough!
; Activate the RESUME signal for 5ms
	BSF	UCON,RESUME		; Yes!  Activate the 
	l_time	.5			; RESUME signal for 5ms
	RCALL	STA_LT
	BRA	USB_RT
; Initialize the USB buffer address
UEP_AD	BSF	COM_ST,CO_ACT		; Communication started!
	MOVLW	high CMD_CD		; Initialize the buffer
	MOVWF	POSTDEC0		; address.
	MOVLW	low CMD_CD
	MOVWF	POSTDEC0
; Set the time-out for the USB Communication Stage to 1s
	BSF	TIM_ST,TM_USB		; Signal: USB timer running!
	RCALL	TIM_1S			; Start timer
; Establish the packet size
UEP_SZ	MOVLW	-(1<<PA_DAT)
	ANDWF	CNT_LO,W		; Clear Z if there is enough data
	IORWF	CNT_HI,W		; for a full-length packet
	MOVLW	1<<PA_DAT		; Full-length
; Allow full-length for the first packet to be received
	BTFSC	COM_ST,CO_HED		; Have Header Block?
	BRA	UEP_ST			; Not yet!
	BNZ	UEP_ST			; Full packet!
	MOVF	CNT_LO,W		; Short packet!
UEP_ST	MOVWF	POSTDEC0		; Set the packet size
; Pass control of the endpoint to SIE
	MOVLW	1<<UOWN|1<<DTSEN
	BTFSC	INDF0,DTS		; Copy the DTS bit			
	BSF	WREG,DTS
	MOVWF	INDF0			; Pass control to SIE
USB_RT	BCF	STATUS,Z		; Signal: No Error!
USB_ND	BCF	STATUS,C		; Signal: Not done!
	RETURN
COM_ER	MOVLW	E_COMM			; 'Communication error'
USB_ER	BSF	STATUS,Z		; Signal: USB Error!
	BRA	USB_ND

; Calculate the combined length of header and data
TOT_LN	MOVLW	DATA_B-CMD_CD		; Length of header block
	ADDWF	BLN_LO,W		; Add to the length of the
	MOVWF	CNT_LO			; data block to get the
	MOVLW	0			; total length and copy it
	ADDWFC	BLN_HI,W		; to the byte counter
	MOVWF	CNT_HI
	BCF	COM_ST,CO_HED		; Signal: Header Block done!
	RETURN

; Attach / Detach USB
USB_SE	RCALL	EP_DIS			; Disable all Endpoints
	MOVLW	1<<SB_SPW		; Signal: Self powered!
	MOVWF	USB_ST
	BTFSC	CHG_ST,B_USEN		; USB cable connected to host?
	BRA	USB_AT			; Yes!
; Detach USB
	BCF	UCON,SUSPND		; End Suspend Mode (17.2.1)
	BCF	UCON,USBEN		; Disable USB
	BCF	PIE2,USBIE		; Disable the USB interrupt
	BRA	USB_CE
; Attach USB
USB_AT	BSF	UCON,USBEN		; Enable USB
	MOVLW	1<<IDLEIE|1<<TRNIE|1<<URSTIE
	MOVWF	UIE			; Unmask needed USB interrupts
	BSF	PIE2,USBIE		; Enable USB interrupts
WT_SE0	BTFSC	UCON,SE0		; Single Ended Zero State?
	BRA	WT_SE0			; Yes! Wait until it ends.
USB_CE	CLRF	UIR			; Clear all USB event flags
	BTFSC	UIR,ACTVIF		; Wait until the Bus Activity
	BRA	USB_CE			; event flag is stable
USB_CF	BCF	PIR2,USBIF		; Clear the USB interrupt flag
	RETURN

;------------------------------------------------------------------------------
; COMMON SUBROUTINES

;
; Copy  data from a table in the program memory.
; The table header has the following structure:
; (Word)Destination Address 
; (Byte)Data Length
; (Byte)Selector: 0 = Descriptor, >0 = Other
;
; Continue copying the same table from address (TBLPTR) if called
; from a Control IN Transfer interrupt, or otherwise copy from a new 
; descriptor table at address: (TBLPTR)
TAB_RD	BTFSC	EP_OFS,DIR		; Control IN Transfer?
	BRA	TB_TST			; Yes!
; Copy from the table at address: (TBLPTR)
TAB_NW	TBLRD*+	
	MOVFF	TABLAT,FSR1L		; Load the destination
	TBLRD*+				; address from the table
	MOVFF	TABLAT,FSR1H
	TBLRD*+				; Load the size of the
	MOVFF	TABLAT,TB_CNT		; table in the counter
; Descriptors are copied only (WREG) bytes at a time
	TBLRD*+				; Read the Table Selector
	TSTFSZ	WREG			
	BTFSC	TABLAT,TB_SEL		; Copy the entire table?
	MOVF	TB_CNT,W		; Yes!
TB_TST	TSTFSZ	TB_CNT			; End of table reached?
	BRA	TB_NXT			; Not yet!
	RETURN				; Yes!
; (WREG) returns the number of unavailable bytes
TB_NXT	TBLRD*+				; Copy one byte and decrement
	MOVFF	TABLAT,POSTINC1		; the 'available bytes' counter
	DECF	TB_CNT			; and the 'copied bytes'
	DECFSZ	WREG			; counter
	BRA	TB_TST			; All done?
	RETURN				; Yes!
;
; Index (WREG) into the table at (TBLPTR)
TB_IND	RLNCF	WREG
	RLNCF	WREG
	RCALL	ADD_TP
	TBLRD*+
	MOVFF	TABLAT,PRODL
	TBLRD*+
	MOVFF	TABLAT,PRODH
	TBLRD*+
	MOVFF	TABLAT,TBLPTRU
	MOVFF	PRODH,TBLPTRH
	MOVFF	PRODL,TBLPTRL
	RETURN
;
ADD_TP	ADDWF	TBLPTRL
	MOVLW	0
	ADDWFC	TBLPTRH
	ADDWFC	TBLPTRU
	RETURN
;
; Interrupt on change
;
INT_CG	MOVF	PORTB,W			; Get the new PORTB value
	XORWF	CHG_ST,W		; Set the changed bits
	XORWF	CHG_ST			; Change old to new value
	BCF	INTCON,RBIF		; Clear interrupt flag
	BTFSS	WREG,B_USEN		; USB power?
	BRA	SDC_CH			; No!
; USB cable (dis)connected
	BSF	TIM_ST,TM_SBD		; Signal: USB power!
TIM_50	l_time	.50			; and wait for 50ms
	BRA	STA_LT
SDC_CH	BTFSS	WREG,B_NCRD		; SD card change?
	RETURN				; No!
; SD card inserted / ejected 
	BSF	TIM_ST,TM_CRD		; Signal: SD card change!
;
; Set timer 0 to 1s
;
TIM_1S	l_time	.1000			; Start timer 0 for 1s
; Start Timer 0
STA_LT	BCF	T0CON,TMR0ON		; Stop timer 0
	BCF	INTCON,TMR0IF		; Clear interrupt flag
	MOVWF	TMR0L			; Load timing constant
	BSF	T0CON,TMR0ON		; Start timer 0
	RETURN
; 
INT_TM	BCF	T0CON,TMR0ON		; Stop timer
;
; Timer 0 interrupt for asynchronous events
;
TIM_IT	BCF	INTCON,TMR0IF		; Clear interrupt flag
	BSF	STATUS,C		; Signal: No event!
	BTFSS	UCON,RESUME		; USB resume?
	BRA	USB_TO			; No!
	BCF	UCON,RESUME		; Stop RESUME signal
	BTFSS	UCON,SUSPND		; USB still in Suspended Mode?
	BRA	UEP_AD			; No!!
USB_NP	MOVLW	E_NPRE			; Error: 'Not present'
	BRA	USB_ER
USB_TO	BTFSS	TIM_ST,TM_SBD		; USB power?
	BRA	CRD_TO			; No!
; The USB cable connection state is assessed after 50ms
	BCF	TIM_ST,TM_SBD		; Signal: Not USB power!
	RCALL	USB_SE			; Attach / detach USB
	BRA	NO_DTR
CRD_TO	BTFSS	TIM_ST,TM_CRD		; SD card change?
	BRA	SDC_TO			; No!
; The SD card state is assessed after 1s
	BCF	TIM_ST,TM_CRD		; Signal: No SD card change!
	BSF	SDC_ST,SD_DET		; Signal: Detect SD card!
	BRA	USB_RT
SDC_TO	BTFSS	TIM_ST,TM_SDC		; SD card timer running?
	RETURN				; No!
; SD card time-out
	BCF	TIM_ST,TM_SDC		; Signal: Time-out!
	BRA	USB_RT
;
; Spare self-programming block program memory
;
	fill	0xffff,0x0800-$


;------------------------------------------------------------------------------
; FIXED START ADDRESSES


;------------------------------------------------------------------------------
; MAIN PROGRAM

MAIN_C	GOTO	CONF_D			; Jump to main code

;------------------------------------------------------------------------------
; (HIGH PRIORITY) INTERRUPT

HI_INT	BTFSC	PIR1,PMPIF		; PSP interrupt?
	BRA	PSP_IT			; Yes!
	BTFSC	INTCON,TMR0IF		; Timer 0 Interrupt?
	BRA	LTM_IT			; Yes!
	BTFSC	PIR1,TMR1IF		; Timer 1 Interrupt?
	BRA	STM_IT			; Yes!
	BTFSS	INTCON,RBIF		; PORTB Interrupt on change?
	BRA	EI1_IT			; No!
; PORTB Interrupt on Change can be triggered by:
; - Connecting or disconnecting the USB cable
; - Inserting or ejecting the SD card
	RCALL	INT_CG
	RETFIE	FAST
;------------------------------------------------------------------------------
; MAIN PROGRAM (CONTINUED)

CONF_D	
	banksel	PPSCON
; Assign periferial pins
	MOVLW	.2			; Pin RA5/RP2 mapped to
	MOVWF	RPINR21			; input function SDI2
	MOVLW	.0			; Pin RA0/RP0 mapped to
	MOVWF	RPINR22			; input function SCK2
	MOVLW	.11			; Output function SCK2
	MOVWF	RPOR0			; mapped to pin RA0/RP0
	MOVLW	.10			; Output function SDO2
	MOVWF	RPOR1			; mapped to pin RA1/RP1
	MOVLW	.17			; Pin RC6/RP17 mapped to
	MOVWF	RPINR1			; input function INT1
	MOVLW	.18			; Pin RC7/RP18 mapped to
	MOVWF	RPINR2			; input function INT2
	MOVLW	.17			; Pin RC6/RP17 mapped to
	MOVWF	RPINR17			; input function CK2
	MOVLW	.18			; Pin RC7/RP18 mapped to
	MOVWF	RPINR16			; input function DT2
	MOVLW	.6			; Output function CK2
 	MOVWF   RPOR17                  ; mapped to pin RC6/RP17
	MOVLW	.7			; Output function DT2
	MOVWF   RPOR18			; mapped to pin RC7/RP18
	MOVLW	0x55
	MOVWF	EECON2
	MOVLW	0xAA
	MOVWF	EECON2
	BSF	PPSCON,IOLOCK		; Lock pin mapping
	banksel	DV_STA
; Port A configuration
	SETF	PORTA
	MOVLW	1<<RA5			; RA0 is input SDI2
	MOVWF	TRISA
; Port B configuration
	BCF	INTCON2,NOT_RBPU	; Enable PORTB pull-up
; Port C configuration
	CLRF	PORTC
	MOVLW	1<<DT1|1<<CK1		; PS/2 clock and data 
	MOVWF	TRISC                   ; are inputs
; Joystick button configuration
	BSF	TRIS_J,J_JOY4		; Joystick button is input
; External interrupt 2 configuration
	BCF	INTCON2,INTEDG2		; Interrupt on falling DT line
; PSP configuration
	BSF	PMCONH,PMPEN		; PSP in addressable mode
	MOVLW	1<<IRQM1|1<<IRQM0|1<<MODE0
	MOVWF	PMMODEH
	SETF	PSP_DO			; Clear the 'buffer empty' flag
	BSF	PIE1,PMPIE		; Unmask the PSP interrupt
; Timer 0 configuration
	MOVLW	1<<T0PS2|1<<T0PS1|1<<T0PS0
	MOVWF	T0CON			; Timer 0 prescaler is 1:256
; Timer 1 configuration
	MOVLW	1<<T1CKPS1|1<<T1CKPS0|1<<RD16
	MOVWF	T1CON			; Timer 1 prescaler is 1:8
	BSF	PIE1,TMR1IE		; Unmask the timer 1 interrupt
; SPI configuration
	MOVLW	1<<CKE			; Shift on falling edge
	MOVWF	SSP2STAT
; USB configuration
	MOVLW	1<<UPUEN|1<<FSEN	; Full speed, on-chip transreceiver,
	MOVWF	UCFG			; pull-up enabled, no ping-pong
; USART configuration
	BSF	ODCON2,U2OD		; Open drain,
	BSF	BAUDCON2,TXCKP		; clock idle high,
	BSF	TXSTA2,SYNC		; synchronous slave,
	BSF	RCSTA2,RX9		; 9-bit reception
; Enable SPI
	MOVLW	1<<SSPEN		; Enable SPI at 12MHz
	MOVWF	SSP2CON1
	RCALL	SPI_RD			; Set SDO high
; Start-up delay
	BSF	OSCCON,IDLEN		; Set Idle mode on SLEEP
	BSF	WDTCON,SWDTEN		; Start Watch-dog timer
	SLEEP				; Wait for 2 s
	NOP
	BCF	WDTCON,SWDTEN		; Stop Watch-dog timer
	BSF	PORTC,C_NSRV		; Switch off the Server LED
	BSF	PORTC,C_NSDC		; Switch off the SDC LED
	BSF	PORTC,C_NMSE		; Switch off the mouse LED
; Initialize joystick and mouse parameters
	tab_ad	JM_INI			; Initialisation table address
	RCALL	TAB_NW			; Initialise variables
; Clear interrupt-on-change mismatch
	MOVF	PORTB,W			; Clear RB port mismatch condition
  	MOVWF	CHG_ST			; Save the value of the RB port.
; Attach USB
	RCALL	USB_SE			; Attach USB
; Initialize the device status
	MOVLW	1<<DV_MSE|1<<DV_REZ|1<<MSE_BL|1<<MSE_BR
	MOVWF	DV_STA                  ; Initialize the device status
; Enable interrupts
	CLRF	PIR1			; Clear interrupt flags
	MOVLW	1<<GIE|1<<PEIE|1<<TMR0IE|1<<RBIE
	MOVWF	INTCON			; Enable needed interrupts
; Detect PS/2 mouse
	RCALL	MOUS_I			; Detect PS/2 mouse
; Start joystick timer
	RCALL	TIME_J
; Signal: Detect SD Card
	BSF	SDC_ST,SD_DET
; Main loop
M_LOOP	BTFSC	COM_ST,CO_ACT		; Communication in progress?
	BRA	M_LOOP			; Yes! Wait until it ends
	BTFSC	SDC_ST,SD_DET		; Detect SD card?
	RCALL	SDC_ID			; Yes! Detect SD card
	BTFSC	SDC_ST,SD_EXE		; SD card command pending?
	RCALL	SDC_OP			; Yes! Execute SD card command
	SLEEP				; Wait for an interrupt
	NOP				; to loop back
	BRA	M_LOOP	

;------------------------------------------------------------------------------
; HIGH PRIORITY INTERRUPT (CONTINUED)

; Timer 1 Overflow interrupt
; Used with the PS/2 mouse
STM_IT	BCF	T1CON,TMR1ON		; Stop timer 1
	BCF	PIR1,TMR1IF		; Clear interrupt flag
	BTFSS	MSE_ST,MS_PAK		; Movement packet in progress?
	RETFIE	FAST			; No!
	BCF	MSE_ST,MS_PAK		; Clear packet in progress flag
	LFSR	FSR2,MOVE_B		; Discard partial packet
	RETFIE	FAST
;
EI1_IT	BTFSS	INTCON3,INT1IE		; INT1 enabled?
	BRA	EI2_IT			; No!
	BTFSS	INTCON3,INT1IF		; INT1 interrupt?
	BRA	EI2_IT			; No!
; PS/2 begin reception interrupt
	BCF	INTCON3,INT1IE		; Disable CK interrupt
	BSF	PIE3,RC2IE		; Enable USART interrupt
	BSF	RCSTA2,CREN		; Start USART reception
	RETFIE	FAST
EI2_IT	BTFSS	INTCON3,INT2IE		; INT2 enabled?
	BRA	SRT_IT			; No!
	BTFSS	INTCON3,INT2IF		; INT2 interrupt?
	BRA	SRT_IT			; No!
; PS/2 start bit interrupt
	BCF	INTCON3,INT2IE		; Disable DT interrupt
	BCF	INTCON3,INT1IF		; Clear CK interrupt flag
	BSF	INTCON3,INT1IE		; Enable CK interrupt
	RETFIE	FAST
SRT_IT	BTFSS	PIE3,RC2IE		; USART interrupt enabled? 
	BRA	SDC_IT			; No!
	BTFSS	PIR3,RC2IF		; USART interrupt?
	BRA	SDC_IT			; No!
; USART reception interrupt
	BCF	PIE3,RC2IE		; Disable USART interrupt
	BCF	RCSTA2,CREN		; Stop reception
	MOVF	RCREG2,W		; Get received byte
	MOVWF	POSTINC2		; Store received byte
; Check parity
	MOVWF	MS_POS			; Copy received byte	
	SWAPF	MS_POS,W		; Calculate: abcdefgh^
	XORWF	MS_POS			;            efghabcd^
	RRNCF	MS_POS,W		;            habcdefg^
	XORWF	MS_POS,W                ;            defghabc
	BTFSC	WREG,2			; Bit0^Bit2 gives the 
	INCF	WREG			; parity of the byte
	XORWF	RCSTA2,W		; XOR with Bit9
	BTFSS	WREG,RX9D		; If the parity is not odd,
	CLRF	MOVE_B			; Signal: Parity error!
	BCF	INTCON3,INT2IF		; Clear DT interrupt flag
	BSF	INTCON3,INT2IE		; Enable DT interrupt
	MOVLW	low MS_POS
	CPFSEQ	FSR2L			; Movement packet done?
	BRA	WT_PAK			; Not yet!
; A movement packet has been received
	BCF	MSE_ST,MS_PAK		; Clear packet in progress flag
; Update the mouse coordinates
	MOVFF	MWIN_Y,POSTDEC2		; Mouse window height
	MOVF	MPOS_Y,W		; Y-coordinate
	MOVF	MOVE_B			; Set N-flag for Y-overflow
	RCALL	ADD_MV			; Add Y-movement
	MOVFF	MWIN_X,MS_POS		; Mouse window width
	MOVF	MPOS_X,W		; X-coordinate
	RLNCF	MOVE_B			; Set N-flag for X-overflow
	RCALL	ADD_MV			; Add X-movement
	RRNCF	MOVE_B			; Discard the packet if it had
	BZ	PAR_ER			; parity errors
	MOVFF	MOVE_X,MPOS_X		; Save new X-coordinate
	MOVFF	MOVE_Y,MPOS_Y		; Save new Y-coordinate
	BTFSS	MSE_ST,MS_ENB		; Mouse enabled?
	BRA	MOUS_E			; No! Any buttons pressed?
; If enabled and both button are pressed, disable the mouse
	BTFSS	MOVE_B,MV_MBL		; Mouse left button pressed?
	BRA	MOUS_S			; No!
	BTFSC	MOVE_B,MV_MBR		; Mouse right button also pressed?
	BRA	MOUS_D			; Yes! Disable the mouse
MOUS_S	BTFSC	MSE_ST,MS_JOY		; Mouse as joystick?
PAR_ER	BRA	STM_IT			; Yes!
; Update the Kempston mouse ports
	BSF	MOUS_B,MSE_BR		; Update the right button
	BTFSC	MOVE_B,MV_MBR		; bit of the Kempston
	BCF	MOUS_B,MSE_BR		; mouse buttons port
	BSF	MOUS_B,MSE_BL		; Update the left button
	BTFSC	MOVE_B,MV_MBL		; bit of the Kempston
	BCF	MOUS_B,MSE_BL		; mouse buttons port
	MOVFF	MPOS_Y,MOUS_Y		; Update the Y-coordinate
	MOVFF	MPOS_X,MOUS_X		; Update the X-coordinate
	BRA	STM_IT			; Stop timer 1
; Expect the next byte of the packet within 7ms
WT_PAK	BSF	MSE_ST,MS_PAK		; Signal: packet in progress
	s_time	.7000			; Set time-out to 7ms
	RCALL	STA_ST			; Start timer 1
	RETFIE	FAST
SDC_IT	BTFSS	PIE3,SSP2IE		; SPI interrupt enabled?
	BRA	USB_IT			; No!
	BTFSS	PIR3,SSP2IF		; SPI interrupt?
	BRA	USB_IT			; No!
; SPI Interrupt signals completion of a SD card operation
	BCF	PIE3,SSP2IE		; Disable the SPI interrupt
	BRA	STRT_T			; Send response
USB_IT	BTFSS	PIE2,USBIE		; USB interrupt enabled?
	RETFIE	FAST			; No!
	BTFSS	PIR2,USBIF		; USB interrupt?
	RETFIE	FAST			; No!
	RCALL	INT_SB			; Service the USB interrupt
	BC	END_CO			; CC data transfer done
	BZ	ERR_RT			; Error
        RETFIE	FAST
; Timer 0 Overflow Interrupt :
; - USB RESUME signal timing
; - USB cable (dis)connect de-bounce
; - SD card insert/eject de-bounce
; - SD card time-out
; - Joystick read
; - USB Communication Stage time-out
; - PSP Communication Stage time-out
; The events asynchronous to the communication state machine let the
; timer running for another 1.4s in order not to disrupt the joystick 
; reading cycle and the PSP or USB time-outs
LTM_IT	RCALL	TIM_IT			; Handle asynchronous events.
	BC	STOP_T			; No synchronous event.
	BZ	ERR_RT			; Error
	RETFIE	FAST
; Stop the timer
STOP_T	BCF	T0CON,TMR0ON		; Stop timer
	BTFSC	TIM_ST,TM_PSP		; PSP time-out error?
; PSP Communication Stage time-out occurs after 1s
	BRA	PSP_ER			; Yes!
	BTFSC	TIM_ST,TM_USB		; USB time-out error?
; USB Communication Stage time-out occurs after 1s
	BRA	CO_ERR			; Yes! 
	BTFSC	TIM_ST,TM_JOY		; Joystick read?
; The joystick is read every 100 - 1000ms
	BRA	JOY_RD			; Yes!
	RETFIE	FAST			; No other timer events!
; PSP Interrupt
PSP_IT	BCF	PIR1,PMPIF		; Clear the interrupt flag		
	BTFSC	PMSTATL,OB3E		; Data output?
	BRA	TST_CN			; Yes!
	BTFSS	PMSTATH,IB3F		; Data input?
	RETFIE	FAST			; No!
; One byte has been received from host
	BTFSC	COM_ST,CO_ACT		; New Communication Cycle?
	BRA	CON_CO			; No!
; Start of a new Communication Cycle
; Joystick and mouse are disabled during the cycle
	BTFSC	DV_STA,DV_MSE		; Mouse connected? If yes
	RCALL	STO_RC			; disable PS/2 reception
	MOVLW	DATA_B-CMD_CD-1		; Header Block length - 1
; Start of a new Communication Stage
STA_CO	BSF	COM_ST,CO_ACT		; Communication started!
	BSF	COM_ST,CO_HED		; Signal: Header Block needed!
	MOVWF	CNT_LO			; Copy the Header Block length
	CLRF	CNT_HI			; to the byte counter
	LFSR	FSR0,CMD_CD		; Reset the buffer address
	CLRF	TIM_ST			; Clear all timer events
	BSF	TIM_ST,TM_PSP		; Signal: PSP timer running
	RCALL	TIM_1S			; Start timer
; One byte has been transferred over the PSP
CON_CO	BTFSS	COM_ST,CO_TRA		; Transmission in progress?
; One byte has been received over the PSP
	MOVFF	PSP_DI,POSTINC0	        ; Copy received byte to buffer
; Test if the byte counter has reached 0
TST_CN	TSTFSZ	CNT_LO			; Test if the byte counter has
	BRA	DEC_LO			; reached 0.
	TSTFSZ	CNT_HI
	BRA	DEC_HI
; All data has been transferred. Check if there is a Data Block, of up
; to 512 bytes, to be transferred after the 3-byte Header Block
	BTFSS	COM_ST,CO_HED		; Data Block finished?
	BRA	END_CO			; Yes!
; Prepare to transfer the Data Block
	BCF	COM_ST,CO_HED		; Signal: Header Block done!
	MOVFF	BLN_LO,CNT_LO		; Copy the Data Block size
	MOVFF	BLN_HI,CNT_HI		; to the byte counter and
	BRA	TST_CN			; test again
; Communication Error
; Send an error message to the host
CO_ERR	MOVLW	E_COMM			; Communication error code
ERR_RT	MOVWF	CMD_CD			; Set the error code
	CLRF	BLN_LO			; Set the Data Block size
	CLRF	BLN_HI			; to 0
	BSF	COM_ST,CO_SER           ; Parallel transmission
	BCF	COM_ST,CO_TRA		; follows!
; End of Communication Stage
END_CO	BCF	COM_ST,CO_ACT		; End of Communication Stage!
	CLRF	TIM_ST			; Clear all timer events
; Prepare the next Communication Stage
	BTG	COM_ST,CO_TRA		; Toggle Transmit/Receive flag
	BTFSS	COM_ST,CO_TRA		; Receive to follow?
	BRA	PREP_R			; Yes!
; A Transmit Stage is triggered immediately after each Receive Stage,
; within the same interrupt
	BTFSC	COM_ST,CO_SER		; Reception was from host?
	BRA	PREP_T			; No!
	MOVLW	NS_ATC			; Is the received code a
	CPFSLT	CMD_CD			; Host Command?
	BRA	TS_HOC			; Yes!
; Prepare for Transmission
PREP_T	BTG	COM_ST,CO_SER		; Toggle Serial/Parallel flag
	BTFSS	COM_ST,CO_SER		; Serial transmission?
	BRA	STRT_T			; No!
	RCALL	USB_TX			; Yes!
	BZ	ERR_RT
	RETFIE	FAST
STRT_T	MOVLW	DATA_B-CMD_CD		; Size of the Header Block
	BRA	STA_CO			; Start transmission to host
; Execute a host command and send the response
TS_HOC	MOVLW	SD_SWA			; 'Set Write Address' code
	CPFSGT	CMD_CD			; SD card command?
	BRA	EX_HOC			; No!
	BSF	SDC_ST,SD_EXE		; Signal: Execute SDC command!
	RETFIE	FAST
; Host commands are executed immediately, within the same interrupt
EX_HOC	RCALL	HO_CMD			; Execute the command
	CLRF	CMD_CD			; Clear the error code
	BRA	STRT_T			; Send response to host
; Decrement the byte counter
DEC_HI	DECF	CNT_HI			; Decrement the byte counter
DEC_LO	DECF	CNT_LO
	BTFSS	COM_ST,CO_TRA		; Receive?
; One byte has already been received and copied to the buffer
	RETFIE	FAST			; Done!
; Transmit one byte over the PSP
	MOVFF	POSTINC0,PSP_DO  	; Byte to transmit
	BSF	DV_STA,DV_DAT		; Signal: Data ready!
	RETFIE	FAST
; PSP error
PSP_ER	CLRF	TIM_ST			; Clear all timer flags
	CLRF	COM_ST			; Clear communication status
; Prepare for reception
PREP_R	BTFSS	COM_ST,CO_SER		; Serial receive?
	BRA	CY_END			; No!
	RCALL	USB_RX			; Yes!
	BZ	ERR_RT
	RETFIE	FAST
CY_END	SETF	PSP_DO			; Clear the 'buffer empty' flag
	BCF	DV_STA,DV_DAT		; Signal: Data not ready!
; End of the Communication Cycle
; Joystick and mouse are re-enabled
	BTFSC	DV_STA,DV_MSE		; Mouse connected? If yes
	RCALL	STA_RC			; enable PS/2 reception
; Read joystick
JOY_RD	BTFSS	MSE_ST,MS_ENB		; Mouse enabled?
	BRA	JOY_ST			; No!	
	BTFSS	PORT_J,J_JOY4		; Disable mouse?
	BRA	MOUS_D			; Yes!
	BTFSS	MSE_ST,MS_JOY		; Mouse as joystick?
	BRA	JS_DIS			; No!
; Read mouse as Kempston joystick
	BTG	PORTC,C_NMSE		; Mouse LED blink
	MOVFF	KEM_JS,MOUS_Y		; Work on the mouse Y register
; Get fire button state
	BCF	MOUS_Y,JOY_FR		; Copy the mouse left button
	BTFSC	MOVE_B,MV_MBL		; bit to the joytick fire
	BSF	MOUS_Y,JOY_FR		; button bit
; Get the vertical movement bits: 'Up' and 'Down'
	MOVFF	MPOS_Y,MS_POS		; Get the current Y-coordinate
	MOVLW	0x80			; Center the  Y-coordinate
	MOVWF	MPOS_Y			; for the next read
	RRNCF	MOUS_Y			; Shift vertical movement bits
	RRNCF	MOUS_Y			; in place of horizontal bits
	RCALL	SW_DIR			; Swap 'Up' and 'Down' bits 
	RCALL	MOUS_A			; Update vertical movement bits
	RCALL	SW_DIR			; Swap back 'Up' and 'Down' bits 
	RLNCF	MOUS_Y			; Shift back vertical movement
	RLNCF	MOUS_Y			; bits to their normal position
; Get the horizontal movement bits: 'Left' and 'Right'
	MOVFF	MPOS_X,MS_POS		; Get the current X-coordinate
	MOVLW	0x80			; Center the X-coordinate
	MOVWF	MPOS_X			; for the next read
	RCALL	MOUS_A			; Update horizontal movement bits
	MOVFF	MOUS_Y,KEM_JS		; Copy new state to joystick port
	BRA	JS_ARM
; Read Kempston joystick port
JOY_ST	COMF	PORTB,W
	ANDLW	1<<B_JOY0|1<<B_JOY1|1<<B_JOY2|1<<B_JOY3
	BTFSS	PORT_J,J_JOY4
	BSF	WREG,JOY_FR
	MOVWF	KEM_JS
	BRA	JS_ARM
; Disable mouse when both buttons or joystick fire are pressed
MOUS_D	BSF	MSE_ST,MS_HLD		; Signal: Mouse on hold!
	BCF	MSE_ST,MS_ENB		; Disable mouse
	BSF	PORTC,C_NMSE		; Mouse LED off
; Re-arm the joystick timer
JS_ARM	RCALL	TIME_J			; Re-arm joystick timer
	RETFIE	FAST
; Enable mouse if only one button is pressed
MOUS_E	BTFSS	MOVE_B,MV_MBL		; Mouse left button pressed?
	BRA	OTH_MB			; No!
	BTFSS	MOVE_B,MV_MBR		; Right button also pressed?
	BRA	MSE_JS			; No!
; Set the 'mouse on hold' flag if both buttons are pressed
	BSF	MSE_ST,MS_HLD		; Signal: Mouse on hold!
	BRA	STM_IT			; Stop timer 1
OTH_MB	BTFSC	MOVE_B,MV_MBR		; Mouse right button pressed?
	BRA	KEM_MS			; Yes!
; Clear the 'mouse on hold' flag if no button is pressed
	BCF	MSE_ST,MS_HLD		; Clear 'hold' flag
	BRA	STM_IT			; Stop timer 1
; Enable mouse as Kempston joystick if left button pressed
MSE_JS	BTFSC	MSE_ST,MS_HLD		; Mouse on hold?
	BRA	STM_IT			; Yes!
	BSF	MSE_ST,MS_ENB		; Enable mouse
	BSF	MSE_ST,MS_JOY		; Mouse as joystick
	CLRF	KEM_JS			; Clear joystick port
	BRA	STM_IT			; Stop timer 1
; Enable mouse as Kempston mouse if right button pressed
KEM_MS	BTFSC	MSE_ST,MS_HLD		; Mouse on hold?
	BRA	STM_IT			; Yes!
	BCF	MSE_ST,MS_JOY		; Mouse not as joystick
	RCALL	KM_INI			; Enable mouse
JS_DIS	BCF	TIM_ST,TM_JOY		; Disable joystick timer
	BRA	STM_IT			; Stop timer 1

;------------------------------------------------------------------------------
; INTERRUPT SUBROUTINES

; Host Command 
HO_CMD	CPFSEQ	CMD_CD			; SD card 'Set Write Address'?
	BRA	HOST_C			; No!
; SD card Set Write Block Address
	CLRF	BLN_LO			; Set reponse length to 0
	tab_ad	CMD_24			; Command: WRITE_BLOCK
; Set up a SD card command string in RAM
CM_SET	RCALL	TAB_NW			; Copy the command string to RAM
	DECF	SD_CRC,W		; A CRC-7 value of 0, indicates
	BNZ	CDT_EN			; that a data address is required
; Copy the data address to the command string
; The data address must be given in bytes, for standard capacity SD cards
; and in blocks of 512 bytes for high capacity SD cards
	LFSR	FSR1,SD_AD3		; Point to the highest byte and
	BTFSC	SDC_ST,SD_HCS		; copy it only for a high capacity
	MOVFF	DATA_B+3,POSTINC1	; SD card
	MOVFF	DATA_B+2,POSTINC1	; In this way, the block address
	MOVFF	DATA_B+1,POSTINC1	; of standard capacity SD cards 
	MOVFF	DATA_B+0,INDF1		; is implicitly multiplied by 256
	BTFSC	SDC_ST,SD_HCS		; For standard capacity cards,
	RETURN				; multiply the address again by 2
	RLCF	SD_AD0,W		; This clears the Carry Flag
	RLCF	SD_AD1			; Multiply by 2
	RLCF	SD_AD2
	RLCF	SD_AD3
CDT_EN	RETURN
; Execute a host command
HOST_C	TSTFSZ	BLN_LO			; Return if there is no command
	BRA	PARS_C			; string
HOC_EN	RETURN
PARS_C	LFSR	FSR0,DATA_B		; Point FSR0 to the command string
	MOVF	POSTINC0,W		; Check the command type
	XORLW	'm'			; Mouse/joystick command?
	BNZ	HOC_EN			; No!
; Parse and execute a mouse/joystick command string
TST_CL	DECFSZ	BLN_LO			; Test the remaining length of
	DCFSNZ	BLN_LO			; the command string
	RETURN				; Finished!
	MOVF	POSTINC0,W		; Check the command literal
	XORLW	'l'			; 'Legacy mode'?
	BNZ	MC_WIN			; No!
	BSF	MSE_ST,MS_LEG		; Switch to legacy mode
	BRA	TST_CL
MC_WIN	XORLW	'l'^'w'			; 'Windowed mode'?
	BNZ	MC_SEN			; No!
	BCF	MSE_ST,MS_LEG		; Switch to windowed mode
	BRA	TST_CL
MC_SEN	XORLW	'w'^'s'			; 'Joystick sensitivity'?
	BNZ	MC_RTE			; No!
	MOVLW	.2			; Check lower limit
	CPFSGT	INDF0
	BRA	MC_END
	MOVLW	.32			; Check upper limit
	CPFSGT	INDF0	
	MOVFF	INDF0,JS_SEN		; OK!
	BRA	MC_END
MC_RTE	XORLW	's'^'r'			; 'Joystick read rate [Hz]'?
	BNZ	MC_WSX			; No!
	MOVLW	(JM_INI-JS_TMC)/2+1	; Check if the command parameter
	CPFSLT	INDF0			; is:
	BRA	MC_END			; - too large
	tab_ad	(JS_TMC-2)
	RLNCF	INDF0,W			;
	BZ	MC_END			; - zero
	RCALL	ADD_TP
	TBLRD*+				; Copy the timing constant
	MOVFF	TABLAT,JS_RTC		; from the table
	TBLRD*+
	MOVFF	TABLAT,JS_RTC+1
	BRA	MC_END
MC_WSX	XORLW	'r'^'x'			; 'Mouse window width'?
	BNZ	MC_WSY			; No!
	MOVLW	.16			; Check lower limit
	CPFSLT	INDF0
	MOVFF	INDF0,MWIN_X		; OK!
	BRA	MC_CEN
MC_WSY	XORLW	'x'^'y'			; 'Mouse window height'?
	BNZ	HOC_EN			; No! Abort parsing
	MOVLW	.176			; Check upper limit
	CPFSLT	INDF0
	BRA	MC_END
	MOVLW	.16			; Check lower limit
	CPFSLT	INDF0
	MOVFF	INDF0,MWIN_Y		; OK!
MC_CEN	RCALL	MSE_CE			; Center the mouse coordinates
MC_END	MOVF	POSTINC0,W		; Advance pointer to the next
	BRA	TST_CL			; command and loop back
; Center the mouse coordinates
MSE_CE	RRNCF	MWIN_X,W
	BCF	WREG,7
	MOVWF	MPOS_X			; Update the X-coordinate
	RRNCF	MWIN_Y,W
	BCF	WREG,7
	MOVWF	MPOS_Y			; Update the Y-coordinate
	RETURN
; Start the SD card time-out counter
SDC_TM	BSF	TIM_ST,TM_SDC		; Signal: SD card timer is running
	BRA	TIM_1S
; Set timer 1 to 20ms
TIM_20	s_time	.20000			; Set time-out to 20ms
; Start timer 1
STA_ST	BCF	T1CON,TMR1ON		; Stop timer 1
	BCF	PIR1,TMR1IF		; Clear interrupt flag
	MOVWF	TMR1L			; Load timing constant
	BSF	T1CON,TMR1ON		; Start timer 1
	RETURN
; Add movement value (WREG) to coordinate (INDF2)
ADD_MV	BTFSC	MSE_ST,MS_LEG		; Legacy mode?
	BRA	LG_MOV			; Yes!
	BTFSC	MOVE_B,MV_SGN		; Positive sign?
	BRA	N_SIGN			; No!
; Positive movement
	BN	P_OVFL			; Movement overflow!
	ADDWF	INDF2,W			; Add movement to coordinate
	BNC	MOV_EN			; In case of overflow, set
P_OVFL	SETF	WREG			; new coordinate to 255
	BRA	MOV_EN
; Negative movement
N_SIGN	BN	N_OVFL			; Movement overflow!
	ADDWF	INDF2,W			; Add movement to coordinate
	BC	MOV_EN			; In case of overflow, set
N_OVFL	CLRF	WREG			; new coordinate to 0
MOV_EN	BTFSC	MSE_ST,MS_JOY
	BRA	STO_CO			; If not in joystick mode,
	CPFSGT	MS_POS			; check new coordinate against
	MOVF	MS_POS,W		; the maximum allowed value
STO_CO	MOVWF	POSTDEC2		; Store new coordinate
	RETURN
LG_MOV	ADDWF	POSTDEC2		; Add movement in legacy mode
	RETURN
; Mouse as Kempston joystick horizontal axis bits update
MOUS_A	BTFSC	MOVE_B,MV_MBR		; The right mouse button
	BRA	MOUS_Z			; centers the joystick
	RLNCF	JS_SEN,W		; Double the sensitivity
	SUBLW	0x80			; Subtract from center
	CPFSGT	MS_POS			; If the coordinate is off-center
	BRA	MOUS_L			; by more than double the sensi-
	ADDWF	JS_SEN,W		; tivity threshold the joystick
	CPFSGT	MS_POS			; bits are set to 'Left' or 'Right'
	BRA	MOUS_M			; If the deviation is of between
	ADDWF	JS_SEN,W		; single and double the sensitivity
	ADDWF	JS_SEN,W		; threshold, the jostick bits are
	CPFSGT	MS_POS			; set depending on their previous
	RETURN				; state, so that the move from one
	ADDWF	JS_SEN,W		; direction to another passes 
	CPFSGT	MS_POS			; through the centered state
	BTFSS	MOUS_Y,JOY_LT
	BRA	MOUS_R
MOUS_Z	BCF	MOUS_Y,JOY_LT		; Set joystick to: 'Centered'
	BCF	MOUS_Y,JOY_RT
	RETURN
MOUS_R	BCF	MOUS_Y,JOY_LT		; Set joystick to: 'Right'
	BSF	MOUS_Y,JOY_RT
	RETURN
MOUS_M	BTFSC	MOUS_Y,JOY_RT
	BRA	MOUS_Z
MOUS_L	BCF	MOUS_Y,JOY_RT		; Set joystick to: 'Left'
	BSF	MOUS_Y,JOY_LT
	RETURN
; Swap bits 0 and 1
SW_DIR	RRCF	MOUS_Y			; C7654321 0
	RRNCF	MOUS_Y			; 1C765432 0
	RLCF	MOUS_Y			; C7654320 1
	RLCF	MOUS_Y			; 76543201 C
	RETURN
; Start joystick timer
TIME_J	BSF	TIM_ST,TM_JOY		; Signal: Joystick timer running
	MOVFF	JS_RTC+1,TMR0H
	MOVF	JS_RTC,W
	BRA	STA_LT

;------------------------------------------------------------------------------
; PS/2 MOUSE SUBROUTINES
;
; Wait while timer 1 is running
WAIT_S	RCALL	STA_ST
WST_LP	BTFSC	T1CON,TMR1ON
	BRA	WST_LP
	RETURN
; Send command (WREG) to PS/2 mouse
PS_SAK	BSF	MSE_ST,MS_ACK		; Signal: Expect 'Acknowledge'
	MOVWF	MS_POS			; Store command
	RCALL	STO_RC			; Disable PS/2 reception
; Signal: 'Request to send'
	BCF	LATC,CK1
	BCF	TRISC,CK1		; Assert the clock line
	s_time	.100			; for 100us
	RCALL	WAIT_S
; Start transmission
	RCALL	TIM_20			; Time-out in 20ms
	BCF	LATC,DT1
	BCF	TRISC,DT1		; Produce the start bit
	BSF	TRISC,CK1		; Release the clock line
	BSF	MOVE_B,7		; Set parity to 'odd'
	MOVLW	.10			; Count 10 bits
	BCF	INTCON2,INTEDG1		; Interrupt on falling CK
; Transmission loop
BIT_LP	BCF	INTCON3,INT1IF		; Clear CK interrupt flag
WAI_CK	BTFSS	T1CON,TMR1ON		; Check for time-out
	BRA	PS_ERR			; Time-out!
	BTFSS	INTCON3,INT1IF		; Wait for a falling edge
	BRA	WAI_CK			; of the CK line
	DCFSNZ	WREG			; The stop bit is the
	BRA	SET_DT			; last one
	MOVF	MOVE_B			; Parity is the
	DECFSZ	WREG			; last-but-one bit
	RRNCF	MS_POS			; Get the next data bit
	INCFSZ	WREG			; Adjust the bit counter
	BN	SET_DT			; Test the current bit
	BCF	TRISC,DT1		; Send a '0' bit
	BRA	NX_BIT
SET_DT	BTG	MOVE_B,7		; Update the parity bit
	BSF	TRISC,DT1		; Send a '1' bit
NX_BIT	TSTFSZ	WREG			; Done?
	BRA	BIT_LP			; Not yet!
; Wait for the acknowledge pulse
	BCF	INTCON3,INT2IF		; Clear DT interrupt flag
WAI_AK	BTFSS	T1CON,TMR1ON		; Check for time-out
	BRA	PS_ERR			; Time-out!
	BTFSS	INTCON3,INT2IF		; Wait for a falling edge
	BRA	WAI_AK			; of the DT line
	BSF	INTCON2,INTEDG1		; Interrupt on rising CK
	RCALL	STA_RC			; Start PS/2 reception
; Get PS/2 mouse response
	RCALL	TIM_20			; Time-out in 20ms
	MOVLW	low MOVE_X		; Expect 1 byte
RSP_LP	BTFSS	T1CON,TMR1ON		; Check for time-out
	BRA	PS_ERR			; Time-out!
	CPFSEQ	FSR2L		        ; Response received?
	BRA	RSP_LP			; Not yet!
	BCF	MSE_ST,MS_PAK		; Clear 'packet in progress' flag
	MOVF	POSTDEC2,W		; Adjust pointer
	MOVF	INDF2,W		        ; Collect received byte
	BTFSS	MSE_ST,MS_ACK		; Expect "Acknowledge'?
	RETURN				; No!
	BCF	MSE_ST,MS_ACK		; Do not expect 'Acknowledge'
	XORLW	PS_ACK			; 'Acknowledge'?
	BNZ	PS_ERR			; No!
	RETURN
PS_ERR	POP				; Drop return address
	BSF	TRISC,DT1		; Release the data line
	BCF	DV_STA,DV_MSE		; Signal: Mouse not present!	
; Disable PS/2 reception
STO_RC	BCF	INTCON3,INT2IE		; Disable DT interrupt
	BCF	INTCON3,INT1IE		; Disable CK interrupt
	BCF	PIE3,RC2IE		; Disable USART interrupt
	BCF	RCSTA2,CREN		; Stop USART reception
	BCF	RCSTA2,SPEN		; Disable USART
	RETURN				; 
; Enable PS/2 reception
STA_RC	LFSR	FSR2,MOVE_B		; Set movement packet pointer
	BSF	RCSTA2,SPEN		; Enable USART
	BCF	INTCON3,INT2IF		; Clear DT interrupt flag
	BSF	INTCON3,INT2IE		; Enable DT interrupt
	RETURN
; Enable PS/2 mouse data reporting
MOUS_I	MOVLW	PS_REP			; Send 'Enable data reporting'
	RCALL	PS_SAK			; command
KM_INI	BSF	MSE_ST,MS_ENB		; Enable mouse
	RCALL	MSE_CE			; Center mouse coordinates
	BCF	PORTC,C_NMSE		; Mouse LED on
	RETURN

;------------------------------------------------------------------------------
; SD CARD SUBROUTINES

; Read SPI buffer to (WREG) 
SPI_RD	SETF	WREG
; Write (WREG) to SPI buffer
SPI_WR	MOVWF	SSP2BUF			; Write the byte to the SPI buffer
SPI_LP	BTFSS	SSP2STAT,BF		; Wait until the transmission ends
	BRA	SPI_LP
	MOVF	SSP2BUF,W		; Read the received byte
	RETURN
; Send a SD card Command from the table at (TABLPTR)
SDC_CM	RCALL	CM_SET			; Set up the command string in RAM
CM_SND	BCF	LATA,A_NFCS		; Select the SD card	
	LFSR	FSR1,SD_CMD		; Command string address
; Send the command string
SCM_LP	MOVF	POSTINC1,W		; Command byte to WREG
	RCALL	SPI_WR			; Send command bytes
	DECFSZ	CM_CNT			
	BRA	SCM_LP	
	SETF	CM_CNT			; Number of trials: 255
; Wait for the response
SRW_LP	RCALL	SPI_RD
	INCFSZ	WREG			; First response byte?
	BRA	SCM_RP			; Yes!
	DECFSZ	CM_CNT			; No!
	BRA	SRW_LP			; Retry	
SCM_RP	DECF	WREG	
	LFSR	FSR1,SD_CMD		; Response store address
; Get the response
SRP_LP	MOVWF	POSTINC1		; Store response bytes	
	RCALL	SPI_RD			; Get next response byte
	DECFSZ	RP_CNT			; Any more response bytes?
	BRA	SRP_LP			; Yes!
	TSTFSZ	DT_CNL			; Done!
	RETURN
	TSTFSZ	DT_CNH			; If the command is a
	RETURN				; 'no-data' operation,
	BRA	SD_DSL			; de-select the SD card
; Detect the SD card
SDC_ID	CLRF	SDC_ST			; Clear the SD card status byte
	BCF	DV_STA,DV_SDC	        ; Device state: SDC not present
	BSF	PORTC,C_NSDC		; Switch off the SDC LED
	BTFSC	CHG_ST,B_NCRD		; Is a card present?
	RETURN				; No!
; Start the SD card in SPI mode
	MOVLW	.10			; Generate 80 clock pulses
	MOVWF	CM_CNT
CLK_LP	RCALL	SPI_RD
	DECFSZ	CM_CNT			
	BRA	CLK_LP
	BCF	LATA,A_NFCS		; Select the SD card	
	tab_ad	CMD_00			; Command: GO_IDLE_STATE
	RCALL	SDC_CM			; Send command
	MOVLW	1<<R1_IDL		; Idle state?
	CPFSEQ	SD_CMD
	RETURN				; No! Abort card detection
; Check the SD card's version and operating voltage
	tab_ad	CMD_08			; Command: SEND_IF_COND
	RCALL	SDC_CM			; Send command
	BTFSC	SD_CMD,R1_ILC		; Illegal command?
	BRA	SD_V1X			; Yes! Version 1.x card
	MOVLW	C8_PAT			; Check pattern			
	CPFSEQ	SD_AD0			; Does it match?
	RETURN				; No! Abort card detection
	MOVLW	C8_VHS			; Check Voltage
	CPFSEQ	SD_AD1			; Does it match?
	RETURN				; No! Abort card detection
	BSF	SDC_ST,SD_VER		; Signal: Version 2 SD card
; Initialize the SD card
SD_V1X	RCALL	SDC_TM			; Start timer
C41_LP	tab_ad	CMD_55			; Command: APP_CMD
	RCALL	SDC_CM			; Send command
	tab_ad	ACM_41			; Command: SEND_OP_COND
	RCALL	SDC_CM			; Send command
	BTFSC	SD_CMD,R1_ILC		; Illegal command?
	RETURN				; Yes! Not a SD card
	BTFSS	TIM_ST,TM_SDC		; Time-out?
	RETURN				; Yes! Abort card detection
	TSTFSZ	SD_CMD			; Card still in 'Idle State'?
	BRA	C41_LP			; Yes! Repeat command
; Differentiate between standard and high capacity SD cards
	RCALL	SDC_TM			; Restart timer
C58_LP	tab_ad	CMD_58			; Command: READ_OCR
	RCALL	SDC_CM			; Send command
	BTFSS	TIM_ST,TM_SDC		; Time-out?
	RETURN				; Yes! Abort card detection
	BTFSS	SD_AD3,OC_PWU		; Power-up finished?
	BRA	C58_LP			; Not yet! Try again
	BTFSC	SD_AD3,OC_CCS		; High Capacity card?
	BSF	SDC_ST,SD_HCS		; Yes!
; Disable the CRC check
	tab_ad	CMD_59			; Command: CRC_ON_OFF
	RCALL	SDC_CM			; Send command
; Set block length to 512 bytes for standard capacity SD cards
	tab_ad	CMD_16			; Command: SET_BLOCKLEN
	BTFSS	SDC_ST,SD_HCS		; Execute only for standard
	RCALL	SDC_CM			; capacity cards
; Get the SD card's capacity
	tab_ad	CMD_09			; Command: SEND_CSD
	RCALL	SDC_RD			; Send command
	BTFSS	TIM_ST,TM_SDC		; Time-out?
	RETURN				; Yes! Abort card detection
	BTFSC	DATA_B,CSD_HC		; Version 2 CSD structure?
	BRA	CSD_V2			; Yes!
; Version 1 CSD structure
; Get C_SIZE*2^6 and multiply by 2^(C_SIZE_MULT+2+READ_BL_LEN-9-6)
	MOVF	DATA_B+0x08,W
	ANDLW	b'11000000'		; Bits <63:62>
	MOVWF	SD_SZ0
	MOVF	DATA_B+0x07,W		; Bits <71:64>  
	MOVWF	SD_SZ1
	MOVF	DATA_B+0x06,W
	ANDLW	b'00000011'		; Bits <73:72>
	MOVWF	SD_SZ2
	CLRF	SD_SZ3
	MOVF	DATA_B+0x05,W
	ANDLW	b'00001111'		; Bits <83:80>
	MOVWF	CM_CNT			; READ_BL_LEN
	RLCF	DATA_B+0x0a,W		; Bit  <47> to Carry Flag
	RLCF	DATA_B+0x09,W		; Bits <49:48>
	ANDLW	b'00000111'		; C_SIZE_MULT
	ADDLW	-.13			; Add 2-6-9 = -13
	ADDWF	CM_CNT			; Add READ_BL_LEN
	BZ	C_SMAL
	BC	CMP_SZ			; SD Cards capacities of less
C_SMAL	RETURN				; than 256 MB are not accepted
; Version 2 CSD structure
; Get C_SIZE*2^8 and multiply by 2^2
CSD_V2	CLRF	SD_SZ0
	MOVFF	DATA_B+0x09,SD_SZ1	; Bits <55:48>
	MOVFF	DATA_B+0x08,SD_SZ2	; Bits <63:56>
	MOVFF	DATA_B+0x07,SD_SZ3	; Bits <71:64> (<71:70> are 00)
	MOVLW	2
	MOVWF	CM_CNT
; Calculate the SD card's capacity, in blocks of 512 bytes
CMP_SZ	MOVLW	0
	INCF	SD_SZ0			; Add 1
	ADDWFC	SD_SZ1
	ADDWFC	SD_SZ2
	ADDWFC	SD_SZ3			; This clears the Carry Flag
SHL_LP	RLCF	SD_SZ0			; Multiply by 2^(CM_CNT)
	RLCF	SD_SZ1
	RLCF	SD_SZ2
	RLCF	SD_SZ3			; This clears the Carry Flag
	DECFSZ	CM_CNT			; ( Borrow = set )
	BRA	SHL_LP
	SUBWFB	SD_SZ0			; Subtract 1
	SUBWFB	SD_SZ1
	SUBWFB	SD_SZ2
	SUBWFB	SD_SZ3
; Get the SD card's ID
	tab_ad	CMD_10			; Command: SEND_CID
	RCALL	SDC_RD			; Send command
	BTFSS	TIM_ST,TM_SDC		; Time-out?
	RETURN				; Yes! Abort card detection
	MOVF	SDC_ST,W		; Set first response byte to:
	ANDLW	b'00000011'		; '1' - Ver 1.x standard SD card
	ADDLW	'1'			; '2' - Ver 2.0 standard SD card
	MOVWF	DATA_B			; '4' - High capacity SD card
	copy_b  SD_SZ0,DATA_B+8,4	; Copy the SD card's capacity
	BSF	SDC_ST,SD_PRE		; Signal: The SD card is present
	BSF	DV_STA,DV_SDC	        ; Device state: SDC identified
	BCF	PORTC,C_NSDC		; Switch on the SDC LED
	RETURN
; SD card Read ( Commands: CMD9, CMD10 and CMD17 )
SDC_RD	RCALL	SDC_TM			; Start timer
	RCALL	SDC_CM			; Send command
	LFSR	FSR1,DATA_B
; Wait for the DATA_START_TOKEN
TS_FBU	INCFSZ	WREG			; Is the SDI line still high?
	BRA	TS_DST			; Not any more!
	BTFSS	TIM_ST,TM_SDC		; Time-out?
	BRA	SD_DSL			; Yes! Abort read operation
	RCALL	SPI_RD			; Wait for non-floating line
	BRA	TS_FBU
TS_DST	DECF	WREG			; Is the received byte a
DST_LP	XORLW	SD_DST			; DATA_START_TOKEN?
	BZ	RD_DAT			; Yes
	BTFSS	TIM_ST,TM_SDC		; Time-out?
	BRA	SD_DSL			; Yes! Abort read operation
	RCALL	SPI_RD			; Wait for DATA_START_TOKEN
	BRA	DST_LP
; Receive data
DRC_HI	DECF	DT_CNH			; Decrement the byte counter
DRC_LO	DECF	DT_CNL
	MOVWF	POSTINC1		; Save the received byte
RD_DAT	RCALL	SPI_RD			; Read next byte
	TSTFSZ	DT_CNL			; Check the byte counter
	BRA	DRC_LO
	TSTFSZ	DT_CNH			; Done?
	BRA	DRC_HI			; Not yet!
SD_DSL	BSF	LATA,A_NFCS		; Deselect the SD card
	RETURN
; Execute SD card host commands
SDC_OP	BCF	SDC_ST,SD_EXE		; Signal: No SDC commands!
	MOVF	CMD_CD,W
	CLRF	CMD_CD			; Set Error Code to 0
	CLRF	BLN_LO			; Set the response length to 0
	CLRF	BLN_HI
	XORLW	SD_IDF			; 'Identify Card' command ?
	BNZ	TST_PR			; No!
; Identify card
	RCALL	SDC_ID			; Detect the SD card
	BTFSS	SDC_ST,SD_PRE		; Is the SD card present?
	BRA	NOT_PR			; No! Signal error
	MOVLW	0x0c			; Send 8 bytes of SD card
	MOVWF	BLN_LO			; identifier followed by 
	BRA	SD_END			; 4 bytes of card capacity
; Check if the SD card is present
TST_PR	BTFSC	SDC_ST,SD_PRE		; Is the SD card present?
	BRA	TST_RD			; Yes!
; Error condition: 'Not present'
NOT_PR	MOVLW	E_NPRE			; Error: 'Not present'
SD_ERR	MOVWF	CMD_CD			; Set the error code
SD_END	BSF	LATA,A_NFCS		; Deselect the SD card
	BSF	PIE3,SSP2IE		; Signal: SDC command done!
	RETURN
TST_RD	XORLW	SD_IDF^SD_BRD		; 'Read Block' Command?
	BNZ	TST_WR			; No!
; Read Block
	tab_ad	CMD_17			; Command: READ_SINGLE_BLOCK
	RCALL	SDC_RD			; Send command
	BTFSS	TIM_ST,TM_SDC		; Time-out?
	BRA	NOT_PR			; Yes! Signal error
	MOVLW	0x02			; Return 512 bytes of data
	MOVWF	BLN_HI
	BRA	SD_END
TST_WR	XORLW	SD_BRD^SD_BWR		; 'Write Block' Command?
	BNZ	TST_EF			; No!
; Write Block
; The command string and data address have already been set up by a
; previous 'Set Write Address' command
	RCALL	CM_SND			; Send command
	TSTFSZ	SD_CMD			; Are any error bits set?
	BRA	WR_PRO			; Yes! Signal error
	LFSR	FSR1,DATA_B		; Address of data to be written
	MOVLW	SD_DST			; First send a DATA_START_TOKEN
	BRA	WR_DAT
; Send data
DWC_HI	DECF	DT_CNH			; Decrement the byte counter
DWC_LO	DECF	DT_CNL
	MOVF	POSTINC1,W		; Get the next byte to transmit
WR_DAT	RCALL	SPI_WR			; Send one byte
	TSTFSZ	DT_CNL			; Test the byte counter
	BRA	DWC_LO
	TSTFSZ	DT_CNH			; Done?
	BRA	DWC_HI			; Not yet!
; Check if the data was accepted
	RCALL	SPI_RD
	ANDLW	0x0f			; Did the card respond with
	XORLW	SD_DAC			; a DATA_ACCEPTED token?
	MOVLW	2			; Yes! Wait for 1 second while
	BZ	W_BUSY			; the card is busy
; Error condition: 'Write protected'
WR_PRO	MOVLW	E_WRPR			; Error: 'Write protected'
	BRA	SD_ERR
TST_EF	XORLW	SD_BWR^SD_EFA		; 'Set First Erase Address'?
	BNZ	TST_EL			; No!
; Set First Erase Address
	tab_ad	CMD_32			; Command: ERASE_WR_BLK_START_ADDR
	BRA	SND_EA			; Send command
TST_EL	XORLW	SD_EFA^SD_ELA		; 'Set Last Erase Address'?
	BNZ	TST_ER			; No!
; Set Last Erase Address
	tab_ad	CMD_33			; Command: ERASE_WR_BLK_END_ADDR
SND_EA	RCALL	SDC_CM			; Send command
	TSTFSZ	SD_CMD			; Are any error bits set?
	BRA	WR_PRO			; Yes! Signal error
	BRA	SD_END
TST_ER	XORLW	SD_ELA^SD_ERA		; 'Erase Blocks' Command?
	BNZ	NOT_SU			; No other commands are supported
; Erase
	tab_ad	CMD_38			; Command: ERASE
	RCALL	SDC_CM			; Send command
	TSTFSZ	SD_CMD			; Are any error bits set?
	BRA	WR_PRO			; Yes! Signal error
	MOVLW	6			; Set time-out to 5 seconds
; Wait while the SD card is busy
W_BUSY	MOVWF	CM_CNT			; Timing duration + 1 (in seconds)
CHK_TM	DCFSNZ	CM_CNT			; Count seconds
	BRA	WR_PRO			; Time-out! Signal error
	RCALL	SDC_TM			; Restart timer
BSY_LP	BTFSS	TIM_ST,TM_SDC		; Is one second over?
	BRA	CHK_TM			; Yes! 
	RCALL	SPI_RD			; Card still busy?
	BZ	BSY_LP			; Yes! Check again!
	RCALL	SPI_RD			; Done! Send 8 clock pulses
	BRA	SD_END			; Deselect the SD card
; Unimplemented command
NOT_SU	MOVLW	E_NSUP			; Error: 'Feature not supported'
	BRA	SD_ERR			; Set the error code


;------------------------------------------------------------------------------
; End of program
	END