;******************************************************************************
;                                                                             *
;   This file is a basic code template for code generation on the             *
;   PIC18F14K22. This file contains the basic code building blocks to build   *
;   upon.                                                                     *
;                                                                             *
;   Refer to the MPASM User's Guide for additional information on features    *
;   of the assembler.                                                         *
;                                                                             *
;   Refer to the respective data sheet for additional information on the      *
;   instruction set.                                                          *
;                                                                             *
;******************************************************************************
;                                                                             *
;    Filename:                                                                *
;    Date:                                                                    *
;    File Version:                                                            *
;    Author:                                                                  *
;    Company:                                                                 *
;                                                                             *
;******************************************************************************
;                                                                             *
;    Files Required: P18F14K22.INC                                            *
;                                                                             *
;******************************************************************************
;                                                                             *
;    Notes:                                                                   *
;                                                                             *
;******************************************************************************
;                                                                             *
;    Revision History:                                                        *
;                                                                             *
;******************************************************************************

;------------------------------------------------------------------------------
; PROCESSOR DECLARATION
;------------------------------------------------------------------------------

     LIST      P=PIC18F14K22          ; list directive to define processor
     #INCLUDE <P18F14K22.INC>         ; processor specific variable definitions

;------------------------------------------------------------------------------
;
; CONFIGURATION WORD SETUP
;
; The 'CONFIG' directive is used to embed the configuration word within the 
; .asm file. The lables following the directive are located in the respective 
; .inc file.  See the data sheet for additional information on configuration 
; word settings.
;
;------------------------------------------------------------------------------

	CONFIG	FOSC = HS	; HS oscillator
	CONFIG	PLLEN = ON	; Oscillator multiplied by 4	
	CONFIG	PCLKEN = ON     ; Primary clock enabled	
	CONFIG	FCMEN = OFF	; Fail-Safe Clock Monitor disabled
	CONFIG	IESO = OFF	; Oscillator Switchover mode disabled
	CONFIG	PWRTEN = OFF	; PWRT disabled
	CONFIG	BOREN = OFF     ; Brown-out Reset disabled 
	CONFIG	BORV = 19       ; VBOR set to 1.9 V nominal
	CONFIG	WDTEN = OFF     ; WDT is controlled by SWDTEN bit 
	CONFIG	WDTPS = 256     ; 1:256
	CONFIG	HFOFST = ON     ; HFINTOSC starts clocking the CPU immediately
	CONFIG	MCLRE = OFF     ; RA3 input pin enabled; MCLR disabled
	CONFIG	STVREN = OFF    ; Stack full/underflow will not cause Reset
	CONFIG	LVP = OFF       ; Single-Supply ICSP disabled
	CONFIG	BBSIZ = OFF     ; 1kW boot block size
	CONFIG	XINST = OFF     ; Instruction set extension disabled 
	CONFIG	CP0 = OFF       ; Block 0 not code-protected
	CONFIG	CP1 = OFF       ; Block 1 not code-protected
	CONFIG	CPB = OFF       ; Boot block not code-protected
	CONFIG	CPD = OFF       ; Data EEPROM not code-protected
	CONFIG	WRT0 = OFF      ; Block 0 not write-protected
	CONFIG	WRT1 = OFF      ; Block 1 not write-protected
	CONFIG	WRTB = OFF      ; Boot block not write-protected
	CONFIG	WRTC = OFF      ; Configuration registers not write-protected
	CONFIG	WRTD = OFF      ; Data EEPROM not write-protected
	CONFIG	EBTR0 = OFF     ; Block 0 not protected from table reads
	CONFIG	EBTR1 = OFF     ; Block 1 not protected from table reads
	CONFIG	EBTRB = OFF     ; Boot block not protected from table reads

;------------------------------------------------------------------------------
;
; CONSTANTS
;
;------------------------------------------------------------------------------

FOSC_4	EQU	.12000000
BAUD_R	EQU	.57600
B_NRDY	EQU	RB4
B_NSTB	EQU	RB6
SRB_RZ	EQU	.16

;------------------------------------------------------------------------------
;
; VARIABLES
;
;------------------------------------------------------------------------------


	CBLOCK 0x100

D_BUFF:0x100

	ENDC

;------------------------------------------------------------------------------
; CODE
;------------------------------------------------------------------------------

	ORG     0x0000
	CLRF	ANSEL
	CLRF	ANSELH
	BSF	OSCCON,IDLEN
        BRA     START_

;------------------------------------------------------------------------------
; INTERRUPT SERVICE ROUTINE
;------------------------------------------------------------------------------

	BTFSC	INTCON3,INT2IE
	BTFSS	INTCON3,INT2IF
	BRA	CHK_RX
	BCF	INTCON3,INT2IE
	MOVF	FSR1L,W
	SUBWF	FSR0L,W
	BNZ	DT_OUT
CHK_RX	BTFSC	PIE1,RCIE
	BTFSS	PIR1,RCIF
	RETFIE	FAST
	MOVF	RCSTA,W
	MOVFF	RCREG,POSTINC0		; Copy received byte to buffer
	ANDLW	1<<FERR|1<<OERR		; Error?
	BZ	ARX_PT
	BCF	RCSTA,CREN
	BSF	RCSTA,CREN
ARX_PT	BTFSC	FSR0H,1
	LFSR	FSR0,D_BUFF
	BTFSC	INTCON3,INT2IE
	BRA	CK_RDY
DT_OUT	MOVFF	POSTINC1,PORTC
	BTFSC	FSR1H,1
	LFSR	FSR1,D_BUFF
	BTFSS	PIE1,RCIE
	BSF	PIE1,RCIE
	BCF	INTCON3,INT2IF
	MOVLW	4
	BCF	PORTB,B_NSTB
STB_LP	DECFSZ	WREG
	BRA	STB_LP
	BSF	INTCON3,INT2IE
	BSF	PORTB,B_NSTB
CK_RDY	MOVF	FSR1L,W
	SUBWF	FSR0L,W
	INFSNZ	WREG
	BRA	STP_RX
CK_REZ	ADDLW	SRB_RZ-1	
	BC	NOT_RY
	BCF	TRISB,B_NRDY		; Signal: Ready!
	RETFIE 	FAST
STP_RX	BCF	PIE1,RCIE		; Buffer full, stop reception
NOT_RY	BSF	TRISB,B_NRDY		; Signal: Not ready!
	RETFIE  FAST

;------------------------------------------------------------------------------
; MAIN PROGRAM
;------------------------------------------------------------------------------

START_	MOVLW	1<<TX|1<<B_NSTB		; Set nSTB and clear nRDY
	MOVWF	PORTB
	MOVLW	1<<RX
	MOVWF	TRISB		        ; Set RX as input
	CLRF	TRISC			; Set DAT(0-7) as output
	BCF	INTCON2,RABPU		; Enable pull-up
	BSF	TXSTA,BRGH
	BSF	BAUDCTL,BRG16		; 16-bit Baud rate generator
	MOVLW	low (FOSC_4/BAUD_R-1)
	MOVWF	SPBRG
	MOVLW	high (FOSC_4/BAUD_R-1)
	MOVWF	SPBRGH			; Baud rate: BAUD_R bit/s
	BSF	TXSTA,TXEN		; Enable USART transmit
	MOVLW	1<<SPEN|1<<CREN		; Enable USART receive
	MOVWF	RCSTA
	BSF	PIE1,RCIE		; Enable USART RC interrupt
	BCF	INTCON2,INTEDG2		; Interrupt on falling BUSY
	LFSR	FSR0,D_BUFF		; Initialize write pointer
	LFSR	FSR1,D_BUFF		; Initialize read pointer
	BSF	INTCON,PEIE		; Enable peripheral interrupts
	BSF	INTCON,GIE		; Enable interrupts
M_LOOP	BTFSC	INTCON3,INT2IE
	BRA	M_LOOP
	SLEEP
	NOP
	BRA	M_LOOP	

;------------------------------------------------------------------------------
; END
;------------------------------------------------------------------------------

	END