;******************************************************************************
;                                                                             *
;   This file is a basic code template for code generation on the             *
;   PIC18F46J53. This file contains the basic code building blocks to build   *
;   upon.                                                                     *
;                                                                             *
;   Refer to the MPASM User's Guide for additional information on features    *
;   of the assembler.                                                         *
;                                                                             *
;   Refer to the respective data sheet for additional information on the      *
;   instruction set.                                                          *
;                                                                             *
;******************************************************************************

;------------------------------------------------------------------------------
; PROCESSOR DECLARATION

	LIST      P=PIC18F46J53 ; List directive to define processor
	#INCLUDE <P18F46J53.INC>; Processor specific variable definitions
        errorlevel    -311,-202 ; Suppress the message: Operand of HIGH
                                ;      operator was larger than H'FFFF'

;------------------------------------------------------------------------------
;
; CONFIGURATION WORD SETUP
;
; The 'CONFIG' directive is used to embed the configuration word within the
; .asm file. The lables following the directive are located in the respective
; .inc file.  See the data sheet for additional information on configuration
; word settings.

     CONFIG WDTEN = OFF         ;Watchdog Timer: Disabled - Controlled by SWDTEN bit
     CONFIG PLLDIV = 3          ;Divide by 3 (12 MHz oscillator input)
     CONFIG CFGPLLEN = ON       ;PLL Enabled
     CONFIG STVREN = OFF        ;Stack Overflow/Underflow Reset: Disabled
     CONFIG XINST = OFF         ;Extended Instruction Set: Disabled
     CONFIG CPUDIV = OSC1       ;No CPU system clock divide
     CONFIG CP0 = OFF           ;Program memory is not code-protected
     CONFIG OSC = HSPLL         ;HSPLL oscillator with PLL software controlled
     CONFIG SOSCSEL = DIG       ;Digital (SCLKI) mode selected
     CONFIG CLKOEC = OFF        ;CLKO output disabled on the RA6 pin
     CONFIG FCMEN = OFF         ;Fail-Safe Clock Monitor: Disabled
     CONFIG IESO = OFF          ;Internal External Oscillator Switch Over Mode: Disabled
     CONFIG WDTPS = 512         ;Watch-dog timer set to 512*4ms
     CONFIG DSWDTOSC = INTOSCREF;DSWDT Reference Clock uses INTRC
     CONFIG RTCOSC = INTOSCREF  ;RTCC Reference Clock INTRC
     CONFIG DSBOREN = OFF       ;Deep Sleep BOR: Disabled
     CONFIG DSWDTEN = OFF       ;Deep Sleep Watchdog Timer: Disabled
     CONFIG DSWDTPS = 2         ;Deep Sleep Watchdog Postscaler:
     CONFIG IOL1WAY = ON        ;The IOLOCK bit (PPSCON<0>) can be set once
     CONFIG ADCSEL = BIT10      ;ADC 10 or 12 Bit Select: 10 - Bit ADC Enabled
     CONFIG MSSP7B_EN = MSK5    ;MSSP address masking: 5 Bit address masking mode
     CONFIG WPFP = PAGE_1       ;Write Protect Program Flash Page 1
     CONFIG WPCFG = ON          ;Configuration Words page erase/write-protected
     CONFIG WPDIS = ON          ;WPFP[5:0], WPEND, and WPCFG bits enabled
     CONFIG WPEND = PAGE_0      ;Page 0 to WPFP<6:0> erase/write-protected
     CONFIG LS48MHZ = SYS48X8   ;System clock at 48 MHz USB CLKEN divide-by is set to 8

;------------------------------------------------------------------------------
; CONSTANTS

; Board number and date
BRD_RV	equ	.2
BRD_LT	equ	.19
BRD_SR	equ	.16
BRD_CE	equ	.20
BRD_YR	equ	.19
BRD_MO	equ	.2
BRD_DY	equ	.3

; Clocks
FOSC_4  equ     .12000000               ; Instruction cycle (Hz)
BAUD_R  equ     .115200                 ; Maximum Baud rate (Hz)
BAUD_W	equ	.4000000		; Wifi module rate  (Hz)

; I/O Ports
MOUS_Y  equ     PMDOUT1L                ; Kempston mouse Y port
EXT_KB  equ     PMDOUT1L		; External keyboard port
MOUS_X  equ     PMDOUT1H                ; Kempston mouse X port
KEM_JS  equ     PMDOUT1H                ; Kempston joystick port
DEV_ST  equ     PMDOUT2L                ; Device state
PSP_DO  equ     PMDOUT2H                ; PSP out
BANK_R	equ	PMDIN1L			; 128k memory bank control
AY_ADR	equ	PMDIN1H			; 128k sound chip address
BANK_P	equ	PMDIN2L			; +3 memory bank control
PSP_DI  equ     PMDIN2H                 ; PSP in

; Bits
; I/O Port bits
A_NFCS  equ     RA2                     ; SD card deselect
A_JOYF  equ     RA3                     ; Joystick FIRE
A_NCTS  equ     RA3                     ; Serial port nCTS
B_JOYR  equ     RB0                     ; Joystick RIGHT
B_JOYL  equ     RB1                     ; Joystick LEFT
B_JOYD  equ     RB2                     ; Joystick DOWN
B_JOYU  equ     RB3                     ; Joystick UP
B_NRTS  equ     RB3                     ; Serial port nRTS
B_USEN  equ     RB6                     ; USB power sense
B_NCRD  equ     RB7                     ; Flash card sense
C_NSRV  equ     RC0                     ; LED: Server not connected
C_NRCS	equ	RC0			; Serial RAM chip deselect
C_NMSE  equ     RC1                     ; LED: Mouse not enabled
C_NEIT  equ     RC1                     ; Ethernet modudle interrupt
C_WFRX	equ	RC1			; Wifi module RX
C_NSDC  equ     RC2                     ; LED: SD card not present
C_NECS  equ     RC2                     ; Ethernet module deselect
C_NCTS	equ	DT1			; Printer nCTS
; Joystick port bits
JOY_RT  equ     0                       ; Joystick 'Right' bit
JOY_LT  equ     1                       ; Joystick 'Left' bit
JOY_DN	equ	2                       ; Joystick 'Down' bit
JOY_UP	equ	3                       ; Joystick 'Up' bit
JOY_FR  equ     4                       ; Joystick 'Fire' bit
; Device State (DEV_ST) bits
DV_BSY  equ     7                       ; Device busy
DV_DAT  equ     6                       ; Data ready
DV_SRV  equ     5                       ; Server connected
DV_SDC  equ     4                       ; SD card identified
DV_MSE  equ     3                       ; Mouse connected
DV_ETH  equ     2                       ; Ethernet module configured
DV_MBL  equ     1                       ; Mouse left button inactive
DV_NPR  equ     1                       ; Printer not connected
DV_BFE  equ     1                       ; Serial receive buffer empty
DV_MBR  equ     0                       ; Mouse right button inactive
DV_NSR	equ	0			; Serial device not connected
; Communication State (COM_ST) bits
CO_ACT  equ     0                       ; Communication in progress
CO_HED  equ     1                       ; Header   /Data
CO_TRA  equ     2                       ; Transmit /Receive
CO_SER  equ     3                       ; Serial   /Parallel
CO_RDT  equ     4                       ; Read-transmit in progress
CO_RWR  equ     5                       ; Receive-write in progress
CO_RXT	equ	6			; Receive-transmit in progress
CO_CYC	equ	7			; Active communication cycle
; Timer State (TIM_ST) bits
TM_SBD  equ     0                       ; USB power debounce
TM_CRD  equ     1                       ; SDC detect debounce
TM_PSP  equ     2                       ; PSP timer running
TM_USB  equ     3                       ; USB timer running
TM_SDC  equ     4                       ; SDC timer running
TM_ASY	equ	5			; Async receive timer running
TM_KYB	equ	6			; Keyboard timer running
TM_ETH	equ	7			; Ethernet timer running
; USB State (USB_ST) bits
SB_SPW  equ     0                       ; USB Self Powered
SB_RWU  equ     1                       ; USB Remote Wake-up enabled
SB_DEF  equ     2                       ; USB in Default state
SB_CFG  equ     3                       ; USB in Configured state
SB_DTR  equ     4                       ; DTR line state
SB_POL	equ	5			; Poll keyboard & mouse
SB_REQ	equ	6			; Polling in progress
SB_CMD	equ	7			; Server command
; SD card State (SDC_ST) bits
SD_VER  equ     0                       ; SD card Version 2.0
SD_HCS  equ     1                       ; SD card High Capacity
SD_DET  equ     2                       ; Detect SD card
SD_EXE  equ     3                       ; Execute SDC command
SD_MBR  equ     4                       ; Multiple block read
SD_MBF  equ     5                       ; Multiple blocks
SD_ABF  equ     6                       ; Alternate buffer
SD_EBF	equ	7			; Ethernet buffer
; Mouse State (MSE_ST) bits
MS_HLD  equ     0                       ; Mouse on hold
MS_LEG  equ     1                       ; Mouse in legacy mode
MS_JOY  equ     2                       ; Mouse as joystick
MS_ENB  equ     3                       ; Mouse enabled
MS_TRA  equ     4                       ; Transmitting to mouse
MS_RSD  equ     5                       ; Resend command sent
MS_KYB	equ	6			; Keyboard as joystick
MS_PRE	equ	7			; PS/2 device present
; Keyboard State (KYB_ST) bits
KB_EXT	equ	0			; Extended scan code
KB_BRK	equ	1			; Break code
KB_KEY	equ	2			; Scan code received
KB_ALT	equ	5			; Keyword flag 
KB_CTL	equ	6			; 'SYMBOL SHIFT' flag
KB_SHF	equ	7			; 'CAPS SHIFT' flag
; System flags (SYS_FL) bits
; Serial Asyncronous port flags
AS_ENB	equ	0			; Port enabled
AS_PRN	equ	1			; Transmit only mode
AS_RTS	equ	2			; Assert RTS
; Serial RAM State flags
SR_SEL	equ	3			; Serial RAM selected
SR_PRE	equ	4			; Serial RAM present
; Command flags
CM_CAS	equ	5			; Command: Case insensitive
CM_SKP	equ	6			; Command: Skip
CM_DNS	equ	7			; Command: DNS query
; Tape player State (PLY_ST) bits
PL_PRE  equ     0			; Tape player cable present
PL_ACT	equ	1			; Tape player is active
PL_PLY	equ	2			; Play /Pause
PL_STD	equ	3			; Standard speed tape block
PL_LDR	equ	4			; Leader signal done
PL_SYN	equ	5			; Sync pulse done
; Ethernet adaptor State (ETH_ST) bits
ET_PRE  equ     0			; Ethernet adaptor present
ET_LNK	equ	1			; Ethernet adaptor linked
ET_WFI	equ	2			; Wifi module present
ET_CYC	equ	3			; Hold communication cycle
ET_DIS	equ	4			; DHCP Discover sent
ET_OFR	equ	5			; DHCP Offer received
ET_REQ	equ	6			; DHCP Request sent
ET_ADR	equ	7			; IP address aquired
; Ethernet flags (ETH_FL) bits
EF_TMO	equ	0			; Time-out
EF_RDA	equ	1			; Read alternate
EF_WRA	equ	2			; Write alternate
EF_RPL	equ	3			; Sending reply
EF_STZ	equ	4			; Station 0
EF_REQ	equ	5			; Request sent
EF_DNS	equ	6			; DNS query sent
EF_HED	equ	7			; Packet header

; Commands
; Server host commands
SV_RDS  equ     0x10                    ; Read sector
SV_WRS  equ     0x20                    ; Write sector
SV_CRE	equ	0x40			; Create temporary file
SV_OPN	equ	0x60			; Open temporary file
; Server MCU commands
POL_KY	equ	0xf0			; Keyboard/mouse polling
GET_KY	equ	0xf1			; Read keyboard/mouse
NS_FWR	equ	0xf9			; PIC firmware request
WF_FW1	equ	0xfa			; ESP firmware request
WF_FW2	equ	0xfb			; ESP firmware request
; Server Receive-Transmit
S_RXTX	equ	0xe0                    ; Receive-Transmit
; Asynchronous serial port commands
ASY_TX	equ     0xe1                    ; Serials Transmit
ASY_RX	equ     0xe2                    ; Serial Receive
; Networking commands
NET_TX	equ	0xe3			; Net Transmit
NET_RX	equ	0xe4			; Net Receive
ST_LNK	equ	0xe5			; Link station
ST_CHK	equ	0xe6			; Check station
; Periferial host command
PER_CM  equ     0xf0                    ; Periferial host command
; SD card Commands
SD_PLY	equ	0xf3			; Play tape file
SD_RWR  equ     0xf4                    ; Receive-write
SD_SWA  equ     0xf5                    ; Set Write Address
SD_RDM  equ     0xf6                    ; Read Multiple Blocks
SD_RDT  equ     0xf7                    ; Read-Transmit
SD_RDB  equ     0xf8                    ; Read Single Block
SD_RDA  equ     0xf9                    ; Read to alternate buffer
SD_WRB  equ     0xfa                    ; Write Block
SD_WRA  equ     0xfb                    ; Write from alternate buffer
SD_IDF  equ     0xfc                    ; Identify SD media
SD_EFA  equ     0xfd                    ; Set First Erase Address
SD_ELA  equ     0xfe                    ; Set Last Erase Address
SD_ERA  equ     0xff                    ; Erase Blocks

; Error codes
EER	equ	0x2a
E_ISTA	equ	0x05+EER		; Error 'Invalid station'
E_COMM  equ     0x09+EER		; Error 'Communication error'
E_WRPR  equ     0x0e+EER		; Error 'Drive write protected'
E_NPRE  equ     0x10+EER		; Error 'Device not present'
E_ENOF	equ	0x07			; Error 'End of file'

; USB related constants
; Processor specific
; BDnSTAT bits (DS39632E Register 17-5, 17-6) missing from .INC file
BSTALL  equ     2                       ; Stall
DTSEN   equ     3                       ; Data Sync Enable
DTS     equ     6                       ; Data Sync Parity
UOWN    equ     7                       ; USB Owned
PID0    equ     2                       ; PID bit 0
PID1    equ     3                       ; PID bit 1
PID2    equ     4                       ; PID bit 2
PID3    equ     5                       ; PID bit 3
; Application specific
; General
CF_MAX  equ     1                       ; Maximum configuration number
IF_MAX  equ     1                       ; Maximum interface number
EP_MAX  equ     2                       ; Maximum endpoint number
; CC Endpoint Numbers
EP_CTL  equ     1                       ; CC Control Endpoint number
EP_DAT  equ     EP_MAX+1-EP_CTL         ; CC Data Endpoint number
; Maximum packet size (binary logarithm)
PA_CTL  equ     3                       ; LOG2 of control packet size
PA_DAT  equ     6                       ; LOG2 of data packet size
; Other
TB_SEL  equ     0                       ; Table selector bit
; Standard USB Constants
; Token PID Types (USB 2.0 Table 8-1)
PID_SU  equ     b'1101'                 ; SETUP Token
; Request Type Bits (USB 2.0 Table 9-2)
RT_ITF  equ     0                       ; Interface
RT_ENP  equ     1                       ; Endpoint
RT_CLA  equ     5                       ; Class
RT_VEN  equ     6                       ; Vendor
RT_DIR  equ     7                       ; Data Direction IN
; Standard Request Codes (USB 2.0 Table 9-4)
R_GSTA  equ     0x00                    ; Get Status
R_CFTR  equ     0x01                    ; Clear Feature
R_SFTR  equ     0x03                    ; Set Feature
R_SADR  equ     0x05                    ; Set Address
R_GDES  equ     0x06                    ; Get Descriptor
R_GCFG  equ     0x08                    ; Get Configuration
R_SCFG  equ     0x09                    ; Set Configuration
R_GITF  equ     0x0a                    ; Get Interface
R_SITF  equ     0x0b                    ; Set Interface
; Standard Descriptor Types (USB 2.0 Table 9-5)
DT_DEV  equ     0x01                    ; Device
DT_CFG  equ     0x02                    ; Configuration
DT_STR  equ     0x03                    ; String
DT_ITF  equ     0x04                    ; Interface
DT_ENP  equ     0x05                    ; Endpoint
; Feature Selectors (USB 2.0 Table 9-6)
DF_RWU  equ     0x01                    ; Device Remote Wake-up
EF_HLT  equ     0x00                    ; Endpoint HALT
; Endpoint Status bits (USB 2.0 Figure 9-6)
GS_HLT  equ     0                       ; HALT
; Standard Configuration Attributes (USB 2.0 Table 9-10)
CA_RWU  equ     1<<7|1<<5               ; Remote Wake-up
CA_SPW  equ     1<<7|1<<6               ; Self powered
; Standard Endpoint Address bits (USB 2.0 Table 9-13)
EA_DIR  equ     7                       ; Direction IN
; Standard Endpoint Attributes (USB 2.0 Table 9-13)
EA_BLK  equ     0x02                    ; Bulk
EA_INT  equ     0x03                    ; Interrupt
; CC Device and Interface (Sub)Classes (CDC 1.2 Table 2,3,4,6)
DC_NUS  equ     0x00                    ; Not Used
DC_COM  equ     0x02                    ; Communication Device Class
IC_COM  equ     0x02                    ; Communication Interface Class
IC_DAT  equ     0x0a                    ; Data Interface Class
IS_ACM  equ     0x02                    ; Abstract Control Interface Subclass
; CC Protocols (CDC 1.2 Table 5,7)
PC_NOP  equ     0x00                    ; No protocol
PC_ATC  equ     0x01                    ; AT Commands V.250
; CC Descriptor Types (CDC 1.2 Table 12)
DT_CSI  equ     0x24                    ; CS Interface Descriptor
; CC Functional Descriptor Subtypes (CDC 1.2 Table 13)
FD_HED  equ     0x00                    ; Header Descriptor
FD_CMA  equ     0x01                    ; Call Management Descriptor
FD_ACM  equ     0x02                    ; Abstract Control Descriptor
FD_UNI  equ     0x06                    ; Union Descriptor
; CC Request Codes (CDC 1.2 Table 19)
C_SECO  equ     0x00                    ; Send Encapsulated Command
C_GERP  equ     0x01                    ; Get Encapsulated Response
C_SLCO  equ     0x20                    ; Set Line Coding
C_GLCO  equ     0x21                    ; Get Line Coding
C_SCLS  equ     0x22                    ; Set Control Line State
; CC Notifications (CDC 1.2 Table 20)
N_SSTA  equ     0x20                    ; Serial State
; CC Abstract Control Management Capability Bits (PSTN 1.2 Table 4)
CA_AC1  equ     1                       ; C_SLCO, C_GLCO, C_SCLS, N_SSTA
; CC Control Line State Bits (CLN_ST) (PSTN 1.2 Table 18)
CS_DTR  equ     0                       ; DTR
CS_RTS  equ     1                       ; RTS
; CC Serial Line State Bits (PSTN 1.2 Table 31)
SS_DCD  equ     0                       ; DCD
SS_DSR  equ     1                       ; DSR

; SD card related constants
CB_TRN  equ     6                       ; Command Transfer bit
CB_HCS  equ     6                       ;
R1_IDL  equ     0                       ; R1 'Idle State' bit
R1_ILC  equ     2                       ; R1 'Illegal Command' bit
C8_VHS  equ     b'0001'                 ; CMD8 VHS field
C8_PAT  equ     0xaa                    ; CMD8 check pattern
OC_CCS  equ     6
OC_PWU  equ     7
SD_DST  equ     0xfe                    ; DATA_START_TOKEN
SD_DAC  equ     0x05                    ; DATA_ACCEPTED
CSD_HC  equ     6
SD_STR  equ     .12|1<<CB_TRN           ; Stop transmission
SD_WBK	equ	.24|1<<CB_TRN		; Write block

; PS/2 mouse related constants
MV_MBL  equ     0                       ; PS/2 mouse left button
MV_MBR  equ     1                       ; PS/2 mouse right button
MV_MBM  equ     2                       ; PS/2 mouse middle button
MV_REZ  equ     3                       ; PS/2 mouse rezerved = 1
MV_SGN  equ     5                       ; Y-movement sign bit
; PS/2 mouse/keyboard commands and resposes
PS_RST  equ     0xff                    ; Reset
PS_RSD  equ     0xfe                    ; Resend
PS_ERR  equ     0xfc                    ; Error
PS_ACK  equ     0xfa                    ; Acknowledge
PS_REP  equ     0xf4                    ; Enable data reporting
PS_TST  equ     0xaa                    ; Self-test passed

; Serial RAM related constants
; Serial RAM commands
;SR_MWR	equ	1			; Write Mode
SR_DWR	equ	2			; Write Data
SR_DRD	equ	3			; Read Data
SR_MRD	equ	5			; Read Mode 
; Serial Ram status bits
SR_SEQ	equ	6			; Sequential mode
;SR_PAG	equ	7			; Page mode

; Asynchronous serial port related constants
SRB_RZ	equ	.16			; Receive buffer reserve
SBF_EN	equ	0			; End of buffer 0xd00

; Tape player constants
RBF_SZ	equ	3			; Ring buffer size
NON_DT	equ	5			; Non data TZX block bit
TZX_HD	equ	6			; TZX header block bit
TAP_SL	equ	.0			; Tape sector (low)
TAP_SH	equ	.1			; Tape sector (high)
TAP_LL	equ	.4			; Tape length (low)
TAP_LH	equ	.5			; Tape length (high)
TAP_LU	equ	.6			; Tape length (upper)
FILE_T	equ	.7			; Tape file/block type
PAUS_L	equ	.8			; Pause length (low)
PAUS_H	equ	.9			; Pause length (high)
LAST_B	equ	.10			; Last byte bits number
TY_TAP	equ	0x0a			; .TAP file type
BK_TAP  equ     0x10			; Standard block
BK_TRB	equ	0x11			; Turbo block
BK_TON	equ	0x12			; Pure tone
BK_PLS	equ	0x13			; Pulse sequence
BK_DAT	equ	0x14			; Data block
BK_PAU	equ	0x20			; Pause
BK_STA	equ	0x21			; Group start
BK_END	equ	0x22			; Group end
BK_STP  equ     0x2a			; Stop the tape
BK_LEV	equ	0x2b			; Set signal level
BK_TXT  equ     0x30			; Text description
BK_MSG	equ	0x31			; Message
BK_ARC  equ     0x32			; Archive info
BK_HWR	equ	0x33			; Hardware info
;BK_INF	equ	0x35			; Custom info

; Ethernet controller related constants
; Ethernet controller buffer memory
BF_RPL	equ	0x2000-.576		; Reply buffer start
BF_REQ	equ	BF_RPL-.576		; Request buffer start
BF_RXE	equ	(BF_REQ-2)|1		; Receive buffer end              
BF_RXS	equ	0			; Receive buffer start
; General Ethernet constants
MAC_LN	equ	6			; Length of MAC address
IPN_LN	equ	4			; Length of IP number
IP_PRO	equ	0x0800			; IP protocol
ARP_PR	equ	0x0806			; ARP protocol
ETH_TY	equ	1			; Ethernet network type
IP_VER	equ	0x40			; IP version code
IP_SRV	equ	0			; IP service type
IP_TTL	equ	.64			; Frame time-to-live
OT_REQ	equ	1			; Request operation
OT_REP	equ	2			; Reply operation
UDP_PN	equ	.17			; UDP protocol number
HC_UDL	equ	.308			; DHCP packet UDP length
HC_OPL	equ	.32			; DHCP options length
UD_PYL	equ	.520			; Maximum UDP payload
; Network datagram control bits
PAR	equ	7			; Parity
REQ	equ	6			; Request
FDV	equ	5			; File device
EOF	equ	4			; End of file
CLN	equ	3			; Client
; Station table constants
MST	equ	low(RPOR0)>>4		; Total stations limit
SST	equ	1<<CLN			; Server stations limit			
CST	equ	MST-SST+1		; Client stations limit
SRL	equ	.16			; Station table record length
SHL	equ	4			; Station table header length
; UDP ports
DNS_SV	equ	.53			; DNS server
DHC_SV	equ	.67			; DHCP server
UDP_CL	equ	.41772			; Device UDP
; DHCP message types
MS_DIS	equ	1			; Discovery
MS_OFR	equ	2			; Offer
MS_REQ	equ	3			; Request
MS_ACK	equ	5			; Acknowledge
MS_NAK	equ	6			; NAK
; DHCP options
OP_MSK	equ	.1			; Subnet mask
OP_GAT	equ	.3			; Router
OP_DNS	equ	.6			; DNS server
OP_HNA	equ	.12			; Host name
OP_RIP	equ	.50			; Request IP address
OP_MTY	equ	.53			; Message type
OP_SVI	equ	.54			; Server identifier
OP_LIS	equ	.55			; Parameter request list
OP_END	equ	.255			; Option end
;DSC	equ	2			; Check all sections
DSC	equ	1			; Check only the 'Ansers' section
; Ethernet controller instruction set
;RCR    equ     0x00                    ; Read Control Register
RBM     equ     0x3a                    ; Read Buffer Memory
WCR     equ     0x40                    ; Write Control Register
WBM     equ     0x7a                    ; Write Buffer Memory
BFS     equ     0x80                    ; Set Control Register Bit
BFC     equ     0xa0                    ; Clear Control Register Bit
;RST    equ     0xff                    ; Reset
; Command interpreter instructions
;	equ	0x20			; 
WRB	equ	0x24			; Write from (LFSR1) to buffer
CLB	equ	0x28			; Write zero to buffer	
CMP	equ	0x2c			; Compare string	
CPY	equ	0x30			; Copy string	
CLR	equ	0x34			; Fill with '0'-s	
;	equ	0x60			; + 11-bit
;	equ	0x68			; + 11-bit
;	equ	0x70			; + 11-bit
LX1	equ	0xc0			; Load (FSR1)
LX2	equ	0xd0			; Load (FSR2)
;	equ	0xe0			; + 12-bit
RXI	equ	0xf0			; Read reg to (LFSR1) & inc
RXD	equ	0xf1			; Read reg to (LFSR1) & dec
WDI	equ	0xf2			; Write direct to (LFSR1) & inc
WDD	equ	0xf3			; Write direct to (LFSR1) & dec
WXI	equ	0xf4			; Write reg from (LFSR1) & inc
WXD	equ	0xf5			; Write reg from (LFSR1) & dec
WPH	equ	0xf6			; Write PHY register
TCY	equ	0xf7			; Copy from table to (LFSR1)
SKP	equ	0xf8			; Relative forward skip if set
;	equ	0xf9			;
GTO	equ	0xfa			; Relative forward Jump to table
;	equ	0xfb			;
CAL	equ	0xfc			; Relative forward call table
;	equ	0xfd			;
SBR	equ	0xfe			; Execute subroutine
RET	equ	0xff			; End of command

; Wifi module related constants
WF_VER	equ	1
APC_LN	equ	.16			; Length of AP ID and password
; Wifi module communication tokens
TK_ACK	equ	'A'			; Acknowledge
TK_DNS	equ	'D'			; DNS
TK_END	equ	'E'			; End
TK_FLA	equ	'F'			; Flash
TK_HED	equ	'H'			; Header
TK_INI	equ	'I'			; Initialization
TK_PLD	equ	'L'			; Payload
TK_RPL	equ	'P'			; Reply
TK_REQ	equ	'Q'			; Request
TK_RRQ	equ	'R'			; Resend request
TK_RDN	equ	'S'			; Resend DNS
TK_VER	equ	'V'			; Version
TK_WIF	equ	'W'			; Start Wifi
TK_RDY	equ	'Y'			; Ready

; Keyboard related constants
; ZX Spectrum key numbers
KN_B	equ	0x00
KN_H	equ	0x01
KN_Y	equ	0x02
KN_6	equ	0x03
KN_5	equ	0x04
KN_T	equ	0x05
KN_G	equ	0x06
KN_V	equ	0x07
KN_N	equ	0x08
KN_J	equ	0x09
KN_U	equ	0x0a
KN_7	equ	0x0b
KN_4	equ	0x0c
KN_R	equ	0x0d
KN_F	equ	0x0e
KN_C	equ	0x0f
KN_M	equ	0x10
KN_K	equ	0x11
KN_I	equ	0x12
KN_8	equ	0x13
KN_3	equ	0x14
KN_E	equ	0x15
KN_D	equ	0x16
KN_X	equ	0x17
N_SS	equ	0x18			; Symbol Shift
KN_L	equ	0x19
KN_O	equ	0x1a
KN_9	equ	0x1b
KN_2	equ	0x1c
KN_W	equ	0x1d
KN_S	equ	0x1e
KN_Z	equ	0x1f
N_SP	equ	0x20			; Space
N_EN	equ	0x21			; Enter
KN_P	equ	0x22
KN_0	equ	0x23
KN_1	equ	0x24
KN_Q	equ	0x25
KN_A	equ	0x26
N_CS	equ	0x27			; Caps Shift
; Set 2 keyboard scan codes
; Extended codes have bit 7 set
SC_A	equ	0x1c
SC_B	equ	0x32
SC_C	equ	0x21
SC_D	equ	0x23
SC_E	equ	0x24
SC_F	equ	0x2b
SC_G	equ	0x34
SC_H	equ	0x33
SC_I	equ	0x43
SC_J	equ	0x3b
SC_K	equ	0x42
SC_L	equ	0x4b
SC_M	equ	0x3a
SC_N	equ	0x31
SC_O	equ	0x44
SC_P	equ	0x4d
SC_Q	equ	0x15
SC_R	equ	0x2d
SC_S	equ	0x1b
SC_T	equ	0x2c
SC_U	equ	0x3c
SC_V	equ	0x2a
SC_W	equ	0x1d
SC_X	equ	0x22
SC_Y	equ	0x35
SC_Z	equ	0x1a
SC_0	equ	0x45
SC_1	equ	0x16
SC_2	equ	0x1e
SC_3	equ	0x26
SC_4	equ	0x25
SC_5	equ	0x2e
SC_6	equ	0x36
SC_7	equ	0x3d
SC_8	equ	0x3e
SC_9	equ	0x46
C_AC	equ	0x0e			; `
C_MI	equ	0x4e			; -
C_EQ	equ	0x55			; =
C_IS	equ	0x5d			; \
C_CO	equ	0x4c			; ;
C_AP	equ	0x52			; '
C_CM	equ	0x41			; ,
C_DO	equ	0x49			; .
C_SL	equ	0x4a			; /
C_PL	equ	0x54			; [
C_PR	equ	0x5b			; ]
C_SP	equ	0x29			; Space
C_EN	equ	0x5a			; Enter
C_ES	equ	0x76			; Esc
C_BS	equ	0x66			; Backspace
C_CL	equ	0x58			; Caps Lock
C_TB	equ	0x0d			; Tab
C_AL	equ	0x11			; Alt
C_RA	equ	0x91			; Right Alt
C_SH	equ	0x12			; Shift
C_RS	equ	0x59			; Right Shift
C_CT	equ	0x14			; Ctrl
C_RC	equ	0x94			; Right Ctrl
C_WD	equ	0x9f			; Win
C_RW	equ	0xa7			; Right Win
C_ME	equ	0xaf			; Menu
C_LT	equ	0xeb			; Arrow left
C_RT	equ	0xf4			; Arrow right
C_UP	equ	0xf5			; Arrow up
C_DN	equ	0xf2			; Arrow down
C_HO	equ	0xec			; Home
C_ED	equ	0xe9			; End
C_IN	equ	0xf0			; Insert
C_PU	equ	0xfd			; PgUp
C_PD	equ	0xfa			; PgDn
C_DE	equ	0xf1			; Delete
C_NL	equ	0x77			; Num Lock
C_EX	equ	0xe0			; Extended code
C_BK	equ	0xf0			; Break code
C_ST	equ	0x7a

; ENC28J60 registers
; Bank 0 registers
ERDPTL		equ     0x00
ERDPTH		equ     0x01
EWRPTL		equ     0x02
EWRPTH		equ     0x03
ETXSTL		equ     0x04
ETXSTH		equ     0x05
ETXNDL		equ     0x06
ETXNDH		equ     0x07
ERXSTL		equ     0x08
ERXSTH		equ     0x09
ERXNDL		equ     0x0A
ERXNDH		equ     0x0B
ERXRDPTL	equ     0x0C
ERXRDPTH	equ     0x0D
ERXWRPTL	equ     0x0E
ERXWRPTH	equ     0x0F
;EDMASTL	equ     0x10
;EDMASTH	equ     0x11
;EDMANDL	equ     0x12
;EDMANDH	equ     0x13
;EDMADSTL	equ     0x14
;EDMADSTH	equ     0x15
;EDMACSL	equ     0x16
;EDMACSH	equ     0x17
EIE		equ     0x1B
EIR		equ     0x1C
ESTAT		equ     0x1D
ECON2		equ     0x1E
ECON1		equ     0x1F
; Bank 1 registers
;EHT0		equ     0x00
;EHT1		equ     0x01
;EHT2		equ     0x02
;EHT3		equ     0x03
;EHT4		equ     0x04
;EHT5		equ     0x05
;EHT6		equ     0x06
;EHT7		equ     0x07
;EPMM0		equ     0x08
;EPMM1		equ     0x09
;EPMM2		equ     0x0A
;EPMM3		equ     0x0B
;EPMM4		equ     0x0C
;EPMM5		equ     0x0D
;EPMM6		equ     0x0E
;EPMM7		equ     0x0F
;EPMCSL		equ     0x10
;EPMCSH		equ     0x11
;EPMOL		equ     0x14
;EPMOH		equ     0x15
;ERXFCON	equ     0x18
EPKTCNT         equ     0x19
; Bank 2 registers
MACON1		equ     0x00
MACON3		equ     0x02
MACON4		equ     0x03
MABBIPG		equ     0x04
MAIPGL		equ     0x06
MAIPGH		equ     0x07
;MACLCON1	equ     0x08
MACLCON2	equ     0x09
MAMXFLL		equ     0x0A
MAMXFLH		equ     0x0B
MICMD		equ     0x12
MIREGADR	equ     0x14
MIWRL		equ     0x16
MIWRH		equ     0x17
MIRDL		equ     0x18
MIRDH		equ     0x19
; Bank 3 registers
MAADR5		equ     0x00
MAADR6		equ     0x01
MAADR3		equ     0x02
MAADR4		equ     0x03
MAADR1		equ     0x04
MAADR2		equ     0x05
;EBSTSD		equ     0x06
;EBSTCON	equ     0x07
;EBSTCSL	equ     0x08
;EBSTCSH	equ     0x09
MISTAT		equ     0x0A
;EREVID		equ     0x12
ECOCON		equ     0x15
;EFLOCON	equ     0x17
;EPAUSL		equ     0x18
;EPAUSH		equ     0x19
; PHY Registers
PHCON1		equ     0x00
;PHSTAT		equ     0x01
;PHID1		equ     0x02
;PHID2		equ     0x03
PHCON2		equ     0x10
PHSTAT2		equ     0x11
PHIE		equ     0x12
PHIR		equ     0x13
PHLCON		equ     0x14
; EIE bits
INTIE		equ     7
PKTIE		equ     6
;DMAIE		equ     5
LINKIE		equ     4
;ETXIE		equ     3
TXERIE		equ     1
RXERIE		equ     0
; EIR bits
PKTIF		equ     6
;DMAIF		equ     5
LINKIF		equ     4
;ETXIF		equ     3
TXERIF		equ     1
;RXERIF		equ     0
; ESTAT bits
;INT		equ     7
;BUFER		equ     6
LATECOL		equ     4
;RXBUSY		equ     2
TXABRT		equ     1
;CLKRDY		equ     0
; ECON2 bits
;AUTOINC	equ     7
PKTDEC		equ     6
;PWRSV		equ     5
;VRPS		equ     3
; ECON1 bits
TXRST		equ     7
;RXRST		equ     6
;DMAST		equ     5
;CSUMEN		equ     4
TXRTS		equ     3
RXEN		equ     2
BSEL1		equ     1
BSEL0		equ     0
; ERXFCON bits
;UCEN		equ     7
;ANDOR		equ     6
;CRCEN		equ     5
;PMEN		equ     4
;MPEN		equ     3
;HTEN		equ     2
;MCEN		equ     1
;BCEN		equ     0
; MACON1 bits
TXPAUS		equ     3
RXPAUS		equ     2
;PASSALL	equ     1
MARXEN		equ     0
; MACON3 bits
;PADCF2		equ     7
;PADCF1		equ     6
PADCF0		equ     5
TXCRCEN		equ     4
;PHDREN		equ     3
;HFRMEN		equ     2
FRMLNEN		equ     1
;FULDPX		equ     0
; MACON4 bits
DEFER           equ     6
;BPEN		equ     5
;NOBKOFF	equ     4
; MICMD bits
;MIISCAN	equ     1
MIIRD		equ     0
; EBSTCON bits
;PSV2		equ     7
;PSV1		equ     6
;PSV0		equ     5
;PSEL		equ     4
;TMSEL1		equ     3
;TMSEL0		equ     2
;TME		equ     1
;BISTST		equ     0
; MISTAT bits
;NVALID		equ     2
;SCAN		equ     1
MIBUSY          equ     0
; ECOCON bits
;COCON2		equ     2
;COCON1		equ     1
;COCON0		equ     0
; EFLOCON bits
;FULDPXS	equ     2
;FCEN1		equ     1
;FCEN0		equ     0
; PHCON1 bits
;PRST		equ     0x0F
;PLOOPBK	equ     0x0E
;PPWRSV		equ     0x0B
;PDPXMD		equ     0x08
; PHSTAT1 bits
;PFDPX		equ     0x0C
;PHDPX		equ     0x0B
;LLSTAT		equ     0x02
;JBSTAT		equ     0x01
; PHID2 bits
;PID24		equ     0x0F
;PID23		equ     0x0E
;PID22		equ     0x0D
;PID21		equ     0x0C
;PID20		equ     0x0B
;PID19		equ     0x0A
;PPN5		equ     0x09
;PPN4		equ     0x08
;PPN3		equ     0x07
;PPN2		equ     0x06
;PPN1		equ     0x05
;PPN0		equ     0x04
;PREV3		equ     0x03
;PREV2		equ     0x02
;PREV1		equ     0x01
;PREV0		equ     0x00
; PHCON2 bits
;FRCLNK		equ     0x0E
;TXDIS		equ     0x0D
;JABBER		equ     0x0A
HDLDIS		equ     0x08
; PHSTAT2 bits
;TXSTAT		equ     0x0D
;RXSTAT		equ     0x0C
;COLSTAT	equ     0x0B
LSTAT		equ     0x0A
;DPXSTAT	equ     0x09
;PLRITY		equ     0x05
; PHIE bits
PLNKIE          equ     0x04
PGEIE           equ     0x01
; PHIR bits
;PLNKIF		equ     0x04
;PGIF		equ     0x02
; PHLCON bits
;LACFG3		equ     0x0B
;LACFG2		equ     0x0A
;LACFG1		equ     0x09
;LACFG0		equ     0x08
;LBCFG3		equ     0x07
;LBCFG2		equ     0x06
;LBCFG1		equ     0x05
;LBCFG0		equ     0x04
;LFRQ1		equ     0x03
;LFRQ0		equ     0x02
;STRCH		equ     0x01

; Other constants
ESC	equ	0x1b
PGS	equ	0x100			; Page size
SCS	equ	0x200			; Sector size
FLA_DV	equ	0x80			; Flash device
FIL_DV	equ	0xb0			; File device

;------------------------------------------------------------------------------
; MACRO INSTRUCTIONS

l_time  macro   ms
        MOVLW   high (~(FOSC_4/.1000*ms/PGS))
        MOVWF   TMR0H
        MOVLW   low  (~(FOSC_4/.1000*ms/PGS))
        endm

s_time  macro   us
        MOVLW   high (~(FOSC_4/.1000000*us/8))
        MOVWF   TMR3H
        MOVLW   low  (~(FOSC_4/.1000000*us/8))
        endm

copy_b  macro   src,dst,num
        local   i = 0
        while   i < num
        MOVFF   src+i,dst+i
i ++
        endw
        endm

tab_ad  macro   adr
        MOVLW   low(adr)
        MOVWF   TBLPTRL
        MOVLW   high(adr)
        MOVWF   TBLPTRH
        MOVLW   upper(adr)
        MOVWF   TBLPTRU
        endm

table_	macro   adr
	MOVLW   low(adr)
        MOVWF   TBLPTRL
        MOVLW   high(adr)
        MOVWF   TBLPTRH
	endm

swap_r  macro   r1,r2
        MOVF    r1,W
        MOVFF   r2,r1
        MOVWF   r2
        endm

eth_wr  macro   reg,dat
        MOVLW   reg
	RCALL	WR_ADR
        MOVLW   dat
	RCALL	WR_DAT
        endm

eth_bk	macro	bank
	if bank < 3
	eth_wr	BFC|ECON1,1<<BSEL1|1<<BSEL0
	endif
	if bank > 0
	eth_wr	BFS|ECON1,bank
	endif
        endm

be_wrd	macro	wrd
	db	high(wrd),low(wrd)
	endm

d_byte	macro	byt
	db	byt/.10+'0',byt%.10+'0'
	endm

et_lfs	macro	cmd,adr
	db	cmd|high(adr),low(adr)
	endm

et_rel	macro	cmd,adr
	if	adr-$-2 < PGS
	db	cmd,adr-$-2
	else
	db	RELERR
	endif
	endm

et_sbr	macro	adr
	db	SBR,upper(adr),high(adr),low(adr)
	endm

et_wph	macro	reg,dat
	db	WPH,reg,low(dat),high(dat)
	endm

;------------------------------------------------------------------------------
; VARIABLES

        cblock  0x000                   ; Variables in access RAM

; General variables
COM_ST                                  ; Communication State
TIM_ST                                  ; Timer 0 State
CHG_ST                                  ; Last PORTB value
USB_ST                                  ; USB State
SDC_ST                                  ; SD card State
CNT_LO                                  ; Byte counter low
CNT_HI                                  ; Byte counter high
EP_OFS                                  ; Offset of last active EP
TB_CNT                                  ; Table read byte counter
; Mouse variables
MSE_ST                                  ; Mouse state
JS_SEN                                  ; Mouse as joystick sensitivity
JSTC_L,JSTC_H                           ; and sampling timing constant
MWIN_X,MWIN_Y                           ; Mouse window size
MPOS_X,MPOS_Y                           ; Mouse coordinates
MOVE_B,MOVE_X,MOVE_Y                    ; Mouse movement packet
MS_REG                                  ; Mouse work register
; SD card variables
SD_SZ0,SD_SZ1,SD_SZ2,SD_SZ3             ; SD card size
CM_CNL,CM_CNH,DT_CNL,DT_CNH             ; SD Command counters
; SD card Command
SD_CMD,SD_AD3,SD_AD2,SD_AD1,SD_AD0,SD_CRC
; USB Request Header
RQ_TYP,RQ_NUM,RQ_VAL,RQ_DTY,RQ_IDX:2,RQ_LNL,RQ_LNH
; System flags
SYS_FL
; Serial asynchronous port variables
WRPT_L,WRPT_H				; Write buffer pointer
RDPT_L,RDPT_H				; Read buffer pointer
; Keyboard variables
KYB_ST					; Keyboard state
; Tape player variables
PLY_ST					; Tape player state
; Ethernet controller variables
ETH_ST,ETH_FL				; Ethernet controller state
TO_PER,TICK_N,TICK_L,TICK_H		; Ethernet timer variables
MSG_TM,REQ_ST
; Spare access RAM
AC_RAM
        endc

        cblock  0x050
; Main communication buffer
CMD_CD,CTL_CD,BLN_LO,BLN_HI             ; Communication Data Header
DATA_B:SCS-CMD_CD-4                  	; Communication Data Block
RING_B:DATA_B
ALT_DB:SCS
; Ethernet receive buffer
; Received packet header
PH_NPK:2,PH_FRL:2,PH_STA:2		; Next packet pointer and status
; Ethernet header
ER_DMA:MAC_LN,ER_SMA:MAC_LN,ER_TYH,ER_TYL
; IP Header
IR_VER,IR_SRV,IR_LEN:2
IR_LDH,IR_LDL,IR_FRH,IR_FRL
IR_TTL,IR_PRO,IR_CKS:2
IR_SIP:IPN_LN
IR_DIP:IPN_LN
; UDP Header
UR_SPO:2,UR_DPO:2
UR_LNH,UR_LNL,UR_CKS:2
; DHCP data
HR_OPR,HR_HWR,HR_MCL,HR_HOP
HR_XID:4
HR_SEC:2,HR_FLG:2
HR_CIA:IPN_LN
HR_YIA:IPN_LN
HR_SIA:IPN_LN
HR_GIA:IPN_LN
HR_CHA:MAC_LN
HR_SHN:.10+.64
HR_BFN:.128
HR_COK:4
HR_OPT:HR_OPR+UD_PYL-HR_COK-4
; Ethernet transmit buffer
; Ethernet header
ET_DMA:MAC_LN,ET_SMA:MAC_LN,ET_TYH,ET_TYL
; IP Header
IT_VER,IT_SRV,IT_LNH,IT_LNL
IT_LDH,IT_LDL,IT_FRH,IT_FRL
IT_TTL,IT_PRO,IT_CKS:2
IT_SIP:IPN_LN
IT_DIP:IPN_LN
; UDP Header
UT_SPO:2,UT_DPO:2
UT_LNH,UT_LNL,UT_CKS:2
; DHCP data
HT_OPR,HT_HWR,HT_MCL,HT_HOP
HT_XID:4
HT_CHA:MAC_LN
HT_COK:4
HT_OPT:2,HT_MSG:HC_OPL-2
HT_END:(PGS<<RBF_SZ)-HT_OPT-HC_OPL
; Sector buffers
FAT_BF:SCS
AUX_BF:SCS
; Asyncronous serial receive buffer
S_BUFF:2				; Serial input buffer, also used for
SYNC_1:2, SYNC_2:2 			; the printer initialization string,
BIT__0:2, BIT__1:2			; the keyboard token buffer and the
TONE_0:2, TONE_1:2			; tape player timing constants
        endc

; ARP variables
AR_HWR	equ	IR_VER
;AR_PRO	equ	IR_LEN
;AR_MCL	equ	IR_LDH
;AR_IPL	equ	IR_LDL
AR_OPH	equ	IR_FRH
;AR_OPL	equ	IR_FRL
AR_SMA	equ	IR_TTL
AR_SIA	equ	IR_SIP+2
;AR_DMA	equ	IR_DIP+2
AR_DIA	equ	UR_LNH
AT_HWR	equ	IT_VER
;AT_PRO	equ	IT_LNH
;AT_MCL	equ	IT_LDH
;AT_IPL	equ	IT_LDL
;AT_OPH	equ	IT_FRH
;AT_OPL	equ	IT_FRL
AT_SMA	equ	IT_TTL
AT_SIA	equ	IT_SIP+2
AT_DMA	equ	IT_DIP+2
AT_DIA	equ	UT_LNH

; DNS variables
NR_IDN	equ	HR_OPR
NR_PAR	equ	HR_MCL
NR_QDC	equ	HR_XID
NR_NAM	equ	HR_CIA
NT_IDN	equ	HT_OPR
NT_PAR	equ	HT_MCL
NT_ANC	equ	HT_XID+2
NT_NAM	equ	HT_CHA+4

; Data packet variables
PR_HED	equ	UR_SPO
PR_LEN	equ	UR_DPO
PR_SIP	equ	UR_LNH
PR_CMD	equ	HR_OPR
PR_CTL	equ	HR_HWR
PR_LNL	equ	HR_MCL
;PR_LNH	equ	HR_HOP
PR_BLK	equ	HR_XID
PR_DAT	equ	HR_SEC
PT_CMD	equ	HT_OPR
PT_CTL	equ	HT_HWR
PT_LNL	equ	HT_MCL
;PT_LNH	equ	HT_HOP
PT_BLK	equ	HT_XID
PT_DAT	equ	HT_CHA

; Derived variables
IP_HLN	equ	UR_SPO-IR_VER		; Minimum IP header length
UD_HLN	equ	HR_OPR-UR_SPO		; UDP header length
PL_HLN	equ	PR_BLK-PR_CMD		; Payload header length
PL_BLN	equ	PR_DAT-PR_BLK		; Payload block number length
AP_CRD	equ	PR_DAT+.72		; AP credentials address

        cblock  0xd00
; USB Buffer Descriptor Table
BDT0_O,EP0O_L:3,BDT0_I,EP0I_L:0x13
; USB Endpoint 0 buffers
BUF0_O:1<<PA_CTL                        ; EP0 OUT buffer
BUF0_I:1<<PA_CTL                        ; EP0 IN buffer
; CC Notification Header
NT_TYP,NT_NUM,NT_VAL:2,NT_IDX:2,NT_LEN:2
; CC Notification Data
NT_DAT:4
; CC Line Coding data
LN_COD:7
ECM_SP					; Command interpreter stack pointer
I_FSR1:2				; FSR1 saved during interrupt
SD_IDS:.16				; SD card ID string
WR_BLN:4				; Write address for block devices
SRC_ST,DST_ST				; Source and destination station
TXE_PT:2				; Transmission buffer end pointer
CRT_ST:2				; Current station ID
; Wifi AP credentials
WAP_ID:APC_LN				; Wireless access point ID
WAP_PW:APC_LN				; Wireless access point password
; Ethernet configuration data
HOST_N:4,HOST_I:4			; Unique Host name
MA_OWN:MAC_LN-2, MAC_NR:2		; Unique MAC address
IP_OWN:IPN_LN-1,IP_LDG			; Host IP number
IP_MSK:IPN_LN				; Subnet mask
IP_GAT:IPN_LN				; Default gateway IP number
IP_DNS:IPN_LN 				; DNS server IP number
MA_DNS:MAC_LN				; DNS server MAC address
MA_GAT:MAC_LN-2,GAT_NR:2		; Default gateway MAC address
IP_STA:IPN_LN
ETH_BF:RPOR0-SRL*MST-IP_STA-IPN_LN
; Stations table
SST_TB:SRL*(SST-1)
CST_TB:SRL*(CST-1)
NA_SFR
        endc

;------------------------------------------------------------------------------
; RESET

        org     0x0000
        CLRF    COM_ST                  ; Clear Communication State
        CLRF    TIM_ST                  ; Clear Timer State
        GOTO    INIT_D

;------------------------------------------------------------------------------
; INTERRUPT

        GOTO    HI_INT
        fill    0xffff,0x0018-$
        GOTO    HI_INT

;------------------------------------------------------------------------------
; TABLES

; USB initialization data
BD_EP0  dw      BDT0_O,BD_CTL-BD_EP0-4+(PGS<<TB_SEL)
; Buffer Descriptor Initial Data for EP0
        db      1<<UOWN|1<<BSTALL,1<<PA_CTL
        dw      BUF0_O
        db      1<<UOWN|1<<BSTALL,1<<PA_CTL
        dw      BUF0_I
BD_CTL  dw      BDT0_O+(1<<ENDP0)*EP_CTL,BD_DAT-BD_CTL-4+(PGS<<TB_SEL)
; Buffer Descriptor Initial Data for the CC Control EP
        dw      0,0
        db      1<<UOWN|1<<DTSEN,0x0a
        dw      NT_TYP
BD_DAT  dw      BDT0_O+(1<<ENDP0)*EP_DAT,NT_INI-BD_DAT-4+(PGS<<TB_SEL)
; Buffer Descriptor Initial Data for the CC Data EP
        db      0,0
        dw      CMD_CD
        db      0,0
        dw      CMD_CD
NT_INI  dw      NT_TYP,LC_INI-NT_INI-4+(PGS<<TB_SEL)
; Notification buffer Initial Data
        db      1<<RT_DIR|1<<RT_CLA|1<<RT_ITF,N_SSTA
        dw      0,IF_MAX-1
        dw      2,1<<SS_DSR|1<<SS_DCD
LC_INI  dw      LN_COD,ECM_SP-LN_COD+(PGS<<TB_SEL)
; CC Line Coding Initial Data
        db      low BAUD_R, high BAUD_R
        db      upper BAUD_R, 0
        db      0,0,8

; USB Descriptors Index Table
DE_PNT  db      low DE_DEV,high DE_DEV,upper DE_DEV,0
        db      low DE_CFG,high DE_CFG,upper DE_CFG,0
        db      low DE_ST0,high DE_ST0,upper DE_ST0,0
        db      low DE_ST1,high DE_ST1,upper DE_ST1,0
        db      low DE_ST2,high DE_ST2,upper DE_ST2,0
        db      low DE_ST3,high DE_ST3,upper DE_ST3,0

; USB Descriptors
DE_DEV  dw      BUF0_I,DE_CFG-DE_DEV-4  ; Table header
; Device Descriptor
        db      0x12,DT_DEV             ; bLength               18
                                        ; bDescriptorType       Device
        dw      0x0200                  ; bcdUSB                USB 2.0
        db      DC_COM,IC_COM           ; bDeviceClass          Communication
                                        ; bDeviceSubClass       
        db      DC_NUS,1<<PA_CTL        ; bDeviceProtocol       None
                                        ; bMaxPacketSize0       Packet Size
        dw      0x04D8                  ; idVendor              Microchip
        dw      0xF6FC                  ; idProduct             IF1BIS
        dw      0x0111                  ; bcdDevice             PIC18F47J53
        db      1,2                     ; iManufacturer         String 1
                                        ; iProduct              String 2
        db      3,CF_MAX                ; iSerialNum            String 3
                                        ; bNumConfigurations    1

DE_CFG  dw      BUF0_I,DE_ST0-DE_CFG-5  ; Table header
; Configuration (1) Descriptor
        db      0x09,DT_CFG             ; bLength               9
                                        ; bDescriptorType       Configuration
        dw      0x0043                  ; wTotalLength          67
        db      IF_MAX+1,CF_MAX         ; bNumInterfaces        2
                                        ; bConfigurationValue   1
        db      0,CA_SPW|CA_RWU         ; iConfiguration        None
                                        ; bmAttributes          Self-Pow,Rem Wake-up
        db      1,9                     ; bMaxPower             2mA
; CDC Control Interface (0) Descriptor
                                        ; bLength               9
        db      DT_ITF,IF_MAX-1         ; bDescriptorType       Interface
                                        ; bInterfaceNumber      0
        db      0,1                     ; bAlternateSetting     None
                                        ; bNumEndpoints         1
        db      IC_COM,IS_ACM           ; bInterfaceClass       Communication
                                        ; bInterfaceSubClass    Abstract
        db      PC_ATC,0                ; bInterfaceProtocol    AT Command
                                        ; iInterface            None
; CDC Header Functional Descriptor
        db      5,DT_CSI                ; bFunctionLength       5
                                        ; bDescriptorType       CS Interface
        db      FD_HED,0x10             ; bDescriptorSubtype    Header
                                        ; bcdCDC                CDC 1.10
        db      0x01,4                  ;
; CDC Abstract Control Model Functional Descriptor
                                        ; bFunctionLength       4
        db      DT_CSI,FD_ACM           ; bDescriptorType       CS Interface
                                        ; bDescriptorSubtype    Abstract Control
        db      1<<CA_AC1,5             ; bmCapabilities        D1
; CDC Union Functional Descriptor
                                        ; bFunctionLength       5
        db      DT_CSI,FD_UNI           ; bDescriptorType       CS Interface
                                        ; bDescriptorSubtype    Union
        db      IF_MAX-1,IF_MAX         ; bmMasterInterface     0
                                        ; bSlaveInterface0      1
; CDC Call Management Functional Descriptor
        db      5,DT_CSI                ; bFunctionLength       5
                                        ; bDescriptorType       CS Interface
        db      FD_CMA,0                ; bDescriptorSubtype    Call Management
                                        ; bmCapabilities        None
        db      IF_MAX,7                ; bDataInterface        1
; CDC Notification Endpoint Descriptor
                                        ; bLength               7
        db      DT_ENP,EP_CTL|1<<EA_DIR ; bDescriptorType       Endpoint
                                        ; bEndpointAddress      Address
        db      EA_INT,LN_COD-NT_TYP    ; bmAttributes          Interrupt
                                        ; wMaxPacketSize        Packet Size
        db      0,2                     ; bInterval             2ms
; CDC Data Interface (1) Descriptor
        db      9,DT_ITF                ; bLength               9
                                        ; bDescriptorType       Interface
        db      IF_MAX,0                ; bInterfaceNumber      1
                                        ; bAlternateSetting     None
        db      2,IC_DAT                ; bNumEndpoints         2
                                        ; bInterfaceClass       Data
        db      DC_NUS,PC_NOP           ; bInterfaceSubClass    None
                                        ; bInterfaceProtocol    None
        db      0,7                     ; iInterface            None
; CDC Data OUT Endpoint Descriptor
                                        ; bLength               7
        db      DT_ENP,EP_DAT           ; bDescriptorType       Endpoint
                                        ; bEndpointAddress      Address
        db      EA_BLK,1<<PA_DAT        ; bmAttributes          Bulk
                                        ; wMaxPacketSize        Packet Size
        db      0,0                     ; bInterval             None
; CDC Data IN Endpoint Descriptor
        db      7,DT_ENP                ; bLength               7
                                        ; bDescriptorType       Endpoint
        db      EP_DAT|1<<EA_DIR,EA_BLK ; bEndpointAddress      Address
                                        ; bmAttributes          Bulk
        db      1<<PA_DAT,0             ; wMaxPacketSize        Packet Size
        db      0                       ; bInterval             None

DE_ST0  dw      BUF0_I,DE_ST1-DE_ST0-4  ; Table header
; Language ID String Descriptor
        db      DE_ST1-DE_ST0-4,DT_STR  ; bLength               Length
                                        ; bDescriptorType       String
        dw      0x0409                  ; wLangID               English (US)

DE_ST1  dw      BUF0_I,DE_ST2-DE_ST1-4  ; Table header
; Manufacturer String Descriptor
        db      DE_ST2-DE_ST1-4,DT_STR  ; bLength               Length
                                        ; bDescriptorType       String
        dw      'M','i','c','r','o','c','h','i','p',' '
        dw      'T','e','c','h','n','o','l','o','g','y'
        dw      ' ','I','n','c','.'

DE_ST2  dw      BUF0_I,DE_ST3-DE_ST2-4  ; Table header
; Product String Descriptor
        db      DE_ST3-DE_ST2-4,DT_STR  ; bLength               Length
                                        ; bDescriptorType       String
        dw      'I','n','t','e','r','f','a','c','e','1','b','i','s'

DE_ST3  dw      BUF0_I,DE_END-DE_ST3-4  ; Table header
; Serial Number String Descriptor
        db      DE_END-DE_ST3-4,DT_STR  ; bLength               Length
                                        ; bDescriptorType       String
        dw      'P','I','C','1','8','F','4','7','J','5','3'
DE_END

; SD card command strings
; GO_IDLE_STATE
CMD_00  dw      CM_CNL,0x0a+(PGS<<TB_SEL)
        db      6,1                                     ; Response R1
        dw      0
        db      .0|1<<CB_TRN,0,0,0,0,0x4a<<1|1
; SEND_IF_COND
CMD_08  dw      CM_CNL,0x0a+(PGS<<TB_SEL)
        db      6,5                                     ; Response R3
        dw      0
        db      .8|1<<CB_TRN,0,0,C8_VHS,C8_PAT,0x43<<1|1
; SEND_CSD
CMD_09  dw      CM_CNL,0x0a+(PGS<<TB_SEL)
        db      6,1                                     ; Response R1
        dw      0x0012                                  ; Data size: 16+2
        db      .9|1<<CB_TRN,0,0,0,0,0x57<<1|1
; SEND_CID
CMD_10  dw      CM_CNL,0x0a+(PGS<<TB_SEL)
        db      6,1                                     ; Response R1
        dw      0x0012                                  ; Data size: 16+2
        db      .10|1<<CB_TRN,0,0,0,0,0x0d<<1|1
; SET_BLOCKLEN
CMD_16  dw      CM_CNL,0x0a+(PGS<<TB_SEL)
        db      6,1                                     ; Response R1
        dw      0
        db      .16|1<<CB_TRN,0,0,2,0,0x0a<<1|1
; READ_SINGLE_BLOCK
CMD_17  dw      CM_CNL,0x0a+(PGS<<TB_SEL)
        db      6,1                                     ; Response R1
        dw      0x0202                                  ; Data size: 512+2
        db      .17|1<<CB_TRN,0,0,0,0,0x00<<1|1         ; Data address
; WRITE_BLOCK
CMD_24  dw      CM_CNL,0x0a+(PGS<<TB_SEL)
        db      6,1                                     ; Response R1
        dw      0x0202                                  ; Data size: 512+2
        db      .24|1<<CB_TRN,0,0,0,0,0x00<<1|1         ; Data address
; ERASE_WR_BLK_START_ADDR
CMD_32  dw      CM_CNL,0x0a+(PGS<<TB_SEL)
        db      6,1                                     ; Response R1
        dw      0
        db      .32|1<<CB_TRN,0,0,0,0,0x00<<1|1         ; Data address
; ERASE_WR_BLK_END_ADDR
CMD_33  dw      CM_CNL,0x0a+(PGS<<TB_SEL)
        db      6,1                                     ; Response R1
        dw      0
        db      .33|1<<CB_TRN,0,0,0,0,0x00<<1|1         ; Data address
; ERASE
CMD_38  dw      CM_CNL,0x0a+(PGS<<TB_SEL)
        db      6,1
        dw      0xffff                                  ; Response R1b
        db      .38|1<<CB_TRN,0,0,0,0,0x52<<1|1
; SD_SEND_OP_COND
ACM_41  dw      CM_CNL,0x0a+(PGS<<TB_SEL)
        db      6,1                                     ; Response R1
        dw      0
        db      .41|1<<CB_TRN,1<<CB_HCS,0,0,0,0x3b<<1|1
; APP_CMD
CMD_55  dw      CM_CNL,0x0a+(PGS<<TB_SEL)
        db      6,1                                     ; Response R1
        dw      0
        db      .55|1<<CB_TRN,0,0,0,0,0x32<<1|1
; READ_OCR
CMD_58  dw      CM_CNL,0x0a+(PGS<<TB_SEL)
        db      6,5                                     ; Response R7
        dw      0
        db      .58|1<<CB_TRN,0,0,0,0,0x7e<<1|1
; CRC_ON_OFF
CMD_59  dw      CM_CNL,0x0a+(PGS<<TB_SEL)
        db      6,1                                     ; Response R1
        dw      0
        db      .59|1<<CB_TRN,0,0,0,0,0x48<<1|1

; Joystick read rate timing constants
JS_TMC  dw      ~(FOSC_4/PGS/.1)                      ;  1Hz
        dw      ~(FOSC_4/PGS/.2)                      ;  2Hz
        dw      ~(FOSC_4/PGS/.3)                      ;  3Hz
        dw      ~(FOSC_4/PGS/.4)                      ;  4Hz
        dw      ~(FOSC_4/PGS/.5)                      ;  5Hz
        dw      ~(FOSC_4/PGS/.6)                      ;  6Hz
        dw      ~(FOSC_4/PGS/.7)                      ;  7Hz
        dw      ~(FOSC_4/PGS/.8)                      ;  8Hz
        dw      ~(FOSC_4/PGS/.9)                      ;  9Hz
        dw      ~(FOSC_4/PGS/.10)                     ; 10Hz

; Joystick / mouse initial settings
JM_INI  dw      MSE_ST,MPOS_X-MSE_ST+(PGS<<TB_SEL)
        db      0,.10                                 ; State, Sensitivity
        dw      ~(FOSC_4/PGS/.5)                      ; Timing constant
        db      .255,.175                             ; Window size

; Board ID string
BRD_ID  dw      CMD_CD,STR_EN-BRD_ID-4+(PGS<<TB_SEL)
        dw      0,STR_EN-BRD_ID-8
        db      '4','b'
	d_byte	BRD_RV
	d_byte	BRD_LT
	d_byte	BRD_SR
	d_byte	BRD_CE
	d_byte	BRD_YR
	d_byte	BRD_MO
	d_byte	BRD_DY
STR_EN

;------------------------------------------------------------------------------
; DEVICE INITALIZATION

INIT_D	BCF     INTCON2,NOT_RBPU        ; Enable PORTB pull-up
        banksel ANCON0
; Set all inputs to digital
        SETF    ANCON0                  ; No analog inputs AN0-AN7
        MOVLW   1<<PCFG12|1<<PCFG11|1<<PCFG10|1<<PCFG9|1<<PCFG8
        MOVWF   ANCON1                  ; No analog inputs AN8-AN12
        banksel DEV_ST
        LFSR    FSR0,NT_DAT             ; No CC data request active!
; Read PORTB
        MOVF    PORTB,W                 ; Clear RB port mismatch condition
        MOVWF   CHG_ST                  ; Save the value of the RB port.
; Check if line JOY0 is low
        BTFSC   CHG_ST,B_JOYR           ; Joystick to the right?
        BRA     MAIN_C                  ; No! Continue initialization
;
; The self-programming sequence is triggered by holding the
; joystick to the right during power-up

; Port C configuration
        SETF    PORTC                   ; LEDs are off
        MOVLW   1<<DT1|1<<CK1           ; PS/2 clock and data
        MOVWF   TRISC                   ; are inputs
; PSP configuration
        BSF     PMCONH,PMPEN            ; Configure PSP in addressable mode
        MOVLW   1<<IRQM1|1<<IRQM0|1<<MODE0
        MOVWF   PMMODEH
; Timer 0 configuration
        MOVLW   1<<T0PS2|1<<T0PS1|1<<T0PS0
        MOVWF   T0CON                   ; Timer 0 prescaler is 1:256
; USB configuration
        MOVLW   1<<UPUEN|1<<FSEN        ; Full speed, on-chip transreceiver,
        MOVWF   UCFG                    ; pull-up enabled, no ping-pong
; Attach USB
        RCALL   USB_CG                  ; Attach USB
        CLRF    DEV_ST                  ; Signal: Server not connected!
FW_UPD  CLRF    INTCON                  ; Clear Interrupt flags
; Wait for connection to server
SRV_LP  RCALL   POLL_E
        BTFSS   DEV_ST,DV_SRV           ; Server connected?
        BRA     SRV_LP                  ; Not yet!
; The server is connected. Wait for 50ms
        RCALL   TIM_50
C_LOOP  BTFSS   INTCON,TMR0IF
        BRA     C_LOOP
; Send a data request to the server
B_LOOP  MOVLW   NS_FWR                  ; Firmware request code
        MOVWF   CMD_CD
        CLRF    BLN_LO
        CLRF    BLN_HI
        RCALL   USB_TX
T_LOOP  RCALL   POLL_E                  ; Transmission complete?
        BNC     T_LOOP                  ; Not yet!
        RCALL   USB_RX                  ; Wait for data from server
R_LOOP  RCALL   POLL_E                  ; Receipt complete?
        BNC     R_LOOP                  ; Not yet!
; Check if a block of 1024 bytes was received
        MOVLW   high .1024
        CPFSEQ  BLN_HI
        BRA     R_LOOP                  ; No more data. Loop indefinitely
; A data block was received
        CPFSLT  CMD_CD                  ; The command code field holds
        BRA     ST_LIN                  ; the destination page number,
        BRA     R_LOOP                  ; multiplied by 2.
;
; Program a flash memory block of 1024 bytes
;
ST_LIN  CLRF    TBLPTRU                 ; Set the destination pointer
        RLCF    CMD_CD,W                ; to an address between:
        RLCF    TBLPTRU                 ; 0x00800 and 0x1fc00
        ANDLW   high -.1024
        MOVWF   TBLPTRH
        CLRF    TBLPTRL
        BCF     PORTC,C_NSDC            ; Switch the SDC LED on
; Erase a flash memory block of 1024 bytes
        BSF     EECON1,FREE             ; Enable flash memory erase
        RCALL   WR_MEM                  ; Erase block
        TBLRD*-                         ; Adjust destination address
        LFSR    FSR0,DATA_B             ; Set source address
; Write a flash memory block of 1024 bytes
        MOVLW   .16
        MOVWF   CM_CNL                  ; There are 16 lines
NX_LIN  MOVLW   .64                     ; of 64 bytes each
        MOVWF   CM_CNH
WR_LIN  MOVFF   POSTINC0,TABLAT
        TBLWT+*
        DECFSZ  CM_CNH                  ; Line done?
        BRA     WR_LIN                  ; Not yet!
        RCALL   WR_MEM                  ; Write line
        DECFSZ  CM_CNL                  ; All lines done?
        BRA     NX_LIN                  ; Not yet!
; The 1024 bytes block has been written
        BSF     PORTC,C_NSDC            ; Switch the SDC LED off
        BRA     B_LOOP                  ; Loop back

; Write to flash memory
WR_MEM  BSF     EECON1,WREN             ; Enable write to flash memory
        MOVLW   0x55
        MOVWF   EECON2
        MOVLW   0xAA
        MOVWF   EECON2
        BSF     EECON1,WR               ; Start erase operation
        NOP
        BCF     EECON1,WREN             ; Disable write to flash memory
        RETURN

;------------------------------------------------------------------------------
; USB SUBROUTINES

; USB Interrupt Service Routine
USB_IS  SETF    EP_OFS                  ; Clear the last EP offset
        BTFSS   UCON,SUSPND             ; USB in Suspend Mode?
        BRA     TS_IDL                  ; No!
; USB in Suspend Mode
; Only Bus Activity can generate an interrupt (DS39632e 17.5.1)
; Bus Activity interrupt
        BCF     UCON,SUSPND             ; End USB Suspend Mode
        BCF     UIE,ACTVIE              ; Disable Bus Activity Interrupt
        BRA     USB_SE                  ; Return
TS_IDL  BTFSS   UIR,IDLEIF              ; Idle Detect Interrupt?
        BRA     TS_RST                  ; No!
; Idle Detect interrupt
        BSF     UCON,SUSPND             ; Enter the USB Suspend Mode
        BSF     UIE,ACTVIE              ; Enable Bus Activity Interrupt
        BRA     USB_CE                  ; Return
TS_RST  BTFSS   UIR,URSTIF              ; Reset interrupt?
        BRA     TS_TRN                  ; No!
; Reset interrupt
; Empty the USTAT FIFO. If the stack is not empty it takes 5 instruction
; cycles after clearing UIR for the TRNIF flag to be re-asserted
CL_UIR  CLRF    UIR                     ; Clear all USB event flags and
        NOP                             ; wait for 5 instruction cycles
        NOP
        NOP
        NOP
        NOP
        BTFSC   UIR,TRNIF               ; Is the USTAT FIFO empty?
        BRA     CL_UIR                  ; Not yet!
        RCALL   EP_DIS                  ; Disable all endpoints
        tab_ad  BD_EP0                  ; Initialize the EP0 buffer
        RCALL   TAB_NW                  ; descriptor
        MOVLW   1<<EPHSHK|1<<EPOUTEN|1<<EPINEN
        MOVWF   UEP0                    ; Enable EP0 for control
        BCF     UCON,PKTDIS             ; Enable packet processing
        MOVLW   1<<SB_DEF|1<<SB_SPW     ; Signal: Default State!
        MOVWF   USB_ST
TS_TRN  BTFSS   UIR,TRNIF               ; Transfer Complete interrupt?
NO_TRN  BRA     USB_CF                  ; No!
; Transfer Complete interrupt
; Identify the active endpoint
        MOVF    USTAT,W
        ANDLW   1<<ENDP3|1<<ENDP2|1<<ENDP1|1<<ENDP0|1<<DIR
        MOVWF   EP_OFS                  ; Offset of endpoint in BDT
        BCF     UIR,TRNIF               ; Clear transfer complete flag
        LFSR    FSR1,BDT0_O             ; Point FSR1 to BDT
        BTG     PLUSW1,DTS              ; Toggle the data sync bit.
        ANDLW   1<<ENDP3|1<<ENDP2|1<<ENDP1|1<<ENDP0
        BNZ     NO_TRN                  ; Not a Control Transfer!
; Control Transfer Complete interrupt
        banksel BDT0_O
        BTFSS   EP_OFS,DIR              ; Control IN Transfer?
        BRA     CK_STU                  ; No!
; Control IN Transfer Complete interrupt
        BTFSC   RQ_TYP,RT_DIR           ; Control Read?
        BRA     CDT_IN                  ; Yes!
; Status Stage of a Control Write or Non-Data Control Transfer
; Only 'Set Address' requests need to be handled
        MOVLW   R_SADR
        CPFSEQ  RQ_NUM                  ; Set Address Request?
        BRA     TS_TRN                  ; No!
; Status Stage of a 'Set Address' Setup Request
        MOVFF   RQ_VAL,UADDR            ; Copy the new address
        BRA     TS_TRN
; Control Read Transfer Data Stage
; All implemented requests except for 'Get Descriptor' need to respond
; with only one (short) IN packet. After completion of the corresponding
; transfer, returning to this point results in the EP0 IN endpoint being
; stalled, which prevents the host from requesting any further data
; The 'Get Descriptor' request uses this entry point repeatedly, after
; each completed transfer of a packet, until all required data has been
; sent, following which it also stalls the endpoint.
CDT_IN  LFSR    FSR1,BUF0_I             ; Initialize the buffer address
; Check if all requested data has been transmitted
        MOVF    EP0I_L,W                ; Size of the transmitted packet
        SUBWF   RQ_LNL                  ; Subtract the size of the sent
        MOVLW   0                       ; packet from the requested
        SUBWFB  RQ_LNH                  ; remaining length
        BNC     RQ_ERR                  ; This should never happen!
        BNZ     CDI_SZ
        CPFSEQ  RQ_LNL                  ; Any more data requested?
        BRA     CDI_SZ                  ; Yes!
; All data has been sent
; If the last packet was full-length, send a zero-length packet
; or else stall the endpoint
        BTFSS   EP0I_L,PA_CTL           ; Last packet was full-length?
        BRA     RQ_ERR                  ; No! Stall EP0 IN
; Establish the size of the packet to be sent
CDI_SZ  MOVLW   -(1<<PA_CTL)            ; Clear Z if there is enough data
        ANDWF   RQ_LNL,W                ; for a full-length packet
        IORWF   RQ_LNH,W
        MOVLW   1<<PA_CTL               ; Full-length
        BNZ     CDI_FL
        MOVF    RQ_LNL,W                ; Short packet
CDI_FL  MOVWF   EP0I_L                  ; Set the packet size
        BZ      CDI_ST                  ; Zero-length packet
; Copy the next (WREG) descriptor bytes from the table to the EP0 IN buffer
        RCALL   TAB_RD                  ; Copy (WREG) bytes to the buffer
        SUBWF   EP0I_L                  ; Reduce the packet size by the
        BNZ     CDI_ST                  ; number of bytes not available
; If no more data is available set the remaining response length to zero
CDI_ZR  CLRF    RQ_LNL                  ; More data requested than
        CLRF    RQ_LNH                  ; available!
; Pass control of the endpoint to SIE
CDI_ST  MOVLW   1<<UOWN|1<<DTSEN
        BTFSC   BDT0_I,DTS              ; Copy the DTS bit
        BSF     WREG,DTS
CIN_EX  MOVWF   BDT0_I                  ; Pass control of EP0 IN to SIE
; Enable Packet Processing because this exit point is also used by
; every SETUP transfer
        BCF     UCON,PKTDIS             ; Enable Packet Processing
        BRA     TS_TRN                  ; Done!
; Differentiate between an OUT and a SETUP packet
CK_STU  MOVF    BDT0_O,W                ; Get the Token PID
        ANDLW   1<<PID3|1<<PID2|1<<PID1|1<<PID0
        XORLW   PID_SU<<PID0            ; SETUP Token?
        BZ      SET_UP                  ; Yes!
; Control OUT Transfer Complete interrupt
; Only 'Set Line Coding' Data OUT packets need to be processed
        MOVLW   C_SLCO
        CPFSEQ  RQ_NUM                  ; Set Line Coding?
        BRA     CDO_ST                  ; No!
; Copy the received Line Coding Data
        copy_b  BUF0_O,LN_COD,ECM_SP-LN_COD
; Prepare EP0 OUT for the next transfer
CDO_ST  RCALL   EP0O_A                  ; Re-arm EP0 OUT
        BRA     TS_TRN
; SETUP Transfer Complete interrupt
; Copy the Request Header to the access RAM
SET_UP  copy_b  BUF0_O,RQ_TYP,8
        MOVLW   1<<DTS                  ; Reclaim control and
        MOVWF   BDT0_O                  ; set DATA1 for both EP0
        MOVWF   BDT0_I                  ; OUT and IN endpoints
; Setting EP0 OUT to DATA1 is necessary for both Control Read and
; Control Write transfers. Non-Data Control Transfers are not
; affected while the next Setup Token will be accepted irrespective
; of the DTS bit state ( DS39632e Register 17-5)
        RCALL   EP0O_A                  ; Re-arm EP0 OUT and set the
        CLRF    EP0I_L                  ; EP0 IN packet size to 0
        CLRF    BUF0_I                  ; Clear the first response byte
; Differentiate between Setup Request types
        BTFSC   RQ_TYP,RT_CLA           ; Class request?
        BRA     CLA_RQ                  ; Yes!
        BTFSS   RQ_TYP,RT_VEN           ; Standard request?
        BRA     STD_RQ                  ; Yes!
; Vendor Setup Requests are not supported
EP_HLT  MOVLW   1<<UOWN|1<<BSTALL       ; Stall EP0
        MOVWF   BDT0_O                  ; Stall EP0 OUT
        BRA     CIN_EX                  ; Stall EP0 IN
; Check the Request Recipient: Device, Interface or Endpoint
CK_RQR  BTFSC   RQ_TYP,RT_ITF           ; Interface?
        BRA     CKR_OR                  ; Yes!
        BTFSS   RQ_TYP,RT_ENP           ; Device?
; Device Recipient: nothing to check      Yes!
        RETURN
; Endpoint Recipient
        RLNCF   RQ_IDX,W                ; Point to the status byte
        RLNCF   WREG                    ; of the endpoint descriptor
        RLNCF   WREG                    ;
        MOVWF   FSR1L
        MOVLW   (1<<ENDP0)*(EP_MAX+1)   ; Lowest invalid offset
        BTFSS   USB_ST,SB_CFG           ; If not in Configured State
        MOVLW   1<<ENDP0                ; only EP0 is valid
        CPFSLT  FSR1L                   ; Valid endpoint index?
        BRA     CKR_ER                  ; No! Signal: Request Error!
        MOVLW   (1<<ENDP0)*EP_CTL       ; The Control EP OUT is not
        CPFSEQ  FSR1L                   ; a valid endpoint
        RETURN
        BRA     CKR_ER                  ; Signal: Request Error!
; The recipient must be an interface, otherwise signal: Request Error!
CKR_IF  BTFSS   RQ_TYP,RT_ITF           ; Interface?
        BRA     CKR_ER                  ; No! Signal: Request Error!
; Setup Request Recipient of type 'Other' is not supported
CKR_OR  BTFSC   RQ_TYP,RT_ENP           ; Other?
        BRA     CKR_ER                  ; Signal: Request Error!
; Interface Recipient
        BTFSS   USB_ST,SB_CFG           ; Configured state?
        BRA     CKR_ER                  ; No! Signal: Request Error!
        MOVLW   IF_MAX                  ; Maximum interface number
        CPFSGT  RQ_IDX                  ; Valid interface number?
        RETURN                          ; Yes!
CKR_ER  POP                             ; Discard the return address
; Request Error
RQ_ERR  MOVLW   1<<UOWN|1<<BSTALL       ; Request Error action:
        BRA     CIN_EX                  ; Stall EP0 IN
; Communication Class Requests
CLA_RQ  RCALL   CKR_IF                  ; Recipient must be an interface
        TSTFSZ  RQ_IDX                  ; Recipient must be the
        BRA     RQ_ERR                  ; CC Control Interface (0)
        MOVF    RQ_NUM,W                ; Request Code to WREG
        XORLW   C_SECO                  ; Send Encapsulated Command Request?
; Send Encapsulated Command
; Accept the Request but discard the received data
        BZ      CDI_ST                  ; Yes! Prepare Status
        XORLW   C_SECO^C_GERP           ; Get Encapsulated Response Request?
; Get Encapsulated Response
; Clear the requested response length and send a zero-length packet
        BZ      CDI_ZR                  ; Yes!
        XORLW   C_GERP^C_SLCO           ; Set Line Coding Request?
; Set Line Coding
        BZ      CDI_ST                  ; Yes! Prepare Status
        XORLW   C_SLCO^C_GLCO           ; Get Line Coding Request?
        BNZ     TS_SLS                  ; No!
; Get Line Coding
; Copy the Line Coding data to the EP0 IN buffer
        copy_b  LN_COD,BUF0_I,ECM_SP-LN_COD
        BRA     IN_SHO                  ; Send the Line Coding data
TS_SLS  XORLW   C_GLCO^C_SCLS           ; Set Control Line State Request?
; No other requests are supported
NO_OTH  BNZ     EP_HLT                  ; Stall EP0
; Set Control Line State
        BCF     USB_ST,SB_DTR           ; Copy DTR bit to
        BTFSC   RQ_VAL,CS_DTR           ; USB state
        BSF     USB_ST,SB_DTR
        BRA     IN_SHO                  ; Send Status
; Standard SETUP Requests
STD_RQ  MOVF    RQ_NUM,W                ; Request Code to WREG
        XORLW   R_SADR                  ; Set Address Request?
; Set Address
        BZ      IN_SHO                  ; Yes! Send Status
        XORLW   R_SADR^R_GDES           ; Get Descriptor Request?
        BNZ     TS_SCF                  ; No!
; Get Descriptor
        MOVLW   DT_ITF                  ; Only supported types: Device,
        CPFSLT  RQ_DTY                  ; Configuration and String
        BRA     RQ_ERR                  ; Signal: Request Error!
        MOVLW   DT_CFG                  ; Configuration Descriptor
        CPFSEQ  RQ_DTY                  ; requested?
        BRA     GT_DES                  ; No!
        MOVLW   CF_MAX                  ; If yes, check the index
        CPFSLT  RQ_VAL                  ; Valid Configuration index?
        BRA     RQ_ERR                  ; No! Signal: Request Error!
GT_DES  tab_ad  (DE_PNT-4)              ; Index into the table DE_PNT
        MOVF    RQ_DTY,W                ; containing the descriptor
        ADDWF   RQ_VAL,W                ; addresses
        RCALL   TB_IND                  ; Get the descriptor's address
        BRA     CDI_SZ                  ; and send the first packet
TS_SCF  XORLW   R_GDES^R_SCFG           ; Set Configuration Request?
        BNZ     TS_GCF                  ; No!
; Set Configuration
        MOVLW   CF_MAX+1                ; Test the Configuration number
        CPFSLT  RQ_VAL                  ; If it is invalid,
        BRA     RQ_ERR                  ; signal: Request Error!
        CLRF    WREG
        MOVFF   WREG,UEP0+EP_CTL        ; Disable the CC Control EP
        MOVFF   WREG,UEP0+EP_DAT        ; Disable the CC Data EP
        TSTFSZ  RQ_VAL                  ; Configuration number >0 ?
        BRA     SCF_CO                  ; Yes!
; Go back to Addressed Sate
        BCF     USB_ST,SB_CFG           ; Signal: Not Configured State!
        BRA     IN_SHO                  ; Send Status
; Advance to Configured State
SCF_CO  tab_ad  BD_CTL                  ; Initialize the CC Control
        RCALL   TAB_NW                  ; EP Buffer Descriptor
        MOVLW   1<<EPHSHK|1<<EPINEN|1<<EPCONDIS
        MOVFF   WREG,UEP0+EP_CTL        ; Enable the CC Control EP
        RCALL   TAB_NW                  ; Initialize the CC Data EP
        MOVLW   1<<EPHSHK|1<<EPOUTEN|1<<EPINEN|1<<EPCONDIS
        MOVFF   WREG,UEP0+EP_DAT        ; Enable the CC Data EP
        RCALL   TAB_NW                  ; Initialize the Notification
        RCALL   TAB_NW                  ; Header and Line Coding Data
        BSF     USB_ST,SB_CFG           ; Signal: Configured State!
        BRA     IN_SHO                  ; Send Status
TS_GCF  XORLW   R_SCFG^R_GCFG           ; Get Configuration Request?
        BNZ     TS_GST                  ; No!
; Get Configuration
        BTFSC   USB_ST,SB_CFG           ; Configured State?
        MOVLW   CF_MAX                  ; Yes! Return the configuration
IN_VAL  MOVWF   BUF0_I                  ; number or else return 0
; Send one short packet
IN_SHO  MOVFF   RQ_LNL,EP0I_L           ; Requested packet size
        BRA     CDI_ST                  ; Send packet
TS_GST  XORLW   R_GCFG^R_GSTA           ; Get Status Request?
        BNZ     TS_CFT                  ; No!
; Get Status
        RCALL   CK_RQR                  ; Check Recipient
        CLRF    BUF0_I+1                ; Clear the second response byte
        BTFSC   RQ_TYP,RT_ITF           ; Interface?
        BRA     IN_SHO                  ; Yes! Nothing to report
        BTFSC   RQ_TYP,RT_ENP           ; Endpoint?
        BRA     GST_EP                  ; Yes!
; Get Device Status
        MOVF    USB_ST,W                ; Copy the 'Self Powered'
        ANDLW   1<<SB_RWU|1<<SB_SPW     ; and 'Remote Wake-up' bits
        BRA     IN_VAL                  ; to the EP0 IN buffer
; Get Endpoint status
GST_EP  BTFSC   INDF1,BSTALL            ; If the endpoint is stalled,
        BSF     BUF0_I,GS_HLT           ; set the HLT bit
        BRA     IN_SHO                  ; Send the response
TS_CFT  XORLW   R_GSTA^R_CFTR           ; Clear Feature Request?
        BNZ     TS_SFT                  ; No!
; Clear / Set Feature
CHG_FT  RCALL   CK_RQR                  ; Check Recipient
        MOVF    RQ_NUM,W                ; Set Z for 'Clear' and
        XORLW   R_CFTR                  ; clear Z for 'Set'
        BTFSC   RQ_TYP,RT_ITF           ; Interface?
        BRA     RQ_ERR                  ; Yes! Signal: Request Error!
        BTFSC   RQ_TYP,RT_ENP           ; Endpoint?
        BRA     CSF_EP                  ; Yes!
; Clear / Set Remote Wake-up
        MOVLW   DF_RWU                  ; Device Remote Wake-up Feature?
        CPFSEQ  RQ_VAL
        BRA     RQ_ERR                  ; No! Signal: Request Error!
        BCF     USB_ST,SB_RWU           ; Clear Remote Wake-up bit
        BZ      IN_SHO                  ; Set?
        BSF     USB_ST,SB_RWU           ; Set Remote Wake-up bit
        BRA     IN_SHO                  ; Send status
; Clear / Set Endpoint HALT
CSF_EP  MOVLW   EF_HLT                  ; Endpoint Halt Feature?
        CPFSEQ  RQ_VAL
        BRA     RQ_ERR                  ; No! Signal: Request Error!
CSF_HT  RCALL   CHG_HT                  ; Change HALT and clear DTS bit
        BRA     IN_SHO                  ; Send status
TS_SFT  XORLW   R_CFTR^R_SFTR           ; Set Feature Request?
; Set Feature
        BZ      CHG_FT                  ; Yes!
        XORLW   R_SFTR^R_GITF           ; Get Interface Request?
        BNZ     TS_SIF                  ; No!
; Get Interface
        RCALL   CKR_IF                  ; Recipient must be an interface
        BRA     IN_SHO                  ; Return 0
TS_SIF  XORLW   R_GITF^R_SITF           ; Set Interface Request?
        BNZ     NO_OTH                  ; No other requests are supported
; Set Interface
        RCALL   CKR_IF                  ; Recipient must be an interface
        TSTFSZ  RQ_VAL                  ; Accept only the alternate
        BRA     RQ_ERR                  ; interface 0
        MOVLW   (1<<ENDP0)*EP_CTL       ; CC Control Interface (0)
        TSTFSZ  RQ_IDX                  ; CC Data Interface (1)?
        MOVLW   (1<<ENDP0)*EP_DAT       ; Yes!
        MOVWF   FSR1L                   ; Point to corresponding OUT EP
        CLRF    WREG                    ; Clear! ( Set Z )
        RCALL   CHG_HT                  ; Clear the OUT EP HALT and DTS
        BSF     FSR1L,DIR               ; Switch to the IN EP
        BRA     CSF_HT                  ; Clear the IN EP HALT and DTS
; Disable all used USB Endpoints and reset the Device Address
EP_DIS  CLRF    UEP0                    ; Disable EP0
        CLRF    UEP0+EP_CTL             ; Disable CC Control EP
        CLRF    UEP0+EP_DAT             ; Disable CC Data EP
        CLRF    UADDR                   ; Set default address
        RETURN
; Arm EP0 OUT
EP0O_A  MOVLW   1<<PA_CTL               ; Set the packet length to the
        MOVWF   EP0O_L                  ; maximum declared size
        MOVLW   1<<UOWN|1<<DTSEN
        BTFSC   BDT0_O,DTS              ; Copy the DTS bit
        BSF     WREG,DTS
        MOVWF   BDT0_O                  ; Pass control of EP0 OUT to SIE
        RETURN
; Clear ( Z==1 ) / Set ( Z==0 ) the endpoint HALT and clear the DTS bit
CHG_HT  MOVLW   1<<DTSEN                ; Clear the HALT and DTS bits
        BZ      CHT_UO                  ; Set?
        BSF     WREG,BSTALL             ; Yes! Set the HALT bit
CHT_UO  BTFSC   INDF1,UOWN              ; Copy the UOWN bit
        BSF     WREG,UOWN
        MOVWF   INDF1
        RETURN

; Poll relevant events during self-programming
POLL_E  BTFSC   INTCON,RBIF             ; PORTB change?
        RCALL   INT_CG                  ; Handle event
        BTFSC   INTCON,TMR0IF           ; Timer 0 time-out?
        RCALL   INT_TM                  ; Handle event
        BCF     STATUS,C                ; Signal: USB transfer not done!
        BTFSS   UCON,USBEN              ; USB enabled?
        RETURN                          ; No!
        BTFSS   PIR2,USBIF              ; USB event?
        RETURN                          ; No!
;
; USB Interrupt
;
; Call the USB Interrupt Service Routine to handle all USB events
; except for the CC Data Transfers Complete interrupts
INT_SB  RCALL   USB_IS                  ; Service USB interrupt
        banksel DEV_ST
; Handle CC Data Transfers Complete interrupts
; Return the Carry flag set on completion
        BTFSS   USB_ST,SB_CFG           ; USB in 'Configured State'?
        BRA     NO_DTR                  ; No!
; FSR0 was initialized at the start of the USB Communication Stage with
; the address of the corresponding CC Data IN / OUT endpoint descriptor
; Only transfers of that same endpoint are handled until the stage ends
        MOVF    FSR0L,W                 ; Expected EP offset
        CPFSEQ  EP_OFS                  ; Does it match?
        BRA     CK_DTR                  ; No!
        INCF    FSR0L                   ; Point to the packet size
; The first 4 bytes of the first packet received are the Communication
; Header, containing the size of an eventual Data Block
        BTFSS   COM_ST,CO_HED           ; Header Block needed?
        BRA     USB_CO                  ; No!
        MOVF    INDF0,W                 ; Size of the received packet
        BZ      UEP_SZ                  ; Allow for zero-length packets
        SUBLW   PL_HLN-1         	; The first received packet must
        BC      COM_ER                  ; be at least 4 bytes long
        RCALL   TOT_LN                  ; Get the total length
; Move the buffer pointer after the last byte transferred
USB_CO  MOVF    POSTINC0,W              ; Update the buffer address
        ADDWF   POSTINC0                ; by adding the length of the
        MOVLW   0                       ; current transfer.
        ADDWFC  POSTDEC0
        DECF    FSR0L                   ; Point again to the packet size
; Check if there is more data to be transferred
        MOVF    INDF0,W                 ; Subtract the packet size of
        SUBWF   CNT_LO                  ; the current transfer from the
        MOVLW   0                       ; remaining length.
        SUBWFB  CNT_HI
        BNC     COM_ER                  ; This should never happen!
        BNZ     UEP_SZ
        CPFSEQ  CNT_LO                  ; If there is still data to be
        BRA     UEP_SZ                  ; transferred, arm the endpoint
; All data has been transferred. If the last transmitted packet was
; full-sized, send a zero-length packet.
        BSF     STATUS,C                ; Signal: Transfer done!
        BTFSC   EP_OFS,DIR              ; IN transfer ?
        BTFSS   INDF0,PA_DAT            ; If the last packet was short,
        RETURN                          ; end communication stage, else
        BRA     UEP_ST                  ; send a zero-length packet

; Check if the server has set the DTR line
CK_DTR  BTFSS   USB_ST,SB_DTR           ; DTR set?
        BRA     NO_DTR                  ; No!
        BSF     DEV_ST,DV_SRV           ; Signal: Server connected
        BCF     PORTC,C_NSRV            ; Switch the USB LED on
        BRA     USB_RT
NO_DTR  BCF     DEV_ST,DV_SRV           ; Signal: Server not ready
        BSF     PORTC,C_NSRV            ; Switch the USB LED off
        BRA     USB_RT

; Start USB Transmission
; Point FSR0 to the last byte of the CC Data IN endpoint BDT entry
USB_TX  LFSR    FSR0,BDT0_I+(1<<ENDP0)*EP_DAT+(1<<DIR)-1
        RCALL   TOT_LN                  ; Get the total length
        BRA     SRV_CK
; Start USB Reception
; Point FSR0 to the last byte of the CC Data OUT endpoint BDT entry
USB_RX  LFSR    FSR0,BDT0_O+(1<<ENDP0)*EP_DAT+(1<<DIR)-1
        BSF     COM_ST,CO_HED           ; Signal: Need Header Block!
SRV_CK  BTFSS   DEV_ST,DV_SRV           ; Check if the server is connected
        BRA     USB_NP                  ; Signal error if not
        BTFSS   UCON,SUSPND             ; USB in Suspended Mode?
        BRA     UEP_AD                  ; No!
; If the USB is in Suspend Mode, try Remote Wake-up
        BTFSS   USB_ST,SB_RWU           ; Remote Wake-up enabled?
        BRA     USB_NP                  ; No! Tough!
; Activate the RESUME signal for 5ms
        BSF     UCON,RESUME             ; Yes!  Activate the
        l_time  .5                      ; RESUME signal for 5ms
        RCALL   STA_LT
        BRA     USB_RT
; Initialize the USB buffer address
UEP_AD  BSF     COM_ST,CO_ACT           ; Communication started!
        MOVLW   high CMD_CD             ; Initialize the buffer
        MOVWF   POSTDEC0                ; address.
        MOVLW   low CMD_CD
        MOVWF   POSTDEC0
; Set the time-out for the USB Communication Stage to 1s
        BSF     TIM_ST,TM_USB           ; Signal: USB timer running!
        RCALL   TIM_1S                  ; Start timer
; Establish the packet size
UEP_SZ  MOVLW   -(1<<PA_DAT)
        ANDWF   CNT_LO,W                ; Clear Z if there is enough data
        IORWF   CNT_HI,W                ; for a full-length packet
        MOVLW   1<<PA_DAT               ; Full-length
; Allow full-length for the first packet to be received
        BTFSC   COM_ST,CO_HED           ; Have Header Block?
        BRA     UEP_ST                  ; Not yet!
        BNZ     UEP_ST                  ; Full packet!
        MOVF    CNT_LO,W                ; Short packet!
UEP_ST  MOVWF   POSTDEC0                ; Set the packet size
; Pass control of the endpoint to SIE
        MOVLW   1<<UOWN|1<<DTSEN
        BTFSC   INDF0,DTS               ; Copy the DTS bit
        BSF     WREG,DTS
        MOVWF   INDF0                   ; Pass control to SIE
USB_RT  BCF     STATUS,Z                ; Signal: No Error!
USB_ND  BCF     STATUS,C                ; Signal: Not done!
        RETURN
COM_ER  MOVLW   E_COMM                  ; 'Communication error'
USB_ER  BSF     STATUS,Z                ; Signal: USB Error!
        BRA     USB_ND

; Calculate the combined length of header and data
TOT_LN  MOVLW   PL_HLN		        ; Length of header block
        ADDWF   BLN_LO,W                ; Add to the length of the
        MOVWF   CNT_LO                  ; data block to get the
        MOVLW   0                       ; total length and copy it
        ADDWFC  BLN_HI,W                ; to the byte counter
        MOVWF   CNT_HI
        BCF     COM_ST,CO_HED           ; Signal: Header Block done!
        RETURN

; Attach / Detach USB
USB_CG  RCALL   EP_DIS                  ; Disable all Endpoints
        MOVLW   1<<SB_SPW               ; Signal: Self powered!
        MOVWF   USB_ST
        BTFSC   CHG_ST,B_USEN           ; USB cable connected to host?
        BRA     USB_AT                  ; Yes!
; Detach USB
        BCF     UCON,SUSPND             ; End Suspend Mode (17.2.1)
        BCF     UCON,USBEN              ; Disable USB
        BCF     PIE2,USBIE              ; Disable the USB interrupt
        BRA     USB_CE
; Attach USB
USB_AT  BSF     UCON,USBEN              ; Enable USB
        MOVLW   1<<IDLEIE|1<<TRNIE|1<<URSTIE
        MOVWF   UIE                     ; Unmask needed USB interrupts
        BSF     PIE2,USBIE              ; Enable USB interrupts
USB_SE  BTFSC   UCON,SE0                ; Single Ended Zero State?
        BRA     USB_SE                  ; Yes! Wait until it ends.
USB_CE  CLRF    UIR                     ; Clear all USB event flags
        BTFSC   UIR,ACTVIF              ; Wait until the Bus Activity
        BRA     USB_CE                  ; event flag is stable
USB_CF  BCF     PIR2,USBIF              ; Clear the USB interrupt flag
        RETURN

;------------------------------------------------------------------------------
; COMMON SUBROUTINES

;
; Copy  data from a table in the program memory.
; The table header has the following structure:
; (Word)Destination Address
; (Byte)Data Length
; (Byte)Selector: 0 = Descriptor, >0 = Other
;
; Continue copying the same table from address (TBLPTR) if called
; from a Control IN Transfer interrupt, or otherwise copy from a new
; descriptor table at address: (TBLPTR)
TAB_RD  BTFSC   EP_OFS,DIR              ; Control IN Transfer?
        BRA     TB_TST                  ; Yes!
; Copy from the table at address: (TBLPTR)
TAB_NW  TBLRD*+
        MOVFF   TABLAT,FSR1L            ; Load the destination
        TBLRD*+                         ; address from the table
        MOVFF   TABLAT,FSR1H
        TBLRD*+                         ; Load the size of the
        MOVFF   TABLAT,TB_CNT           ; table in the counter
; Descriptors are copied only (WREG) bytes at a time
        TBLRD*+                         ; Read the Table Selector
        TSTFSZ  WREG
        BTFSC   TABLAT,TB_SEL           ; Copy the entire table?
        MOVF    TB_CNT,W                ; Yes!
TB_TST  TSTFSZ  TB_CNT                  ; End of table reached?
        BRA     TB_NXT                  ; Not yet!
        RETURN                          ; Yes!
; (WREG) returns the number of unavailable bytes
TB_NXT  TBLRD*+                         ; Copy one byte and decrement
        MOVFF   TABLAT,POSTINC1         ; the 'available bytes' counter
        DECF    TB_CNT                  ; and the 'copied bytes'
        DECFSZ  WREG                    ; counter
        BRA     TB_TST                  ; All done?
        RETURN                          ; Yes!
;
; Index (WREG) into the table at (TBLPTR)
TB_IND  RLNCF   WREG
        RLNCF   WREG
        RCALL   ADD_TP
        TBLRD*+
        MOVFF   TABLAT,PRODL
        TBLRD*+
        MOVFF   TABLAT,PRODH
        TBLRD*+
        MOVFF   TABLAT,TBLPTRU
        MOVFF   PRODH,TBLPTRH
        MOVFF   PRODL,TBLPTRL
        RETURN
;
ADD_TP  ADDWF   TBLPTRL
        MOVLW   0
        ADDWFC  TBLPTRH
        ADDWFC  TBLPTRU
        RETURN
;
; Port B change
;
INT_CG  MOVF    PORTB,W                 ; Get the new PORTB value
        XORWF   CHG_ST,W                ; Set the changed bits
        XORWF   CHG_ST                  ; Change old to new value
        BCF     INTCON,RBIF             ; Clear interrupt flag
        BTFSS   WREG,B_USEN             ; USB power?
        BRA     SDC_CH                  ; No!
; USB cable (dis)connected
        BSF     TIM_ST,TM_SBD           ; Signal: USB power!
TIM_50  l_time  .50                     ; and wait for 50ms
        BRA     STA_LT
SDC_CH  BTFSS   WREG,B_NCRD             ; SD card change?
        RETURN                          ; No!
; SD card inserted / ejected
        BSF     TIM_ST,TM_CRD           ; Signal: SD card change!
;
; Set timer 0 to 1s
;
TIM_1S  l_time  .1000                   ; Start timer 0 for 1s
; Start Timer 0
STA_LT  BCF     T0CON,TMR0ON            ; Stop timer 0
        BCF     INTCON,TMR0IF           ; Clear interrupt flag
        MOVWF   TMR0L                   ; Load timing constant
        BSF     T0CON,TMR0ON            ; Start timer 0
        RETURN
;
INT_TM  BCF     T0CON,TMR0ON            ; Stop timer
;
; Timer 0 interrupt for asynchronous events
;
TIM_IT  BCF     INTCON,TMR0IF           ; Clear interrupt flag
        BSF     STATUS,C                ; Signal: No event!
        BTFSS   UCON,RESUME             ; USB resume?
        BRA     USB_TO                  ; No!
        BCF     UCON,RESUME             ; Stop RESUME signal
        BTFSS   UCON,SUSPND             ; USB still in Suspended Mode?
        BRA     UEP_AD                  ; No!!
USB_NP  MOVLW   E_NPRE                  ; Error: 'Not present'
        BRA     USB_ER
USB_TO  BTFSS   TIM_ST,TM_SBD           ; USB power?
        BRA     CRD_TO                  ; No!
; The USB cable connection state is assessed after 50ms
        BCF     TIM_ST,TM_SBD           ; Signal: Not USB power!
        RCALL   USB_CG                  ; Attach / detach USB
        BRA     NO_DTR
CRD_TO  BTFSS   TIM_ST,TM_CRD           ; SD card change?
        BRA     SDC_TO                  ; No!
; The SD card state is assessed after 1s
        BCF     TIM_ST,TM_CRD           ; Signal: No SD card change!
        BSF     SDC_ST,SD_DET           ; Signal: Detect SD card!
        BRA     USB_RT
SDC_TO  BTFSS   TIM_ST,TM_SDC           ; SD card timer running?
        RETURN                          ; No!
; SD card time-out
        BCF     TIM_ST,TM_SDC           ; Signal: Time-out!
        BRA     USB_RT
;
; Spare self-programming block program memory
;
        fill    0xffff,0x0800-$

;------------------------------------------------------------------------------
; FIXED START ADDRESSES

MAIN_C  GOTO    CONF_D                  ; Jump to main program

;------------------------------------------------------------------------------
; (HIGH PRIORITY) INTERRUPT

HI_INT  BTFSS   PIR1,PMPIF              ; PSP interrupt?
	BRA	CK_PST			; No!
        BCF     PIR1,PMPIF              ; Clear the interrupt flag
        BTFSS   PMSTATL,OB3E            ; Data output?
	BRA     PSP_IN                  ; No!
        BTFSC   SDC_ST,SD_MBR           ; Multiple block read?
        BRA     ADJ_RP                  ; Yes!
        SETF    PSP_DO                  ; Clear OB3F
        BTFSC   COM_ST,CO_TRA           ; Transmission stage?
        BRA     TST_CN                  ; Yes!
PSP_IN  BTFSS   PMSTATH,IB3F            ; Data input?
        BRA	PSP_FL                  ; No!
; One byte has been received from the host
	BTFSC   COM_ST,CO_ACT           ; New Communication Cycle?
        BRA     CON_CO                  ; No!
; Do not start a new Communication cycle
; during a PS/2 or server polling cycle
	BTFSS	PLY_ST,PL_PRE		;
	BTFSS	T3CON,TMR3ON		; Wait for the PS/2 cycle
	BTFSC	USB_ST,SB_REQ		; or server polling cycle
	BRA	D_BUSY			; to finish
	BTFSC	ETH_ST,ET_CYC		; Hold cycle?
	BRA	D_BUSY			; Yes!
; Start of a new Communication Cycle
	CLRF    COM_ST                  ; Clear communication state
        BSF	COM_ST,CO_CYC		; Signal: New cycle
        BCF     DEV_ST,DV_DAT           ; Signal: Data not ready!
	BTFSC	MSE_ST,MS_PRE		; Inhibit the PS/2 device
	RCALL	PS_INH			; if present
	MOVFF   FSR0L,PRODL             ; Save the buffer pointer
        MOVFF   FSR0H,PRODH
        MOVLW   PL_HLN-1         	; Header Block length - 1
	BRA	STA_CO			; Start communication stage	
; Clear buffer full flags 0 and 1 to avoid an overflow condition which
; blocks any further data input (undocumented).
PSP_FL	MOVF	BANK_R,W		; Clear IB0F
	MOVF	AY_ADR,W		; Clear IB1F
	MOVF	BANK_P,W		; Clear IB2F
	RETFIE	FAST
D_BUSY	MOVF	PSP_DI,W		; Discard received byte
	BSF	DEV_ST,DV_BSY		; Signal: Busy
CK_PST	BTFSS	T3CON,TMR3ON		; PS/2 timer running?
	BRA     LTM_IT                  ; No!
; PS/2 interrupts
	BTFSS   PIR2,TMR3IF             ; PS/2 timer interrupt?
        BRA     PCK_IT                  ; No!
	BTFSC	PLY_ST,PL_PRE		; Tape player cable present?
	GOTO	PLY_IT			; Yes
; PS/2 timer overflow interrupt
	BTFSC   TRISC,CK1               ; PS/2 clock line asserted? 
        BRA     PS_TMO                  ; No! PS/2 time-out!
; Start transmission to Mouse
        BCF     LATC,DT1 		; Assert the DT line to
        BCF     TRISC,DT1               ; produce the start bit
        BSF     MOVE_B,7                ; Set parity to 'odd'
        MOVLW   .10			; Number of bits = 10
        MOVWF   MOVE_X
        s_time  .15000                  ; Set time-out period
        BCF     INTCON2,INTEDG1         ; Interrupt on falling CK line
        BSF     TRISC,CK1               ; Release the clock line
PCK_EI  BCF     INTCON3,INT1IF          ; Clear CK interrupt flag
        BSF     INTCON3,INT1IE          ; Enable CK interrupt
; Start PS/2 timer
PTM_ST  BCF     T3CON,TMR3ON            ; Stop PS/2 timer
        BCF     PIR2,TMR3IF             ; Clear interrupt flag
        MOVWF   TMR3L                   ; Load timing constant
        BSF     T3CON,TMR3ON            ; Start PS/2 timer
        BSF     DEV_ST,DV_BSY           ; Signal: Busy!
        RETFIE  FAST
; PS/2 time-out
PS_TMO  BSF     TRISC,DT1               ; Release the DT line
; Force restart PS/2 reception
PS_FRX  RCALL   PS_STP
	RCALL	PS_STA
	RETFIE	FAST
PCK_IT  BTFSC   INTCON3,INT1IE          ; CK interrupt enabled?
        BTFSS   INTCON3,INT1IF          ; CK interrupt?
        BRA     SRX_IT                  ; No!
; PS/2 CK interrupt: Receive byte or Transmit bit
        BTFSS   INTCON2,INTEDG1         ; Transmitting?
        BRA     TX_BIT                  ; Yes!
; Receive byte from PS/2 device
        BCF     INTCON3,INT1IE          ; Disable CK interrupt
        BSF     RCSTA1,SPEN             ; Enable USART
        BSF     PIE1,RC1IE              ; Enable USART RX interrupt
        BSF     RCSTA1,CREN             ; Enable USART reception
        MOVF    RCREG1,W                ; Clear interrupt flag
        RETFIE  FAST
; Transmit bit to mouse
TX_BIT  BCF     INTCON3,INT1IF          ; Clear interrupt flag
        DCFSNZ  MOVE_X                  ; Stop bit?
        BRA     SET_DT                  ; Yes!
        MOVF    MOVE_B                  ; Prepare parity bit.
        DECFSZ  MOVE_X                  ; Parity bit ?
        RRNCF   MS_REG                  ; No! Get the next data bit
        INCFSZ  MOVE_X                  ; Adjust the bit counter
        BN      SET_DT                  ; Test the current bit
        BCF	LATC,DT1
	BCF     TRISC,DT1               ; Send a '0' bit
        RETFIE  FAST
SET_DT  BTG     MOVE_B,7                ; Update the parity bit
        BSF     TRISC,DT1               ; Send a '1' bit
        TSTFSZ  MOVE_X
        RETFIE  FAST
; Wait for an 'Ack' pulse after transmitting the parity bit
        BCF     INTCON3,INT1IE          ; Disable CK interrupt
	RCALL	PDT_EI
	RETFIE  FAST
PDT_IT  BTFSC   INTCON3,INT2IE          ; DT interrupt enabled?
        BTFSS   INTCON3,INT2IF          ; DT interrupt?
        RETFIE  FAST                    ; No!
; PS/2 DT interrupt: Start-receive or end-transmit
	BSF     INTCON2,INTEDG1         ; Interrupt on rising CK line
        BTFSC   MSE_ST,MS_TRA           ; 'Ack' pulse?
        BRA     PS_FRX                  ; Yes!
        BCF     INTCON3,INT2IE          ; Disable DT interrupt
        s_time  .2000                   ; Set time-out period
        BRA     PCK_EI                  ; Start receiving
; Asynchronous serial port interrupts
ASY_IT	BTFSS	SYS_FL,AS_ENB		; Port enabled?
	BRA	PDT_IT			; No! Check PS/2 port               
        BTFSC   PIE1,RC1IE              ; USART RC interrupt enabled?
        BTFSS   PIR1,RC1IF              ; USART RC interrupt?
        BRA     ATX_IT                  ; No!
; Asynchronous serial receive
        BSF     DEV_ST,DV_BSY           ; Signal: Busy!
	RCALL	WR_PTR			; Get buffer write pointer
	MOVF	RCSTA1,W		; USART Receive Status
	MOVFF	RCREG1,POSTINC1		; Copy received byte to buffer
	ANDLW	1<<FERR|1<<OERR		; Error?
	BZ	ARX_PT
	BCF	RCSTA1,CREN
	BSF	RCSTA1,CREN
ARX_PT	RCALL	WRPT_A			; Save buffer write pointer
	MOVF	RDPT_L,W
	SUBWF	WRPT_L,W
	INCFSZ	WREG
	BRA	CK_REZ
	BCF     PIE1,RC1IE              ; Disable reception
	BRA	NO_RTS
CK_REZ	ADDLW	SRB_RZ-1	
	BC	NO_RTS
	BTFSC   COM_ST,CO_ACT           ; Communication in progress?
NO_RTS	BSF	PORTB,B_NRTS		; Signal: Do not send!
	BSF	TIM_ST,TM_ASY		; Signal: timer running
	RCALL	KYB_TM			; Start timer
	RETFIE	FAST
ATX_IT	BTFSC   PIE1,TX1IE              ; USART TX interrupt enabled?
        BTFSS   PIR1,TX1IF              ; USART TX interrupt?
        RETFIE	FAST			; No!
; End of asynchronous serial transmission
	BCF	PIE1,TX1IE		; Disable USART TX interrupt
	BRA	STRT_T			; Send response to host	
SRX_IT  BTFSC   PIE1,RC1IE              ; USART RC interrupt enabled?
        BTFSS   PIR1,RC1IF              ; USART RC interrupt?
        BRA     PDT_IT                  ; No!
; USART synchronous reception buffer full interrupt
        BCF     RCSTA1,CREN             ; Stop reception
        MOVFF   RCREG1,MS_REG           ; Received byte to MS_REG
; Calculate the parity bit of ( MS_REG )
        SWAPF   MS_REG,W                ; Calculate: abcdefgh^
        XORWF   MS_REG                  ;            efghabcd^
        RRNCF   MS_REG,W                ;            habcdefg^
        XORWF   MS_REG,W                ;            defghabc
        BTFSC   WREG,2                  ; Bit0^Bit2 gives the
        INCF    WREG                    ; parity of the byte
        XORWF   RCSTA1,W                ; XOR with Bit 9
; Check parity
        BTFSS   WREG,RX9D
        BRA     PAR_ER                  ; Parity error!
        BCF     MSE_ST,MS_RSD
        MOVF    RCREG1,W
	MOVWF	MS_REG			; Received byte to MS_REG
        BCF     PIE1,RC1IE              ; Disable USART RX interrupt
        BCF     RCSTA1,SPEN             ; Disable USART
        BCF     INTCON3,INT2IF          ; Clear DT interrupt flag
        BSF     INTCON3,INT2IE          ; Enable DT interrupt
        BTFSS   DEV_ST,DV_MSE           ; Data reporting enabled?
        BRA     MSE_RX                  ; No!
        BTFSC   WRPT_L,1                ; First byte of packet?
        BRA     GET_PP                  ; No!
        XORLW   PS_RSD                  ; 'Resend' from mouse?
        BZ      PS_FRX                  ; Yes! Discard
        XORLW   PS_RSD^PS_ERR           ; 'Error' from mouse?
        BZ      PS_FRX                  ; Yes! Discard
GET_PP  RCALL   WR_PTR                  ; Packet pointer to FSR1
        MOVFF   MS_REG,POSTINC1         ; Save received byte
        BTFSS   FSR1L,2                 ; Packet done?
        BRA     SAV_PP                  ; Not yet!
; A movement packet has been received
        BCF     T3CON,TMR3ON            ; Stop PS/2 timer
        BCF     PIR2,TMR3IF             ; Clear interrupt flag
	BCF     DEV_ST,DV_BSY           ; Signal: Not Busy!
; Update the mouse coordinates
MS_MPK  MOVFF   MWIN_Y,POSTDEC1         ; Mouse window height
        MOVF    MPOS_Y,W                ; Y-coordinate
        MOVF    MOVE_B                  ; Set N-flag for Y-overflow
        RCALL   ADD_MV                  ; Add Y-movement
        MOVFF   MWIN_X,MS_REG           ; Mouse window width
        MOVF    MPOS_X,W                ; X-coordinate
        RLNCF   MOVE_B                  ; Set N-flag for X-overflow
        RCALL   ADD_MV                  ; Add X-movement
        RRNCF   MOVE_B                  ; Restore first byte
        MOVFF   MOVE_X,MPOS_X           ; Save new X-coordinate
        MOVFF   MOVE_Y,MPOS_Y           ; Save new Y-coordinate
        RCALL   WR_PTR                  ; Reset packet pointer
        BTFSS   MSE_ST,MS_ENB           ; Mouse enabled?
        BRA     MOUS_E                  ; No! Any buttons pressed?
; If enabled and both buttons are pressed, disable the mouse
        MOVF    MOVE_B,W
        ANDLW   1<<MV_REZ|1<<MV_MBM|1<<MV_MBR|1<<MV_MBL
        XORLW   1<<MV_REZ|1<<MV_MBR|1<<MV_MBL
        BZ      MOUS_D                  ; Disable mouse
        BTFSC   MSE_ST,MS_JOY           ; Mouse as joystick?
        BRA	MS_RET                  ; Yes!
; Update the Kempston mouse output ports
        BTFSS   MOVE_B,MV_MBR           ;
        BSF     DEV_ST,DV_MBR           ; Update the right button
        BTFSC   MOVE_B,MV_MBR           ; bit of the Kempston
        BCF     DEV_ST,DV_MBR           ; mouse buttons port
        BTFSS   MOVE_B,MV_MBL           ; 
        BSF     DEV_ST,DV_MBL           ; Update the left button
        BTFSC   MOVE_B,MV_MBL           ; bit of the Kempston
        BCF     DEV_ST,DV_MBL           ; mouse buttons port
        MOVFF   MPOS_Y,MOUS_Y           ; Update the Y-coordinate
        MOVFF   MPOS_X,MOUS_X           ; Update the X-coordinate
MS_RET	BTFSC	USB_ST,SB_POL
	BRA	KY_RET
        RETFIE  FAST
SAV_PP  RCALL   WR_PTR                  ; Save packet pointer
        RETFIE  FAST
SDC_IT  BTFSC   PIE3,SSP2IE             ; SPI interrupt enabled?
        BTFSS   PIR3,SSP2IF             ; SPI interrupt?
        BRA     USB_IT                  ; No!
; SPI Interrupt signals completion of a SD card operation
        BCF     PIE3,SSP2IE             ; Disable the SPI interrupt
        BTFSC   SDC_ST,SD_MBR           ; Start multiple block read?
        BRA     MBR_ST                  ; Yes!
	MOVLW	POL_KY
	CPFSLT	CMD_CD
	BRA	PREP_T
        BTFSS   COM_ST,CO_RDT           ; 'Read - Transmit' command?
        BRA     STRT_T                  ; No! Send response to host
; Read - Transmit command, second stage
; The specified SD card block has been read and is now sent to the server
        BCF     COM_ST,CO_RDT           ; Clear the 'Read-Transmit' flag
	MOVF	CMD_CD,W
	BNZ	RET_ER
SRV_TX  MOVLW   SV_WRS                  ; Server command: 'Write sector'
        MOVWF   CMD_CD
        MOVFF	DST_ST,CTL_CD		; Set the destination station
	BRA     PREP_T
USB_IT  BTFSC   PIE2,USBIE              ; USB interrupt enabled?
        BTFSS   PIR2,USBIF              ; USB interrupt?
        BRA     ASY_IT                  ; No!
; USB interrupt
        copy_b	FSR1L,I_FSR1,2		; Save FSR1
        RCALL   INT_SB                  ; Service the USB interrupt
	BTFSC	SYS_FL,SR_PRE		; If the RAM drive is present
	BSF	LATC,C_NSRV		; switch the USB LED off
	copy_b	I_FSR1,FSR1L,2		; Restore FSR1
	BC	END_CO			; CC data transfer done
        BZ      RET_ER                  ; Error
        RETFIE  FAST
; Timer 0 Overflow Interrupt :
; - USB RESUME signal timing
; - USB cable (dis)connect de-bounce
; - SD card insert/eject de-bounce
; - SD operation card time-out
; - PSP Communication Stage time-out
; - USB Communication Stage time-out
; - End of asynchronous reception
; - End of key output period
; - Server polling
; - Joystick sampling
; - Ethernet time-out
LTM_IT  BTFSS   INTCON,TMR0IF           ; Timer 0 Interrupt?
        BRA	SDC_IT			; No!
	MOVF	MSG_TM,W
	BZ	CK_ETM
	ADDWF	TICK_N,W
	BNC	SET_MT
	BTG	LATC,C_NSRV       
	MOVLW	-.10
SET_MT	MOVWF	MSG_TM
CK_ETM	BTFSS	TIM_ST,TM_ETH		; Ethernet timer running?
	BRA	DEF_TK			; No!
; Handle the Ethernet timer
	MOVF	TICK_N,W		; Update the tick counter
	ADDWF	TICK_L
	MOVLW	0
	ADDWFC	TICK_H			; Time out?
	BNC	DEF_TK			; No!
	BCF	TIM_ST,TM_ETH		; Stop Ethernet timer
	BSF	ETH_FL,EF_TMO		; Signal: Ethernet time-out
DEF_TK	MOVLW	.28
	MOVWF	TICK_N
	RCALL   TIM_IT                  ; Handle asynchronous events.
	BTFSC	TIM_ST,TM_ASY
; Asynchronous reception stage time-out occurs after 50ms
	BRA	ASY_TO
	BTFSC	TIM_ST,TM_KYB
; Keyboard time-out occurs after 50ms
	BRA	KYB_TO
	BTFSC   TIM_ST,TM_PSP           ; PSP time-out error?
; PSP Communication Stage time-out occurs after 1s
        BRA     PSP_ER                  ; Yes!
;	BTFSC   TIM_ST,TM_USB           ; USB time-out error?
; USB Communication Stage time-out occurs after 1s
;	BRA     CO_ERR                  ; Yes!
; Poll the server keyboard and mouse, if enabled
	BTFSC	USB_ST,SB_POL		; Server polling enabled?
	BRA	POL_RQ
; If the mouse is disabled the joystick is sampled every 50ms
	BTFSS   MSE_ST,MS_ENB           ; Mouse enabled?
	BRA	JOY_RD			; No!
; Otherwise, only the fire button is sampled
        BTFSS   PORTA,A_JOYF            ; Disable mouse?
MOUS_D  RCALL   MOUS_S                  ; Yes!
	BTFSC	USB_ST,SB_REQ		; Server polling enabled?
	BRA	M_DONE			; Yes!
; The mouse as jostick is sampled every 0.1 - 1.0 seconds
        BTFSC   MSE_ST,MS_JOY           ; Mouse as joystick?
        RCALL   JOY_MS                  ; Yes!
; Check if
; - The USB cable was connected or disconnected
; - The SD card was inserted or ejected
CK_CHG  RCALL   INT_CG
        RETFIE  FAST
; Signal end of asynchronous reception stage
ASY_TO	BTFSS	PORTB,B_NRTS		; RTS asserted?
	BRA	NO_RTS			; Yes! 
	BCF	TIM_ST,TM_ASY		; Signal: Timer not running
	BCF     DEV_ST,DV_BSY           ; Signal: Not Busy!
	BCF	DEV_ST,DV_BFE		; Signal: Data available!
        RETFIE  FAST
; PSP Interrupt (continued)
; Start of a new Communication Stage with the host over the PSP
STA_CO  BSF     COM_ST,CO_ACT           ; Communication stage started!
        BSF     COM_ST,CO_HED           ; Signal: Header Block needed!
        MOVWF   CNT_LO                  ; Copy the Header Block length
        CLRF    CNT_HI                  ; to the byte counter
        LFSR    FSR0,CMD_CD             ; Reset the buffer pointer
        BSF     TIM_ST,TM_PSP           ; Signal: PSP timer running
        RCALL   TIM_1S                  ; Start timer
CON_CO  BTFSS   COM_ST,CO_TRA           ; Transmission in progress?
; One byte has been received over the PSP
        MOVFF   PSP_DI,POSTINC0         ; Copy received byte to buffer
; Test if the byte counter has reached 0
TST_CN  TSTFSZ  CNT_LO                  ; Test if the byte counter has
        BRA     DEC_LO                  ; reached 0.
        TSTFSZ  CNT_HI
        BRA     DEC_HI
; All data has been transferred. Check if there is a Data Block, of up
; to 512 bytes, to be transferred after the 4-byte Header Block
        BTFSS   COM_ST,CO_HED           ; Data Block finished?
        BRA     END_CO                  ; Yes!
; Prepare to transfer the Data Block
        BCF     COM_ST,CO_HED           ; Signal: Header Block done!
        MOVFF   BLN_LO,CNT_LO           ; Copy the Data Block size
        MOVFF   BLN_HI,CNT_HI           ; to the byte counter and
        BRA     TST_CN             	; test again
; Decrement the byte counter
DEC_HI  DECF    CNT_HI                  ; Decrement the byte counter
DEC_LO  DECF    CNT_LO
        BTFSS   COM_ST,CO_TRA           ; Receive?
; One byte has already been received and copied to the buffer
        RETFIE  FAST                    ; Done!
; Transmit one byte over the PSP
PSP_TX  MOVFF   POSTINC0,PSP_DO         ; Byte to transmit
        BSF     DEV_ST,DV_DAT           ; Signal: Data ready!
        RETFIE  FAST
; End of the Communication Cycle
CY_END  BCF     DEV_ST,DV_DAT           ; Signal: Data not ready!
	BCF	COM_ST,CO_CYC		; Signal: Cycle end
	CLRF	CTL_CD			; Clear the control byte
	BTFSS	USB_ST,SB_POL
	BTFSC	MSE_ST,MS_PRE		; Enable the PS/2 device
	RCALL	PS_STA			; if present
	BRA	CK_RTS
; Communication Error
; Send an error message to the host
CO_ERR  MOVLW   E_COMM                  ; Communication error code
RET_ER	BSF     COM_ST,CO_SER           ; Parallel
SND_CM	BCF     COM_ST,CO_TRA           ; Transmission follows!
	MOVWF   CMD_CD                  ; Set the error code
	CLRF    BLN_LO                  ; Set the Data Block size
        CLRF    BLN_HI                  ; to 0
; End of Communication Stage
END_CO  BCF     COM_ST,CO_ACT           ; End of Communication Stage!
        BCF     TIM_ST,TM_PSP		; Stop the PSP and USB timers
	BCF	TIM_ST,TM_USB
	MOVLW	.20			; Set tick to 1s
	MOVWF	TICK_N
; Prepare the next Communication Stage
        BTG     COM_ST,CO_TRA           ; Toggle Transmit/Receive flag
        BTFSS   COM_ST,CO_TRA           ; Receive to follow?
        BRA     PREP_R                  ; Yes!
; A Transmit Stage is triggered immediately after each Receive Stage,
; within the same interrupt
        BTFSC   COM_ST,CO_SER           ; Reception was from host?
        BRA     PREP_T                  ; No!
        MOVLW   S_RXTX                  ; Is the received code a
        CPFSLT  CMD_CD                  ; Server command?
	BRA	CK_RT1			; No!
; Prepare for Transmission
PREP_T  BTG     COM_ST,CO_SER           ; Toggle Serial/Parallel flag
        BTFSS   COM_ST,CO_SER           ; Serial transmission?
        BRA     STRT_T                  ; No!
	MOVF	CMD_CD,W
	ANDLW	0xf0
	XORLW	SV_OPN			; 'Open file' command?
        BNZ	CK_CRE			; No!
	MOVFF	CTL_CD,SRC_ST		; Yes! store source station
	BRA	CK_CTL
CK_CRE	XORLW	SV_OPN^SV_CRE		; 'Create file' command?
	BNZ	CK_CTL			; No!
	MOVFF	CTL_CD,DST_ST		; Yes! Store destination station
CK_CTL	TSTFSZ	CTL_CD
	BRA	NET_SV
SER_TX	RCALL   USB_TX
USB_EN  BZ      RET_ER
POL_TM  BTFSC	USB_ST,SB_REQ		; If polling the server
	BRA	JOY_RD			; restart the polling timer
	RETFIE	FAST
NET_SV	BSF	SDC_ST,SD_EXE		; Network server request
	BCF     COM_ST,CO_SER           ; Parallel communication follows
        BRA	POL_TM
; Asynchrous Serial port operation
ASY_CM	BTFSS	SYS_FL,AS_ENB		; Port enabled?
	BRA	CO_ERR			; No! Signal error 	
	CPFSEQ	CMD_CD			; Asynchronous serial transmit?	
	BRA	EX_SDC			; Yes!
	BTFSC	SYS_FL,AS_PRN		; Transmission only?
	BRA	CO_ERR			; Yes! Signal error 	
	CLRF	BLN_LO			; No response data
	TSTFSZ	DATA_B
	BRA	ARX_CM
	BSF	SYS_FL,AS_RTS
	BRA	DC_RET
; 'Asynchronous Serial Receive' command
ARX_CM	MOVF	RDPT_L,W
	SUBWF	WRPT_L,W
	BZ	ARX_ND			; Zero length
	MOVWF	MS_REG			; Available data length
	CPFSGT	DATA_B			; Compare with requested
	MOVF	DATA_B,W		; length
	MOVWF	BLN_HI			; Length of data to be
	MOVWF	BLN_LO			; transferred
	SUBWF	MS_REG			; Remaining data lemgth
	LFSR    FSR0,DATA_B             ; Destination pointer
	RCALL	RD_PTR			; Get the buffer read pointer
ARX_LP	MOVFF	POSTINC1,POSTINC0
        BTFSC   FSR1H,SBF_EN            ; Buffer end?
	LFSR	FSR1,S_BUFF             ; Wrap around buffer pointer
	DECFSZ	BLN_HI                  ; Done?
	BRA	ARX_LP			; Not yet!
	RCALL	RD_PTR			; Save the buffer read pointer
	TSTFSZ	MS_REG			; Any data left unread?
	BRA	DC_RET			; Yes! 
ARX_ND	BSF	DEV_ST,DV_BFE		; Signal: Buffer empty!
DC_RET  CLRF    CMD_CD                  ; Clear the error code
STRT_T  BTFSS	USB_ST,SB_CMD		; MCU server command?
	BRA	CK_RW2			; No!
	BCF	USB_ST,SB_CMD		; Signal: Command completed
	RETFIE	FAST
CK_RW2	MOVLW   PL_HLN           	; Size of the Header Block
	BTFSS   COM_ST,CO_RWR           ; 'Receive - Write' command?
	BRA	CK_RT2			; No!
; 'Receive - Write' command, second stage
; The sector received from the server is written to the SD card
        BCF     COM_ST,CO_RWR           ; Clear flag
        MOVLW   SD_WRB                  ; SD card command: 'Write block'
        MOVWF   CMD_CD
	MOVFF	DST_ST,CTL_CD		; Set destination host
	BRA	EX_SDC			; Execute
CK_RT2	BTFSS   COM_ST,CO_RXT           ; 'Receive - Transmit' command?
	BRA     CK_POL                  ; No!
; 'Receive - Transmit' command, second stage
; The received sector is transmitted to another server
        BCF     COM_ST,CO_RXT           ; Clear flag
        BRA	SRV_TX			; Transmit sector
CK_RT1	CPFSEQ	CMD_CD			; 'Receive-Transmit'
	BRA	CK_ASY			; No!
; 'Receive - Transmit' command, first stage
; A sector is requested from the server
	BSF     COM_ST,CO_RXT           ; Set the 'Receive-Transmit' flag
	BRA	SRV_RX			; Receive sector
CK_ASY	MOVLW   ASY_RX
	CPFSGT	CMD_CD			; Asynchronous port command?
	BRA	ASY_CM			; Yes!
; Execute a SD card or Serial RAM command
	TSTFSZ	BLN_LO			; If a block number is specified	
	BTFSS	DATA_B+3,7		; and its highest byte is larger
	BRA	CK_CMD			; than 0x80,
	BSF	SYS_FL,SR_SEL		; signal: Serial RAM command!
CK_CMD	MOVLW   SD_SWA
        CPFSEQ  CMD_CD                  ; 'Set Write Address' command?
        BRA     CK_RWR                  ; No!
; Execute a 'Set Write Address' command
        RCALL   SET_WA                  ; Set the write address
        CLRF	BLN_LO			; No response data
        BRA     DC_RET                  ; Send response
CK_RWR	MOVLW   SD_RWR
        CPFSEQ  CMD_CD                  ; 'Receive - Write' command?
	BRA	EX_SDC			;  No!
; 'Receive - Write' command, first stage
; A sector is requested from the server
	RCALL   SET_WA                  ; Set the write address
        MOVFF	CTL_CD,DST_ST		; Save destination station
	MOVFF	SRC_ST,CTL_CD		; Set source station
	BSF     COM_ST,CO_RWR           ; Set the 'Receive-Write' flag
SRV_RX  MOVLW   SV_RDS                  ; Server command: 'Read sector'
	MOVWF   CMD_CD			; Send command to server
	CLRF	BLN_LO			; No data
        BRA     PREP_T                  ; Send request to server
; Execute a regular SD card command
EX_SDC  BSF     SDC_ST,SD_EXE           ; Signal: Execute SDC command!
        MOVLW   SD_RDM
        CPFSEQ  CMD_CD                  ; Read multiple block command?
        RETFIE  FAST                    ; No!
; Multiple block read command
	BTFSS   SDC_ST,SD_MBR           ; Multiple block read in
        RETFIE  FAST                    ; progress?
        TSTFSZ  BLN_LO                  ; Block address specified?
        BRA     MBR_EN
; Advance the multiple block read pointer to the beginnig of a sector
        BCF     T0CON,TMR0ON            ; Stop timer 0
        BCF     INTCON,TMR0IF           ; Clear interrupt flag
        MOVF    PRODH,W                 ; Saved FSR0H value
        DECFSZ  PRODL                   ; Saved FSR0L value
        INCF    WREG
        INCF    WREG                    ; Advance the read pointer to
        ANDLW   ~1                      ; the beginning of a sector
        MOVWF   FSR0H
        CLRF    FSR0L
ADJ_RP  BTFSC   FSR0H,RBF_SZ            ; Wrap around the ring pointer
MBR_ST  LFSR    FSR0,RING_B
        BRA     PSP_TX                  ; Read next byte from buffer
MBR_EN  BCF     SDC_ST,SD_MBR           ; Stop multiple block read
        RETFIE  FAST
; PSP error
PSP_ER  BCF     TIM_ST,TM_PSP		; Stop the PSP timer
        CLRF    COM_ST                  ; Clear communication status
	MOVF	CMD_CD,W		; If the command is 'Play tape
	XORLW	SD_PLY			; file',
	BNZ	PREP_R			; play the operating system
	BSF     SDC_ST,SD_EXE		; loader
; Prepare for reception
PREP_R  BTFSS   COM_ST,CO_SER           ; Serial receive?
        BRA     CY_END                  ; No!
	RCALL   USB_RX                  ; Yes!
	BRA	USB_EN
; Parity error. Send a 'Resend' or 'Error' command
PAR_ER  MOVLW   PS_ERR
        BTFSC   MSE_ST,MS_RSD
        BRA     MSE_CH
        BSF     MSE_ST,MS_RSD
        MOVLW   PS_RSD                  ; 'Resend' comand
MSE_CH  BCF     MSE_ST,MS_HLD
; Transmit byte ( WREG ) to the mouse
MSE_TX  MOVWF   MS_REG                  ; Store mouse command
        RCALL   PS_DIS			; Disable all PS/2 interrupts
; Signal: 'Request to send' to the mouse
        BCF     LATC,CK1
        BCF     TRISC,CK1               ; Assert the clock line
        BSF     MSE_ST,MS_TRA           ; Signal: Transmitting!
        s_time  .200                    ; Set time period
        BRA     PTM_ST                  ; Start transmission
; Enable mouse data reporting or receive from keyboard
MSE_RX  BCF     T3CON,TMR3ON            ; Stop PS/2 timer
        BCF     PIR2,TMR3IF             ; Clear interrupt flag
        BSF	MSE_ST,MS_PRE		; Signal PS/2 device present
	BNZ     CK_ACK                  ; Mouse ID ( 0 )?
; Send an 'Enable data reporting' command
        BSF     MSE_ST,MS_HLD		; Signal: 'Enable reporting'
        MOVLW   PS_REP			; command sent!
        BRA     MSE_TX			; Send command
CK_ACK  XORLW   PS_ACK                  ; Acknowledge?
        BNZ     CK_ERR                  ; No!
        BTFSC   MSE_ST,MS_HLD           ; After 'Enable reporting'?
        RCALL   MOUS_K                  ; Yes! Enable mouse
MS_NBY	BCF	DEV_ST,DV_BSY		; Signal: Not Busy!
	RETFIE  FAST
CK_ERR	XORLW   PS_ACK^PS_ERR           ; Error?
	BNZ	CK_RSD			; If yes, send a 
MS_RST  MOVLW   PS_RST			; 'Reset' command
        BRA     MSE_CH
CK_RSD	XORLW   PS_ERR^PS_RSD           ; Resend?
	BZ	MS_RST			; Yes!
        XORLW   PS_RSD^PS_TST           ; Self-test passed?
        BZ      MS_NBY			; Yes!
; A byte has been received from the PS/2 keyboard or the server
	XORLW	PS_TST^C_EX		; Extended scan code prefix?
CK_EXT	BNZ	CK_BRK			; No!
	BSF	KYB_ST,KB_EXT		; Else set flag
	BRA	NXT_KY
CK_BRK	XORLW	C_EX^C_BK		; Break scan code prefix?
	BNZ	ST_KEY			; No!
	BSF	KYB_ST,KB_BRK		; Else set flag
	BRA	NXT_KY
ST_KEY	XORLW	C_BK
; A scan code has been received from the PS/2 keyboard or the server
	MOVWF	JS_SEN			; Store the scan code
	BSF	KYB_ST,KB_KEY		; Signal: Scan code received!
	RETFIE  FAST
; Poll the server for keyboard and mouse data
POL_RQ	BTFSS	COM_ST,CO_CYC		; During communication cycle or
	BTFSC	USB_ST,SB_REQ		; polling in progress?
	BRA	JOY_RD			; Yes!
        BTFSS	ETH_ST,ET_CYC		; Cycle on hold or
	BTFSS   DEV_ST,DV_SRV		; server not connected?
	BRA	JOY_RD			; Yes!
	BSF     DEV_ST,DV_BSY           ; Signal: Busy!
	BSF	USB_ST,SB_REQ		; Signal: Polling in progress!
NXT_KY	BTFSS	USB_ST,SB_REQ		; External PS/2 keyboard?
	RETFIE	FAST			; Yes!
	MOVLW	GET_KY			; Send a polling request
	BRA	SND_CM			; to the server
CK_RTS	BTFSS	SYS_FL,AS_RTS		; Assert RTS?
	BRA	UPD_JS			; No!
; Assert RTS
	BCF	SYS_FL,AS_RTS		; Signal: Do not assert RTS!
	BSF     PIE1,RC1IE              ; Enable reception interrupt
	BCF	PORTB,B_NRTS		; Assert RTS
JOY_RD  RCALL	KYB_TM			; Restart the joystick timer
        BTFSS	MSE_ST,MS_KYB		; Mouse enabled or keyboard
UPD_JS	BTFSC	MSE_ST,MS_ENB		; as Kempston joystick?
	BRA	CK_CHG			; Yes!
; Read Kempston joystick port
	COMF    PORTB,W
        ANDLW   1<<B_JOYR|1<<B_JOYL|1<<B_JOYD|1<<B_JOYU
        BTFSS   PORTA,A_JOYF
        BSF     WREG,JOY_FR
        MOVWF   KEM_JS
        BRA     CK_CHG
; Read mouse as Kempston joystick
JOY_MS  BTFSS   ETH_ST,ET_PRE           ; If ethernet module not attached,
        BTG     LATC,C_NMSE             ; toggle mouse LED
        BTFSC	SYS_FL,SR_PRE
        BRA	SMP_JS
	BTFSC	ETH_ST,ET_PRE
	BTG	LATC,C_NSRV
SMP_JS  MOVFF   KEM_JS,RDPT_H
; Get fire button state
        BCF     RDPT_H,JOY_FR           ; Copy the mouse left button
        BTFSC   MOVE_B,MV_MBL           ; bit to the joytick fire
        BSF     RDPT_H,JOY_FR           ; button bit
; Get the vertical movement bits: 'Up' and 'Down'
        MOVFF   MPOS_Y,RDPT_L           ; Get the current Y-coordinate
        MOVLW   0x80                    ; Center the  Y-coordinate
        MOVWF   MPOS_Y                  ; for the next read
        RRNCF   RDPT_H                  ; Shift vertical movement bits
        RRNCF   RDPT_H                  ; in place of horizontal bits
        RCALL   SW_DIR                  ; Swap 'Up' and 'Down' bits
        RCALL   MOUS_A                  ; Update vertical movement bits
        RCALL   SW_DIR                  ; Swap back 'Up' and 'Down' bits
        RLNCF   RDPT_H                  ; Shift back vertical movement
        RLNCF   RDPT_H                  ; bits to their normal position
; Get the horizontal movement bits: 'Left' and 'Right'
        MOVFF   MPOS_X,RDPT_L           ; Get the current X-coordinate
        MOVLW   0x80                    ; Center the X-coordinate
        MOVWF   MPOS_X                  ; for the next read
        RCALL   MOUS_A                  ; Update horizontal movement bits
        MOVFF   RDPT_H,KEM_JS           ; Copy new state to joystick port
MJ_TIM  MOVLW	.10			; Set tick to 0.5s
	MOVWF	TICK_N
	MOVFF   JSTC_H,TMR0H            ; Get the mouse as joystick
        MOVF    JSTC_L,W                ; sampling rate and
	BRA     STA_LT                  ; restart joystick timer
; Enable the mouse if only one button is pressed
MOUS_E  BTFSS   MOVE_B,MV_MBL           ; Mouse left button pressed?
        BRA     OTH_MB                  ; No!
        BTFSS   MOVE_B,MV_MBR           ; Right button also pressed?
        BRA     MSE_JS                  ; No!
; Set the 'mouse on hold' flag if both buttons are pressed
        BSF     MSE_ST,MS_HLD           ; Signal: Mouse on hold!
        BRA	MS_RET
OTH_MB  BTFSC   MOVE_B,MV_MBR           ; Mouse right button pressed?
        BRA     KEM_MS                  ; Yes!
; Clear the 'mouse on hold' flag if no button is pressed
        BCF     MSE_ST,MS_HLD           ; Clear 'hold' flag
        BRA	MS_RET
; Enable the mouse as Kempston joystick if the left button pressed
MSE_JS  BTFSS   MSE_ST,MS_HLD           ; Mouse on hold?
        RCALL   MOUS_J
        BRA	MS_RET
; Enable the mouse as Kempston mouse if the right button pressed
KEM_MS  BTFSS   MSE_ST,MS_HLD           ; Mouse on hold?
        RCALL   MOUS_K                  ; Enable mouse
        BRA	MS_RET
; Swap bits 0 and 1			; 76543210 C
SW_DIR  RRCF    RDPT_H                  ; C7654321 0
        RRNCF   RDPT_H                  ; 1C765432 0
        RLCF    RDPT_H                  ; C7654320 1
        RLCF    RDPT_H                  ; 76543201 C
        RETURN

;------------------------------------------------------------------------------
; MAIN PROGRAM

CONF_D
        banksel PPSCON
; Assign periferial pins
        MOVLW   .0                      ; Pin RA0/RP0 mapped to
        MOVWF   RPINR21                 ; input function SDI2
        MOVLW   .1                      ; Pin RA1/RP1 mapped to
        MOVWF   RPINR22                 ; input function SCK2
        MOVLW   .11                     ; Output function SCK2
        MOVWF   RPOR1                   ; mapped to pin RA1/RP1
        MOVLW   .10                     ; Output function SDO2
        MOVWF   RPOR2                   ; mapped to pin RA5/RP2
        MOVLW   .12                     ; Pin RC1/RP12 mapped to
        MOVWF   RPINR3                  ; input function INT3
        MOVLW   .17                     ; Pin RC6/RP17 mapped to
        MOVWF   RPINR1                  ; input function INT1
        MOVLW   .18                     ; Pin RC7/RP18 mapped to
        MOVWF   RPINR2                  ; input function INT2
        MOVLW   .12                     ; Pin RC1/RP12 mapped to
        MOVWF   RPINR16                 ; input function RX2
        MOVLW   .6                      ; Output function TX2
        MOVWF   RPOR13                  ; mapped to pin RC2/RP13
        MOVLW   0x55
        MOVWF   EECON2
        MOVLW   0xAA
        MOVWF   EECON2
        BSF     PPSCON,IOLOCK           ; Lock pin mapping
        banksel DEV_ST
; Port A configuration
        SETF    PORTA
        MOVLW   1<<A_JOYF|1<<RA0        ; SDI and joystick button
        MOVWF   TRISA                   ; are inputs
; Port B is input
	SETF	TRISB
; Port C configuration
	SETF	LATC
	BCF	LATC,CK1		; Mouse CK is asserted
        MOVLW   1<<DT1|1<<C_NEIT	; Mouse Data and Ethernet
        MOVWF   TRISC			; interrupt are inputs
; Ports D and E are configured as Parallel Slave Port
        RCALL	REG_CL
	MOVLW   1<<PMPEN|1<<PTWREN|1<<PTRDEN
        MOVWF   PMCONH
        MOVLW   1<<PTEN1|1<<PTEN0
        MOVWF   PMEL
        MOVLW   1<<IRQM0|1<<MODE0
        MOVWF   PMMODEH
        BSF     PIE1,PMPIE              ; Enable the PSP interrupt
	SETF	EXT_KB			; Signal: No keyboard
; External interrupt 2 configuration
        BCF     INTCON2,INTEDG2         ; Interrupt on falling DT line
; External interrupt 3 configuration
        BCF     INTCON2,INTEDG3         ; Interrupt on falling ETIT line
; Timer 0 configuration
        MOVLW   1<<T0PS2|1<<T0PS1|1<<T0PS0
        MOVWF   T0CON                   ; Timer 0 prescaler is 1:256
; Timer 3 configuration
        MOVLW   1<<T3CKPS1|1<<T3CKPS0|1<<RD16
        MOVWF   T3CON                   ; Timer 3 prescaler is 1:8
        BSF     PIE2,TMR3IE             ; Enable timer 3 interrupt
; SPI configuration
        MOVLW   1<<CKE			; SPI mode 0,0
        MOVWF   SSP2STAT
; USB configuration
        MOVLW   1<<UPUEN|1<<FSEN        ; Full speed, on-chip transreceiver,
        MOVWF   UCFG                    ; pull-up enabled, no ping-pong
; Enable SPI
        MOVLW   1<<SSPEN                ; Enable SPI at 12MHz
        MOVWF   SSP2CON1
; Start-up delay
        BCF     LATC,C_NSRV             ; Switch the USB LED on
        BSF     OSCCON,IDLEN            ; Set Idle mode on SLEEP
        BSF     WDTCON,SWDTEN           ; Start Watch-dog timer
        SLEEP                           ; Wait for 2 s
        NOP
        BCF     WDTCON,SWDTEN           ; Stop Watch-dog timer
        BSF     LATC,C_NSRV             ; Switch the USB LED off
; Initialize joystick and mouse parameters
        table_	JM_INI			; Get the table address
        RCALL   TAB_NW                  ; Initialise the variables
; Initialize device status
        MOVLW   1<<DV_MBL|1<<DV_MBR
        MOVWF   DEV_ST                  ; Initialize the device status
	CLRF	KYB_ST			; Initialize the keyboard status
	CLRF	PLY_ST			; Initialize the tape player status
	CLRF    ETH_ST                  ; Initialize the Ethernet status
	CLRF	SYS_FL			; Initialize system flags
; Clear interrupt-on-change mismatch
        MOVF    PORTB,W                 ; Clear RB port mismatch condition
        MOVWF   CHG_ST                  ; Save the value of the RB port.
; Check presence of additional hardware
        CALL	CK_SRM
; USART configuration
	BSF     ODCON2,U1OD             ; Open drain,
        BTFSS	CHG_ST,B_JOYD		; If joystick 'Up' and 'Down' are
	BTFSC	CHG_ST,B_JOYU		; asserted simultaneusly, the
        BRA	ASY_HD			; serial cable is connected
; Configure the USART for full duplex asynchronous communication
	RCALL	AS_CFG			; Enable USART transmit
	BSF	RCSTA1,CREN		; Enable USART receive
	BSF	PIE1,RC1IE		; Enable USART RX interrupt
	BSF	PORTB,B_NRTS		; Signal: Do not send!
	BCF	TRISB,B_NRTS		; Set nRTS as output
	BCF	DEV_ST,DV_NSR		; Signal: Serial device connected!
	BRA	STA_TM
ASY_HD	BTFSC	PORTC,DT1		; If the mouse DT line is asserted,
        BRA	MOUS_P			; a printer cable is connected
	BSF     TRISC,CK1               ; unless DT and CK are shorted in
	BRA	CK_EAR
CK_EAR	BTFSS	PORTC,DT1		; which case a tape player cable
        BRA	PR_CFG			; is connected
; Reconfigure timer 3 for the EAR port
        MOVLW   1<<RD16
        MOVWF   T3CON                   ; Timer 3 prescaler is 1:1
	BSF	PLY_ST,PL_PRE		; Signal: Tape player cable present
	BRA	STA_TM
; Configure the USART for asynchronous transmission to printer
PR_CFG	BCF	LATC,CK1
	BCF     TRISC,CK1
	RCALL	AS_CFG			; Enable USART transmit
	BSF	SYS_FL,AS_PRN		; Signal: Asynchronous TX only!
	BCF	DEV_ST,DV_NPR		; Signal: Printer connected!
	BRA	STA_TM
; Configure the USART for synchronous communication with mouse or keyboard
MOUS_P	BSF     BAUDCON1,TXCKP          ; Clock idle high,
        BSF     TXSTA1,SYNC             ; Synchronous slave,
        BSF     RCSTA1,RX9              ; 9-bit reception
; Enable mouse or keyboard
	CLRF	RDPT_L			; Signal: No scan code received
	BCF     INTCON3,INT2IF          ; Clear the DT interrupt flag
        BSF	INTCON3,INT2IE          ; Enable the DT interrupt
        BSF     TRISC,CK1               ; Release the CK line
; Start the system timer
STA_TM  RCALL   SDC_TM                  ; Start timer 0
; Clear the peripheral interrupt flags
	CLRF    PIR1                    ; Clear interrupt flags
        CLRF    PIR2                    ; Clear interrupt flags
        CLRF    PIR3                    ; Clear interrupt flags
; Enable interrupts
        MOVLW   1<<GIE|1<<PEIE|1<<TMR0IE
        MOVWF   INTCON                  ; Enable needed interrupts
; Detect the SD card
        RCALL	SDC_ID                  ; Detect SD card
; Attach USB
        RCALL   USB_CG                  ; Attach USB
; Main loop
M_LOOP  BTFSC   SDC_ST,SD_DET           ; Detect SD card request?
        RCALL	SDC_ID                  ; Yes! Detect SD card
        BTFSC   SDC_ST,SD_EXE           ; SD card command pending?
        RCALL   SDC_EX                  ; Yes! Execute SD card command
	BTFSS	ETH_ST,ET_PRE		; Ethernet module present?
	BRA	RD_KYB			; No!
	BTFSC	ETH_ST,ET_WFI		; Wifi plug-in present?
	BRA	CK_WFI			; Yes!
	BTFSC   INTCON3,INT3IF          ; Ethernet interrupt pending?
	CALL	ETH_IS			; Service Ethernet interrupt
	BRA	CK_TMO
CK_WFI	BTFSC	PIR3,RC2IF          	; Wifi reception pending?
	CALL	WFI_IS			; Wifi reception
CK_TMO	BTFSC	ETH_FL,EF_TMO		; Ethernet timer time-out?
	CALL	ETH_TO			; Service Ethernet time-out
RD_KYB  BTFSC	KYB_ST,KB_KEY		; Key scan code received?
	RCALL	KEY_DC			; Yes! Decode key
	BTFSS	ETH_ST,ET_PRE		; Ethernet module present?
	SLEEP				; No!
	BRA     M_LOOP

;------------------------------------------------------------------------------
; GENERAL SUBROUTINES

; Set up a SD card command string in RAM
CM_SET  RCALL   TAB_NW                  ; Copy the command string to RAM
        DECFSZ  SD_CRC,W                ; A CRC-7 value of 0, indicates
        RETURN                          ; that a data address is required
; Copy the data address from the data buffer to the command string. For
; command WRITE_BLOCK the data address was previously stored at (WR_BLN).
; The data address must be given in bytes, for standard capacity SD cards
; or Serial RAM and in blocks of 512 bytes for high capacity SD cards
	LFSR	FSR1,WR_BLN
	MOVLW	SD_WBK
	CPFSEQ	SD_CMD
	LFSR	FSR1,DATA_B
	BTFSC	SDC_ST,SD_EBF
	LFSR	FSR1,PR_BLK
	BTFSC	SYS_FL,SR_SEL		; Serial RAM command?
	BRA	NO_MSB			; Yes!
	BTFSC   SDC_ST,SD_HCS           ; Copy the most significant byte only
	MOVFF   POSTINC1,SD_AD0		; for a high capacity SD card
NO_MSB  MOVFF   POSTINC1,SD_AD1		; In this way, the block address
        MOVFF   POSTINC1,SD_AD2		; of standard capacity SD cards
        MOVFF   POSTINC1,SD_AD3		; and Serial RAM is multiplied by 256
        BTFSC	SYS_FL,SR_SEL		;
	BRA	DBL_AD           	; For standard capacity cards and
	BTFSC   SDC_ST,SD_HCS		; Serial RAM, multiply the address
        RETURN                          ; again by 2
DBL_AD  RLCF    SD_AD0,W                ; This clears the Carry Flag
        RLCF    SD_AD1                  ; Multiply by 2
        RLCF    SD_AD2
        RLCF    SD_AD3
        RETURN
; Configure USART for asynchronous communication
; at half the maximum Baud rate
AS_CFG	BSF	SYS_FL,AS_ENB		; Signal: Port enabled!
	BSF	BAUDCON1,BRG16		; 16-bit Baud rate generator
	MOVLW	low (2*FOSC_4/BAUD_R-1)
	MOVWF	SPBRG1
	MOVLW	high (2*FOSC_4/BAUD_R-1)
	MOVWF	SPBRGH1			; Baud rate: BAUD_R/2 bit/s
	MOVLW	1<<TXEN|1<<BRGH
	MOVWF	TXSTA1			; Enable USART transmit
	BSF	RCSTA1,SPEN		; Enable USART
        table_  PR_INI                  ; Load table address
        RCALL   TAB_NW                  ; Printer initialization string
	RCALL	WRPT_I			; Initialize input write pointer 
; Initialize the serial input buffer read pointer 
	LFSR	FSR1,S_BUFF
; Load the serial input buffer read pointer 
RD_PTR  swap_r  FSR1L,RDPT_L
        swap_r  FSR1H,RDPT_H
        RETURN
; Adjust and save the input buffer write pointer 
WRPT_A	BTFSC   FSR1H,SBF_EN		; Buffer end?
; Initialize the input buffer write pointer 
WRPT_I	LFSR	FSR1,S_BUFF
; Load the serial input buffer write pointer 
WR_PTR  swap_r  FSR1L,WRPT_L
        swap_r  FSR1H,WRPT_H
        RETURN
; Start the SD card time-out counter
SDC_TM  BSF     TIM_ST,TM_SDC           ; Signal: SD card timer is running
	MOVLW	.20			; Set tick to 1s
	MOVWF	TICK_N
        BRA     TIM_1S
; Set Write Block Address
SET_WA	copy_b	DATA_B,WR_BLN,PL_HLN
	RETURN

;------------------------------------------------------------------------------
; PS/2 MOUSE SUBROUTINES
;
; Inhibit the PS/2 device
PS_INH	BCF     LATC,CK1
	BCF     TRISC,CK1               ; Assert the clock line
; Stop the PS/2 timer and disable all PS/2 interrupts
PS_STP	BCF     T3CON,TMR3ON            ; Stop the PS/2 timer
        BCF     PIR2,TMR3IF             ; Clear the interrupt flag
	BCF     DEV_ST,DV_BSY           ; Signal: Not Busy!
; Disable all PS/2 interrupts
PS_DIS  BCF     INTCON3,INT2IE          ; Disable DT interrupt
        BCF     INTCON3,INT1IE          ; Disable CK interrupt
        BCF     PIE1,RC1IE              ; Disable USART RX interrupt
        BCF     RCSTA1,CREN             ; Disable USART reception
        BCF     RCSTA1,SPEN             ; Disable USART
        RETURN
; (Re)start PS/2 reception
PS_STA	BCF     MSE_ST,MS_TRA           ; Signal: Reception!
        MOVLW   low MOVE_B              ; Set movement packet pointer
	MOVWF   WRPT_L
	CLRF    WRPT_H
        BSF     DEV_ST,DV_MBR           ; Update the right button
        BSF     DEV_ST,DV_MBL           ; Update the left button
        BSF     TRISC,CK1               ; Release the clock line
PDT_EI  BCF     INTCON3,INT2IF          ; Clear DT interrupt flag
        BSF     INTCON3,INT2IE          ; Enable DT interrupt
	RETURN        
CK_POL	BTFSS	USB_ST,SB_REQ		; Server polling response?
	BRA	STA_CO			; No!
; Keyboard and mouse data returned by the server
	TSTFSZ	DATA_B			; Any mouse movement?
	BRA	MS_INP			; Yes!
; Return point after processing the mouse movement packet
KY_RET	MOVF	CMD_CD,W		; Get the key scan code
	BTFSS	DEV_ST,DV_MSE		; Mouse selected?
	BZ	NO_KEY			; Server key buffer empty!
	XORLW	C_ST			; Server input window closed?
	BZ	STOP_P			; Yes! Stop polling
	BTFSC	DEV_ST,DV_MSE		; Mouse selected?
	BRA	M_DONE			; Yes!
	XORLW	C_ST^C_EX		; Else handle the key
	BRA	CK_EXT			; scan code
; Process the mouse movement packet received from the server
MS_INP	RCALL	WR_PTR			; Get packet pointer
	MOVFF	DATA_B,POSTINC1
	MOVFF	DATA_B+1,POSTINC1
	MOVFF	DATA_B+2,POSTINC1	; Copy the movement packet
	BRA	MS_MPK			; and process it
; The server input window has been closed
STOP_P	RCALL	MOUS_S			; Disable the mouse
	BCF	USB_ST,SB_POL		; Disable server polling
; Mouse polling done
M_DONE	BCF	USB_ST,SB_REQ		; Signal: Polling done!
	BCF     DEV_ST,DV_BSY           ; Signal: Not Busy!
	RETFIE	FAST
; Keyboard time-out
KYB_TO	MOVF	EXT_KB,W
	MOVFF	JSTC_H,EXT_KB
	CPFSEQ	EXT_KB
	BRA	JOY_RD			; Read joystick
NO_KEY	BCF	TIM_ST,TM_KYB		; Signal: Timer not running
	RCALL	CS_KEY			; Signal: No key pressed
	BRA	JOY_RD			; Read joystick
; Enable Kempston mouse
MOUS_K  BSF     DEV_ST,DV_MSE           ; Signal: Mouse present!
	BSF     MSE_ST,MS_ENB           ; Signal: Mouse enabled
	RCALL	MSL_ON			; Switch on mouse LED
; Set mouse coordinates to origin
        CLRF	MOUS_Y
CLR_MS	CLRF	MOUS_X
        BSF     DEV_ST,DV_MBR           ; Update the right button
        BSF     DEV_ST,DV_MBL           ; Update the left button
        RETURN
; Enable mouse as joystick
MOUS_J  BTFSC	USB_ST,SB_POL		; Not when server polling
	RETURN				; is enabled
	BSF     MSE_ST,MS_ENB           ; Enable mouse
        BSF     MSE_ST,MS_JOY           ; Mouse as joystick
        CLRF    KEM_JS                  ; Clear joystick port
        BRA     MJ_TIM			; Restart mouse as joystick timer
; Disable the mouse when both buttons or joystick fire are pressed
MOUS_S  SETF	EXT_KB			; Signal: No keyboard
	RCALL	CLR_MS
	BTFSC	USB_ST,SB_POL		; Server polling enabled?
	BCF	DEV_ST,DV_MSE		; Signal: Mouse not present!
	BSF     MSE_ST,MS_HLD           ; Signal: Mouse on hold!
        BCF     MSE_ST,MS_ENB           ; Disable mouse
        BCF     MSE_ST,MS_JOY           ; Mouse not as joystick
	RCALL	ML_OFF			; Switch off mouse LED
KYB_TM	MOVLW	1			; Set tick to 0.05s
	MOVWF	TICK_N
	GOTO    TIM_50			; Start joystick timer
; Add coordinate (WREG) to movement value (INDF1)
ADD_MV  BTFSC   MSE_ST,MS_LEG           ; Legacy mode?
        BRA     LG_MOV                  ; Yes!
        BTFSC   MOVE_B,MV_SGN           ; Positive sign?
        BRA     N_SIGN                  ; No!
; Positive movement
        BN      P_OVFL                  ; Movement overflow!
        ADDWF   INDF1,W                 ; Add movement to coordinate
        BNC     MOV_EN                  ; In case of overflow, set
P_OVFL  SETF    WREG                    ; new coordinate to 255
        BRA     MOV_EN
; Negative movement
N_SIGN  BN      N_OVFL                  ; Movement overflow!
        ADDWF   INDF1,W                 ; Add movement to coordinate
        BC      MOV_EN                  ; In case of overflow, set
N_OVFL  CLRF    WREG                    ; new coordinate to 0
MOV_EN  BTFSC   MSE_ST,MS_JOY
        BRA     STO_CO                  ; If not in joystick mode,
        CPFSGT  MS_REG                  ; check new coordinate against
        MOVF    MS_REG,W                ; the maximum allowed value
STO_CO  MOVWF   POSTDEC1                ; Store new coordinate
        RETURN
LG_MOV  ADDWF   POSTDEC1                ; Add movement in legacy mode
        RETURN
; Mouse as Kempston joystick horizontal axis bits update
MOUS_A  BTFSC   MOVE_B,MV_MBR           ; The right mouse button
        BRA     MOUS_Z                  ; centers the joystick
        RLNCF   JS_SEN,W                ; Double the sensitivity
        SUBLW   0x80                    ; Subtract from center
        CPFSGT  RDPT_L                  ; If the coordinate is off-center
        BRA     MOUS_L                  ; by more than double the sensi-
        ADDWF   JS_SEN,W                ; tivity threshold the joystick
        CPFSGT  RDPT_L                  ; bits are set to 'Left' or 'Right'
        BRA     MOUS_C                  ; If the deviation is of between
        ADDWF   JS_SEN,W                ; single and double the sensitivity
        ADDWF   JS_SEN,W                ; threshold, the jostick bits are
        CPFSGT  RDPT_L                  ; set depending on their previous
        RETURN                          ; state, so that the move from one
        ADDWF   JS_SEN,W                ; direction to another passes
        CPFSGT  RDPT_L                  ; through the centered state
        BTFSS   RDPT_H,JOY_LT
        BRA     MOUS_R
MOUS_Z  BCF     RDPT_H,JOY_LT           ; Set joystick to: 'Centered'
        BCF     RDPT_H,JOY_RT
        RETURN
MOUS_R  BCF     RDPT_H,JOY_LT           ; Set joystick to: 'Right'
        BSF     RDPT_H,JOY_RT
        RETURN
MOUS_C  BTFSC   RDPT_H,JOY_RT
        BRA     MOUS_Z
MOUS_L  BCF     RDPT_H,JOY_RT           ; Set joystick to: 'Left'
        BSF     RDPT_H,JOY_LT
        RETURN

;------------------------------------------------------------------------------
; SD CARD SUBROUTINES
;
; Read SPI buffer to (WREG)
SPI_RD  SETF    WREG
; Write (WREG) to SPI buffer
SPI_WR  MOVWF   SSP2BUF                 ; Write the byte to the SPI buffer
SPI_LP  BTFSS   SSP2STAT,BF             ; Wait until the transmission ends
        BRA     SPI_LP
        MOVF    SSP2BUF,W               ; Read the received byte
        RETURN
; Send a SD card Command from the table at (TABLPTR)
SDC_CM  BCF	SYS_FL,SR_SEL		; Signal: Not a Serial RAM command
	RCALL   CM_SET                  ; Set up the command string in RAM
        MOVLW   SD_STR                  ; 'STOP TRANSMISSION' command?
        CPFSEQ  SD_CMD			; Do not deselect the SD card
CM_SND  RCALL	DE_SEL			; Deselect the SPI memory devices
        LFSR    FSR1,SD_CMD             ; Command string address
	BTFSS	SYS_FL,SR_SEL		; Serial RAM command?
	BRA	SD_SEL			; No!
	MOVWF	PREINC1			; Set Serial RAM command from (WREG)
	MOVLW	4			; Length of command + address
	MOVWF	CM_CNL			; to counter
	BCF	LATC,C_NRCS		; Select the Serial RAM chip
	BRA	SCM_LP			; Send command and sddress
SD_SEL	BCF     PORTA,A_NFCS		; Select the SD card
; Send the command string
SCM_LP  MOVF    POSTINC1,W              ; Command byte to WREG
        RCALL   SPI_WR                  ; Send command bytes
        DECFSZ  CM_CNL
        BRA     SCM_LP
	BTFSC	SYS_FL,SR_SEL		; Serial RAM command?
	RETURN				; Yes!
        SETF    CM_CNL                  ; Number of trials: 255
        MOVLW   SD_STR                  ;
        CPFSEQ  SD_CMD                  ; 'STOP TRANSMISSION' command?
        BRA     SRW_LP                  ; No!
        RCALL   SPI_RD                  ; Discard first byte
; Wait for the response
SRW_LP  RCALL   SPI_RD
        INCFSZ  WREG                    ; First response byte?
        BRA     SCM_RP                  ; Yes!
        DECFSZ  CM_CNL                  ; No!
        BRA     SRW_LP                  ; Retry
SCM_RP  DECF    WREG
        LFSR    FSR1,SD_CMD             ; Response store address
; Get the response
SRP_LP  MOVWF   POSTINC1                ; Store response bytes
        RCALL   SPI_RD                  ; Get next response byte
        DECFSZ  CM_CNH                  ; Any more response bytes?
        BRA     SRP_LP                  ; Yes!
	TSTFSZ  DT_CNL                  ; Done!
        RETURN
        TSTFSZ  DT_CNH                  ; If the command is a
        RETURN                          ; 'no-data' operation, deselect
        BRA     DE_SEL                  ; the SPI memory devices
;
; Execute SD card commands
SDC_EX	BCF     SDC_ST,SD_EXE           ; Signal: No SDC commands!
	MOVF    CMD_CD,W
SDC_CD  BCF     SDC_ST,SD_MBF
	XORLW	PER_CM
	BZ	TST_PE
	TSTFSZ	CTL_CD			; Station specified?
	GOTO	ETH_EX			; Yes!
TST_PE	RCALL	RW_CPT
	CLRF    POSTINC1                ; Set Error Code to 0
        MOVF	WREG			; Host command?
	BNZ	TST_AT			; No!
	RCALL	HOST_C			; Execute host command
	BRA	SD_END			; and return
TST_AT	XORLW	PER_CM^ASY_TX		; Asynchronous serial port transmit?
	BNZ	TST_ID			; No!
	LFSR    FSR1,DATA_B
	MOVF	BLN_LO,W
	IORWF	BLN_HI,W
	BNZ	ATX_TC
	LFSR	FSR1,S_BUFF
	MOVFF	POSTINC1,BLN_LO
	BRA	ATX_TC
TST_ID	CLRF    PREINC1                 ; Set the response length to 0
        CLRF    PREINC1
	XORLW   ASY_TX^SD_IDF           ; 'Identify Card' command ?
        BNZ     TST_PR                  ; No!
; Identify card
        BTFSS   DEV_ST,DV_SDC           ; Is the SD card present?
        BRA     NOT_PR                  ; No! Signal error
	RCALL	RW_CPT
        MOVF	PREINC1
	MOVLW   .12                     ; Send 8 bytes of SD card
        MOVWF   PREINC1                 ; identifier followed by
	RCALL	RW_BUF                  ; 4 bytes of card capacity
	LFSR	FSR2,SD_IDS
	CALL	STR_CY
        BRA     SD_END
; Transmit to the asynchronous serial port
ATX_HI	DECF	BLN_HI
ATX_LO	DECF	BLN_LO
        MOVLW   5+1                     ; Set time-out to 5 seconds
        MOVWF   CM_CNL                  ; Timing duration (in seconds)
ATX_SC  RCALL   SDC_TM                  ; Restart timer
	DECFSZ  CM_CNL                  ; Count seconds
        BRA     ATX_WT
	MOVLW	E_COMM                  ; Time-out! Signal error
	MOVWF	CMD_CD
        CLRF    BLN_LO                  ; Set the response length to 0
        CLRF    BLN_HI
	BRA	ATX_EN	
ATX_PR	BTFSS	PORTC,C_NCTS		; Printer ready?	
	BRA	ATX_LP			; Yes!
ATX_WT  BTFSS   TIM_ST,TM_SDC           ; Is one second over?
        BRA     ATX_SC                  ; Yes!
	BTFSC	SYS_FL,AS_PRN		; Printer?
	BRA	ATX_PR                  ; Yes!
	BTFSC	PORTA,A_NCTS		; Is CTS asserted?
        BRA     ATX_WT                  ; No!
ATX_LP	BTFSS	PIR1,TX1IF	
	BRA	ATX_LP
	MOVFF	POSTINC1,TXREG1
ATX_TC  TSTFSZ  BLN_LO                  ; Test the byte counter
        BRA     ATX_LO
        TSTFSZ  BLN_HI                  ; Done?
        BRA     ATX_HI                  ; Not yet!
ATX_EN	BSF	PIE1,TX1IE		; Trigger an interrupt
	RETURN
; Check if the SD card or Serial RAM is present
TST_PR  BTFSS	SYS_FL,SR_SEL		; Serial RAM command?
	BRA	CK_SDC			; No!
	BTFSC	SYS_FL,SR_PRE		; Is the Serial RAM present?
	BRA	TST_RD			; Yes!
	BRA	NOT_PR			; No! Signal error
CK_SDC	BTFSC   DEV_ST,DV_SDC           ; Is the SD card present?
        BRA     TST_RD                  ; Yes!
; Error condition: 'Not present'
NOT_PR  MOVLW   E_NPRE                  ; Error: 'Not present'
SD_ERR  RCALL	RW_CPT
	MOVWF   INDF1                   ; Set the error code
SD_END  RCALL	DE_SEL			; Deselect the SPI memory devices
	BCF     SDC_ST,SD_ABF           ; Clear the 'Alternate' flag
	BCF	SYS_FL,SR_SEL		; Signal: Not a Serial RAM command!
	BTFSS	SDC_ST,SD_EBF
	BRA	SD_RSP
	BCF	SDC_ST,SD_EBF
	RETURN
SD_RSP  BTFSC	COM_ST,CO_CYC
        BSF     PIE3,SSP2IE             ; Trigger an interrupt
        RETURN
; Read operation
TST_RD  XORLW   SD_IDF^SD_RDM           ; Read multiple blocks?
        BZ      RD_MBK                  ; Yes!
        XORLW   SD_RDM^SD_RDB           ; 'Read Block' Command?
        BZ      RD_SBK                  ; Yes!
        XORLW   SD_RDB^SD_RDT           ; 'Read-Transmit' Command?
        BNZ     TST_RA                  ; No!
	copy_b	DATA_B+PL_BLN,CNT_LO,2	; Save the data length
        BSF     COM_ST,CO_RDT           ; Set 'Read-Transmit' flag
        BRA     RD_SBK
TST_RA  XORLW   SD_RDT^SD_RDA           ; 'Read Alternate' Command?
        BNZ     TST_PY                  ; No!
        BSF     SDC_ST,SD_ABF           ; Set the 'Alternate' flag
; Read Block
RD_SBK  RCALL   RD_CMD                  ; Send command
        BTFSS   TIM_ST,TM_SDC           ; Time-out?
        BRA     NOT_PR                  ; Yes! Signal error
	RCALL	RW_CPT
        MOVF	POSTINC1,W
        MOVF	POSTINC1,W
	BTFSS   COM_ST,CO_RDT           ; 'Read-Transmit'?
        BRA     SET_RL                  ; No!
        MOVF    CNT_LO,W                ; For 'Read-Transmit' the
        MOVWF   INDF1                   ; data length is supplied
        MOVF    CNT_HI,W                ; by the host
        BRA     SET_LH
SET_RL  MOVLW   high SCS                ; Return 512 bytes of data
        BTFSS   SDC_ST,SD_ABF           ; if not in 'Alternate' mode
SET_LH  MOVWF   PREINC1
        BRA     SD_END
TST_PY	XORLW   SD_RDA^SD_PLY
	BNZ	TST_WR
	BTFSS	PLY_ST,PL_PRE		; Tape player cable present?
	BRA	NOT_PR			; No!
	BRA	RD_TAP			; Start playing
; Write operation
TST_WR  XORLW   SD_PLY^SD_WRB           ; 'Write Block' Command?
        BZ      WSB_CM                  ; Yes!
        XORLW   SD_WRB^SD_WRA           ; 'Write Alternate' command?
        BNZ     TST_EF                  ; No!
        BSF     SDC_ST,SD_ABF           ; Set the 'Alternate' flag
        RCALL   SET_WA                  ; Set the write address
; Write Block
WSB_CM  table_  CMD_24                  ; Command: WRITE_BLOCK
	RCALL   CM_SET                  ; Copy the command string to RAM
	MOVLW	SR_DWR			; Serial RAM 'Write' Code
	RCALL   RW_CMD                  ; Send command
	BTFSC	SYS_FL,SR_SEL		; Serial RAM command?
	BRA	WSB_SR			; Yes!
        TSTFSZ  SD_CMD                  ; Are any error bits set?
        BRA     WR_PRO                  ; Yes! Signal error
        MOVLW   SD_DST                  ; First send a DATA_START_TOKEN
        BRA     WR_DST
WSB_SR	CLRF	DT_CNL			; No CRC bytes!
; Send the data and two CRC bytes
DWC_HI  DECF    DT_CNH                  ; Decrement the byte counter
DWC_LO  DECF    DT_CNL
        MOVF    POSTINC1,W              ; Get the next byte to transmit
WR_DST  MOVWF   SSP2BUF                 ; Send one byte
SPW_WT  BTFSS   SSP2STAT,BF             ; Wait until the transmission ends
        BRA     SPW_WT
        MOVF    SSP2BUF,W               ; Clear buffer full flag
        TSTFSZ  DT_CNL                  ; Test the byte counter
        BRA     DWC_LO
        TSTFSZ  DT_CNH                  ; Done?
        BRA     DWC_HI                  ; Not yet!
        BTFSC	SYS_FL,SR_SEL		; Serial RAM command?
	BRA	SD_END			; Yes!
; Check if the data was accepted
        RCALL   SPI_RD
        ANDLW   0x0f                    ; Did the card respond with
        XORLW   SD_DAC                  ; a DATA_ACCEPTED token?
        MOVLW   1                       ; Yes! Wait for 1 second while
        BZ      W_BUSY                  ; the card is busy
; Error condition: 'Write protected'
WR_PRO  MOVLW   E_WRPR                  ; Error: 'Write protected'
        BRA     SD_ERR
TST_EF  XORLW   SD_WRA^SD_EFA           ; 'Set First Erase Address'?
        BNZ     TST_EL                  ; No!
; Set First Erase Address
        table_  CMD_32                  ; Command: ERASE_WR_BLK_START_ADDR
        BRA     SND_EA                  ; Send command
TST_EL  XORLW   SD_EFA^SD_ELA           ; 'Set Last Erase Address'?
        BNZ     TST_ER                  ; No!
; Set Last Erase Address
        table_  CMD_33                  ; Command: ERASE_WR_BLK_END_ADDR
SND_EA  RCALL   SDC_CM                  ; Send command
        TSTFSZ  SD_CMD                  ; Are any error bits set?
        BRA     WR_PRO                  ; Yes! Signal error
        BRA     SD_END
TST_ER  XORLW   SD_ELA^SD_ERA           ; 'Erase Blocks' Command?
        BNZ     TST_ST                  ; No!
; Erase
        table_  CMD_38                  ; Command: ERASE
        BRA     R1B_CM
TST_ST  XORLW   SD_ERA^SD_STR           ; 'Stop transmission' Command?
        BNZ     SD_END                  ; No other commands are supported
; Stop Transmission
        table_  CMD_12                  ; Command: 'STOP TRANMISSION'
; Execute command with R1b response
R1B_CM  RCALL   SDC_CM                  ; Send command
        TSTFSZ  SD_CMD                  ; Are any error bits set?
        BRA     WR_PRO                  ; Yes! Signal error
        MOVLW   5                       ; Set time-out to 5 seconds
; Wait while the SD card is busy
W_BUSY  MOVWF   CM_CNL                  ; Timing duration (in seconds)
        INCF    CM_CNL
CHK_TM  DCFSNZ  CM_CNL                  ; Count seconds
        BRA     WR_PRO                  ; Time-out! Signal error
        RCALL   SDC_TM                  ; Restart timer
BSY_LP  BTFSS   TIM_ST,TM_SDC           ; Is one second over?
        BRA     CHK_TM                  ; Yes!
        RCALL   SPI_RD                  ; Card still busy?
        BZ      BSY_LP                  ; Yes! Check again!
	RCALL   SPI_RD                  ; Done! Send 8 clock pulses
        BRA     SD_END                  ; Deselect the SD card
; Read multiple blocks ( Command CMD18 )
RD_MBK  BSF     SDC_ST,SD_MBR           ; Set flags
        BSF     SDC_ST,SD_MBF
        BCF     T0CON,TMR0ON            ; Stop timer 0
        BCF     INTCON,TMR0IF           ; Clear interrupt flag
; Read single block ( Command CMD17 )
RD_CMD  BTFSC	PLY_ST,PL_ACT		; Save the tape player's
	MOVFF	FSR1H,WRPT_L		; write pointer
	table_  CMD_17                  ; Single block read
; SD card Read ( Commands: CMD9, CMD10, CMD17 and CMD18 )
SDC_RD  RCALL   CM_SET                  ; Set-up command string
        DECF    DT_CNL                  ; Correction for the data
        DECF    DT_CNL                  ; length in the table
        BTFSC   SDC_ST,SD_MBF           ; Adjust command code for
        INCF    SD_CMD                  ; multiple block read
        BTFSC   PLY_ST,PL_ACT		; Do not start timer for
        BRA	SDR_CM			; the tape player
        BTFSS   SDC_ST,SD_MBF           ; If not a multiple block read,
        RCALL   SDC_TM                  ; Start timer
SDR_CM	MOVLW	SR_DRD			; Serial RAM 'Read' code
	RCALL   RW_CMD                  ; Send command
        BTFSS	PLY_ST,PL_ACT		; Tape player?
	BRA	CK_RAM			; No!
	MOVFF	WRPT_L,FSR1H		; Restore the tape player's
	CLRF	FSR1L			; write pointer
CK_RAM	BTFSC	SYS_FL,SR_SEL		; Serial RAM command?
	BRA	CK_CNT			; Yes!
	BTFSC   SDC_ST,SD_MBF           ; For a multiple block read,
        LFSR    FSR1,RING_B             ; initialize the ring pointer
; Wait for the DATA_START_TOKEN
CK_DST  XORLW   SD_DST                  ; DATA_START_TOKEN?
        BZ	CK_CNT                  ; Yes!
DST_LP  RCALL   SPI_RD
	BTFSC   PLY_ST,PL_ACT		; Do not check time-out for
        BRA     CK_DST			; the tape player
	BTFSS   SDC_ST,SD_MBF           ; If not a multiple block read,
        BTFSC   TIM_ST,TM_SDC           ; Check for time-out
        BRA     CK_DST                  ; No time-out! Loop back
; Deselect the SPI memory devices: SD card and RAM drive
DE_SEL	BSF     PORTA,A_NFCS		; Deselect the SD card
	BTFSC	SYS_FL,SR_PRE		; Deselect the the Serial RAM
	BSF	LATC,C_NRCS		; if present
        RETURN
; Receive the data and two CRC bytes
DRC_HI  DECF    DT_CNH                  ; Decrement the byte counter
DRC_LO  DECF    DT_CNL
	SETF    SSP2BUF                 ; Start reception
SPR_WT  BTFSS   SSP2STAT,BF             ; Wait until the reception ends
        BRA     SPR_WT
        MOVF    SSP2BUF,W               ; Read the received byte
        MOVWF   POSTINC1                ; Save the received byte
CK_CNT  TSTFSZ  DT_CNL                  ; Check the byte counter
        BRA     DRC_LO
        TSTFSZ  DT_CNH                  ; Done?
        BRA     DRC_HI                  ; Not yet!
	BTFSC	SYS_FL,SR_SEL		; Serial RAM command?
	RETURN				; Yes!
        RCALL   SPI_RD			; Read the first CRC byte
        RCALL   SPI_RD                  ; Read the second CRC byte
; One block has been read
        BTFSS   SDC_ST,SD_MBF           ; For a single block read,
	BRA	RD_DSL			; send 8 clocks and return
	BTFSC   FSR1H,RBF_SZ            ; If necessary, wrap the write
        LFSR    FSR1,RING_B             ; ring pointer around
        BTFSC   DEV_ST,DV_DAT           ; Data ready?
        BRA     MBK_WT                  ; Yes!
        BTFSC   FSR1H,2                 ; If not, wait until two
        BTFSS   FSR1H,1                 ; blocks have been filled
        BRA     MBK_WT                  ; before signaling:
        BSF     PIE3,SSP2IE             ; 'Data available'
MBK_WT  MOVLW   SD_STR
        BTFSS   SDC_ST,SD_MBR           ; Stop reading?
        BRA     SDC_CD                  ; Yes! Send 'STOP TRANSMISSION'
        MOVF    FSR0H,W                 ; While aligning the read pointer,
        BNZ     MBK_PT                  ; FSR0H is 0 and the real value
        MOVF    PRODH,W                 ; is stored in PRODH
MBK_PT  XORWF   FSR1H,W                 ; Compare read and write pointers
        ANDLW   ~1                      ; Do not start writing the block
        BZ      MBK_WT                  ; that is being read, but wait
        BSF     DT_CNH,1                ; Set the byte counter to 512
        BRA     DST_LP                  ; Read next block
RD_DSL  RCALL   SPI_RD
	BRA	DE_SEL			; Deselect the SPI memory devices
; Detect the SD card
SDC_ID  CLRF	SDC_ST                  ; Clear the SD card status byte
	BSF	SDC_ST,SD_DET		; Signal: Detecting SD card
        BCF     DEV_ST,DV_SDC           ; Device state: SDC not present
        BTFSS   ETH_ST,ET_PRE           ; If ethernet module not attached,
        BSF     LATC,C_NSDC             ; switch the SDC LED off
	RCALL	SL_OFF
	BTFSC   CHG_ST,B_NCRD           ; Is a card present?
        BRA	SID_EN                  ; No!
; Start the SD card in SPI mode
        MOVLW   .10                     ; Generate 80 clock pulses
        MOVWF   CM_CNL
CLK_LP  RCALL   SPI_RD
        DECFSZ  CM_CNL
        BRA     CLK_LP
        table_  CMD_00                  ; Command: GO_IDLE_STATE
        RCALL   SDC_CM                  ; Send command
        MOVLW   1<<R1_IDL               ; Idle state?
        CPFSEQ  SD_CMD
        BRA	SID_EN                  ; No! Abort card detection
; Check the SD card's version and operating voltage
        table_  CMD_08                  ; Command: SEND_IF_COND
        RCALL   SDC_CM                  ; Send command
        BTFSC   SD_CMD,R1_ILC           ; Illegal command?
        BRA     SD_V1X                  ; Yes! Version 1.x card
        MOVLW   C8_PAT                  ; Check pattern
        CPFSEQ  SD_AD0                  ; Does it match?
        BRA	SID_EN                  ; No! Abort card detection
        MOVLW   C8_VHS                  ; Check Voltage
        CPFSEQ  SD_AD1                  ; Does it match?
        BRA	SID_EN                  ; No! Abort card detection
        BSF     SDC_ST,SD_VER           ; Signal: Version 2 SD card
; Initialize the SD card
SD_V1X  RCALL   SDC_TM                  ; Start timer
C41_LP  table_  CMD_55                  ; Command: APP_CMD
        RCALL   SDC_CM                  ; Send command
        table_  ACM_41                  ; Command: SEND_OP_COND
        RCALL   SDC_CM                  ; Send command
        BTFSC   SD_CMD,R1_ILC           ; Illegal command?
        BRA	SID_EN                  ; Yes! Not a SD card
        BTFSS   TIM_ST,TM_SDC           ; Time-out?
        BRA	SID_EN                  ; Yes! Abort card detection
        TSTFSZ  SD_CMD                  ; Card still in 'Idle State'?
        BRA     C41_LP                  ; Yes! Repeat command
; Differentiate between standard and high capacity SD cards
        RCALL   SDC_TM                  ; Restart timer
C58_LP  table_  CMD_58                  ; Command: READ_OCR
        RCALL   SDC_CM                  ; Send command
        BTFSS   TIM_ST,TM_SDC           ; Time-out?
        BRA	SID_EN                  ; Yes! Abort card detection
        BTFSS   SD_AD3,OC_PWU           ; Power-up finished?
        BRA     C58_LP                  ; Not yet! Try again
        BTFSC   SD_AD3,OC_CCS           ; High Capacity card?
        BSF     SDC_ST,SD_HCS           ; Yes!
; Disable the CRC check
        table_  CMD_59                  ; Command: CRC_ON_OFF
        RCALL   SDC_CM                  ; Send command
; Set block length to 512 bytes for standard capacity SD cards
        table_  CMD_16                  ; Command: SET_BLOCKLEN
        BTFSS   SDC_ST,SD_HCS           ; Execute only for standard
        RCALL   SDC_CM                  ; capacity cards
; Get the SD card's capacity
        table_  CMD_09                  ; Command: SEND_CSD
        RCALL   SDC_RD                  ; Send command
	BTFSS   TIM_ST,TM_SDC           ; Time-out?
        BRA	SID_EN                  ; Yes! Abort card detection
        RCALL	RW_DPT
	BTFSS   INDF1,CSD_HC		; Version 1 CSD structure?
        BRA	SID_EN                  ; Yes! Abort card detection
; Get C_SIZE * 256
	MOVLW	9
	ADDWF	FSR1L
	CLRF    SD_SZ0
        MOVFF   POSTDEC1,SD_SZ1      	; Bits <55:48>
        MOVFF   POSTDEC1,SD_SZ2		; Bits <63:56>
        MOVFF   POSTDEC1,SD_SZ3      	; Bits <71:64> (<71:70> are 00)
; Calculate (C_SIZE+1) * 256
	MOVLW   0
        INCF    SD_SZ1
        ADDWFC  SD_SZ2
        ADDWFC  SD_SZ3                  ; This clears the Carry Flag
; Calculate the card capacity = (C_SIZE+1) * 1024 sectors
        RLCF    SD_SZ1
        RLCF    SD_SZ2
        RLCF    SD_SZ3                  ; This clears the Carry Flag
        RLCF    SD_SZ1
        RLCF    SD_SZ2
        RLCF    SD_SZ3			; Carry = 0, Borrow = 1
; Calculate the number of the last usable sector
        SUBWFB  SD_SZ0                  ; Subtract 1
        SUBWFB  SD_SZ1
        SUBWFB  SD_SZ2
        SUBWFB  SD_SZ3
; Get the SD card's ID
        table_  CMD_10                  ; Command: SEND_CID
        RCALL   SDC_RD                  ; Send command
	BTFSS   TIM_ST,TM_SDC           ; Time-out?
        BRA	SID_EN                  ; Yes! Abort card detection
        MOVF    SDC_ST,W                ; Set first response byte to:
        ANDLW   b'00000011'             ; '1' - Ver 1.x standard SD card
        ADDLW   '1'                     ; '2' - Ver 2.0 standard SD card
 	RCALL	RW_DPT
        MOVWF   INDF1                   ; '4' - High capacity SD card
	MOVLW	8
	ADDWF	FSR1L
	MOVFF	SD_SZ0,POSTINC1
	MOVFF	SD_SZ1,POSTINC1
	MOVFF	SD_SZ2,POSTINC1
	MOVFF	SD_SZ3,POSTINC1
        BSF     DEV_ST,DV_SDC           ; Device state: SDC identified
        BTFSS   ETH_ST,ET_PRE           ; If ethernet module not attached,
        BCF     LATC,C_NSDC             ; switch the SDC LED on
        RCALL	SVL_ON
SID_EN	BCF	SDC_ST,SD_DET
        RETURN                          ; Yes! Abort card detection
; Send command and set buffer addres for Read and Write
RW_CMD 	RCALL   CM_SND                  ; Send command
RW_BUF	LFSR    FSR1,DATA_B
	BTFSC   SDC_ST,SD_ABF
	LFSR    FSR1,ALT_DB
        BTFSC	SDC_ST,SD_EBF
	LFSR	FSR1,PR_DAT
        BTFSC	SDC_ST,SD_DET
RW_DPT	LFSR	FSR1,SD_IDS
        RETURN
RW_CPT	LFSR    FSR1,CMD_CD
        BTFSC	SDC_ST,SD_EBF
	LFSR	FSR1,PR_CMD
        RETURN
; Copy the standard tape signal parameters
TAP_PA	RCALL	RD_PTR
	table_	STD_TP
	RCALL	TAB_JP
	BRA	RD_PTR
;------------------------------------------------------------------------------
; EXECUTE A HOST COMMAND

HOST_C  CLRF    BLN_HI
        TSTFSZ	BLN_LO			; Return if there is no command
	BRA	PARS_C			; string        
	RETURN
PARS_C	LFSR    FSR0,DATA_B             ; Point FSR0 to the command string
        MOVF    POSTINC0,W              ; Check the command
        DECFSZ	BLN_LO
	BRA	SET_BR			; Multicharacter command
	XORLW	'i'
        BNZ     FW_DIS
GET_ID  table_	BRD_ID
	BRA     TAB_JP
FW_DIS  XORLW   'i'^'f'
        BNZ     BNK_RD
        table_  FWR_ID
TAB_JP  GOTO    TAB_NW
BNK_RD	XORLW   'f'^'r'
        BNZ     POLL_S
	MOVFF	BANK_R,DATA_B		; 128k Memory bank control
	MOVFF	BANK_P,DATA_B+1		; +3 Memory bank control
	CLRF	DATA_B+2
	MOVFF	AY_ADR,DATA_B+3		; 128k Sound chip address
	MOVLW	4
	MOVWF	BLN_LO
REG_CL	SETF	BANK_R
	SETF	AY_ADR
	SETF	BANK_P
	RETURN
POLL_S	XORLW	'r'^'k'
        BNZ     PIC_FW
	BTFSS	MSE_ST,MS_PRE		; PS/2 or serial
	BTFSC	SYS_FL,AS_ENB		; asynchronous device present?
	RETURN				; Return if yes!
	BTFSS	PLY_ST,PL_PRE		; Tape player cable present?
	BTFSS   DEV_ST,DV_SRV		; Server not connected?
	RETURN				; Return if yes!
	BSF	USB_ST,SB_POL		; Enable server polling
	MOVLW	POL_KY			; Send a 'Toggle polling'
	MOVWF	CMD_CD			; command to the server
	CLRF	CTL_CD
	RETURN
PIC_FW	XORLW	'k'^'@'
        BNZ	ESP_FW
	BTFSS   DEV_ST,DV_SRV		; Server connected?
	RETURN
	GOTO	FW_UPD			; Yes! Update firmware
ESP_FW	XORLW	'@'^'&'
	BNZ	RST_CM
	BTFSC	ETH_ST,ET_WFI		; Wifi module present?
	BTFSS   DEV_ST,DV_SRV		; Server connected?
	RETURN
	CALL	WF_UPD			; Yes! Update firmware
RST_CM	XORLW	'&'^'!'
	BNZ	NO_ECH
	RCALL	SD_END			; Send response and wait
WT_RST	BTFSC	COM_ST,CO_CYC		; until the cycle ends
	BRA	WT_RST
	RESET
SET_BR	XORLW	'b'
        BNZ     SET_PI
	BTFSS	SYS_FL,AS_ENB
	BRA	NO_ECH
        table_  BAUD_D
	BCF	RCSTA1,SPEN		; Disable USART
SBR_LP	MOVF	POSTINC0,W
	TBLRD*+
	SUBWF	TABLAT,W
	MOVF	POSTDEC0,W
	TBLRD*+
	SUBWFB	TABLAT,W
	TBLRD*+
	MOVFF	TABLAT,SPBRG1
	TBLRD*+
	MOVFF	TABLAT,SPBRGH1
	BNC	SBR_LP
	BSF	RCSTA1,SPEN		; Enable USART
NO_ECH	CLRF	BLN_LO			; No response data	
	RETURN
SET_PI	XORLW	'b'^'p'
        BNZ     MUL_CM
	LFSR	FSR1,S_BUFF
	MOVFF	BLN_LO,POSTINC1
PRI_LP	MOVFF	POSTINC0,POSTINC1
	DECFSZ	BLN_LO		
	BRA	PRI_LP
	RETURN
MUL_CM	XORLW	'p'^'e'			; Ethernet command?	
	BTFSC	STATUS,Z
	GOTO	ET_INF			; Yes!
	XORLW	'e'^'m'			; Mouse/joystick command?
	BNZ	NO_ECH			; No!	
; Mouse driver configuration
MCM_LP	MOVF    POSTINC0,W              ; Check the command literal
        XORLW   'e'                     ; Enable mouse?
        BNZ     MC_JOY                  ; No!
	BTFSS	USB_ST,SB_POL		; Server polling or mouse
        BTFSC   DEV_ST,DV_MSE           ; data reporting enabled?
        RCALL   MOUS_K                  ; If yes, enable mouse
        BRA     TST_SL
MC_JOY  XORLW   'e'^'j'                 ; Mouse as joystick?
        BNZ     MC_DIS                  ; No!
        BTFSC   DEV_ST,DV_MSE           ; Data reporting enabled?
        RCALL   MOUS_J                  ; If yes, enable mouse as joystick
        BRA     TST_SL
MC_DIS  XORLW   'j'^'d'                 ; Disable mouse?
        BNZ     MC_LEG                  ; No!
        RCALL   MOUS_S                  ; Disable mouse
        BRA     TST_SL
MC_LEG  XORLW   'd'^'l'                 ; 'Legacy mode'?
        BNZ     MC_WIN                  ; No!
        BSF     MSE_ST,MS_LEG           ; Switch to legacy mode
        BRA     TST_SL
MC_WIN  XORLW   'l'^'w'                 ; 'Windowed mode'?
        BNZ     MC_RTE                  ; No!
        BCF     MSE_ST,MS_LEG           ; Switch to windowed mode
        BRA     TST_SL
MC_RTE  XORLW   's'^'r'                 ; 'Joystick read rate [Hz]'?
        BNZ     MC_SEN                  ; No!
        MOVLW   (JM_INI-JS_TMC)/2+1     ; Check if the command parameter
        CPFSLT  INDF0                   ; is:
        BRA     MC_END                  ; - too large
        table_	JS_TMC-2
        RLNCF   INDF0,W                 ;
        BZ      MC_END                  ; - zero
        CALL	ADD_TP
        TBLRD*+                         ; Copy the timing constant
        MOVFF   TABLAT,JSTC_L           ; from the table
        TBLRD*+
        MOVFF   TABLAT,JSTC_H
        BRA     MC_END
MC_SEN  XORLW   'w'^'s'                 ; 'Joystick sensitivity'?
        BNZ     MC_WSX                  ; No!
        MOVLW   .2                      ; Check lower limit
        CPFSGT  INDF0
        BRA     MC_END
        MOVLW   .32                     ; Check upper limit
        CPFSGT  INDF0
        MOVFF   INDF0,JS_SEN            ; OK!
        BRA     MC_END
MC_WSX  XORLW   'r'^'x'                 ; 'Mouse window width'?
        BNZ     MC_WSY                  ; No!
        MOVLW   .16                     ; Check lower limit
        CPFSLT  INDF0
        MOVFF   INDF0,MWIN_X            ; OK!
	CLRF	MPOS_X
        BRA     MC_END
MC_WSY  XORLW   'x'^'y'                 ; 'Mouse window height'?
        BNZ     TST_SL                  ; No! loop back
        MOVLW   .192                    ; Check upper limit
        CPFSLT  INDF0
        BRA     MC_POS
        MOVLW   .16                     ; Check lower limit
        CPFSLT  INDF0
        MOVFF   INDF0,MWIN_Y            ; OK!
MC_POS  CLRF	MPOS_Y
MC_END  MOVF    POSTINC0,W              ; Advance to the next command
	DECFSZ  BLN_LO                  ; Test the remaining length of
TST_SL  DCFSNZ  BLN_LO                  ; the command string
        RETURN                          ; Finished!
	BRA	MCM_LP
;------------------------------------------------------------------------------
; KEYBOARD SUBROUTINES

KEY_DC	table_	SBS_TB
	MOVF	JS_SEN,W
	BTFSC	KYB_ST,KB_EXT
	BSF	WREG,7			; Set bit 7 of extended scan codes
	RCALL	CODE_S			; Substitute keyboard scan codes
	MOVWF	RDPT_H
	XORLW	C_NL
	BNZ	SET_CS
	CLRF	KEM_JS
	BTFSC	KYB_ST,KB_BRK
	BTG	MSE_ST,MS_KYB		; Toggle the 'Keyboard as joystick'
	BRA	CLR_FL			; flag
SET_CS	XORLW	C_NL^C_SH		; 'Shift' key pressed / released?
	BNZ	SET_SS			; No!
	BSF	KYB_ST,KB_SHF		; Set 'CAPS SHIFT' flag
	BTFSC	KYB_ST,KB_BRK		; If break scan code
	BCF	KYB_ST,KB_SHF		; clear 'CAPS SHIFT' flag
	BRA	CLR_FL
SET_SS	XORLW	C_SH^C_CT		; 'Ctrl' ket pressed / released?
	BNZ	CK_ALT			; No!
	BSF	KYB_ST,KB_CTL		; Set 'SYMBOL SHIFT' flag
	BTFSC	KYB_ST,KB_BRK		; If break scan code
	BCF	KYB_ST,KB_CTL		; clear 'SYMBOL SHIFT' flag
	BTFSS	MSE_ST,MS_KYB		; 'Keyboard as joystick'?
	BRA	CLR_FL			; No!
	BSF	KEM_JS,JOY_FR		; Set the joystick 'Fire' bit
	BTFSC	KYB_ST,KB_BRK		; If break scan code,
	BCF	KEM_JS,JOY_FR		; clear the joystick 'Fire' bit
	BRA	CLR_FL
CK_ALT	XORLW	C_CT^C_AL		; 'Alt' key pressed / released?
	BNZ	CHK_RT			; No!
	BTFSC	KYB_ST,KB_BRK		; Break scan code?
	BRA	ALT_OF			; Yes!
	BTFSC	KYB_ST,KB_ALT
	BRA	CLR_FL			; Keyword flag already set
	BSF	KYB_ST,KB_ALT		; Set keyword flag
	LFSR	FSR1,S_BUFF		; Initialize the string		
	MOVFF	FSR1L,JSTC_L		; pointer
	CLRF	INDF1			; Set the string end marker
	BRA	CLR_FL
ALT_OF	CLRF	KYB_ST			; Clear all flags
	table_	FLT_TB			; Address of table
	MOVLW	A_TOKN-FLT_TB+1		; Length of table
	MOVWF	RDPT_H
	MOVFF	S_BUFF,WREG		; First character of string
TBI_LP	DCFSNZ	RDPT_H			; End of list?
	BRA	CLR_KY			; Yes!
	TBLRD*+
	CPFSEQ	TABLAT			; Characters match?
	BRA	TBI_LP			; No!
	MOVFF	TBLPTRL,RDPT_H
	table_	IND_TK			; Set address of index table
	MOVLW	low (FLT_TB+1)
	SUBWF	RDPT_H,W		; Calculate the token table index
	RLNCF   WREG
	CALL	ADD_TP
	TBLRD*+
        MOVF	TABLAT,W
        TBLRD*+
	MOVWF   TBLPTRL
        MOVFF   TABLAT,TBLPTRH
NEW_TK	LFSR	FSR1,S_BUFF+1		; Point to the search string
TOK_CP	MOVF	INDF1,W		        ; End of string reached?
	BNZ	CK_TOK			; No!
NXT_TK	TBLRD*+
END_TK	BTFSS	TABLAT,7		; Skip to end of current token
	BRA	NXT_TK
	TBLRD+*
	MOVF	TABLAT,W
	BNZ	NEW_TK			; No! Check the next token
	BRA	CLR_KY
; Find a string in the token table
CK_TOK	TBLRD*+
	MOVF	TABLAT,W
	BCF	WREG,7			; Clear the end marker bit
	CPFSEQ	POSTINC1		; Characters match?
	BRA	END_TK			; No! Try next token
	BTFSS	TABLAT,7		; Yes! Last token character?
	BRA	TOK_CP			; No! Compare next one
	TBLRD*
	MOVF	TABLAT,W		; Found key number
	MOVWF	JSTC_H			; Store key number
	BTFSS	WREG,KB_SHF		; If the 'Shift' bit is not set
	BRA	TOK_MD			; output the code directly, else
; Output a token or a key in extended mode
EXT_MD	BCF	WREG,KB_SHF		; change first to 'Extended Mode'
	MOVWF	JSTC_H			; Store key number
	MOVLW	N_SS|1<<KB_SHF		; 'Extended mode' key number
TOK_MD	RCALL	OUT_KY
	BSF	TIM_ST,TM_KYB		; Signal: timer running
	BRA	KYB_TM			; Start keyboard timer
CHK_RT	BTFSS	MSE_ST,MS_KYB		; 'Keyboard as joystick'?
	BRA	TS_BRK			; No!
; Produce joystick output
	XORLW	C_AL^C_RT
	BNZ	CHK_LT
	BTFSC	KYB_ST,KB_BRK
	BRA	CLR_RT	
	BSF	KEM_JS,JOY_RT
	BRA	CLR_LT
CLR_RT	BCF	KEM_JS,JOY_RT
	BRA	CLR_SS
CHK_LT	XORLW	C_RT^C_LT
	BNZ	CHK_DN
	BTFSC	KYB_ST,KB_BRK
	BRA	CLR_LT
	BSF	KEM_JS,JOY_LT
	BRA	CLR_RT
CLR_LT	BCF	KEM_JS,JOY_LT
	BRA	CLR_SS
CHK_DN	XORLW	C_LT^C_DN
	BNZ	CHK_UP
	BTFSC	KYB_ST,KB_BRK
	BRA	CLR_DN
	BSF	KEM_JS,JOY_DN
	BRA	CLR_UP
CLR_DN	BCF	KEM_JS,JOY_DN
	BRA	CLR_SS
CHK_UP	XORLW	C_DN^C_UP
	BNZ	TS_BRK
	BTFSC	KYB_ST,KB_BRK
	BRA	CLR_UP
	BSF	KEM_JS,JOY_UP
	BRA	CLR_DN
CLR_UP	BCF	KEM_JS,JOY_UP
CLR_SS	BCF	KYB_ST,KB_CTL		; Clear 'SYMBOL SHIFT' flag
	BRA	CLR_KY			; Discard the arrow key code
; Decode a scan code
TS_BRK	BTFSS	KYB_ST,KB_BRK		; Break scan code?
	BRA	NEW_KY			; No!
	MOVF	RDPT_H,W
	CPFSEQ	RDPT_L
	BRA	CLR_FL
CLR_KY	CLRF	RDPT_H			; Clear last scan code
NEW_KY	MOVFF	RDPT_H,RDPT_L		; Store new scan code
CLR_FL	MOVLW	1<<KB_SHF|1<<KB_CTL|1<<KB_ALT
	ANDWF	KYB_ST			; Clear scan code flags
	table_	PRI_TB			; Primary supplementary key table
	MOVF	RDPT_L,W
	BTFSC	KYB_ST,KB_ALT		; Keyword flag set?
	BRA	ALT_KY			; Yes! Add scan code to buffer
	BZ	DCD_AJ			; No new scancode!
	BTFSC	KYB_ST,KB_CTL		; 'Ctrl' key pressed?
	BRA	DCD_KY			; Yes!
	BTFSS	KYB_ST,KB_SHF		; 'Shift' key pressed?
	BRA	NO_SHF			; No!
	table_	SHF_TB			; Shifted supplementary key table
NO_SHF	MOVF	RDPT_L,W	
	RCALL	CODE_S
	CPFSEQ	RDPT_L
	BRA	SPL_KY
DCD_KY	table_	SCN_TB			; Set address of decoding table
	MOVLW	PRI_TB-SCN_TB+1		; Set length of decoding table
	MOVWF	RDPT_H
	MOVF	RDPT_L,W
DCD_LP	DCFSNZ	RDPT_H			; End of table?
	BRA	DCD_AJ			; Yes!
	TBLRD*+
	CPFSEQ	TABLAT			; Scan code found in table?
	BRA	DCD_LP			; No! Continue
DCD_AJ	MOVLW	low (SCN_TB+1)		; Scan code found!
	SUBWF	TBLPTRL,W	 	; Calculate the 'key number' and
	IORWF	KYB_ST,W		; add the shift flags
K_DONE	BCF	USB_ST,SB_REQ		; Signal: Polling done!
OUT_KY	MOVWF	EXT_KB			; Output the resulting code
	XORLW	N_CS			; A value of 0x27 indicates that
	BNZ	MSL_ON			; no key is being pressed
	BCF     DEV_ST,DV_BSY           ; Signal: Not Busy!
ML_OFF  BTFSS   ETH_ST,ET_PRE           ; If Ethernet module not attached,
        BSF     LATC,C_NMSE             ; turn mouse LED off
SL_OFF	BTFSC	SYS_FL,SR_PRE
        RETURN
	BTFSC	ETH_ST,ET_PRE
	BSF	LATC,C_NSRV
        RETURN
; Output CAPS SHIFT ( 'No key' )
CS_KEY	MOVLW	N_CS			; Signal: No key pressed
; Supplementary key
SPL_KY	BTFSC	WREG,KB_SHF
	BTFSS	WREG,KB_CTL
	BRA	K_DONE
	BRA	EXT_MD
; Add a scan code to the string input buffer
ALT_KY	BZ	A_DONE			; Not a new scan code
	LFSR	FSR1,S_BUFF		; Point to the current position
	MOVFF	JSTC_L,FSR1L		; in the input string buffer
	BTFSS	FSR1L,3			; If the buffer is not full
	MOVWF	POSTINC1		; copy the scan code to the buffer
	MOVFF	FSR1L,JSTC_L		; and save the new pointer
	CLRF	INDF1			; Set the buffer end marker
A_DONE	BCF	USB_ST,SB_REQ		; Signal: Polling done!
MSL_ON	BTFSS   ETH_ST,ET_PRE           ; If ethernet module not attached,
        BCF     LATC,C_NMSE             ; turn mouse LED on
SVL_ON	BTFSC	SYS_FL,SR_PRE
        RETURN
	BTFSC	ETH_ST,ET_PRE
	BCF	LATC,C_NSRV
	RETURN
; Search code in a '0'-terminated substitution table
CODE_S	TBLRD*+
	CPFSEQ	TABLAT			; Scan code found?
	BRA	CK_ETB			; No! Continue
	TBLRD*
	MOVF	TABLAT,W		; Return substitute code in (WREG)
	RETURN
CK_ETB	TBLRD*+
	TSTFSZ	TABLAT			; Table end marker?
	BRA	CODE_S			; No! Continue
	RETURN
	
;------------------------------------------------------------------------------
; ETHERNET CONTROLLER PRIMARY SPI FUNCTIONS
;
; Read from Ethernet controller register (WREG)
ETH_RD  BCF     LATC,C_NECS		; Select Ethernet module
RD_REG  RCALL   SPI_WR
        RCALL   SPI_RD
	BRA	ETH_DS
; Write address (WREG) to Ethernet controller
WR_ADR  BCF     LATC,C_NECS		; Select Ethernet module
	BRA     SPI_WR
; Write data from (TABLPTR) to Ethernet controller register (WREG)
WR_REG	RCALL	WR_ADR
	TBLRD*+	
	MOVF	TABLAT,W
; Write data to Ethernet controller from (WREG)
WR_DAT	RCALL   SPI_WR	
; Deselect the Ethernet module
ETH_DS	BSF     LATC,C_NECS		; Deselect the Ethernet module
	RETURN	

;------------------------------------------------------------------------------
; TAPE PLAYER SUBROUTINES
;
; Read a tape file into the ring buffer
RD_TAP	BSF	PLY_ST,PL_ACT		; Signal: Reading a tape file
	BTFSC	COM_ST,CO_CYC		; If triggert by a PSP timeout, like
	BRA	GET_CS			; when doing an OUT 159,243,
        table_  SYS_LD                  ; fetch the operating system loader
        RCALL  	TAB_JP			; parameters
GET_CS  MOVFF	DATA_B+TAP_SL,MOVE_X	; Save the first sector of the
	MOVFF	DATA_B+TAP_SH,MOVE_Y	; tape file
	LFSR	FSR1,RING_B		; Read sector 1 of the logical
	MOVLW	1			; drive to obtain the cluster size
	RCALL	RD_FSC			; Send a response to the host
	RCALL	SD_RSP			; and wait until the communication
WAI_CY	BTFSC	COM_ST,CO_CYC		; cycle ends
	BRA	WAI_CY
	BCF     T0CON,TMR0ON            ; Stop the joystick timer
        BCF     INTCON,TMR0IF           ; Clear the interrupt flag
        LFSR	FSR0,RING_B		; Initiallize the read pointer
	LFSR	FSR1,RING_B		; Initiallize the write pointer	
	MOVFF	INDF1,MOVE_B		; Cluster size - 1
	MOVFF	MOVE_X,DATA_B+TAP_SL    ; Restore the first sector of the
	MOVFF	MOVE_Y,DATA_B+TAP_SH	; tape file
	CLRF	MOVE_X			; Initalize the FAT sector number
RDS_LP	RCALL	RD_CMD			; Read a tape file sector
	BTFSC   FSR1H,RBF_SZ            ; If necessary, wrap the write
        LFSR    FSR1,RING_B             ; ring pointer around        
        BTFSS   T0CON,TMR0ON
        BTFSC   T3CON,TMR3ON            ; Tape playing?
        BRA     TAP_WT                  ; Yes!
        BTFSC   FSR1H,2                 ; If not, wait until two sectors
        BTFSS   FSR1H,1                 ; have been written before starting
        BRA     TAP_WT            	; the tape player timer
	RCALL	TAP_PA			; Copy standard timing tape parameters
	MOVLW	low .1000		; Set the standard pause between
	MOVWF	DATA_B+PAUS_L		; blocks and the initial delay to 1s
	MOVWF	MPOS_X
	MOVLW	high .1000
	MOVWF	DATA_B+PAUS_H
	MOVWF	MPOS_Y
	BCF	PLY_ST,PL_PLY		; Signal: Pause!
	BSF     T3CON,TMR3ON		; Start the tape player timer
TAP_WT  BTFSC   T0CON,TMR0ON            ; Tape player stopped?
        BRA     TAP_EN                  ; Yes!
        MOVF    FSR0H,W
	XORWF   FSR1H,W                 ; Compare read and write pointers
        ANDLW   ~1                      ; Do not start writing the block
        BZ      TAP_WT                  ; that is being read
	INCF	DATA_B+TAP_SL		; Next sector number
	MOVF	DATA_B+TAP_SL,W
	ANDWF	MOVE_B,W		; Whithin same cluster?
	BNZ	RDS_LP			; Yes!	
	DECF	DATA_B+TAP_SL		; Restore sector number
	MOVFF	FSR1H,WRPT_H		; Save the buffer pointer
	COMF	MOVE_B,W
	BRA	DIV_ST			; Divide the sector number by the
DIV_LP	RRCF	DATA_B+TAP_SH		; cluster size to obtain the 
	RRCF	DATA_B+TAP_SL		; cluster number
DIV_ST	RRCF	WREG
	BNC	DIV_LP
	MOVFF	DATA_B+TAP_SL,MOVE_Y	; Save the cluster number (low)
	INCF	DATA_B+TAP_SH,W		; FAT sector number
	CPFSEQ	MOVE_X			; Already loded?
	RCALL	RD_FAT			; No! Load FAT sector
	LFSR	FSR1,FAT_BF
	MOVF	MOVE_Y,W		; The position of the FAT entry
	MOVWF	FSR1L			; is given by double the cluster
	ADDWF	FSR1L			; number low byte
	MOVLW	0
	ADDWFC	FSR1H
	MOVFF	POSTINC1,DATA_B+TAP_SL	; Copy the new sector number
	MOVFF	INDF1,DATA_B+TAP_SH
	CLRF	FSR1L
	MOVFF	WRPT_H,FSR1H		; Restore the buffer pointer
	DECF	DATA_B+TAP_SL,W		; Check the sector number
	IORWF	DATA_B+TAP_SH,W		; No more clusters?
	BNZ	RDS_LP			; Not yet!
TAP_EN	BCF	PLY_ST,PL_ACT		; Signal: Not reading a tape file
	BCF	SDC_ST,SD_EXE
; Clear the message buffers
CLR_BF	MOVLW	high(S_BUFF)-high(FAT_BF)
	LFSR	FSR1,FAT_BF
CBF_LP	CLRF	INDF1
	INCF	FSR1H
	DECFSZ	WREG
	BRA	CBF_LP
	RETURN
; Read a FAT sector
RD_FAT	LFSR	FSR1,FAT_BF		; Read sector to FAT buffer
	MOVWF	MOVE_X			; Store new FAT sector
RD_FSC	MOVWF	DATA_B+TAP_SL		; Set sector number
	CLRF	DATA_B+TAP_SH
	BRA	RD_CMD
;
; Tape player interrupt service routine
PLY_IT	BCF	PIR2,TMR3IF		; Clear the interrupt flag
	BTFSC   T0CON,TMR0ON            ; Tape player stopped?
	BRA	PY_STP			; Yes! Return
	BCF	LATC,CK1
	BTFSC	PLY_ST,PL_PLY		; Playing block?
	BRA	PY_DLY			; Yes!
; Pause 
	BSF	TRISC,CK1
	BRA	CK_STP
CK_STP	BTFSS	PORTC,DT1		; Stop the tape?
	BRA	T_PAUS			; Yes!
PY_PAU	TSTFSZ	MPOS_X
	BRA	PY_PLO
	TSTFSZ	MPOS_Y
	BRA	PY_PHI
; Parse TZX blocks (23)
TZX_LP	MOVF	DATA_B+TAP_LL,W		; Check if the end of the file
	IORWF	DATA_B+TAP_LH,W		; has been reached
	IORWF	DATA_B+TAP_LU,W
	BZ	PY_END			; End of file reached!
	CLRF	MWIN_X			; Initialize the overhead byte
	CLRF	MWIN_Y			; counter
	CLRF	MS_REG			; Block length upper byte
	MOVLW	TY_TAP			; TAP file?
	CPFSEQ	DATA_B+FILE_T
	RCALL	GET_NX			; For a .TZX file, get the (11)
	MOVWF	DATA_B+FILE_T		; block type
	BTFSS	DATA_B+FILE_T,TZX_HD
	BTFSC	DATA_B+FILE_T,NON_DT	; For block types: 0x10 - 0x14,
	BRA	NO_PLS
	BTFSS	PLY_ST,PL_PLY
	BTG	TRISC,CK1		; toggle the tape player output
NO_PLS	BSF	PLY_ST,PL_STD		; Signal: Standard timing!
	XORLW	TY_TAP			; TAP file?
	BZ	TAP_LN			; Yes!
	XORLW	TY_TAP^BK_TAP		; Standard speed data?
	BZ	STD_SP			; Yes!
	BCF	PLY_ST,PL_STD		; Signal: Non-standard timing!
	CLRF	JSTC_H
	XORLW	BK_TAP^BK_DAT		; 'Pure data' block?
	BNZ	CK_PLS			; No!
; Get 'Pure data' block parameters
	RCALL	GET_BP			; Bit pulses length
	BRA	BITS_N
CK_PLS	XORLW	BK_DAT^BK_PLS		; 'Pulse sequence' block?
	BNZ	CK_TON			; No!
; Get 'Pulse sequence' block parameters
	RCALL	GET_NX
	MOVWF	JS_SEN			; Number of pulses
	MOVWF	JSTC_L
	RLCF	JSTC_L
	RLCF	JSTC_H			; Data length
	BRA	CK_LEN
CK_TON	XORLW	BK_PLS^BK_TON		; 'Pure tone' block?
	BZ	GET_TO			; Yes!
	XORLW	BK_TON^BK_TRB		; 'Turbo speed' block?
	BNZ	CK_PAU			; No!
; Get 'Turbo speed'  and 'Pure tone' block parameters
GET_TO	RCALL	GET_TC
	copy_b	DATA_B+PAUS_L,S_BUFF,2	; Leader tone pulse length
	BTFSC	DATA_B+FILE_T,0
	RCALL	GET_SY			; Sync and bit pulses length
	RCALL	GET_LN
	copy_b	JSTC_L,TONE_0,2		; Number of leader pulses
	BTFSS	DATA_B+FILE_T,0		; 'Pure tone' block?
	BRA	O_HEAD			; Yes!
; Get bits number of the last byte
BITS_N	RCALL	GET_NX
	RLCF	WREG
	MOVWF	DATA_B+LAST_B		; Bits in last byte
	BRA	PAU_LN
CK_PAU	XORLW	BK_TRB^BK_PAU		; 'Pause' block?
	BZ	PAU_LN			; Yes!
; Skip non-data blocks
	XORLW	BK_PAU^BK_END		; 'Group end'?
	BZ	RES_TM			; Yes!
	CLRF	JSTC_L
	BSF	JSTC_L,2		; Set bytes to skip = 4
	XORLW	BK_END^BK_STP		; 'Stop the tape' block?
	BZ	SKIP_B			; Yes!
	RLNCF	JSTC_L			; Set bytes to skip = 8
	XORLW	BK_STP^'Z'		; TZX header block?
	BZ	SKIP_N			; Yes!
	XORLW	'Z'^BK_MSG		; 'Message' block?
	BNZ	CK_TXT			; No!
	RCALL	GET_NX			; Discard message display time
	CLRF	WREG
CK_TXT	MOVFF	INDF0,JSTC_L
	BZ	SKIP_N			; Yes!
	XORLW	BK_MSG^BK_TXT		; 'Text description' block?
	BZ	SKIP_N			; Yes!
	XORLW	BK_TXT^BK_STA		; 'Group start' block?
	BZ	SKIP_N			; Yes!
	XORLW	BK_STA^BK_HWR		; 'Hardware info' block?
	BNZ	CK_ARC			; No!
	ADDWF	JSTC_L,W		; This clears Carry
	RLCF	JSTC_L			; Multiply by 3 the number of
	RLCF	JSTC_H			; configurations to get the
	ADDWF	JSTC_L			; overhead
	MOVLW	0
	ADDWFC	JSTC_H
	BRA	SKIP_N
CK_ARC	RCALL	GET_LN
	MOVF	DATA_B+FILE_T,W
	XORLW	BK_ARC			; 'Archive info' block?
	BZ	SKIP_B			; Yes!
	XORLW	BK_ARC^BK_LEV		; 'Set Level' block?
	BNZ	PY_ERR			; No!
; Set the tape player output level
	RCALL	GET_LN
	RCALL	GET_NX
	TSTFSZ	WREG
	BRA	LEV_HI
	BSF	TRISC,CK1		; Set level '0'
	BRA	RES_TM
LEV_HI	BCF	TRISC,CK1		; Set level '1'
	BRA	RES_TM
; Adjust the remaining file length
ADJ_LN	MOVF	MWIN_X,W		; Overhead length (low)
	SUBWF	DATA_B+TAP_LL		; Subtract from tape length (low)
	MOVF	MWIN_Y,W		; Overhead length (high)
	SUBWFB	DATA_B+TAP_LH		; Subtract from tape length(high)
	MOVLW	0			; Overhead length (upper)
ADJ_UP	SUBWFB	DATA_B+TAP_LU		; Tape length (upper)
	BTFSC	STATUS,C
	RETURN 
	POP				; Drop the return address
; Stop playing
PY_ERR	CALL	KYB_TM			; Restart joystick timer
PY_STP	BCF	T3CON,TMR3ON		; Stop timer 3
	BSF	TRISC,CK1		; Set tape output level '0'
	RETFIE	FAST
PY_END	BTFSS	TRISC,CK1		; If the output level is '1'
	BRA	T_PAUS			; wait 1ms
	BRA	PY_ERR			; before resetting it
; Skip non-data block contents
TZX_LH	DECF	JSTC_H
TZX_LL	DECF	JSTC_L
SKIP_N	RCALL	GET_NX
SKIP_B	TSTFSZ	JSTC_L			; Check the overhead byte
	BRA	TZX_LL			; counter (low)
	TSTFSZ	JSTC_H			; Check the overhead byte
	BRA	TZX_LH			; counter (high)
RES_TM	RCALL	ADJ_LN
	CLRF	TMR3H
	MOVLW	.25
	MOVWF	TMR3L
	BRA	TZX_LP			; TZX overhead block done
; TZX standard speed block
STD_SP	RCALL	TAP_PA			; Copy standard timing parameters
; Collect the pause duration
PAU_LN	RCALL	GET_LN			; Get the pause length
	copy_b	JSTC_L,DATA_B+PAUS_L,2
	BTFSC	DATA_B+FILE_T,NON_DT	; 'Pause' block?
	BRA	O_HEAD			; Yes!
; Collect the block data length
TAP_LN	RCALL	GET_LN			; Get the block data length
	BTFSC	PLY_ST,PL_STD
	BRA	CK_LEN
	RCALL	GET_NX			; Collect the block length upper
	MOVWF	MS_REG			; byte for non-standard blocks
; Subtract the data from the total file length
CK_LEN	MOVF	JSTC_L,W		; Check the data length
	SUBWF	DATA_B+TAP_LL		; Subtract from tape length (low)
	IORWF	JSTC_H,W		; Is the data block void?
	IORWF	MS_REG,W
	BZ	PY_ERR			; Yes! Stop playing
	MOVF	JSTC_H,W		; Block length (high)
	SUBWFB	DATA_B+TAP_LH		; Subtract from tape length(high)
	MOVF	MS_REG,W
	RCALL	ADJ_UP
; Subtract the overhead from the total file length
O_HEAD	RCALL	ADJ_LN
; Start playing a block
	MOVF	DATA_B+FILE_T,W		; Block type
	XORLW	BK_PAU			; 'Pause' block?
	BZ	EX_PAU			; Yes!
	BSF	PLY_ST,PL_PLY		; Signal: Started playing
	XORLW	BK_PAU^BK_PLS		; 'Pulse sequence' block?
	BZ	PY_PLS			; Yes!
	XORLW	BK_PLS^BK_DAT		; 'Pure data' block?
	BZ	NW_BYT			; Yes!
; Set the leader signal length
	BCF	PLY_ST,PL_LDR		; Signal: Leader signal not done
	copy_b	TONE_0,MWIN_X,2		; Number of leader pulses
	BTFSC	PLY_ST,PL_STD		; For standard blocks the
	BTFSS	INDF0,7			; number of pulses depends on
	BRA	PY_LDR			; bit 7 of the flag byte
	copy_b	TONE_1,MWIN_X,2
	BRA	PY_LDR
; Continue playing a block
PY_DLY	MOVLW	.11			; Produce a delay in order to
DLY_LP	DECFSZ	WREG			; toggle the tape player output
	BRA	DLY_LP			; at the same time as for a new
	BTG	TRISC,CK1		; block
	MOVF	DATA_B+FILE_T,W
	XORLW	BK_PLS			; 'Pulse sequence' block?
	BNZ	CK_PDT			; No!
; Create a sequence of pulses
	DCFSNZ	JS_SEN			; Last pulse?
LP_RET	BRA	TZX_LP			; Yes!
PY_PLS	RCALL	GET_TC			; Get the pulse timing constant
	copy_b	DATA_B+PAUS_L,PRODL,2
	BRA	PY_TIM
CK_PDT	XORLW	BK_PLS^BK_DAT		; 'Pure data' block?
	BZ	PY_BYT			; Yes!
	BTFSC	PLY_ST,PL_LDR		; Leader signal done?
	BRA	PY_SYN			; Yes!
; Produce the leader signal
PY_LDR	TSTFSZ	MWIN_X
	BRA	PY_LLO
	TSTFSZ	MWIN_Y
	BRA	PY_LHI
	XORLW	BK_DAT^BK_TON		; 'Pure tone' block done?
	BZ	LP_RET			; Yes! 
; Produce the first sync pulse
	BSF	PLY_ST,PL_LDR		; Signal: Leader signal done
	BCF	PLY_ST,PL_SYN		; Signal: Second sync pulse not done
	copy_b	SYNC_1,PRODL,2		; First sync pulse timing constant
	BRA	PY_TIM
PY_SYN	BTFSC	PLY_ST,PL_SYN		; Second sync pulse done?
	BRA	PY_BYT			; Yes!
; Produce the second sync pulse
	BSF	PLY_ST,PL_SYN		; Signal: Second sync pulse done
	CLRF	JS_SEN			; Signal: New byte
	copy_b	SYNC_2,PRODL,2		; Second sync pulse timing constant
	BRA	PY_TIM
PY_BYT	TSTFSZ	JS_SEN			; Byte done?
	BRA	PY_BIT			; Not yet!
; Play new byte
NW_BYT	MOVLW	2*8
	MOVWF	JS_SEN			; There are 2 pulses per bit
	TSTFSZ	JSTC_L			; Check the block byte counter
	BRA	PY_BLO
	TSTFSZ	JSTC_H
	BRA	PY_BHI
	TSTFSZ	MS_REG
	BRA	PY_BUP
; Data block done	
EX_PAU	copy_b	DATA_B+PAUS_L,MPOS_X,2	; Pause length
	BRA	PY_PAU
; Count pause periods
PY_PHI	DECF	MPOS_Y
PY_PLO	DECF	MPOS_X
T_PAUS	BCF	PLY_ST,PL_PLY		; Signal: Pause!
	MOVLW   low  (~(FOSC_4/.1000))	; Set the timing constant for a
        MOVWF   PRODL			; pause period of 1ms
	MOVLW   high (~(FOSC_4/.1000))
        MOVWF   PRODH
        BRA	PY_TIM
; Count leader pulses
PY_LHI	DECF	MWIN_Y
PY_LLO	DECF	MWIN_X
	copy_b	S_BUFF,PRODL,2		; Leader pulse timing constant
	BRA	PY_TIM
; Count data bytes
PY_BUP	DECF	MS_REG
PY_BHI	DECF	JSTC_H			; Decrement the byte counter
PY_BLO	DCFSNZ	JSTC_L
	TSTFSZ	JSTC_H			; Last byte?
	BRA	PY_BIT			; No!
	TSTFSZ	MS_REG			; Last byte?
	BRA	PY_BIT			; No!
	BTFSS	PLY_ST,PL_STD		; Set the number of bits for
	MOVFF	DATA_B+LAST_B,JS_SEN	; non-standard blocks
; Continue playing a byte
PY_BIT	copy_b	BIT__0,PRODL,2		; Bit-0 timing constant
	BTFSS	INDF0,7	
	BRA	PY_PZR
	copy_b	BIT__1,PRODL,2		; Bit-1 timing constant
PY_PZR	BTFSC	JS_SEN,0		; Rotate next bit only on an
	RLNCF	INDF0			; odd pulse
	DCFSNZ	JS_SEN			; Last bit?
; Advance to the next byte
	RCALL	ADV_PT
; Set the timing constant
PY_TIM	MOVF	TMR3L,W			; Add the timing constant in
	ADDWF	PRODL			; PRODL:PRODH to the current
	MOVF	PRODH,W			; value of the tape timer's
	ADDWFC	TMR3H			; counter.
	MOVF	PRODL,W			; The timer counter was read
	ADDLW	8			; 8 instruction cycles before
	BTFSC	STATUS,C		; setting the new value
	INCF	TMR3H
	MOVWF	TMR3L
	RETFIE	FAST
; Collect the next byte	(11)
GET_NX	INCF	MWIN_X			; Increment the overhead counter
	BTFSC	STATUS,C
	INCF	MWIN_Y
ADV_PT	MOVF	POSTINC0,W		; Collect byte
	BTFSC   FSR0H,RBF_SZ		; Wrap around the ring pointer
	LFSR    FSR0,RING_B		; if necessary
	RETURN
; Collect the block length
GET_LN	RCALL	GET_NX
	MOVWF	JSTC_L
	RCALL	GET_NX
	MOVWF	JSTC_H
	RETURN
; Collect the sync pulse timing constants
GET_SY	RCALL	GET_TC
	copy_b	DATA_B+PAUS_L,SYNC_1,2	; First sync pulse length
	RCALL	GET_TC
	copy_b	DATA_B+PAUS_L,SYNC_2,2	; Second sync pulse length	
; Collect the bit timing constants
GET_BP	RCALL	GET_TC
	copy_b	DATA_B+PAUS_L,BIT__0,2	; Bit-0 length
	RCALL	GET_TC
	copy_b	DATA_B+PAUS_L,BIT__1,2	; Bit-1 length
	RETURN
; Calculate a timing constant
GET_TC	RCALL	GET_NX			; Multiply the low byte
	MULLW	FOSC_4*2/.109375	; by 219 = 64*12/3.5
	copy_b	PRODL,DATA_B+PAUS_L,2
	CLRF	DATA_B+LAST_B
	RCALL	GET_NX			; Multiply the high byte
	MULLW	FOSC_4*2/.109375	; by 256*219
	MOVF	PRODL,W			; and add to the result
	ADDWF	DATA_B+PAUS_H
	MOVF	PRODH,W
	ADDWFC	DATA_B+LAST_B
	MOVLW	b'100000'	
GTC_LP	RRCF	DATA_B+LAST_B		; Divide the result by 64
	RRCF	DATA_B+PAUS_H
	RRCF	DATA_B+PAUS_L
	RRCF	WREG
	BNC	GTC_LP
	COMF	DATA_B+PAUS_L		; Complement the result
	COMF	DATA_B+PAUS_H
	RETURN
;
;------------------------------------------------------------------------------
; ETHERNET NETWORKING SUBROUTINES
;
; Check the buffer operation counter 
CK_WCN	MOVLW	WBM
CK_BCN	TSTFSZ	DT_CNL
	BRA	CO_BCN
	TSTFSZ	DT_CNH
	BRA	CO_BCN
	POP
	RETURN
CO_BCN	COMF	DT_CNL
	COMF	DT_CNH
	BRA	WR_ADR
; Read Serial RAM Mode register
CK_SRM	BCF     LATC,C_NRCS             ; Select Serial RAM
	MOVLW	SR_MRD
	RCALL	RD_REG	
	BSF     LATC,C_NRCS             ; Deselect Serial RAM
	XORLW	1<<SR_SEQ		; Sequential mode?
	BNZ	CK_NET			; No!
	BSF	SYS_FL,SR_PRE		; Serial RAM present!
; Check presence of networking hardware
CK_NET	CLRF	ETH_FL
	CLRF	MSG_TM
	RCALL	CLR_BF
	RCALL	GET_ID
	table_	N_INIT
; Check presence of Wifi module
	BTFSC	PORTC,C_WFRX		; Wifi module present?
	BRA	CK_ETH			; No!
; Initialize the Wifi module
	BSF	BAUDCON2,BRG16		; 16-bit Baud rate generator
	MOVLW	low (FOSC_4/BAUD_W-1)
	MOVWF	SPBRG2
	MOVLW	high (FOSC_4/BAUD_W-1)
	MOVWF	SPBRGH2
	MOVLW	1<<TXEN|1<<BRGH
	MOVWF	TXSTA2			; Enable USART transmit
	BSF	RCSTA2,SPEN		; Enable USART
	MOVLW	TK_INI
	MOVWF	TXREG2
	CALL	TIM_50
WIN_WT	BTFSS	PORTC,C_WFRX	
	BRA	WIN_TM
	RCALL	ETH_CM
	BSF	ETH_ST,ET_WFI		; Wifi module present
	BSF	ETH_ST,ET_LNK		; Signal: Ethernet linked
        BSF     RCSTA2,CREN             ; Enable USART reception
        MOVF    RCREG2,W                ; Clear interrupt flag
ETH_DY	MOVLW	5			; Allow 5 seconds for the
	BRA	ETH_TM			; USB configuration
WIN_TM	BTFSS   INTCON,TMR0IF           ; Timer 0 time-out?
	BRA	WIN_WT
	BCF	RCSTA2,SPEN		; Disable USART
        RETURN
; Check presence of Ethernet module
CK_ETH	MOVLW   ESTAT			; Bit 3 of ESTAT must be 0
        RCALL   ETH_RD
        BTFSS	WREG,3
        BRA	ET_INI
	BCF     TRISC,C_NMSE            ; Set mouse LED as output
	RETURN	
; Initialize the Ethernet controller
ET_INI  RCALL	ETH_CM
	RCALL	ETH_CM
; Ethernet link state change
LNK_IT	MOVLW	PHIR			; Clear PHY interrupt flags
	RCALL	PHY_RD
	MOVLW	PHSTAT2			; Read PHY status registers
	RCALL	PHY_RD
	BCF	ETH_ST,ET_LNK
	BTFSS	WREG,LSTAT%8
	BRA	ETH_IE
	BSF	ETH_ST,ET_LNK		; Ethernet controller linked
	BTFSC	DEV_ST,DV_ETH		; Ethernet configured?
	BRA	ETH_IE			; Yes!
	MOVLW	0x0f
	ANDWF	ETH_ST			; Restart configuration
	CLRF	ETH_FL
	RCALL	ETH_DY			; Allow 5 seconds for the
	BRA	ETH_IE			; USB configuration
; Ethernet interrupt service routine
ETH_IS	eth_wr	BFC|EIE,1<<INTIE	; Disable further interrupts
	MOVLW	EIR
	RCALL	ETH_RD
	BTFSC	WREG,LINKIF		; Link State change?
	BRA	LNK_IT			; Yes!
	BTFSS	ETH_ST,ET_LNK		; Is the Ethernet module linked?
	BRA	ETH_IE			; No!
	BTFSS	WREG,TXERIF		; Transmission error?
	BRA	RPK_CK			; No!
; Transmission error
	RCALL	ETH_TX	
	BRA	ETH_IE
; Ethernet packet received
RPK_LP	LFSR	FSR1,PH_NPK
	MOVLW	IR_VER-PH_NPK
	RCALL	BF_RDX			; Read the Ethernet header
	MOVFF	ER_TYH,WREG
	XORLW	high IP_PRO		; IP or ARP paket?
	BNZ	RPK_SK			; No! Discard packet
	MOVFF	ER_TYL,WREG
	TSTFSZ	WREG			; IP packet?
	BRA	AR_PKT			; No! Check if ARP
; Process IP packet		
	MOVLW	IP_HLN			; IP header length
	RCALL	BF_RDX			; Read the IP header
	MOVFF	IR_VER,WREG
	ANDLW	0xf0
	XORLW	IP_VER			; IP version 4?
	BNZ	RPK_SK			; No! Discard
	MOVFF	IR_PRO,WREG
	XORLW	UDP_PN			; UDP protocol?
	BNZ	RPK_SK			; No! Discard
	MOVFF	IR_FRL,WREG		; Check whether the packet
	TSTFSZ	WREG			; is a fragment and
	BRA	RPK_SK			; discard it if so
	MOVFF	IR_FRH,WREG
	ANDLW	0x3f
	BNZ	RPK_SK
	MOVFF	IR_VER,WREG
	ANDLW	0x0f
	RLNCF	WREG
	RLNCF	WREG	
	MOVWF	CM_CNL			; Total header length
	ADDLW	-IP_HLN			; Options field length
	BTFSS	STATUS,Z		; If there are any options read
	RCALL	BF_RDX			; the field into buffer
	LFSR	FSR1,IR_VER		; Address of IP header
	RCALL	CK_SUM			; Calculate the header checksum
	BNZ	RPK_SK			; IP header checksum error!	
; Process UDP packet		
	LFSR	FSR1,UR_SPO		; UDP header address
	MOVLW	PR_BLK-UR_SPO		; UDP + Payload header length
	RCALL	BF_RDX			; Read the UDP and payload header
	MOVFF	UR_LNL,WREG		; UDP packet length(l)
	MOVWF	CM_CNL			; Set the checksum length(l)
	ADDLW	UR_SPO-PR_BLK		; Subtract the header length
	MOVWF	DT_CNL			; Set the UDP data length(l)
	MOVFF	UR_LNH,WREG		; UDP packet length(h)
	MOVWF	CM_CNH			; Set the checksum length(h)
	MOVWF	DT_CNH			; Set the UDP data length(h)
	MOVLW	0			; Subtract the BO flag
	SUBWFB	DT_CNH
	MOVLW	low(UD_PYL-PL_HLN+1)
	SUBWF	DT_CNL,W
	MOVLW	high(UD_PYL-PL_HLN+1)
	SUBWFB	DT_CNH,W		; UDP data length must not
	BC	RPK_SK			; exceed the maximum payload
	LFSR	FSR2,UR_SPO
	TSTFSZ	POSTINC2		; Check the UDP source port
	BRA	UD_PKT			; port
	MOVF	INDF2,W
	XORLW	DHC_SV			; DHCP message?
	BZ	HC_PKT			; Yes!
	XORLW	DHC_SV^DNS_SV		; DNS message?
	BNZ	RPK_SK			; No!
; DNS reply received
	BTFSC	ETH_FL,EF_DNS		; Expecting a DNS reply?			
	RCALL	DNS_RP			; Process the DNS reply	
RPK_SK	table_	DSC_PK			; Discard the current packet
	RCALL	ETH_CM
RPK_CK	eth_bk	1
	MOVLW	EPKTCNT			; Check the packet counter
	RCALL	ETH_RD			; More packets?	
	BNZ	RPK_LP			; Yes! Handle the next packet		
; Enable Ethernet interrupts
ETH_IE	BCF	INTCON3,INT3IF		; Clear the interrupt flag
	eth_wr	BFS|EIE,1<<INTIE|1<<LINKIE|1<<PKTIE|1<<TXERIE
	RETURN
; DHCP packet received
HC_PKT	BTFSC	ETH_ST,ET_DIS		; Discovery message not yet sent
	BTFSC	ETH_ST,ET_ADR		; or IP address already aquired?
	BRA	RPK_SK			; Yes! Discard packet
	RCALL	BUF_RD			; Get packet body
	MOVFF	HR_OPR,WREG
	XORLW	OT_REP			; Reply message?	
	BNZ	RPK_SK			; No! Discard packet
; DHCP reply received
	LFSR	FSR1,MA_OWN		
	LFSR	FSR2,HR_CHA
	MOVLW	MAC_LN
	RCALL	STR_CP			; Check the MAC address
	BNZ	RPK_SK			; Wrong MAC address. Discard packet
	RCALL	DH_OPT			; Process DHCP options
	BTFSC	ETH_ST,ET_ADR		; IP address acknowledged?
	BRA     AR_DNS			; Yes!
	BTFSC	ETH_ST,ET_OFR		; No offer accepted yet or
	BTFSC	ETH_ST,ET_REQ		; DHCP request already sent?
	BRA	RPK_SK			; Yes! Discard packet
	BCF	SYS_FL,CM_SKP		; Signal: DHCP Request!
	RCALL	DH_SND			; Send a DHCP request packet
	BSF	ETH_ST,ET_REQ		; Signal: DHCP request sent!
	BRA	RPK_SK
; ARP packet received
AR_PKT	XORLW	low ARP_PR		; ARP packet?
	BNZ	RPK_SK			; No! Discard packet
	BTFSS	ETH_ST,ET_ADR		; Cannot handle ARP packets if
	BRA	RPK_SK			; the IP address is not assigned
	MOVLW	HR_OPR-IR_VER
	RCALL	BF_RDX			; Read the ARP packet
	table_	ARP_RP
	RCALL	ETH_CM			; Check the packet
	BNZ	RPK_SK			; Wrong ARP parameters!
	DECF	PREINC1,W		; ARP request?
	BNZ	AR_RES			; No! Check if it is a reply
; ARP request received
	INCF	INDF2			; Set: Reply
	RCALL	ETH_CM			; Own address requested?
	BNZ	RPK_SK			; No! Discard packet
	BSF	ETH_FL,EF_RPL		; Signal: Reply!
	RCALL	UDP_PK			; Send the ARP reply
	BRA	RPK_SK
; ARP reply received
AR_RES	DECF	WREG			; ARP reply?
	BNZ	RPK_SK			; No! Discard packet
	BTFSC	DEV_ST,DV_ETH		; Ethernet already configured?
	BRA	RPK_SK			; Yes! Discard packet
	table_	ARP_CK
	RCALL	ETH_CM	
	BNZ	RPK_SK			; Wrong IP
	RCALL	ETH_CM	
	BZ	ET_CFG			; DNS server MAC address aquired
	LFSR	FSR1,IP_DNS
	RCALL	CK_LOC			; Is the DNS server local?	
	BNZ	ET_CFG			; No!
; Send an ARP request to the DNS server
	BSF	SYS_FL,CM_SKP		; Signal: DNS server ARP query
	RCALL	ARP_RQ			; Query DNS server MAC address
	BRA	RPK_SK
; Send an ARP request to the default gateway
AR_DNS	RCALL	AR_GTW
	BRA	RPK_SK	
; Ethernet is configured
ET_CFG	BCF	TIM_ST,TM_ETH		; Stop the Ethernet timer
	BCF	ETH_FL,EF_TMO		; Clear the time-out flag
	BSF	DEV_ST,DV_ETH		; Signal: Ethernet configured
	RCALL	SVL_ON	
	BRA	RPK_SK
; UDP datagram received
UD_PKT	BTFSS	DEV_ST,DV_ETH		; Cannot handle UDP data packets
	BRA	RPK_SK			; if not configured
	MOVLW	high UDP_CL
	CPFSEQ	PREINC2
	BRA	RPK_SK
	MOVLW	low UDP_CL
	CPFSEQ	PREINC2
	BRA	RPK_SK
	RCALL	UDP_DG
	BRA	RPK_SK
;
; Read PHY register (WREG) to (SD_CRC):(WREG)
PHY_RD	MOVWF	SD_CRC
	RCALL	BANK_2
        MOVLW   WCR|MIREGADR
	RCALL	WR_ADR
	MOVF	SD_CRC,W
	RCALL	WR_DAT
	eth_wr	WCR|MICMD,1<<MIIRD
	RCALL	PHY_WT
	eth_wr	WCR|MICMD,0
	MOVLW	MIRDL
	RCALL	MAC_RD
	MOVWF	SD_CRC
	MOVLW	MIRDH
; Read MAC register (WREG)
MAC_RD	RCALL	WR_ADR
	BRA	RD_REG
; Wait for PHY register access
PHY_WT	eth_bk	3			; Select bank 3
PHY_LP	MOVLW	MISTAT
	RCALL	MAC_RD
	BTFSC	WREG,MIBUSY
	BRA	PHY_LP
; Select bank 2
BANK_2	eth_bk	2
	RETURN
; Wait until the transmitter is ready
TXR_LP	RCALL	EC_PSH			; Save (TABLPTR)
	RCALL	ETH_TX			; Restart transmission
TX_RDY	MOVLW	EIR
	RCALL	ETH_RD
	BTFSC	WREG,TXERIF		; Restart a transmission
	BRA	TXR_LP			; that was aborted
	MOVLW	ECON1
	RCALL	ETH_RD
	BTFSC	WREG,TXRTS		; Tranmission in progress?
	BRA	TX_RDY			; Yes! Wait until it ends
	RETURN				; Transmitter is ready
; Start transmission
ETH_TX	table_	TX_CMD
;
; Ethernet command interpreter
;*	WCR|reg,	dat		; Write to register
;*	WBM,		dat		; Write 1 byte to buffer
;*	BFS|reg,	dat		; Set bits of register		
;*	BFC|reg,	dat		; Clear bits of register
;	WDI,		dat		; Write direct to (FSR1) & inc	
;	WDD,		dat		; Write direct to (FSR1) & dec	
;	CPY,		cnt		; Copy bytes from (FSR2) to (FSR1)
;	CLR,		cnt		; Copy '0'-s to (FSR1)
;	CMP,		cnt		; Compare bytes between (FSR1) (FSR2)
;	TCY,		cnt		; Copy bytes from (TABLPTR) to (FRS1)
;	RXI,		reg 		; Read from register to (FSR1) & inc
;	RXD,		reg		; Read from register to (FSR1) & dec
;	WXI,		reg		; Write from (FSR1) to register & inc	
;	WXD,		reg		; Write from (FSR1) to register	& dec
;	WRB,		cnt		; Write bytes from (FSR1) to buffer
;	CLB,		cnt		; Write '0'-s to buffer
;	SKP		dis		; Skip if flag set
;	GTO		dis		; Forward jump to command table
;	CAL		dis		; Forward call into command table
;	LX1|adr(h),	adr(l)		; Load RAM address into (FSR1)	
;	LX2|adr(h),	adr(l)		; Load RAM address into (FSR2)	
;	SBR,	adr(u),	adr(h),	adr(l)	; Execute subroutine
;	WPH,	phr,	dat(l),	dat(h)	; Write word to PHY register
;	RET,		1		; Return from table call
;	RET,		0		; End of command table
;
ETH_CM	TBLRD*+
	MOVF	TABLAT,W
	XORLW	RET
	BNZ	ECK_RX
	TBLRD*+
	MOVFF	ECM_SP,WREG
	INCFSZ	WREG
	RCALL	EC_POP
	TSTFSZ	TABLAT
	BRA	ETH_CM
	RETURN				; Returns ZR set	
ECK_RX	BCF	STATUS,C		; Clear CY
	XORLW	RET^RXI			; Read register to (FSR1) & inc?
	BZ	EC_RDX			; Yes!
	XORLW	RXI^RXD			; Read register to (FSR1) & dec?
	BNZ	ECK_WD			; No!
	BSF	STATUS,C		; Set CY for 'dec'
EC_RDX	TBLRD*+				; Read the register address
	MOVF	TABLAT,W		; from (TABLPTR) to (WREG)
	RCALL	ETH_RD			; Read the register to (WREG)
EC_RDI	BC	EC_RDD
	MOVWF	POSTINC1		; Store (WREG) at (FSR1) & inc	
	BRA	ETH_CM
EC_RDD	MOVWF	POSTDEC1		; Store (WREG) at (FSR1) & dec	
	BRA	ETH_CM
ECK_WD	XORLW	RXD^WDI			; Write direct to (FSR1) & inc? 
	BZ	EC_WRD			; Yes!
	XORLW	WDI^WDD			; Write direct to (FSR1) & dec?
	BNZ	ECK_PW			; No!
	BSF	STATUS,C		; Set CY for 'dec'	
EC_WRD	TBLRD*+				; Read the direct value
	MOVF	TABLAT,W		; from (TABLPTR) to (WREG)
	BRA	EC_RDI			; Store the value at (FSR1)
ECK_PW	XORLW	WDD^WPH			; Write to a PHY register?
	BNZ	ECK_WX			; No!
	RCALL	BANK_2			; Set bank 2
	MOVLW	WCR|MIREGADR		; Set the register address
	RCALL	WR_REG			; from (TABLPTR)
	MOVLW	WCR|MIWRL		; Set the low byte
	RCALL	WR_REG			; from (TABLPTR)
	MOVLW	WCR|MIWRH		; Set the high byte
	RCALL	WR_REG			; from (TABLPTR)
	RCALL	PHY_WT			; Wait for operation to complete
	BRA	ETH_CM
ECK_WX	XORLW	WPH^WXI			; Write from (FSR1) to register & inc?
	BZ	EC_WRX			; Yes!
	XORLW	WXI^WXD			; Write from (FSR1) to register & dec?
	BNZ	ECK_TC			; No!
	BSF	STATUS,C		; Set CY for 'dec'
EC_WRX	TBLRD*+				; Read the register address
	MOVF	TABLAT,W		; from (TABLPTR) to (WREG)
	IORLW	WCR			; Add the operation code WCR
	RCALL	WR_ADR			; Set the register write address
	BC	EC_WXD
	MOVF	POSTINC1,W		; Read byte from (FSR1) to (WREG) & inc
	BRA	EC_WXW
EC_WXD	MOVF	POSTDEC1,W		; Read byte from (FSR1) to (WREG) & inc
EC_WXW	RCALL	WR_DAT			; Write byte from (WREG) to register
	BRA	ETH_CM
ECK_TC	XORLW	WXD^TCY			; Copy bytes from (TABLPTR) to (FRS1)? 
	BNZ	ECK_GT			; No!
	TBLRD*+				; Read the number of bytes to be copied
	MOVF	TABLAT,W		; from (TABLPTR) to (WREG)
	MOVWF	TB_CNT			; Initialize the byte counter
	CALL	TB_TST			; Copy the bytes to (FSR1), incrementing
	BRA	ETH_CM
ECK_GT	XORLW	TCY^GTO			; Jump to table address?
	BNZ	ECK_SK			; No!
	TBLRD*+				; Read the displacement from (TABLPTR) 
	BRA	EC_FJP
ECK_SK	XORLW	GTO^SKP			; Skip if flag clear?
	BNZ	ECK_CA			; No!
	TBLRD*+				; Read the displacement from (TABLPTR) 
	BTFSC	SYS_FL,CM_SKP		; Flag set?
	BRA	EC_FJP			; Skip
	BRA	ETH_CM
ECK_CA	XORLW	SKP^CAL			; Call a table address?
	BNZ	ECK_SB			; No!
	TBLRD*+				; Read the displacement from (TABLPTR)
	RCALL	EC_PSH			; Push the current table address
EC_FJP	MOVF	TABLAT,W		; Displacement to (WREG)
	CALL	ADD_TP			; Add displacement to (TABLPTR)
	BRA	ETH_CM
ECK_SB	XORLW	CAL^SBR			; Execute a subroutine?
	BNZ	ECK_CP			; No!
	RCALL	EC_SBR			; Call the subroutine
	BRA	ETH_CM
ECK_CP	XORLW	SBR^CMP			; Compare bytes?
	BNZ	ECK_CY			; No!
        TBLRD*+
	MOVF	TABLAT,W
	BRA	STR_CP
ECK_CY	XORLW	CMP^CPY			; Copy (WREG) bytes from (FSR2) to (FSR1)?
	BNZ	ECK_CL			; Yes!
	TBLRD*+
	MOVF	TABLAT,W
	RCALL	STR_CY
	BRA	ETH_CM
ECK_CL	XORLW	CPY^CLR			; Clear (WREG) bytes at (FSR1)?
	BNZ	ECK_BW			; No!
	TBLRD*+
	MOVF	TABLAT,W
	RCALL	STR_CL
	BRA	ETH_CM
ECK_BW	XORLW	CLR^WRB			; Write (WREG) bytes from (FSR1) to buffer?
	BNZ	ECK_BC			; Yes!
        TBLRD*+
	MOVF	TABLAT,W
	RCALL	BF_WRX
	BRA	ETH_CM
ECK_BC	XORLW	WRB^CLB			; Clear (WREG) bytes in buffer?
	BNZ	ECK_X1			; No!
        TBLRD*+
	MOVF	TABLAT,W
	RCALL	BF_CLR
	BRA	ETH_CM
ECK_X1	XORLW	CLB^LX1			; Load (FSR1) from (TABLPTR)?
	ANDLW	b'11110000'		; Separate the operation code
	BNZ	ECK_X2			; No!
	MOVF	TABLAT,W
	ANDLW	b'00001111'		; Get the address(h) in (WREG)
	MOVWF	FSR1H			; Set the address(h)
        TBLRD*+				; Read the address(l) to (WREG)
	MOVF	TABLAT,W
	MOVWF	FSR1L			; Set the address(l)
	BRA	ETH_CM
ECK_X2	XORLW	LX1^LX2			; Load (FSR2) from (TABLPTR)?
	BNZ	EC_DEF			; No!
	MOVF	TABLAT,W
	ANDLW	b'00001111'		; Get the address(h) in (WREG)
	MOVWF	FSR2H			; Set the address(h)
        TBLRD*+				; Read the address(l) to (WREG)
	MOVF	TABLAT,W
	MOVWF	FSR2L			; Set the address(l)
	BRA	ETH_CM
EC_DEF	MOVF	TABLAT,W		; Handle standard write operations
	RCALL	WR_REG			; WCR, WBM, BFC, BFS
	BRA	ETH_CM
; POP the (TABLPTR) form the table stack
EC_POP	RCALL	EC_GSP
	MOVFF	PREINC2,TBLPTRH		; POP address(h)
	MOVFF	PREINC2,TBLPTRL		; POP address(l)
	BRA	EC_PSP			; Save the stack pointer
; PUSH the (TABLPTR) to the table stack
EC_PSH	RCALL	EC_GSP
	MOVFF	TBLPTRL,POSTDEC2	; PUSH address(l)
	MOVFF	TBLPTRH,POSTDEC2	; PUSH address(h)
EC_PSP	MOVFF	FSR2L,ECM_SP		; Save the stack pointer
	RETURN
; Get the table stack pointer in (FSR2)
EC_GSP	LFSR	FSR2,SST_TB-1
	MOVFF	ECM_SP,FSR2L		; Load the stack pointer into (FSR2)
	RETURN
; Jump to a program address read from (TBLPTR)
EC_SBR	TBLRD*+				; Read the PC(u) from (TABLPTR)
	MOVF	TABLAT,W		; to (WREG)
	MOVWF	PCLATU			; Set the PC(h)
	TBLRD*+				; Read the PC(h) from (TABLPTR)
	MOVF	TABLAT,W		; to (WREG)
	MOVWF	PCLATH			; Set the PC(h)
	TBLRD*+				; Read the PC(l) from (TABLPTR)
	MOVF	TABLAT,W		; to (WREG)
	MOVWF	PCL			; Set the PC(l)
; Clear (WREG) bytes at (FSR1)
STR_CL	TSTFSZ	WREG
	BRA	CLR_LP
	RETURN
; Compare (WREG) bytes between (FSR1) and (FSR2)
STR_CP	MOVWF	TB_CNT
	COMF	TB_CNT			; Complement the byte counter
SCP_LP	MOVF	POSTINC2,W
	BTFSC	SYS_FL,CM_CAS		; Upper case comparison?
	BTFSS	WREG,6			; Literal?
	BRA	SCP_CN			; No!
	BCF	WREG,5			; Change to upper case
SCP_CN	INCF	TB_CNT			; Done?
	BTFSS	STATUS,Z		; Yes!  Return ZR=1
	CPFSEQ	POSTINC1		; Do the bytes match?
	RETURN				; No! Return ZR=0
	BRA	SCP_LP	
; Read (WREG) bytes from the buffer to (FSR1)
BF_RDX	MOVWF	DT_CNL
	CLRF	DT_CNH
; Read (DT_CNL:DT_CNH) from the buffer to (FSR1)
BUF_RD	MOVLW	RBM	
	RCALL	CK_BCN
BRD_LP	INFSNZ	DT_CNL
	BRA	BRD_BO
BRD_NX	MOVLW	-1
	MOVWF   SSP2BUF                 ; Start SPI operation
BRD_WT  BTFSS   SSP2STAT,BF             ; Wait until the operation ends
        BRA     BRD_WT
        MOVF    SSP2BUF,W               ; Read the received byte
	MOVWF	POSTINC1		; Store the read byte at (FSR1)
	BRA	BRD_LP	
BRD_BO	INCFSZ	DT_CNH
	BRA	BRD_NX
	BRA	BUF_DS			; Deselect the Ethernet module
; Clear (WREG) bytes in the buffer
BF_CLR	MOVWF	DT_CNL
	CLRF	DT_CNH
	RCALL	CK_WCN	
BCL_LP	INFSNZ	DT_CNL
	BRA	BWR_BO
	MOVLW	0
	MOVWF   SSP2BUF                 ; Start SPI operation
BCL_WT  BTFSS   SSP2STAT,BF             ; Wait until the operation ends
        BRA     BCL_WT
        MOVF    SSP2BUF,W               ; Read the received byte
	BRA	BCL_LP
; Write the datagram header and payload
PL_DAT	RCALL	PL_HED			; Write the header
	RCALL	PT_ADR			; Payload address
	BRA	BUF_WR
; Write (SD_SZ2) bytes from (FSR1) to the buffer
PL_HED	MOVF	SD_SZ2,W
; Write (WREG) bytes from (FSR1) to the transmit buffer
BF_WRX	MOVWF	DT_CNL
	CLRF	DT_CNH
; Write (DT_CNL:DT_CNH) from (FSR1) to buffer
BUF_WR	RCALL	CK_WCN
BWR_LP	INFSNZ	DT_CNL
	BRA	BWR_BO
BWR_NX	MOVF	POSTINC1,W
	MOVWF   SSP2BUF                 ; Start SPI operation
BWR_WT  BTFSS   SSP2STAT,BF             ; Wait until the operation ends
        BRA     BWR_WT
        MOVF    SSP2BUF,W               ; Read the received byte
	BRA	BWR_LP	
BWR_BO	INCFSZ	DT_CNH
	BRA	BWR_NX
BUF_DS	BSF     LATC,C_NECS		; Deselect the Ethernet module
	RETURN
; Calculate checksum of (CM_CNL) bytes at (FSR1)
CK_SUM	COMF	CM_CNL
	CLRF	SD_SZ0
	CLRF	SD_SZ1	
CKS_LP	INFSNZ	CM_CNL
	BRA	CKS_EN
	MOVF	POSTINC1,W
	ADDWF	SD_SZ0
	INCFSZ	CM_CNL,W
	MOVF	POSTINC1,W
	ADDWFC	SD_SZ1
	MOVLW	0
	ADDWFC	SD_SZ0
	ADDWFC	SD_SZ1
	INCFSZ	CM_CNL
	BRA	CKS_LP	
CKS_EN	COMF	SD_SZ0
	COMF	SD_SZ1
	MOVF	SD_SZ0,W
	IORWF	SD_SZ1,W
	RETURN
; Process DHCP options
DHO_AK	LFSR	FSR1,IP_OWN
	LFSR	FSR2,HR_YIA		; Copy the confirmed IP
	MOVLW	IPN_LN			; address
	RCALL	CPY_LP	
	BSF	ETH_ST,ET_ADR		; Signal: address aquired!
	BRA	DH_OPT			; Get other configuration data
DHO_OF	BTFSC	ETH_ST,ET_OFR		; Already accepted an offer?
	BRA	DHO_EN			; Yes!
DH_OPT	BTG	ETH_ST,ET_DIS
	LFSR	FSR2,HR_OPT		; Point to options
DHO_LP	MOVF	POSTINC2,W		; Get the option code
	BZ	DHO_LP			; Skip padding
	MOVFF	POSTINC2,TB_CNT		; and length
	XORLW	OP_END			; End of option list?
	BZ	DHO_EN			; Yes!
	BTFSC	ETH_ST,ET_DIS		; Checking message type?
	BRA	DHO_DT			; No!
	XORLW	OP_END^OP_MTY		; Message type?
	BNZ	DHO_NX			; No!
	MOVF	POSTINC2,W		; Get message type
	XORLW	MS_OFR			; Offer?
	BZ	DHO_OF			; Yes!
	XORLW	MS_OFR^MS_ACK		; Acknowledge?
	BZ	DHO_AK			; Yes!
	XORLW	MS_ACK^MS_NAK		; NAK message?
	BNZ	DHO_EN			; No!
	BCF	ETH_ST,ET_REQ
	BTFSC	ETH_ST,ET_OFR		; After offer?
	BSF	ETH_ST,ET_ADR		; Yes! Keep default settings
DHO_EN	BSF	ETH_ST,ET_DIS
	RETURN
DHO_DT	BTFSS	ETH_ST,ET_ADR		; From an offer message get 
	BRA	DHO_SI			; only the server identifier
	XORLW	OP_END^OP_MSK		; Mask?
	LFSR	FSR1,IP_MSK		; Save the subnet mask
	BZ	DHO_CY
	XORLW	OP_MSK^OP_GAT		; Default gateway?
	LFSR	FSR1,IP_GAT		; Save the default gateway
	BZ	DHO_CY
	XORLW	OP_GAT^OP_DNS		; DNS server?
	BNZ	DHO_NX			; Skip other options
	LFSR	FSR1,IP_DNS		; Save the DNS server
	BRA	DHO_CY
DHO_SI	BTFSC	ETH_ST,ET_OFR		; Server identifier already aquired?
	BRA	DHO_EN			; Yes!
	XORLW	OP_END^OP_SVI		; Server identifier?
	BNZ	DHO_NX			; No!
	BSF	ETH_ST,ET_OFR		; Signal: DHCP offer received
	LFSR	FSR1,IP_STA		; Save the server identifier
DHO_CY	MOVLW	IPN_LN	
	SUBWF	TB_CNT
	RCALL	CPY_LP
DHO_NX	MOVF	TB_CNT,W		; Advance pointer to the
	RCALL	ADV_X2			; next entry
	BRA	DHO_LP
;
; Process a network command
ETH_EX	BTFSS	DEV_ST,DV_ETH		; Ethernet must be configured
	BRA	EE_NPR
	MOVLW	SST-1			; The station number must be
	ANDWF	CTL_CD,W		; in the range: 1-7
	BCF	CTL_CD,CLN
	BTFSC	STATUS,Z
	BSF	ETH_FL,EF_STZ		; Signal: Broadcast!
	MOVF	CMD_CD,W
	XORLW	NET_RX			; Network receive?
	BNZ	CK_LNK			; No!
	BTFSC	ETH_FL,EF_STZ
	BRA	NRX_BC
; Network receive (broadcast)
NRX_LP	CLRF	BLN_LO
	MOVF	CTL_CD,W
	RCALL	STA_IX
	TSTFSZ	INDF1			; Station linked?
	BTFSC	POSTINC1,FDV		; Block device?
	BRA	NRX_EN			; No!
	MOVFF	PREINC1,CM_CNL		; Get the stattion ID
	MOVFF	PREINC1,CM_CNH
	RCALL	FND_ID			; Find the client table entry
	BZ	NRX_EN			; No client table entry found
	ADDLW	CST-1
	LFSR	FSR2,FAT_BF
	ADDWF	FSR2H			; Input buffer address
	MOVF	DATA_B,W		; Requested length
	CPFSGT	INDF2			; More than available?
	MOVF	INDF2,W			; Returned data length
	SUBWF	POSTINC2		; Remaining length
	BTFSS	INDF1,EOF		; EOF message?
	BRA	NRX_DT			; No!
	BCF	INDF1,EOF		; Clear EOF flag
	BSF	CTL_CD,EOF		; Signal: EOF message!
NRX_DT	RCALL	ET_SND			; Copy the received data
	BZ	NRX_CK			; No data remaining in buffer
	MOVFF	FSR2H,FSR1H
	CLRF	FSR1L
	MOVF	POSTINC1,W		; Remaining data length
	RCALL	STR_CY			; Copy any remaining data
NRX_CK	BTFSS	CTL_CD,EOF		; to the start of the buffer
	TSTFSZ	BLN_LO			; Check the data length
	BRA	NRX_RT
NRX_EN	BTFSS	ETH_FL,EF_STZ
	BRA	CLR_EF
NRX_BC	INCF	CTL_CD
	BTFSC	CTL_CD,CLN
	BRA	CLR_EF
	BRA	NRX_LP
NRX_RT	CLRF	CMD_CD			; Clear the error code
	CLRF	MSG_TM			; Stop U LED blinking
	BRA	CLR_EF
CK_LNK	XORLW	NET_RX^ST_LNK		; Link station?
	BNZ	CK_NTX			; No!
	BTFSS	ETH_FL,EF_STZ		; Station zero?
	BRA	LNK_ST			; No!
	BCF	ETH_FL,EF_STZ
	LFSR	FSR1,SST_TB		; If no station number
NTX_LP	INCF	CTL_CD			; specified, find the first
	BTFSC	CTL_CD,CLN		; unlinked one
	BRA	EE_IST			; None found!
	MOVF	INDF1,W
	BZ	LNK_ST	
	MOVLW	SRL
	ADDWF	FSR1L
	BRA	NTX_LP	
; Send a station link request
LNK_ST	CLRF	BLN_HI
	LFSR	FSR2,DATA_B
	COMF	BLN_LO,W		; Length of name / IP 
	RCALL	CK_IPL			; IP address specified?
	BNC	CHK_IP			; Yes!
CK_NAM	MOVLW	2
	CPFSGT	BLN_LO			; Check the name length
	BRA	EE_NPR
	LFSR	FSR1,NT_NAM+1		; Copy the specified 
	LFSR	FSR2,DATA_B		; station name
	MOVFF	BLN_LO,DT_CNL		; Name length
	MOVLW	'.'
LNK_LP	CPFSEQ	INDF2
	BRA	LNK_NX
	BSF	ETH_FL,EF_DNS		; Signal: Remote name!
LNK_NX	MOVFF	POSTINC2,POSTINC1
	DECFSZ	DT_CNL
	BRA	LNK_LP	
	CLRF	INDF2			; Set name end marker
	BTFSS	ETH_FL,EF_DNS		; Local name?
	BRA	LNK_RQ			; Yes!		
; Send a DNS query
	table_	DNS_RQ
	RCALL	ETH_CM
	COMF	BLN_LO,W
	MOVWF	DT_CNH
	MOVLW	'.'
DNS_LP	copy_b	FSR1L,FSR2L,2
	SETF	INDF2
DNC_LP	INCF	INDF2
	INFSNZ	DT_CNH
	BRA	DNQ_EN
	CPFSEQ	PREINC1
	BRA	DNC_LP
	BRA	DNS_LP
DNQ_EN	MOVF	PREINC1,W
	CLRF	POSTINC1
	RCALL	ETH_CM
	MOVLW	low NT_IDN
	SUBWF	FSR1L,W
	BTFSS	STATUS,C
	NEGF	WREG
	RCALL	DNS_LN
	BCF	ETH_FL,EF_RPL		; Signal: Request!
	BRA	IP_CKS
; Host name resolved
ETH_IP	RCALL	ET_SND
WIF_IP	BTFSC	SYS_FL,CM_DNS		; Return the result of a
	BRA	ETH_EN			; DNS query command
HN_RES	BCF	ETH_FL,EF_DNS
REM_IP	CLRF	BLN_LO			; Remote IP known
	BRA	LNK_RQ
CHK_IP	LFSR	FSR1,IP_STA		; Specified IP number
	RCALL	CK_LOC			; Local IP
	BNZ	REM_IP			; No!
	BSF	CTL_CD,EOF		; Signal: IP number!
	LFSR	FSR2,IP_STA		; Copy the IP number
	MOVLW	IPN_LN			; to the data block
	RCALL	ET_SND
LNK_RQ	BTFSC	SYS_FL,CM_DNS		; Error: DNS query command
	BRA	NO_ERR			; specifies a local name
	BSF	CTL_CD,REQ
	BCF	SYS_FL,CM_SKP		; Signal: Request
	table_	STA_TP
	RCALL	ETH_CM			; Set up data packet
	MOVLW	0x0f
	ANDWF	CTL_CD
	MOVFF	CTL_CD,PT_BLK
	TSTFSZ	BLN_LO
	BRA	UDP_BC			; Broadcast
; Send data packet
UDP_TX	RCALL	ETH_CM			; Set destination MAC and IP
UDP_BC	MOVFF	PT_CMD,WREG		
	XORLW	SD_WRB
	BTFSS	STATUS,Z
	XORLW	SD_WRB^ST_LNK		; Set the payload header length	
	MOVLW	PL_HLN			; to 4 bytes, except for the
	BTFSC	STATUS,Z		; 'Write block' and 'Link station'
	ADDLW	PL_BLN			; commands, when it is 8 bytes
	RCALL	DNS_LN
	LFSR	FSR1,PT_LNL		; Add the payload data length
	RCALL	UDP_LN
	table_	UD_REQ
	BTFSC	SYS_FL,CM_SKP
	BSF	ETH_FL,EF_RPL		; Signal: Reply!
	BRA	IP_CKS
CK_NTX	BTFSS	ETH_FL,EF_STZ
	BRA	EFL_RQ
	XORLW	ST_LNK^NET_TX
	BNZ	EE_IST
; Network broadcast transmit
TXB_LP	INCF	CTL_CD
	BTFSC	CTL_CD,CLN
	BRA	NO_ERR
; Send a regular netwok request
EFL_RQ	MOVF	CTL_CD,W
	MOVWF	REQ_ST
	RCALL	STA_IX
	TSTFSZ	INDF1			; Station linked?
	BRA	CK_STA			; Yes!
CK_STZ	BTFSC	ETH_FL,EF_STZ
	BRA	TXB_LP
EE_IST	MOVLW	E_ISTA			; Else signal error
	BRA	ETH_ER
CK_STA	MOVLW	ST_CHK
	CPFSEQ	CMD_CD
	BRA	ERQ_CK
; Check if a station is a file or block device
 	MOVF	CTL_CD,W
	BTFSS	WREG,7			; Bit 7 indicates a block device
	BSF	WREG,FDV
	XORWF	INDF1,W
	ANDLW	1<<FDV
	BRA	ETH_ER
ERQ_CK	MOVLW	S_RXTX
        CPFSLT  CMD_CD                  ; Server command?
	BRA	EBL_RQ			; No!
; Send a network file device request
	BTFSS	INDF1,FDV		; Station must be a file device
	BRA	EE_IST
; Send a network request datagram
ERQ_TX	BTFSC	ETH_FL,EF_STZ		; Broadcast?
	BRA	SET_RQ			; Yes!
	BSF	INDF1,REQ
	BTG	INDF1,PAR
	BSF	ETH_FL,EF_REQ
SET_RQ	BCF	SYS_FL,CM_SKP		; Signal: Request
	copy_b	FSR1L,CRT_ST,2		; Save the station table pointer
EPK_TB	table_	UDP_TP
	RCALL	ETH_CM			; Set up data packet
	copy_b	CRT_ST,FSR2L,2		; Restore the station table pointer
	MOVFF	INDF2,PT_CTL		; Set the control byte
	MOVLW	1
	BTFSC	SYS_FL,CM_SKP		; Save the error code of		
	MOVFF	PT_CMD,PLUSW2		; a reply datagram
	BTFSS	SYS_FL,CM_SKP		; Reset the EOF flag after
	BCF	INDF2,EOF		; sending a request
	BTFSC	ETH_FL,EF_STZ		; Broadcast?
	BSF	SYS_FL,CM_SKP		; Yes! No reply expected!
	MOVLW	SHL
	ADDWF	FSR2L			; Point to the station's MAC and IP
	BRA	UDP_TX
; Send a network block device request
EBL_RQ	BTFSC	INDF1,FDV		; Station must be a block device
	BRA	CK_STZ
	MOVF	CMD_CD,W
        XORLW	SD_IDF			; Identify card?
	BZ	ERQ_TX			; Yes!
	XORLW	SD_IDF^NET_TX		; Transmit data?
	BNZ	BRQ_RB			; No!
	BTFSC	CTL_CD,EOF		; EOF message?
	BSF	INDF1,EOF		; Yes!
	BRA	ERQ_TX
BRQ_RB	XORLW   NET_TX^SD_RDB	        ; 'Read Block' Command?
        BZ      ERQ_TX                  ; Yes!
        XORLW   SD_RDB^SD_RDT           ; 'Read-Transmit' Command?
        BNZ     BRQ_RA                  ; No!
	copy_b	DATA_B+PL_HLN,CNT_LO,2	; Save the data length
        BSF     COM_ST,CO_RDT           ; Set 'Read-Transmit' flag
        BRA     BRQ_RD
BRQ_RA	XORLW   SD_RDT^SD_RDA           ; 'Read Alternate' Command?
        BNZ     BRQ_WR                  ; No!
        BSF     ETH_FL,EF_RDA           ; Signal: 'Read Alternate'
BRQ_RD	MOVLW	SD_RDB
	BRA	BRQ_EN
BRQ_WR	XORLW   SD_RDA^SD_WRB           ; 'Write Block' Command?
        BZ      ERQ_TX                  ; Yes!
        XORLW   SD_WRB^SD_WRA           ; 'Write Alternate' command?
        BNZ     EE_IST                  ; No!
        BSF     ETH_FL,EF_WRA           ; Set the 'Alternate' flag
        CALL	SET_WA                  ; Set the write address
	CLRF	BLN_LO			; Set the data block length
	MOVLW	high SCS
	MOVWF	BLN_HI
	MOVLW	SD_WRB
BRQ_EN	MOVWF	CMD_CD
	BRA	ERQ_TX
; Clear (WREG) bytes at (FSR1)
CLR_LP	CLRF	POSTINC1
	DECFSZ	WREG
	BRA	CLR_LP
	RETURN
;
; Ethernet time-out handling
ETH_TO	BCF	ETH_FL,EF_TMO		; Clear the time-out flag
	BTFSS	ETH_ST,ET_LNK		; Ethernet module linked?
	RETURN				; No!
	RLNCF	TO_PER			; Double the time-out period
	BTFSC	DEV_ST,DV_ETH		; Ethernet configured?
	BRA	UDP_TO			; Yes!
	BTFSC	ETH_ST,ET_ADR		; IP address aquired?
	BRA	ARP_TO			; Yes!
	BTFSC	ETH_ST,ET_DIS		; Discovery message sent?
	BRA	DHC_TO			; Yes!
	BSF	LATC,C_NSRV		; Signal: Not configured!
	BTFSS	DEV_ST,DV_SDC		; SD card present?
	BRA	ET_CON			; No!
; Get settings from SD card
	MOVLW	SD_RDB^PER_CM		; Operation: 'Read Block'
	RCALL	ETD_SD			; Read the last sector of the
	MOVLW	'e'			; SD card's logical drive 1
	CPFSEQ	POSTINC2
	BRA	EDT_WF	
	MOVLW	'a'
	CPFSEQ	POSTINC2
	BRA	EDT_WF	
	COMF	POSTINC2,W
	RCALL	GET_IP
	BC	EDT_WF	
	MOVLW	low(IP_STA-1)
	SUBWF	FSR1L,W
	LFSR	FSR2,IP_STA
	LFSR	FSR1,IP_OWN		; Copy the default settings
	RCALL	STR_CY			; retrieved from the SD card
EDT_WF	LFSR	FSR2,AP_CRD
	MOVLW	'e'
	CPFSEQ	POSTINC2
	BRA	ET_CON
	MOVLW	'w'
	CPFSEQ	POSTINC2
	BRA	ET_CON
	MOVFF	POSTINC2,TB_CNT
	LFSR	FSR1,WAP_ID
	MOVLW	HOST_N-WAP_ID
	RCALL	CLR_LP
	LFSR	FSR1,WAP_ID
	MOVLW	' '
EWF_LP	CPFSEQ	INDF2
	BRA	EWF_CY
	LFSR	FSR1,WAP_PW
	CPFSEQ	POSTINC2
EWF_CY	MOVFF	POSTINC2,POSTINC1	
	DECFSZ	TB_CNT
	BRA	EWF_LP
; Connect to network
ET_CON	BTFSC	ETH_ST,ET_WFI		; Wifi present?
	BRA	WF_STA			; Yes!
; Send DHCP discovery message
	BSF	ETH_ST,ET_DIS		; Signal: Discovery message sent!
	BSF	SYS_FL,CM_SKP		; Signal: DHCP Discovery
; Send DHCP packet
DH_SND	table_	DHC_SN
	RCALL	ETH_CM
	MOVLW	MS_REQ
	BTFSS	SYS_FL,CM_SKP
	MOVFF	WREG,HT_MSG
	MOVF	FSR1L,W
	SUBLW	low HT_END
	RCALL	STR_CL
; Set IP header checksum
IP_CKS	LFSR	FSR1,IT_VER		; Address of IP header
	MOVLW	UT_SPO-IT_VER		; Length of IP header
	MOVWF	CM_CNL
	RCALL	CK_SUM
	copy_b	SD_SZ0,IT_CKS,2
	BRA	UDP_PK
; Data request time-out after 4+8=12 seconds
UDP_TO	BTFSS	TO_PER,4
	BRA	TO_RTX
; Signal error
EE_NPR	MOVLW   E_NPRE
ETH_ER	MOVWF	CMD_CD
	CLRF	BLN_LO
	CLRF	BLN_HI
ETH_EN	BCF	TIM_ST,TM_ETH		; Stop the Ethernet timer
	BTFSS	ETH_FL,EF_REQ		; Station request flag set?
	BRA	CLR_EF			; No!
	MOVF	REQ_ST,W
	RCALL	STA_IX
	BCF	INDF1,REQ		; Clear the station request flag
CLR_EF	CLRF	ETH_FL			; Clear all Ethernet flags
	BCF	SYS_FL,CM_DNS		; Clear the DNS query command flag
        BSF     PIE3,SSP2IE             ; Trigger an interrupt
	RETURN	
; DHCP request time-out after 4+8+16 = 28 seconds
DHC_TO	BTFSS	TO_PER,5
	BRA	TO_RTX
	BSF	ETH_ST,ET_ADR		; Keep the default configuration	
; Send an ARP request to the gateway
AR_GTW	BCF	SYS_FL,CM_SKP		; Signal: Default gateway ARP query
ARP_RQ	table_	ARP_QR
UDP_PK	MOVLW	1<<2			; First resend after 4 seconds	
	BTFSS	ETH_FL,EF_RPL		; Reply?
	MOVWF	TO_PER			; No! Set time-out period
	BRA	SND_PK
; ARP request time-out after 4+8+16 = 28 seconds
ARP_TO	BTFSC	TO_PER,5
	RETURN
; Resend the last request packet
TO_RTX	BTFSC	ETH_ST,ET_WFI
	BRA	WF_RTX
	RCALL	TX_RDY
	table_	PK_RSD
SND_PK	BTFSC	ETH_ST,ET_WFI
	RCALL	WF_PTX			; Send Wifi packet
	BTFSS	ETH_ST,ET_WFI
	RCALL	ETH_CM			; Send Ethernet packet
	BTFSS	ETH_FL,EF_RPL		; Reply?
	BRA	SET_TO			; No! Start timer
	BCF	ETH_FL,EF_RPL		; Signal: Request
	BTFSC	ETH_FL,EF_STZ
	BRA	TXB_LP	
	RETURN
; Resend Wifi data request or DNS query
WF_RTX	MOVLW	TK_RRQ			; Data request
	BTFSC	ETH_FL,EF_DNS
	MOVLW	TK_RDN			; DNS query
	RCALL	WF_CTX
; Set Ethernet time-out to (TO_PER) seconds
SET_TO	MOVF	TO_PER,W
; Set Ethernet time-out to (WREG) seconds
ETH_TM	MULLW	.20
	COMF	PRODL,W
	MOVWF	TICK_L
	COMF	PRODH,W
	MOVWF	TICK_H
	BCF	ETH_FL,EF_TMO		; Clear the time-out flag
	BSF	TIM_ST,TM_ETH		; Start the Ethernet timer
	RETURN	
;
; Check if the IP number at (FSR1) is local
CK_LOC	LFSR	FSR2,IP_MSK
	MOVLW	IPN_LN
	MOVWF	TB_CNT
CKL_LP	MOVLW	-IPN_LN
	MOVF	PLUSW2,W		; Own IP number byte
	XORWF	POSTINC1,W		; Target IP number byte
	ANDWF	POSTINC2,W		; Mask byte
	BNZ	CKL_EN
	DECFSZ	TB_CNT
	BRA	CKL_LP
CKL_EN	RETURN
; DNS reply received
DNS_RP	RCALL	BUF_RD			; Get packet body
	LFSR	FSR2,NR_IDN
	LFSR	FSR1,MAC_NR
	MOVLW	NR_PAR-NR_IDN
	RCALL	STR_CP
	BTFSS	STATUS,Z
DNS_RT	RETURN
	LFSR	FSR1,NR_QDC
	LFSR	FSR2,NR_NAM
	MOVLW	1<<DSC
	MOVWF	SD_CMD
REC_LP	TSTFSZ	PREINC1
	BRA	SKP_NA
	DCFSNZ	SD_CMD
	BRA	EE_NPR
	MOVF	PREINC1,W
	BRA	REC_LP
SKP_NA	DECF	POSTDEC1
	MOVLW	0xc0
	CPFSLT	INDF2
	BRA	NAM_PT
NSG_LP	MOVF	POSTINC2,W
	BZ	NAM_EN
	RCALL	ADV_X2
	BRA	NSG_LP
NAM_PT	RCALL	SK_WRD
NAM_EN	MOVLW	4
	BTFSC	SD_CMD,DSC
	BRA	NX_REC
	copy_b	FSR2L,SD_SZ2,2
	TSTFSZ	POSTINC2
	BRA	SK_REC
	DECFSZ	POSTINC2
	BRA	SK_REC
	TSTFSZ	POSTINC2
	BRA	SK_REC
	DECFSZ	POSTINC2
	BRA	SK_REC
	RCALL	ADV_X2
	TSTFSZ	POSTINC2
	BRA	SK_REC
	MOVLW	IPN_LN
	CPFSEQ	POSTINC2
	BRA	SK_REC
	BTFSC	SYS_FL,CM_DNS		; DNS query command?
	BRA	ETH_IP			; Yes!
	RCALL	IP_CPY			; Copy the IP number
	BRA	HN_RES			; Link station
; Skip DNS record
SK_REC	copy_b	SD_SZ2,FSR2L,2
	MOVLW	9
	RCALL	ADV_X2
	MOVF	POSTINC2,W
NX_REC	RCALL	ADV_X2
	BRA	REC_LP
; Set the outgoing payload address and length
PT_ADR	copy_b	PT_LNL,DT_CNL,2		; Payload length
	LFSR	FSR1,DATA_B
	BTFSC	ETH_FL,EF_STZ		; Broadcast?
	RETURN				; Yes!
	BTFSC	ETH_FL,EF_WRA		; Write alternate?
	LFSR	FSR1,ALT_DB		; Yes!
	BTFSC	ETH_FL,EF_RPL		; Reply?
	LFSR	FSR1,PR_DAT		; Yes!
	RETURN
;
; UDP datagram received
UDP_DG	LFSR	FSR2,PR_CTL	
	BTFSC	POSTDEC2,REQ		; Request datagram?
	BRA	CK_SRQ			; Yes!
	MOVF	INDF2,W
	XORLW	ST_LNK
	BNZ	RPL_DG
; Station link reply received
	RCALL	UDP_PL			; Get payload
	MOVF	PREINC2,W
	ANDLW	0x0f
	MOVWF	CTL_CD			; Point (FSR1) to the
	RCALL	STA_IX			; station number table entry 
	LFSR	FSR2,PR_BLK		; Point to payload block number
	MOVF	INDF2,W			; Allocated station number
	BTFSC	STATUS,Z		; Is it 0?
	BRA	EE_IST			; No number allocated!
	MOVLW	FIL_DV
	BTFSS	INDF2,FDV
	MOVLW	FLA_DV
	IORWF	CTL_CD
	RCALL	ST_ENT			; Create a station table entry
NO_ERR	MOVLW	0
	BRA	ETH_ER
RPL_DG	XORLW	ST_LNK^NET_TX		; Net transmit broadcast?
	BZ	CK_SRQ			; Yes!
; General reply datagram received
	MOVF	PREINC2,W		; Point (FSR1) to the
	RCALL	STA_IX			; station number table entry 
	BTFSS	INDF2,CLN		; Server station?
	BTFSS	INDF1,REQ		; Expecting a reply?
	BRA	RPL_RT			; No! Discard packet
	MOVF	POSTDEC2,W		; Point (FSR2) to the payload
	XORWF	INDF1,W			; When receiving a reply,
	ANDLW	1<<PAR			; the parity bits need 	
	BNZ	RPL_RT			; to be equal
	LFSR	FSR1,CMD_CD		; Header destination address
	MOVLW	PL_HLN			; Header length
	RCALL	CPY_LP			; Copy the payload header
        BTFSS   COM_ST,CO_RDT           ; 'Read-Transmit'?
        BRA     RPL_AD                  ; No!
	copy_b	CNT_LO,BLN_LO,2		; Change the payload data length
RPL_AD	BTFSS	ETH_FL,EF_RDA		; Read alternate?
	BRA	RPL_DT			; No!
	CLRF	BLN_LO
	CLRF	BLN_HI
	LFSR	FSR1,ALT_DB
RPL_DT	RCALL	UDP_PL			; Get payload
	BRA	ETH_EN			; Done!
RPL_RT	BTFSC	ETH_ST,ET_WFI
	BRA	WT_PLD
	RETURN
; Request datagram received
CK_SRQ	RCALL	UDP_PL			; Get payload
	MOVLW	ST_LNK
	CPFSEQ	POSTINC2
	BRA	REQ_DG
; Station link request received
	BCF	POSTINC2,REQ		; Clear the 'Request' flag
	MOVF	INDF2,W			; Name / IP number length
	BZ	CHK_ID			; None specified
	CLRF	POSTDEC2		; Set response length	
	LFSR	FSR1,IP_OWN		; Own IP number
	MOVLW	IPN_LN			; IP number length
	BTFSC	INDF2,EOF		; IP number specified?
	BRA	CHK_ST			; Yes!
	LFSR	FSR1,HOST_N		; Own station name
	MOVLW	MA_OWN-HOST_N		; Station name length
	BSF	SYS_FL,CM_CAS		; Signal: Upper case compare
CHK_ST	LFSR	FSR2,PR_DAT		; Requested name / IP address
	RCALL	STR_CP			; Compare name / IP number
	BCF	SYS_FL,CM_CAS		; Signal: Normal compare
	BNZ	REQ_RT			; The name / IP does not match
CHK_ID	BSF	SYS_FL,CM_SKP		; Signal: Reply
	table_	STA_TP
	RCALL	ETH_CM
	copy_b	PR_BLK+2,CM_CNL,2	; Search for the client's ID	
	RCALL	FND_ID			; in the station table
	BNZ	SET_ST			; The client is already recorded
	LFSR	FSR1,CST_TB-SRL		; in the station table
	MOVLW	-CST
	MOVWF	TB_CNT			; Search for an available client
CFR_LP	INCFSZ	TB_CNT			; record in the station table
	BRA	CFR_NX	
	LFSR	FSR1,PT_DAT		; No available client record
	BRA	NO_REC			; found in the station table	
CFR_NX	MOVLW	SRL
	ADDWF	FSR1L
	TSTFSZ	INDF1
	BRA	CFR_LP
SET_ST	MOVLW	MST
	ADDWF	TB_CNT			; Available client station number
	MOVFF	TB_CNT,PT_BLK		; Allocate the station number
NO_REC	LFSR	FSR2,PR_BLK		; Client station and ID address
	RCALL	ST_ENT			; Create a station table entry
	copy_b	FSR1L,FSR2L,2
	MOVLW	SRL-SHL			; Point (FSR2) to the client's
	SUBWF	FSR2L			; MAC address in the station table
	BRA	UDP_TX
; General request datagram received
REQ_DG	BTFSS	INDF2,CLN		; Client station?
REQ_RT	RETURN				; No! Discard packet
	MOVF	INDF2,W
	MOVWF	SD_CMD			; Point (FSR1) to the
	RCALL	STA_IX			; station number table entry 
	MOVF	POSTDEC2,W		; Point (FSR2) to the payload		
	BTFSS	SD_CMD,REQ		; Request?
	BRA	NTX_BC			; No! Net transmit broadcast
	copy_b	FSR1L,CRT_ST,2		; Save the station table pointer
	XORWF	INDF1,W			; When receiving a new request,
	ANDLW	1<<PAR			; the parity bits need to be
	BNZ	NEW_RQ			; different. Otherwise re-send
	MOVF	INDF2,W			; the reply datagram
	XORLW	SD_WRB			; For write requests only the
	BZ	NOP_CM			; saved error code is re-sent
	XORLW	SD_WRB^NET_TX		; Write request?
	BNZ	RE_SND
NOP_CM	CLRF	INDF2			; Yes! Change the command code
	BRA	RE_SND
NEW_RQ	BTG	INDF1,PAR		; Signal: Reply sent!
RE_SND	BSF	SYS_FL,CM_SKP		; Signal: Reply
	MOVF	INDF2,W			; Command code
	BZ	PRV_ER			; NOP!
	XORLW	NET_TX
	BNZ	SDC_OP			; No!
; Reply to a network transmit request
	INCF	INDF2			; Change error code to NET_RX
NTX_BC	BCF	INDF1,EOF
	BTFSC	PREINC2,EOF		; EOF message?
	BSF	INDF1,EOF		; Signal: EOF message received
	MOVLW	0x0f
	ANDWF	INDF2			; Station number
	MOVLW	SST			; First client station
	SUBWF	POSTINC2,W		; Receive buffer number
	LFSR	FSR1,FAT_BF
	ADDWF	FSR1H			; Station's receive buffer
	MOVFF	INDF2,DT_CNL
	CLRF	POSTINC2		; Set reply length to 0
	CLRF	POSTINC2
	MOVF	INDF1,W			; Length of data already 
	MOVWF	DT_CNH			; in buffer
	ADDWF	DT_CNL,W		; Check space for adding
	BTFSC	STATUS,C		; the received message
	BRA	NTX_EN			; Not enough space!
	MOVWF	POSTINC1		; New data length in buffer
	MOVF	DT_CNH,W		; Point (FSR1) to the 
	ADDWF	FSR1L			; write position
	MOVF	DT_CNL,W		; Bytes to be added to buffer
	RCALL	STR_CY			; Copy message to buffer
	MOVLW	0
	MOVFF	WREG,PR_CMD		; Clear the error code
NTX_EN	MOVLW	-.10
	BTFSS	SYS_FL,SR_PRE
	MOVWF	MSG_TM
	BTFSC	SD_CMD,REQ		; Broadcast?
	BRA	EPK_TB			; No! Send reply
	RETURN
PRV_ER	MOVF	PREINC1,POSTINC2	; Copy the previous error code
	CLRF	PREINC2			; Set the response length to 0
	CLRF	PREINC2
	BRA	EPK_TB
; Perform the requested SDC operation
SDC_OP	XORLW	NET_TX^PER_CM
	RCALL	ETD_CM			; Execute SDC operation
	BRA	EPK_TB			; Send response
;
; Ethernet host commands
ET_INF	MOVF    POSTINC0,W		; Get the command code
	XORLW	'i'
	BNZ	ET_STA
	LFSR	FSR2,WAP_ID
	MOVLW	IP_STA-WAP_ID
ET_SND	MOVWF	BLN_LO
	LFSR	FSR1,DATA_B
; Copy (WREG) bytes from (FSR2) to (FSR1)
STR_CY	TSTFSZ	WREG
	BRA	CPY_LP
	RETURN
CPY_LP	MOVFF	POSTINC2,POSTINC1
	DECFSZ	WREG
	BRA	CPY_LP
	RETURN
ET_STA	XORLW	'i'^'s'
	BNZ	ET_VER
	LFSR	FSR2,SST_TB
	MOVLW	NA_SFR-SST_TB
	BRA	ET_SND
ET_VER	XORLW	's'^'v'
	BNZ	ET_DNS
	BTFSS	ETH_ST,ET_WFI
	BRA	ET_RET
FW_VER	MOVLW	TK_VER
	RCALL	WF_CTX
WT_VER	BTFSS	PIR3,RC2IF
	BRA	WT_VER
	MOVF	RCREG2,W
	MOVWF	DATA_B
	RETURN
ET_DNS	XORLW	'v'^'d'
	BNZ	ET_DCF
	BTFSS	ETH_ST,ET_PRE
	BRA	ET_RET
	MOVLW	' '
	CPFSEQ	INDF0
	BRA	ET_RET
	MOVLW	2
	SUBWF	BLN_LO,W		; Adjust length
	BNC	ET_RET
	LFSR	FSR2,DATA_B+3		; Name address
	RCALL	ET_SND			; Move name to start of buffer
	BSF	SYS_FL,CM_DNS		; Signal: DNS query command
	POP				; Drop the host command
	BRA	CK_NAM			; return address
ET_DCF	XORLW	'd'^'a'
	BNZ	ET_APC
	MOVLW	'd'
ET_SDC	CPFSEQ	INDF0
	BRA	ET_RET
	MOVLW	SD_RDB^PER_CM		; Operation: 'Read Block'
	RCALL	ETD_SD
	LFSR	FSR2,DATA_B
	LFSR	FSR1,PR_DAT
	MOVLW	'd'
	CPFSEQ	INDF0
	LFSR	FSR1,AP_CRD
	MOVLW	2
	SUBWF	BLN_LO,W
	BNC	ET_RET
	MOVWF	INDF0
	INCF	BLN_LO,W
	CLRF	BLN_LO
	RCALL	STR_CY
	MOVLW	SD_WRB^PER_CM		; Operation: 'Write Block'	
ETD_SD	LFSR	FSR2,PR_BLK		; Set the block number
	SETF	POSTINC2		; The default settings are saved
	SETF	POSTINC2		; in the last sector of the SD
	CLRF	POSTINC2		; card's logical drive 1,
	CLRF	POSTINC2		; preceded by the itentifier 'ea'
ETD_CM	BSF	SDC_ST,SD_EBF		; Use the Ethernet buffer
	GOTO	TST_PE			; Execute SDC operation
ET_APC	XORLW	'a'^'w'	
	MOVLW	' '
	BZ	ET_SDC
ET_RET	CLRF	BLN_LO
	RETURN
; Create a station table entry
ST_ENT	MOVLW	SHL	
	RCALL	CPY_LP
	LFSR	FSR2,ER_SMA
	MOVLW	MAC_LN
	RCALL	CPY_LP
	LFSR	FSR2,IR_SIP
	BTFSC	ETH_ST,ET_WFI
	LFSR	FSR2,PR_SIP
	MOVLW	IPN_LN
	RCALL	CPY_LP
	LFSR	FSR2,UR_SPO
	MOVLW	2
	BRA	CPY_LP
; Set the length of a DNS packet
DNS_LN	MOVWF	SD_SZ2
	CLRF	SD_SZ3
	LFSR	FSR1,SD_SZ2
; Set the length of an UDP packet
UDP_LN	LFSR	FSR2,IT_LNL
	RCALL	ADD_BL
	LFSR	FSR2,UT_LNL	
ADD_BL	MOVF	POSTINC1,W
	ADDWF	POSTDEC2
	MOVF	POSTDEC1,W
	ADDWFC	POSTINC2
	RETURN
CK_IPL	TSTFSZ	BLN_LO
	BRA	GET_IP
; Return the first IP address that is recorded only
; as client in the station table
	LFSR	FSR2,CST_TB		; Client table address
CST_LP	MOVF	INDF2,W			; Empty table entry?
	BZ	CST_NX			; Yes!
	MOVLW	2
	MOVFF	PLUSW2,CM_CNL
	MOVLW	3
	MOVFF	PLUSW2,CM_CNH
	LFSR	FSR1,SST_TB+2
SST_LP	MOVF	CM_CNL,W
	CPFSEQ	INDF1
	BRA	SST_NX
	INCF	FSR1L
	MOVF	CM_CNH,W
	CPFSEQ	POSTDEC1
	BRA	SST_NX
CST_NX	MOVLW	SRL			; Advance to the next
	ADDWF	FSR2L			; client table entry
	MOVLW	low(RPOR0)
	CPFSEQ	FSR2L			; End of table?
	BRA	CST_LP			; Not yet!
	POP
	BRA	EE_IST
SST_NX	MOVLW	SRL			; Advance to the next
	ADDWF	FSR1L			; client table entry
	MOVLW	low(CST_TB+2)
	CPFSEQ	FSR1L			; End of table?
	BRA	SST_LP			; Not yet!
	MOVLW	IPN_LN
	ADDWF	FSR2L
IP_CPY	LFSR	FSR1,IP_STA		; Return the IP address
	BRA	CPY_LP
; Convert ASCII IP addresses of lengh (~WREG) from (FRS2)
GET_IP	MOVWF	CM_CNL			; Length of string
	LFSR	FSR1,IP_STA
GIP_IL	MOVLW	-4
	MOVWF	CM_CNH
GIP_NL	MOVLW	0
GIP_LP	MOVWF	INDF1	
	INCFSZ	CM_CNL
	BRA	GIP_SS
	INCF	CM_CNH
GIP_CC	BTG	STATUS,C
	RETURN
GIP_SS	MOVLW	'.'			; Segment separator
	CPFSEQ	INDF2
	BRA	GIP_DG
	INCF	CM_CNH			; Segment separtor counter
	BNC	GIP_NX			; 4-th segment separator
	RETURN
GIP_NS	INCF	CM_CNH
	BNC	GIP_CC
GIP_NX	MOVF	POSTINC2,W
	MOVF	POSTINC1,W
	BNC	GIP_NL
	BRA	GIP_IL
GIP_DG	MOVLW	'9'+1
	CPFSLT	INDF2
	BRA	GIP_NS			; IP number separator
	MOVLW	'0'-1
	CPFSGT	INDF2
	BRA	GIP_NS			; IP number separator
	MOVLW	.10
	MULWF	INDF1
	BSF	STATUS,C
	TSTFSZ	PRODH
	RETURN				; Overflow! Return CY=1
	MOVLW	'0'
	SUBWF	POSTINC2,W
	ADDWF	PRODL,W
	BNC	GIP_LP
	RETURN				; Overflow! Return CY=1
; Position (FSR1) at station's (WREG) record
STA_IX	LFSR	FSR1,SST_TB
	ANDLW	0x0f
	BZ	SIX_RT
	DECF	WREG
	RLNCF	WREG
	RLNCF	WREG
	RLNCF	WREG
	RLNCF	WREG
	ADDWF	FSR1L
SIX_RT	RETURN
; Find a station record by ID (CM_CNL:CM_CNH)
FND_ID	LFSR	FSR1,CST_TB+2-SRL
	MOVLW	-CST
	MOVWF	TB_CNT
FID_LP	INFSNZ	TB_CNT
	BRA	FID_EN
	MOVLW	SRL
	ADDWF	FSR1L
	MOVF	CM_CNL,W
	CPFSEQ	INDF1
	BRA	FID_LP
	INCF	FSR1L
	MOVF	CM_CNH,W
	CPFSEQ	POSTDEC1
	BRA	FID_LP
FID_EN	BCF	FSR1L,1
	MOVF	TB_CNT,W
	RETURN
; Advance (FSR2) by 2 bytes
SK_WRD	MOVLW	2
; Advance (FSR2) by (WREG) bytes
ADV_X2	ADDWF	FSR2L
	MOVLW	0
	ADDWFC	FSR2H
	RETURN

;------------------------------------------------------------------------------
; WIFI NETWORKING SUBROUTINES
;
; Reception
WFI_IS	BSF	ETH_ST,ET_CYC		; Lock communication cycle
	BTFSC	COM_ST,CO_ACT		; Wait until the current
	RETURN				; communication stage ends
	BTFSC	COM_ST,CO_CYC		; During cycle?
	BRA	WIS_DI			; Yes!
	BTFSS	T3CON,TMR3ON		; Wait for the PS/2 cycle
	BTFSC	USB_ST,SB_REQ		; or server polling cycle
	RETURN				; to end
        BCF     DEV_ST,DV_DAT           ; Signal: Data not ready!
	BTFSC	MSE_ST,MS_PRE		; Inhibit the PS/2 device
	CALL	PS_INH			; if present
WIS_DI	BSF     DEV_ST,DV_BSY           ; Signal: Busy!
        BCF	INTCON,GIE		; Disable interrupts
	MOVF	RCREG2,W		; Get communication token
; Get data packet
	XORLW	TK_HED			; Packet header?
	BNZ	WIS_IP			; No!
	BSF	ETH_FL,EF_HED		; Signal: Header
	LFSR	FSR1,PR_HED		; Get Header
	MOVLW	PR_BLK-PR_HED
; Receive (WREG) bytes at (FSR1)
WIS_RN	MOVWF	DT_CNL
	CLRF	DT_CNH
; Receive (DT_CNL:DT_CNH) bytes at (FSR1)
WIS_RD	COMF	DT_CNL
	COMF	DT_CNH
	MOVLW	TK_RDY			; Signal: Ready!
	RCALL	WF_CTX
WIS_LP	INFSNZ	DT_CNL
	BRA	WIS_BO
WIS_WT	BTFSS	PIR3,RC2IF
	BRA	WIS_WT
	MOVFF	RCREG2,POSTINC1
	BRA	WIS_LP
WIS_BO	INCFSZ	DT_CNH
	BRA	WIS_WT
	BTFSS	ETH_FL,EF_HED		; Header?
	BRA	WIS_EN			; No! Return
	BCF	ETH_FL,EF_HED
	RCALL	WSI_AK			; Acknowledge header
	BRA	UDP_DG			; Process header
; Get own IP
WIS_IP	XORLW	TK_HED^TK_WIF		; Wifi started?
	BNZ	WIS_DN			; No!
	LFSR	FSR1,MAC_NR		; Get own IP number,
	MOVLW	IP_DNS-MAC_NR		; subnet mask and
	RCALL	WIS_RN			; default gateway
	BSF	DEV_ST,DV_ETH		; Signal Wifi configured
	GOTO	SVL_ON			; Switch on LED
; Get result of DNS query
WIS_DN	XORLW	TK_WIF^TK_DNS		; DNS reply?
	BNZ	WIS_EN			; No!
	BTFSS	ETH_FL,EF_DNS		; Expecting a DNS reply?		
	BRA	WIS_EN			; No!
	LFSR	FSR1,GAT_NR
	BTFSC	SYS_FL,CM_DNS
	LFSR	FSR1,BLN_LO
	MOVLW	ETH_BF-GAT_NR
	RCALL	WIS_RN			; Get the IP number
	BRA	WIF_IP
; Get payload
UDP_PL	BTFSS	ETH_ST,ET_WFI
	GOTO	BUF_RD
	copy_b	PR_LEN,DT_CNL,2		; Set payload length
; Wait for PLD Token
WT_PLD	BTFSS	PIR3,RC2IF
	BRA	WT_PLD
	MOVF	RCREG2,W
	XORLW	TK_PLD
	BNZ	WT_PLD
	MOVF	DT_CNL,W		; Check length
	IORWF	DT_CNH,W
	BNZ	WIS_RD			; Get payload
; End of reception
WIS_EN	BCF	ETH_ST,ET_CYC		; Unlock communication
	BTFSC	COM_ST,CO_CYC		; During cycle?
	BRA	WSI_EI			; Yes!
	BTFSS	USB_ST,SB_POL
	BTFSC	MSE_ST,MS_PRE		; Enable the PS/2 device
	CALL	PS_STA			; if present
WSI_EI	BCF     DEV_ST,DV_BSY           ; Signal: Not Busy!
        BSF	INTCON,GIE		; Enable interrupts
WSI_AK	MOVLW	TK_ACK			; Acknowledge reception
; Transmission
; Send (WREG)
WF_CTX	BTFSS	PIR3,TX2IF
	BRA	WF_CTX
	MOVWF	TXREG2
	RETURN
; Send token (WREG) and wait for reply
WF_TOK	RCALL	WF_CTX
; Wait for RDY Token
WT_RDY	BTFSS	PIR3,RC2IF
	BRA	WT_RDY
	MOVF	RCREG2,W
	XORLW	TK_RDY
	BNZ	WT_RDY
	RETURN
; Send (WREG) as packet length
WF_WTX	MOVWF	SD_SZ0
	CLRF	SD_SZ1
; Send packet length from (SD_SZ0:SD_SZ1)
WF_LTX	LFSR	FSR1,SD_SZ0
	MOVLW	2
; Send (WREG) bytes from (FSR1)
WF_NTX	MOVWF	DT_CNL
	CLRF	DT_CNH
; Send (DT_CNL:DT_CNH) bytes from (FSR1)
WF_DTX	COMF	DT_CNL
	COMF	DT_CNH
WTX_LP	INCFSZ	DT_CNL
	BRA	WTX_NX
	INFSNZ	DT_CNH
	RETURN
WTX_NX	BTFSS	PIR3,TX2IF
	BRA	WTX_NX
	MOVFF	POSTINC1,TXREG2
	BRA	WTX_LP
; Start Wifi
WF_STA	MOVLW	TK_WIF
	RCALL	WF_TOK
	MOVLW	MA_OWN+1-WAP_ID
	RCALL	WF_WTX
	LFSR	FSR1,WAP_ID
	MOVLW	MA_OWN+1-WAP_ID
; Send (WREG) bytes from (FSR1) and wait for reply
WF_ATX	RCALL	WF_NTX
; Wait for ACK Token
WT_ACK	BTFSS	PIR3,RC2IF
	BRA	WT_ACK
	MOVF	RCREG2,W
	XORLW	TK_ACK
	BNZ	WT_ACK
	RETURN
; Send data packet or DNS query
WF_PTX	BTFSC	ETH_FL,EF_DNS
	BRA	WF_DNS
; Send data packet 
	MOVLW	TK_RPL
	BTFSS	ETH_FL,EF_RPL		; Reply?
	MOVLW	TK_REQ			; No!
	RCALL	WF_TOK
	copy_b	PT_LNL,SD_SZ0,2
	MOVLW	IPN_LN			; Destination IP length
	ADDWF	SD_SZ2,W		; add to payload header length
	ADDWF	SD_SZ0			; add to payload length
	MOVLW	0
	ADDWFC	SD_SZ1
	RCALL	WF_LTX
	LFSR	FSR1,IT_DIP		; Destination IP number
	MOVLW	IPN_LN
	RCALL	WF_NTX
	LFSR	FSR1,PT_CMD		; Payload header address
	MOVF	SD_SZ2,W		; Payload header length
	RCALL	WF_NTX			; Send payload header
	RCALL	PT_ADR
WF_DAK	RCALL	WF_DTX			; Send payload data
	BRA	WT_ACK
; Send DNS query
WF_DNS	MOVLW	TK_DNS
	RCALL	WF_TOK
	INCF	BLN_LO,W		; Station name length
	RCALL	WF_WTX
	LFSR	FSR1,DATA_B		; Station name address
	INCF	BLN_LO,W
	BRA	WF_ATX
; Update firmware
WF_UPD	RCALL	FW_VER
	ANDLW	0x3f
	XORLW	WF_VER
	BZ	WUD_ER
	CLRF	CTL_CD
	MOVLW	WF_FW1
	BTFSC	DATA_B,6
RQ_DAT	MOVLW	WF_FW2
	MOVWF	CMD_CD
	CLRF	BLN_LO
	CLRF	BLN_HI
	BSF	USB_ST,SB_CMD		; Signal: Server command
	BSF     PIE3,SSP2IE             ; Trigger an interrupt
WT_DAT	BTFSC	USB_ST,SB_CMD		; Wait for data received
	BRA	WT_DAT			; from server
	MOVF	CMD_CD,W
	BNZ	CK_EOF
	MOVLW	TK_FLA			; Send data to Wifi module
	RCALL	WF_TOK
	LFSR	FSR1,DATA_B
	copy_b	BLN_LO,DT_CNL,2
	RCALL	WF_DAK
	BRA	RQ_DAT
CK_EOF	XORLW	E_ENOF
	BNZ	WUD_ER
	MOVLW	TK_END			; Send reboot command
	RCALL	WF_CTX
	CALL	ETH_DY
WUD_WT	BTFSS	ETH_FL,EF_TMO		; Wait 5 seconds
	BRA	WUD_WT
	MOVLW	'&'^'!'			; Signal: RESET
WUD_ER	CLRF	CMD_CD
	RETURN

;------------------------------------------------------------------------------
; COMMAND INTERPRETER SUBROUTINES
;
; Decrement the buffer pointer at (FSR1)
DEC_BP	DECF	INDF1			; Decrement the next packet
	BC	DP_RET			; pointer, wrapping around to
	DECF	PREINC1			; the receive buffer end,
	MOVF	POSTDEC1,W		; if necessary
	BC	DP_RET			; DECF affects CY in the same
	MOVLW	low BF_RXE		; way as adding 0xff
	MOVWF	POSTINC1
	MOVLW	high BF_RXE
	MOVWF	POSTDEC1
DP_RET	RETURN
; Create a unique MAC address and host name
NET_ID  BSF	ETH_ST,ET_PRE		; Ethernet is present
	LFSR	FSR2,MAC_NR
	LFSR	FSR1,DATA_B+2
	RCALL	GET_NR
	MOVWF	MS_REG
	RCALL	GET_NR
	MOVWF	POSTINC2
	RCALL	GET_NR
	RLCF	WREG
	RLCF	WREG
	RLCF	WREG
	RLCF	WREG
	MOVWF	POSTDEC2
	RLCF	INDF2
	MOVLW	4
DNR_LP	RRCF	MS_REG
	RRCF	POSTINC2
	RRCF	POSTDEC2
	DECFSZ	WREG
	BRA	DNR_LP
	LFSR	FSR1,HOST_I
	RCALL	HX_BYT
	MOVF	PREINC2,W
	ANDLW	.63
	ADDLW	.132
	MOVFF	WREG,IP_LDG
HX_BYT	RCALL	HX_DIG
HX_DIG	RLNCF	INDF2
	RLNCF	INDF2
	RLNCF	INDF2
	RLNCF	INDF2
	MOVF	INDF2,W
	ANDLW	b'00001111'
	ADDLW	-.10
	BNC	HX_NUM
	ADDLW	'A'-1-'9'
HX_NUM	ADDLW	'0'+.10
	MOVWF	POSTINC1
	RETURN
; Collect a two digit number
GET_NR	MOVLW	'0'
	SUBWF	INDF1
	MOVLW	.10
	MULWF	POSTINC1
	MOVLW	'0'
	SUBWF	POSTINC1,W
	ADDWF	PRODL,W
	RETURN

;------------------------------------------------------------------------------
; DATA SECTION
;
; Baud rate constants
BAUD_D	dw	.300,FOSC_4/.300-1
	dw	.600,FOSC_4/.600-1
	dw	.1200,FOSC_4/.1200-1
	dw	.2400,FOSC_4/.2400-1
	dw	.4800,FOSC_4/.4800-1
	dw	.9600,FOSC_4/.9600-1
	dw	.19200,FOSC_4/.19200-1
	dw	.57600,FOSC_4/.57600-1
	dw	0xffff,FOSC_4/BAUD_R-1
;
; Keyboard tables
SCN_TB	db	SC_B,SC_H,SC_Y,SC_6,SC_5,SC_T,SC_G,SC_V	; BHY65TGV
	db	SC_N,SC_J,SC_U,SC_7,SC_4,SC_R,SC_F,SC_C	; NJU74RFC
	db	SC_M,SC_K,SC_I,SC_8,SC_3,SC_E,SC_D,SC_X	; MKI83EDX
	db	C_CT,SC_L,SC_O,SC_9,SC_2,SC_W,SC_S,SC_Z	;  LO92WSZ
	db	C_SP,C_EN,SC_P,SC_0,SC_1,SC_Q,SC_A,C_SH	;   P01QA
;
PRI_TB	db	C_AC,KN_X|1<<KB_CTL		; ` 	 = Pound
	db	C_MI,KN_J|1<<KB_CTL		; -
	db	C_EQ,KN_L|1<<KB_CTL		; =
	db	C_CO,KN_O|1<<KB_CTL		; ;
	db	C_AP,KN_7|1<<KB_CTL		; '
	db	C_CM,KN_N|1<<KB_CTL		; ,
	db	C_DO,KN_M|1<<KB_CTL		; .
	db	C_SL,KN_V|1<<KB_CTL		; /
	db	C_ES,N_SP|1<<KB_SHF		; Esc    = BREAK
	db	C_TB,KN_1|1<<KB_SHF		; Tab    = EDIT
	db	C_CL,KN_2|1<<KB_SHF		; Caps   = CAPS LOCK
	db	C_HO,KN_3|1<<KB_SHF		; Home   = TRUE VIDEO
	db	C_ED,KN_4|1<<KB_SHF		; End    = INVERSE VIDEO
	db	C_LT,KN_5|1<<KB_SHF		; Arrow Left
	db	C_DN,KN_6|1<<KB_SHF		; Arrow Down
	db	C_UP,KN_7|1<<KB_SHF		; Arrow Up
	db	C_RT,KN_8|1<<KB_SHF		; Arrow Right
	db	C_IN,KN_9|1<<KB_SHF		; Insert = GRAPHICS
	db	C_BS,KN_0|1<<KB_SHF		; BkSp   = DELETE
	db	C_WD,N_SS|1<<KB_SHF		; Win    = EXTENDED MODE
	db	C_DE,KN_P|1<<KB_CTL|1<<KB_SHF	; Delete = Copyright
	db	C_PL,KN_Y|1<<KB_CTL|1<<KB_SHF	; [
	db	C_PR,KN_U|1<<KB_CTL|1<<KB_SHF	; ]
	db	C_IS,KN_D|1<<KB_CTL|1<<KB_SHF	; \
	dw	0
SHF_TB	db	SC_1,KN_1|1<<KB_CTL		; !
	db	SC_2,KN_2|1<<KB_CTL		; @
	db	SC_3,KN_3|1<<KB_CTL		; #
	db	SC_4,KN_4|1<<KB_CTL		; $
	db	SC_5,KN_5|1<<KB_CTL		; %
	db	SC_6,KN_H|1<<KB_CTL		; ^
	db	SC_7,KN_6|1<<KB_CTL		; &
	db	SC_8,KN_B|1<<KB_CTL		; *
	db	SC_9,KN_8|1<<KB_CTL		; (
	db	SC_0,KN_9|1<<KB_CTL		; )
	db	C_MI,KN_0|1<<KB_CTL		; _
	db	C_EQ,KN_K|1<<KB_CTL		; +
	db	C_CO,KN_Z|1<<KB_CTL		; :
	db	C_AP,KN_P|1<<KB_CTL		; "
	db	C_CM,KN_R|1<<KB_CTL		; <
	db	C_DO,KN_T|1<<KB_CTL		; >
	db	C_SL,KN_C|1<<KB_CTL		; ?
	db	C_AC,KN_A|1<<KB_CTL|1<<KB_SHF	; ~
	db	C_PL,KN_F|1<<KB_CTL|1<<KB_SHF	; {
	db	C_PR,KN_G|1<<KB_CTL|1<<KB_SHF	; }
	db	C_IS,KN_S|1<<KB_CTL|1<<KB_SHF	; |
	dw	0
SBS_TB	db	C_RS,C_SH			; Shift
	db	C_RC,C_CT			; Ctrl
	db	C_RA,C_AL			; Alt	
	db	C_RW,C_WD			; Win
	db	C_NL|1<<7,C_NL			; Num Lock
	dw	0
;
FLT_TB	db	SC_A,SC_B,SC_C,SC_D,SC_E,SC_F,SC_G,SC_I		; ABCDEFGI
	db	SC_L,SC_M,SC_N,SC_O,SC_P,SC_R,SC_S,SC_T		; LMNOPRST
	db	SC_U,SC_V,SC_W,C_CM,SC_Y,C_DO			; UVW<Y>
;
A_TOKN	db	SC_T,SC_N|1<<7,KN_E|1<<KB_SHF|1<<KB_CTL,SC_T	; ATN
	db	SC_T|1<<7,KN_L|1<<KB_SHF|1<<KB_CTL		; ATTR
	db	SC_B|1<<7,KN_G|1<<KB_SHF			; ABS
	db	SC_C|1<<7,KN_W|1<<KB_SHF|1<<KB_CTL		; ACS
	db	SC_N|1<<7,KN_Y|1<<KB_CTL			; AND
	db	SC_S|1<<7,KN_Q|1<<KB_SHF|1<<KB_CTL		; ASN
	db	SC_T|1<<7,KN_I|1<<KB_CTL,0			; AT
B_TOKN	db	SC_E|1<<7,KN_Z|1<<KB_SHF|1<<KB_CTL		; BEEP
	db	SC_I|1<<7,KN_B|1<<KB_SHF			; BIN
	db	SC_O|1<<7,KN_B					; BORDER
	db	SC_R|1<<7,KN_B|1<<KB_SHF|1<<KB_CTL		; BRIGHT
	db	SC_L,SC_A|1<<7,KN_0,SC_L			; BLACK
	db	SC_U|1<<7,KN_1,0				; BLUE
C_TOKN	db	SC_L,SC_E|1<<7,KN_X,SC_L			; CLEAR
	db	SC_O|1<<7,KN_5|1<<KB_SHF|1<<KB_CTL	 	; CLOSE #
	db	SC_L,SC_S|1<<7					; CLS
	db	KN_V,SC_O,SC_D|1<<7,KN_I|1<<KB_SHF		; CODE
	db	SC_O,SC_N|1<<7,KN_C,SC_O			; CONTINUE
	db	SC_P|1<<7,KN_Z,SC_O,SC_S|1<<7			; COPY
	db	KN_W|1<<KB_SHF,SC_A|1<<7			; COS
	db	KN_9|1<<KB_SHF|1<<KB_CTL,SC_H|1<<7		; CAT
	db	KN_U|1<<KB_SHF,SC_I|1<<7			; CHR$
	db	KN_H|1<<KB_SHF|1<<KB_CTL			; CIRCLE
	db	SC_Y|1<<7,KN_5,0				; CYAN
D_TOKN	db	SC_A|1<<7,KN_D|1<<KB_SHF			; DATA
	db	SC_E|1<<7,KN_1|1<<KB_SHF|1<<KB_CTL		; DEF FN
	db	SC_I|1<<7,KN_D					; DIM
	db	SC_R|1<<7,KN_W,0				; DRAW
E_TOKN	db	SC_R|1<<7,KN_7|1<<KB_SHF|1<<KB_CTL		; ERASE
	db	SC_X|1<<7,KN_X|1<<KB_SHF,0			; EXP
F_TOKN	db	SC_O,SC_R,SC_M|1<<7,KN_0|1<<KB_SHF|1<<KB_CTL	; FORMAT
	db	SC_O|1<<7,KN_F					; FOR
	db	SC_L|1<<7,KN_V|1<<KB_SHF|1<<KB_CTL		; FLASH
	db	SC_N|1<<7,KN_2|1<<KB_SHF|1<<KB_CTL,0		; FN
G_TOKN	db	SC_O,SC_S|1<<7,KN_H,SC_O			; GOSUB
	db	SC_T|1<<7,KN_G					; GOTO
	db	SC_R|1<<7,KN_4,0				; GREEN
I_TOKN	db	SC_N,SC_K,SC_E|1<<7,KN_N|1<<KB_SHF		; INKEY$
	db	SC_N,SC_K|1<<7,KN_X|1<<KB_SHF|1<<KB_CTL,SC_N	; INK
	db	SC_P|1<<7,KN_I					; INPUT
	db	SC_N,SC_T|1<<7,KN_R|1<<KB_SHF,SC_N		; INT
	db	SC_V|1<<7,KN_M|1<<KB_SHF|1<<KB_CTL		; INVERSE
	db	SC_F|1<<7,KN_U					; IF
	db	SC_N|1<<7,KN_I|1<<KB_SHF|1<<KB_CTL,0		; IN
L_TOKN	db	SC_E,SC_N|1<<7,KN_K|1<<KB_SHF,SC_E		; LEN
	db	SC_T|1<<7,KN_L					; LET
	db	SC_I,SC_N|1<<7,KN_3|1<<KB_SHF|1<<KB_CTL,SC_I	; LINE
	db	SC_S|1<<7,KN_K					; LIST
	db	SC_L|1<<7,KN_V|1<<KB_SHF			; LLIST
	db	SC_N|1<<7,KN_Z|1<<KB_SHF			; LN
	db	SC_O|1<<7,KN_J					; LOAD
	db	SC_P|1<<7,KN_C|1<<KB_SHF,0			; LPRINT
M_TOKN	db	SC_E|1<<7,KN_T|1<<KB_SHF|1<<KB_CTL		; MERGE
	db	SC_O|1<<7,KN_6|1<<KB_SHF|1<<KB_CTL		; MOVE
	db	SC_A|1<<7,KN_3,0				; MAGENTA
N_TOKN	db	SC_E,SC_W|1<<7,KN_A,SC_E			; NEW
	db	SC_X|1<<7,KN_N					; NEXT
	db	SC_O|1<<7,KN_S|1<<KB_CTL,0			; NOT
O_TOKN	db	SC_P|1<<7,KN_4|1<<KB_SHF|1<<KB_CTL		; OPEN #
	db	SC_R|1<<7,KN_U|1<<KB_CTL			; OR
	db	SC_U|1<<7,KN_O|1<<KB_SHF|1<<KB_CTL		; OUT
	db	SC_V|1<<7,KN_N|1<<KB_SHF|1<<KB_CTL,0		; OVER
P_TOKN	db	SC_A,SC_P|1<<7,KN_C|1<<KB_SHF|1<<KB_CTL,SC_A	; PAPER
	db	SC_U|1<<7,KN_M					; PAUSE
	db	SC_O,SC_I|1<<7,KN_8|1<<KB_SHF|1<<KB_CTL,SC_O	; POINT
	db	SC_K|1<<7,KN_O					; POKE
	db	SC_E|1<<7,KN_O|1<<KB_SHF			; PEEK
	db	SC_I|1<<7,KN_M|1<<KB_SHF			; PI
	db	SC_L|1<<7,KN_Q					; PLOT
	db	SC_R|1<<7,KN_P,0				; PRINT
R_TOKN	db	SC_E,SC_A|1<<7,KN_A|1<<KB_SHF,SC_E		; READ
	db	SC_M|1<<7,KN_E,SC_E,SC_S|1<<7			; REM
	db	KN_S|1<<KB_SHF,SC_E				; RESTORE
	db	SC_T|1<<7,KN_Y					; RETURN
	db	SC_A|1<<7,KN_T					; RANDOMIZE
	db	SC_N|1<<7,KN_T|1<<KB_SHF			; RND
	db	SC_U|1<<7,KN_R					; RUN
	db	SC_E|1<<7,KN_2,0				; RED
S_TOKN	db	SC_T,SC_E|1<<7,KN_D|1<<KB_CTL,SC_T		; STEP
	db	SC_O|1<<7,KN_A|1<<KB_CTL			; STOP
	db	SC_T,SC_R|1<<7,KN_Y|1<<KB_SHF,SC_A|1<<7		; STR$
	db	KN_S,SC_C|1<<7					; SAVE
	db	KN_K|1<<KB_SHF|1<<KB_CTL,SC_G|1<<7		; SCREEN$
	db	KN_F|1<<KB_SHF,SC_I|1<<7			; SGN
	db	KN_Q|1<<KB_SHF,SC_Q|1<<7			; SIN
	db	KN_H|1<<KB_SHF,0				; SQR
T_TOKN	db	SC_A,SC_B|1<<7,KN_P|1<<KB_SHF,SC_A		; TAB
	db	SC_N|1<<7,KN_E|1<<KB_SHF			; TAN
	db	SC_H|1<<7,KN_G|1<<KB_CTL			; THEN
	db	SC_O|1<<7,KN_F|1<<KB_CTL,0			; TO
U_TOKN	db	SC_S|1<<7,KN_L|1<<KB_SHF,0			; USR
V_TOKN	db	SC_A,SC_L,SC_4|1<<7,KN_J|1<<KB_SHF|1<<KB_CTL	; VAL$
	db	SC_A,SC_L|1<<7,KN_J|1<<KB_SHF,SC_E|1<<7		; VAL
	db	KN_R|1<<KB_SHF|1<<KB_CTL,0			; VERIFY
W_TOKN	db	SC_H|1<<7,KN_7,0				; WHITE
X_TOKN	db	C_EQ|1<<7,KN_Q|1<<KB_CTL			; <=
	db	C_DO|1<<7,KN_W|1<<KB_CTL,0			; <>
Y_TOKN	db	SC_E|1<<7,KN_6,0				; YELLOW
Z_TOKN	db	C_EQ|1<<7,KN_E|1<<KB_CTL,0			; >=
;
IND_TK	dw	A_TOKN
	dw	B_TOKN
	dw	C_TOKN
	dw	D_TOKN
	dw	E_TOKN
	dw	F_TOKN
	dw	G_TOKN
	dw	I_TOKN
	dw	L_TOKN
	dw	M_TOKN
	dw	N_TOKN
	dw	O_TOKN
	dw	P_TOKN
	dw	R_TOKN
	dw	S_TOKN
	dw	T_TOKN
	dw	U_TOKN
	dw	V_TOKN
	dw	W_TOKN
	dw	X_TOKN
	dw	Y_TOKN
	dw	Z_TOKN
;
; Printer intialization string
PR_INI	dw	S_BUFF,SYS_LD-PR_INI-4+(PGS<<TB_SEL)
	db	SYS_LD-PR_INI-5,ESC,'@',ESC,'(','t',3,0
	db	0,3,0,ESC,'t',0,ESC,'l',.12
;
; System loader tape file parameters
SYS_LD	dw	DATA_B,STD_TP-SYS_LD-4+(PGS<<TB_SEL)
	dw	0xfff8,0,SCS,TY_TAP<<8
;
; Standard speed tape signal parameters
STD_TP	dw	S_BUFF,CMD_12-STD_TP-4+(PGS<<TB_SEL)
	dw	~(FOSC_4/.500000*.2168/7)
	dw	~(FOSC_4/.500000*.667/7)
	dw	~(FOSC_4/.500000*.735/7)
	dw	~(FOSC_4/.500000*.855/7)
	dw	~(FOSC_4/.500000*.1710/7)
	dw	0x1f80, 0x0c98
;
; SD card command strings
; STOP TRANSMISSION
CMD_12  dw      CM_CNL,0x0a+(PGS<<TB_SEL)
        db      6,1
        dw      0xffff                                  ; Response R1b
        db      SD_STR,0,0,0,0,0x30<<1|1
;
; Network initialization
N_INIT	et_lfs	LX1,ECM_SP				; Set the command interpreter
	db	WDI,low(SST_TB-1)			; stack pointer
	et_lfs	LX1,WAP_ID
	db	CLR,HOST_N-WAP_ID
	db	TCY,MA_DNS-HOST_N
	db	'I','F','1','B','0','0','0','0'		; Set the default configuration
	db	0x00,0x04,0xa3,0x1b,0,0
	db	.192,.168,.1,.196
	db	.255,.255,.255,.0
	db	.192,.168,.1,.254
	db	.192,.168,.1,.254
	et_lfs	LX1,WAP_ID
	db	TCY,PN_END-PN_STA
PN_STA	db	'S','S','I','D'
PN_END	et_lfs	LX1,WAP_PW
	db	TCY,PP_END-PP_STA
PP_STA	db	'p','a','s','s'
PP_END	et_lfs	LX1,SST_TB
	db	CLR,NA_SFR-SST_TB			; Clear all stations
	et_sbr	NET_ID	
	db	RET
; Ethernet module initialization
	db	BFC|ECON1,1<<BSEL1|1<<BSEL0		; Select bank 0
	db	WCR|ERXSTL,low BF_RXS
	db	WCR|ERXSTH,high BF_RXS
	db	WCR|ERDPTL,low BF_RXS
	db	WCR|ERDPTH,high BF_RXS
	db	WCR|ERXNDL,low BF_RXE
	db	WCR|ERXNDH,high BF_RXE
	db	WCR|ERXRDPTL,low BF_RXE
	db	WCR|ERXRDPTH,high BF_RXE
	db	WCR|EWRPTL,low BF_RPL
	db	WCR|EWRPTH,high BF_RPL
	db	WBM,0
	db	WCR|EWRPTL,low BF_REQ
	db	WCR|EWRPTH,high BF_REQ
	db	WBM,0
	db	BFS|ECON1,1<<BSEL1			; Select bank 2
	db	WCR|MACON1,1<<TXPAUS|1<<RXPAUS|1<<MARXEN
	db	WCR|MACON3,1<<PADCF0|1<<TXCRCEN|1<<FRMLNEN
	db	WCR|MABBIPG,0x12
	db	WCR|MACON4,1<<DEFER
	db	WCR|MACLCON2,.63
	db	WCR|MAIPGL,0x12
	db	WCR|MAIPGH,0x0c
	db	WCR|MAMXFLL,low .1518
	db	WCR|MAMXFLH,high .1518
	db	BFS|ECON1,1<<BSEL1|1<<BSEL0		; Select bank 3
	et_lfs	LX1,MA_OWN
	db	WXI,MAADR1				; Copy the MAC address
	db	WXI,MAADR2
	db	WXI,MAADR3
	db	WXI,MAADR4
	db	WXI,MAADR5
	db	WXI,MAADR6
	db	WCR|ECOCON,0
	et_wph	PHCON2,1<<HDLDIS
	et_wph	PHLCON,0x3472				; Set the LED display mode
	et_wph	PHCON1,0
	et_wph	PHIE,1<<PLNKIE|1<<PGEIE			; Enable the PHY interrupt
	db	RET
; Discard the current packet
DSC_PK	db	BFC|ECON1,1<<BSEL1|1<<BSEL0		; Select bank 0
	et_lfs	LX1,PH_NPK	
	db	WXI,ERDPTL				; Copy the next packet pointer
	db	WXD,ERDPTH				; to the buffer read pointer
	et_sbr	DEC_BP					; Decrement the next packet pointer 	
	db	WXI,ERXRDPTL            		; Set the receive buffer read
	db	WXD,ERXRDPTH				; pointer
	db	BFS|ECON2,1<<PKTDEC			; Decrement the packet counter
	db	RET	
; Transmit a DNS request
DNS_RQ	et_lfs	LX1,ET_DMA
	et_lfs	LX2,MA_DNS
	db	CPY,MAC_LN				; DNS server MAC address
	et_lfs	LX2,MA_OWN
	db	CPY,MAC_LN				; Own MAC address
	db	TCY,IT_SIP-ET_TYH
	be_wrd	IP_PRO					; IP protocol
; IP header
	db	IP_VER|IP_HLN>>2,IP_SRV
	be_wrd	IP_HLN+UD_HLN				; Minimal IP Length
	db	0,0,0,0					; ID and fragment number
	db	IP_TTL,UDP_PN,0,0			; TTL, protocol and checksum
	et_lfs	LX2,IP_OWN				; Source: Own IP
	db	CPY,IPN_LN
	et_lfs	LX2,IP_DNS				; Destination: DNS server IP
	db	CPY,IPN_LN
	db	TCY,NT_IDN-UT_SPO
; UDP header	
	be_wrd	UDP_CL					; UDP source port
	be_wrd	DNS_SV					; UDP destination port
	be_wrd	UD_HLN					; Minimal UDP Length
	dw	0					; UDP Checksum
; DNS header
	et_lfs	LX2,MAC_NR
	db	CPY,NT_PAR-NT_IDN
	db	TCY,NT_ANC-NT_PAR
	db	1,0,0,1
	db	CLR,NT_NAM-NT_ANC
	db	RET
	db	TCY,4
	db	0,1,0,1
	db	RET
	et_rel	CAL,REQ_TX
	et_sbr	PL_HED
	et_rel	GTO,PK_REQ
; Transmit an ARP query
ARP_QR	et_rel	CAL,ARP_RP
	db	CLR,AT_DIA-AT_SMA			; Clear target MAC address
	et_lfs	LX2,IP_DNS				; DNS server query
	et_rel	SKP,ARP_IP
	et_lfs	LX2,IP_GAT				; Default gateway query
ARP_IP	db	CPY,IPN_LN				; Copy target IP number from (FSR2)
	et_rel	CAL,ARP_PK
	et_rel	CAL,REQ_TX
	et_rel	GTO,PK_REQ
; Send an UDP request or reply
UD_REQ	et_rel	SKP,UD_RPL
	et_rel	CAL,REQ_TX
	et_sbr	PL_DAT
	et_rel	GTO,PK_REQ
UD_RPL	et_rel	CAL,RPL_TX
	et_sbr	PL_DAT
	et_rel	GTO,PK_RPL
; Transmit an ARP reply
ARP_RP	et_lfs	LX1,ET_DMA
	db	TCY,MAC_LN
	dw	-1,-1,-1				; Broadcast MAC address
	et_lfs	LX2,MA_OWN				; Copy own MAC address to
	db	CPY,MAC_LN				; Ethernet header source
	db	TCY,AT_SMA-ET_TYH
	be_wrd	ARP_PR
	be_wrd	ETH_TY
	be_wrd	IP_PRO
	db	MAC_LN,IPN_LN
	be_wrd	OT_REQ
	db	RET,1
	et_lfs	LX1,AR_HWR		
	et_lfs	LX2,AT_HWR
	db	CMP,AR_OPH-AR_HWR			; Validate the ARP header
	et_lfs	LX1,AR_DIA
	et_lfs	LX2,IP_OWN
	db	CMP,IPN_LN				; Check requested IP address
	et_lfs	LX2,AR_SMA				; Copy  the sender MAC address
	et_lfs	LX1,ET_DMA				; to the Ethernet header
	db	CPY,MAC_LN				; destination MAC address field
	et_lfs	LX2,AR_SMA
	et_lfs	LX1,AT_DMA				; Copy sender addresses to
	db	CPY,AT_DMA-AT_SMA			; destination addresses
; Finalize an ARP packet
ARP_PK	et_lfs	LX2,MA_OWN
	et_lfs	LX1,AT_SMA				; Copy own MAC and IP addresses
	db	CPY,AT_DMA-AT_SMA			; to ARP source addresses
	db	RET,1
; Transmit a reply packet
RPL_TX	et_sbr	TX_RDY
	db	BFC|ECON1,1<<BSEL1|1<<BSEL0		; Select bank 0
	db	WCR|ETXSTL,low BF_RPL
	db	WCR|ETXSTH,high BF_RPL
	db	WCR|EWRPTL,low(BF_RPL+1)
	db	WCR|EWRPTH,high(BF_RPL+1)
	et_lfs	LX1,ET_DMA
	db	WRB,HT_OPR-ET_DMA
	db	RET,1
PK_RPL	et_lfs	LX1,DT_CNL
	et_rel	GTO,PK_END
; Re-transmit a request packet from the buffer
PK_RSD	db	BFC|ECON1,1<<BSEL1|1<<BSEL0		; Select bank 0
	db	WCR|ETXSTL,low BF_REQ			; Transmit start pointer 
	db	WCR|ETXSTH,high BF_REQ
	et_lfs	LX1,TXE_PT				; Saved transmit end pointer
	et_rel	GTO,PK_RTX
; Send a DHCP discover / request packet
DHC_SN	et_rel	CAL,DHC_TP
	db	BFS|ECON1,1<<RXEN			; Enable receiving
; Transmit a request packet
REQ_TX	et_sbr	TX_RDY
	db	BFC|ECON1,1<<BSEL1|1<<BSEL0		; Select bank 0
	db	WCR|ETXSTL,low BF_REQ
	db	WCR|ETXSTH,high BF_REQ
	db	WCR|EWRPTL,low(BF_REQ+1)
	db	WCR|EWRPTH,high(BF_REQ+1)
	et_lfs	LX1,ET_DMA
	db	WRB,HT_OPR-ET_DMA
	db	RET,1
; Transmit a DHCP packet
	db	WRB,HT_CHA-HT_OPR
	db	CLB,HR_CHA-HR_SEC
	db	WRB,MAC_LN		
	db	CLB,HR_COK-HR_SHN
	db	WRB,HT_END-HT_COK
	db	CLB,HC_UDL+UR_SPO-HR_OPT-HC_OPL
PK_REQ	et_lfs	LX1,TXE_PT
; Set the end pointer and transmit a packet
PK_END	db	BFC|ECON1,1<<BSEL1|1<<BSEL0		; Select bank 0
	db	RXI,EWRPTL				; Read the buffer write pointer
	db	RXD,EWRPTH				; to (FSR1)
	et_sbr	DEC_BP					; Decrement the value and write
PK_RTX	db	WXI,ETXNDL				; it to the transmit end pointer
	db	WXD,ETXNDH
; Transmit a packet
TX_CMD	db	BFC|ECON1,1<<TXRTS			; Stop transmission
	db	BFS|ECON1,1<<TXRST			; Reset transmitter
	db	BFC|ECON1,1<<TXRST
	db	BFC|EIR,1<<TXERIF|1<<TXIF		; Clear error flags
	db	BFS|ECON1,1<<TXRTS			; Start transmission
	db	RET
; DHCP template
DHC_TP	et_rel	CAL,UDP_TP
	db	TCY,HT_XID-ET_TYH
	be_wrd	IP_PRO					; IP protocol
; IP header
	db	IP_VER|IP_HLN>>2,IP_SRV			; Version and service
	be_wrd	IP_HLN+HC_UDL				; IP length
	db	0,0,0,0					; ID and fragment number
	db	IP_TTL,UDP_PN,0,0			; TTL, protocol and checksum
	dw	0,0					; Source IP: undefined
	dw	-1,-1					; Destination IP: broadcast
; UDP header	
	be_wrd	UDP_CL					; UDP source port
	be_wrd	DHC_SV					; UDP destination port
	be_wrd	HC_UDL					; UDP Length
	dw	0					; UDP Checksum
; DHCP header
	db	OT_REQ,ETH_TY,MAC_LN,0
	et_lfs	LX2,HOST_I
	db	CPY,HT_CHA-HT_XID
	et_lfs	LX2,MA_OWN
	db	CPY,MAC_LN
	db	TCY,DHS_OE-DHS_OS
; DHCP options
DHS_OS	db	0x63,0x82,0x53,0x63			; Magic cookie
	db	OP_MTY,1				; Message type 
	db	MS_DIS,OP_LIS
	db	3,OP_MSK
	db	OP_GAT,OP_DNS
	db	OP_RIP,IPN_LN				; Requested IP
DHS_OE	et_lfs	LX2,IP_OWN
	et_rel	SKP,DHS_IP
	et_lfs	LX2,HR_YIA
DHS_IP	db	CPY,IPN_LN
	db	WDI,OP_HNA
	db	WDI,MA_OWN-HOST_N
	et_lfs	LX2,HOST_N
	db	CPY,MA_OWN-HOST_N
	et_rel	SKP,DHS_EN
	db	WDI,OP_SVI
	db	WDI,IPN_LN
	et_lfs	LX2,IP_STA
	db	CPY,IPN_LN
DHS_EN	db	WDI,OP_END
	db	RET
; Check an ARP reply packet
ARP_CK	et_lfs	LX1,IP_DNS				; DNS server query
	et_rel	SKP,ARL_CK
	et_lfs	LX1,IP_GAT				; Default gateway query
ARL_CK	et_lfs	LX2,AR_SIA
	db	CMP,IPN_LN
	et_lfs	LX1,MA_DNS				; DNS server query
	et_rel	SKP,ARL_CP
	et_lfs	LX1,MA_GAT				; Default gateway query
ARL_CP	et_lfs	LX2,AR_SMA
	db	CPY,MAC_LN
	et_rel	SKP,ARL_EN
	et_lfs	LX2,MA_GAT
	et_lfs	LX1,MA_DNS
	db	CPY,MAC_LN
	et_lfs	LX1,IP_DNS
	et_lfs	LX2,IP_GAT
	db	CMP,IPN_LN
ARL_EN	db	RET
; Station link template
STA_TP	et_lfs	LX1,PT_DAT				; Reply
	et_rel	SKP,STA_ZR
	et_lfs	LX1,WR_BLN				; Request
STA_ZR	db	CLR,2					; Copy two 0-s and the
	et_lfs	LX2,MAC_NR				; unique station ID
	db	CPY,PL_BLN-2
; General UDP template
UDP_TP	et_lfs	LX1,ET_DMA
	db	TCY,MAC_LN
; Ethernet header
	dw	-1,-1,-1				; Broadcast MAC address
	et_lfs	LX2,MA_OWN
	db	CPY,MAC_LN				; Own MAC address
	db	RET,1
	db	TCY,IT_SIP-ET_TYH
	be_wrd	IP_PRO					; IP protocol
; IP header
	db	IP_VER|IP_HLN>>2,IP_SRV
	be_wrd	IP_HLN+UD_HLN				; Minimal IP Length
	db	0,0,0,0					; ID and fragment number
	db	IP_TTL,UDP_PN,0,0			; TTL, protocol and checksum
	et_lfs	LX2,IP_OWN				; Source IP: Own IP
	db	CPY,IPN_LN
	db	TCY,PT_CMD-IT_DIP
	dw	-1,-1					; Destination IP: broadcast
; UDP header	
	be_wrd	UDP_CL					; UDP source port
	be_wrd	UDP_CL					; UDP destination port
	be_wrd	UD_HLN					; Minimal UDP Length
	dw	0					; UDP Checksum
	et_lfs	LX2,PR_CMD				; Reply
	et_rel	SKP,CPY_HD
	et_lfs	LX2,CMD_CD				; Request
CPY_HD	db	CPY,PL_HLN
	et_lfs	LX2,PT_DAT				; Reply
	et_rel	SKP,CPY_BK
	et_lfs	LX2,WR_BLN				; Request
CPY_BK	db	CPY,PL_BLN
	et_lfs	LX2,MA_GAT
	db	RET
; Set the destination MAC and IP
	et_lfs	LX1,ET_DMA
	db	CPY,MAC_LN
	et_lfs	LX1,IT_DIP
	db	CPY,IPN_LN
	db	RET
;
; Firmware ID string
	fill    0xffff,(0x18-($%0x10))%0x10
FWR_ID  dw      CMD_CD,CODE_E-FWR_ID-4+(PGS<<TB_SEL)
        dw      0,CODE_E-FWR_ID-8
        db      'B','u','i','l','d','5','4','9'
        db      '2','0','1','7','0','8','2','9'
CODE_E

;------------------------------------------------------------------------------
; End of program
        END